# -*- coding: utf-8 -*-
# -*- coding: utf-8 -*-
# pylint: disable=line-too-long, wildcard-import, unused-wildcard-import
"""
.. _`German Federal Institute of Hydrology (BfG)`: https://www.bafg.de/EN
.. _`LARSIM`: http://www.larsim.de/en/the-model/

One can understand version 3 of the L-Land model as an extended, more
complicated version of the application model |lland_v1|, using precipitation,
temperature, air pressure, wind speed, relative humidity and sunshine duration
as meteorological input parameters.  It relies on the equations after Penman
and Penman-Monteith for calculating evaporation and evapotranspiration, which
require much more input data than the simple Turc-Wendling equation but bring
flexibility regarding the simulation time step (Turc-Wendling is limited to
daily time steps, Penman and Penman-Monteith work for any time step size).
The second main difference is that |lland_v3| models the energy
balance of snow processes in more detail (following Knauf, 2006), taking not
only the air temperature but also forcings like the soil heat flux or global
radiation into account explicitly.  The third difference is that |lland_v3|
models the soil temperature to simulate the effect of soil water freezing on
runoff generation.  We created |lland_v3| on behalf of the `German Federal
Institute of Hydrology (BfG)`_ in the context of optimising the control of
the Kiel Channel (Nord-Ostsee-Kanal).

The following list summarises the main components of |lland_v3|:

 * Multiple routines for adjusting the meteorological input data
 * Transformation of sunshine duration into global radiation
   :cite:`ref-Thompson1981`
 * Evaporation from water surfaces after Penman :cite:`ref-penman1948natural`
 * Evapotranspiration (interception evaporation) after Penman-Monteith
   :cite:`ref-Thompson1981`
 * Actual soil evapotranspiration after Wigmosta
   :cite:`ref-wigmosta1994distributed`
 * Surface roughnesses after Quast and Böhm :cite:`ref-bohm1998ermittlung`
 * Mixed precipitation within a definable temperature-range
 * Energy balance of the snow-layer after Knauf :cite:`ref-LUBW2006a`
 * Dynamical adjustment of the snow albedo
 * Optional refinement of the snow-surface temperature
 * Optional (re)freezing of liquid water within the snow-layer
 * A simple snow retention routine
 * A soil routine based on the Xinanjiang model :cite:`ref-zhao1977flood`
 * Frost-sealing of the upper soil-layer
 * One base flow, two interflow and two direct flow components
 * A freely configurable capillary rise routine
 * Separate linear storages for modelling runoff concentration
 * Additional evaporation from water areas within the subcatchment
 * Optional evaporation from inflowing runoff

Some notes for `LARSIM`_ users: |lland_v3| is similar to many `LARSIM`_
models used for forecasting (relying on a combination of options equivalent
to the features listed above), but not identical.  Often, we strive for more
flexibility. One example is that you are allowed to modify parameter
|KTSchnee| to control if and how |lland_v3| adjusts the snow surface
temperature.  However, there are also differences due to technical reasons.
One relevant difference is that *HydPy* does not allow to include "future
values" into the current simulation time step. For example, `LARSIM`_ fills
nightly gaps regarding cloudiness with the average value of the particular
day, while |lland_v3| uses the average over the 24 hours preceding the
current simulation step.  Hence, do not expect the results of |lland_v3|
and `LARSIM`_ to be identical. When switching from one model system to the
other, the need to fine-tune some parameters via calibration might arise.

.. note::

    Some details of |lland_v3| are still under discussion and might change
    in the future.

Integration tests
=================

.. how_to_understand_integration_tests::

Many of the following integration tests are similar to the integration tests
performed for the application model |lland_v1|.  When there are no substantial
differences, we do not repeat our explanations, so please note the respective
sections of the documentation on |lland_v1|.  Some additional tests are
covering structural differences between land-use types that only exist for
|lland_v3| and others are showing the different possible snow-related
configurations of |lland_v3|.  Furthermore, we perform some tests both on a
daily and an hourly basis to point out some differences and to make sure
|lland_v3| works well for different simulation step sizes.

Note that, while our daily simulations always cover an entire month, our
hourly simulations cover only a period three days, focussing on the "most
interesting" part of the daily simulation.  To ensure the best possible
comparability, we use the same input data (we aggregate the hourly values
to daily values), the same parameter values (possibly adapted to the change
in the time step size), and the same initial conditions (we extract the
required initial conditions of the respective hourly simulation from the
longer daily simulation).

daily simulation
________________

The following general setup is identical with the one of |lland_v1|,
except that we now start with a different period and with a daily
simulation time step:

>>> from hydpy import pub, Timegrid
>>> pub.timegrids = '1997-08-01', '1997-09-01', '1d'
>>> from hydpy.models.lland_v3 import *
>>> parameterstep('1h')
>>> from hydpy import Node, Element
>>> outlet = Node('outlet')
>>> inlet = Node('inlet')
>>> land = Element('land', inlets=inlet, outlets=outlet)
>>> land.model = model
>>> nhru(1)
>>> ft(1.0)
>>> fhru(1.0)
>>> from hydpy import IntegrationTest
>>> IntegrationTest.plotting_options.axis1 = inputs.nied
>>> IntegrationTest.plotting_options.axis2 = fluxes.qah
>>> test = IntegrationTest(land)
>>> test.dateformat = '%Y-%d-%m'

.. _lland_v3_acker_summer_daily:

acre (summer)
-------------

First, we set the values of those parameters also required by the application
model |lland_v1| (see integration test :ref:`lland_v1_acker_summer`):

>>> lnk(ACKER)
>>> kg(0.94)
>>> kt(0.0)
>>> hinz(0.2)
>>> lai.acker_aug = 3.5
>>> tgr(0.0)
>>> trefn(0.0)
>>> tsp(4.0)
>>> pwmax(1.43)
>>> wmax(309.0)
>>> fk(199.0)
>>> pwp(119.4)
>>> kapgrenz(option='0_WMax/10')
>>> kapmax(0.08)
>>> beta(0.001/24)
>>> fbeta(1.0)
>>> rbeta(False)
>>> dmax(r_dmax=0.55)
>>> dmin(r_dmin=10.0)
>>> bsf(0.3)
>>> a1(0.0)
>>> a2(0.5)
>>> tind(0.191056)
>>> eqb(35000.0)
>>> eqi1(400.0)
>>> eqi2(400.0)
>>> eqd1(200.0)
>>> eqd2(50.0)
>>> negq(False)

Next, we set the values of the parameters specific to |lland_v3|, beginning
with those required for calculating the evaporation via Penman and the
evapotranspiration via Penman-Monteith:

>>> latitude(54.1)
>>> longitude(9.7)
>>> measuringheightwindspeed(10.0)
>>> angstromconstant(0.25)
>>> angstromfactor(0.5)
>>> angstromalternative(0.15)
>>> emissivity(0.95)
>>> py(119.4)
>>> cropheight.acker_aug = 0.4
>>> albedo.acker_aug = 0.24
>>> wg2z.aug = -0.25
>>> surfaceresistance.acker = 40.0

The following parameters are related to the energy accounting snow
modelling approach after Knauf :cite:`ref-LUBW2006a`:

>>> p1strahl(0.5)
>>> p2strahl(1.0/35.0)
>>> p1wind(0.6)
>>> p2wind(1.0/70.0)
>>> turb0(0.0072)
>>> turb1(0.00576)
>>> albedo0snow(0.9)
>>> snowagingfactor(0.35)
>>> refreezeflag(True)
>>> ktschnee(inf)

Now we set the values of the parameters controlling the frost sealing
of the upper soil layer:

>>> fvf(0.3)
>>> bsff(2.0)

Due to preparing a simulation with a daily simulation step size, we only need
to define the initial values of the state sequences:

>>> test.inits = (
...     (states.inzp, 0.0),
...     (states.wats, 0.0),
...     (states.waes, 0.0),
...     (states.esnow, 0.0),
...     (states.taus, 0.0),
...     (states.ebdn, 30.0),
...     (states.bowa, 72.0),
...     (states.qdgz1, 0.0),
...     (states.qdgz2, 0.0),
...     (states.qigz1, 0.0),
...     (states.qigz2, 0.0),
...     (states.qbgz, 0.0),
...     (states.qdga1, 0.0),
...     (states.qdga2, 0.0),
...     (states.qiga1, 0.0),
...     (states.qiga2, 0.0),
...     (states.qbga, 0.0))

Finally, we prepare the input time series:

>>> inputs.nied.series = (
...     13.3, 0.1, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0,
...     2.7, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 2.0, 0.3, 0.0, 0.1, 10.5, 6.5,
...     15.0, 9.2, 0.0, 0.0)
>>> inputs.teml.series = (
...     16.1, 17.3, 18.9, 17.8, 18.5, 19.9, 20.8, 20.9, 22.3, 23.0, 23.1,
...     23.2, 23.2, 21.7, 20.8, 19.5, 21.0, 21.6, 21.1, 22.4, 22.5, 19.6,
...     20.3, 24.1, 23.7, 21.1, 20.9, 20.2, 16.1, 17.0, 18.2)
>>> inputs.sunshineduration.series = (
...     6.3, 1.7, 4.5, 12.4, 13.9, 13.0, 13.8, 12.3, 13.1, 12.8, 13.1, 13.3,
...     12.7, 10.2, 9.4, 10.3, 11.1, 11.0, 8.5, 11.3, 12.4, 0.1, 6.7, 10.4,
...     6.5, 4.9, 6.6, 0.3, 0.1, 5.0, 3.8)
>>> inputs.windspeed.series = (
...     2.4, 1.6, 1.5, 2.8, 3.0, 3.5, 3.8, 3.3, 1.8, 1.9, 3.0, 2.5, 1.6, 2.7,
...     3.2, 1.3, 1.9, 2.0, 2.6, 3.6, 2.5, 2.1, 1.5, 2.2, 1.9, 2.3, 2.1, 2.6,
...     2.6, 2.1, 2.1)
>>> inputs.relativehumidity.series = (
...     86.2, 85.8, 82.4, 77.6, 74.9, 77.9, 73.5, 80.1, 72.9, 68.5, 68.6,
...     66.0, 69.0, 75.6, 81.4, 79.0, 75.9, 70.0, 70.3, 69.0, 62.1, 84.5,
...     83.6, 76.5, 76.6, 84.2, 85.8, 86.5, 89.6, 78.3, 78.5)
>>> inputs.atmosphericpressure.series = (
...     100.7, 101.3, 101.6, 101.7, 101.8, 101.8, 101.9, 101.9, 101.9, 102.0,
...     102.1, 102.2, 101.9, 101.5, 101.6, 101.9, 101.9, 101.8, 101.9, 101.9,
...     101.7, 101.5, 101.5, 101.1, 100.8, 100.6, 100.5, 100.0, 100.4, 101.1,
...     101.4)
>>> inlet.sequences.sim.series = 0.0

The following results illustrate the behaviour of |lland_v3| for relatively
dry and hot summer condions. Compared to the result table of integration
test :ref:`lland_v1_acker_summer` of the application model |lland_v1|,
there are much more columns, because of to the higher number of input,
flux, and state sequences.  These are mainly due to the increased data
requirements and the more complex calculations of the Penman-Monteith
equation and the Knauf :cite:`ref-LUBW2006a` approach.  Additionally,
|lland_v3| calculates many daily values (containing the averages or sums
over the last 24 hours), required for the Penman equation, which we generally
apply based on the input data of the previous 24-hours. (Note that we could
save some computation time by not calculating daily values that are never
required.  First, land-use type |ACKER| does only rely on the Penman-Monteith
equation.  Second, the values of the "normal" sequences are already daily
values, due to the selected simulation step size of one day.  Maybe we
perform some related optimisations later.):

.. integration-test::

    >>> conditions_acker_summer = test(
    ...     'lland_v3_acker_summer_daily',
    ...     axis1=(inputs.nied, fluxes.qah),
    ...     axis2=states.bowa,
    ...     get_conditions='1997-08-03')
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |     g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes | sbes |     evpo |   evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg |   wada |      qdb | qib1 | qib2 | qbb |     qkap |     qdgz |      qah |       qa | inzp | wats | waes | esnow | taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 |      qbgz |    qdga1 | qdga2 | qiga1 | qiga2 |      qbga | inlet |   outlet |
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | -0.25 | 10.933333 |   -2.6784 |             12.492674 |                 12.492674 |                  12.492674 |                  1.877708 |                      0.0 |    10.614966 |              0.0 |         10.614966 |             58.847345 |                 100.0 |               111.309423 |               99.889466 | 11.802 |  0.0 | 3.623966 |   0.7 | 1.720207 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 11.802 | 0.761343 |  0.0 |  0.0 | 0.0 |      0.0 | 0.761343 | 0.196005 | 0.002269 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   32.4284 |  81.32045 | 0.761343 |   0.0 |   0.0 |   0.0 |       0.0 | 0.196005 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.002269 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | -0.25 |    19.028 |  0.895795 |              8.382718 |                  8.382718 |                   8.382718 |                   0.96939 |                      0.0 |     7.413327 |              0.0 |          7.413327 |             88.271017 |                 100.0 |                98.302128 |               94.620659 |    0.0 |  0.0 | 2.555713 | 0.094 | 1.697264 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.263667 | 0.003052 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.282605 | 79.623186 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.263667 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.003052 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | -0.25 | 15.208683 | -1.913579 |             10.795247 |                 10.795247 |                  10.795247 |                  1.386217 |                      0.0 |      9.40903 |              0.0 |           9.40903 |             94.155751 |                 100.0 |               100.629116 |               95.574305 |    0.0 |  0.0 | 3.286521 |   0.0 | 2.329549 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.140696 | 0.001628 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.946184 | 77.293637 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.140696 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.001628 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | -0.25 | 20.753946 |  1.531325 |             17.664962 |                 17.664962 |                  17.664962 |                  3.108707 |                      0.0 |    14.556255 |              0.0 |         14.556255 |             50.440581 |                 100.0 |               103.850807 |               96.853335 |    0.0 |  0.0 | 5.847436 | 0.094 | 3.314416 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.075077 | 0.000869 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.164858 | 73.979222 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.075077 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000869 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | -0.25 | 14.816194 | -1.909685 |             18.901295 |                 18.901295 |                  18.901295 |                  3.414868 |                      0.0 |    15.486427 |              0.0 |         15.486427 |             47.077876 |                 100.0 |               108.496391 |               98.617204 |    0.0 |  0.0 | 6.638616 |   0.0 | 3.732358 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.040062 | 0.000464 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.824543 | 70.246864 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.040062 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000464 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | -0.25 | 20.348477 |  0.232491 |             18.036211 |                 18.036211 |                  18.036211 |                  2.845741 |                      0.0 |     15.19047 |              0.0 |          15.19047 |             40.352465 |                 100.0 |               113.828249 |              100.519671 |    0.0 |  0.0 | 6.829116 |   0.0 | 3.652462 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.021377 | 0.000247 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.342053 | 66.594402 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.021377 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000247 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | -0.25 | 18.740175 | -1.067813 |             18.651721 |                 18.651721 |                  18.651721 |                  3.051087 |                      0.0 |    15.600634 |              0.0 |         15.600634 |             37.166744 |                 100.0 |               119.168431 |              102.297453 |    0.0 |  0.0 | 7.945284 |   0.0 | 4.128761 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.011407 | 0.000132 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.159866 | 62.465641 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.011407 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000132 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | -0.25 | 21.466219 |  0.293528 |             17.267745 |                 17.267745 |                  17.267745 |                   2.45312 |                      0.0 |    14.814625 |              0.0 |         14.814625 |             42.798069 |                 100.0 |               125.382697 |              104.231161 |    0.0 |  0.0 | 6.428125 |   0.0 | 3.518477 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.006087 |  0.00007 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.616338 | 58.947163 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.006087 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.00007 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | -0.25 | 19.654459 | -1.371448 |             17.877445 |                 17.877445 |                  17.877445 |                  2.688326 |                      0.0 |    15.189119 |              0.0 |         15.189119 |             78.463126 |                 100.0 |               130.861333 |              105.804695 |    0.0 |  0.0 | 6.036736 |   0.0 | 4.091355 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.003248 | 0.000038 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.737786 | 54.855808 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.003248 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000038 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | -0.25 |  23.39262 |  0.203534 |              17.53355 |                  17.53355 |                   17.53355 |                   2.76572 |                      0.0 |     14.76783 |              0.0 |          14.76783 |             74.333488 |                 100.0 |               137.498449 |               107.59179 |    0.0 |  0.0 | 6.494046 |   0.0 | 4.345849 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.001733 |  0.00002 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.284252 | 50.509959 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.001733 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.00002 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | -0.25 | 21.880839 | -0.632013 |             17.704758 |                 17.704758 |                  17.704758 |                  2.799361 |                      0.0 |    14.905396 |              0.0 |         14.905396 |             47.077876 |                 100.0 |               144.953402 |                109.4495 |    0.0 |  0.0 | 7.914762 |   0.0 | 4.548032 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000925 | 0.000011 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.666265 | 45.961927 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000925 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000011 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | -0.25 | 23.154216 | -0.023734 |             17.786952 |                 17.786952 |                  17.786952 |                  2.986484 |                      0.0 |    14.800468 |              0.0 |         14.800468 |             56.493451 |                 100.0 |               153.359832 |              111.375646 |    0.0 |  0.0 | 7.566718 | 0.094 | 4.554485 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000494 | 0.000006 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.439999 | 41.407442 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000494 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000006 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | -0.25 | 22.399998 | -0.414721 |             17.184083 |                 17.184083 |                  17.184083 |                  2.708378 |                      0.0 |    14.475704 |              0.0 |         14.475704 |             88.271017 |                 100.0 |               162.655293 |                113.3188 |    0.0 |  0.0 | 5.996364 |   0.0 | 4.137143 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000263 | 0.000003 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.604721 | 37.270299 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000263 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000003 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | -0.25 | 22.949068 |  0.647517 |             14.968904 |                 14.968904 |                  14.968904 |                  2.239353 |                      0.0 |    12.729551 |              0.0 |         12.729551 |             52.308751 |                 100.0 |               172.213709 |              115.125967 |    0.0 |  0.0 | 5.983925 |   0.0 | 3.428814 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000141 | 0.000002 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.707203 | 33.841485 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000141 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000002 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | -0.25 | 19.957345 | -0.436832 |             14.206821 |                 14.206821 |                  14.206821 |                  2.021579 |                      0.0 |    12.185243 |              0.0 |         12.185243 |             44.135508 |                 100.0 |               181.325786 |              116.675553 |  1.838 |  0.0 | 5.470984 |   0.7 |  2.50575 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  1.838 | 0.049891 |  0.0 |  0.0 | 0.0 |      0.0 | 0.049891 | 0.012919 |  0.00015 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.894036 | 33.123844 | 0.049891 |   0.0 |   0.0 |   0.0 |       0.0 | 0.012919 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.00015 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | -0.25 |  20.58012 |  0.559934 |             14.881951 |                 14.881951 |                  14.881951 |                  2.487815 |                      0.0 |    12.394137 |              0.0 |         12.394137 |            108.641252 |                 100.0 |               183.415178 |              116.867739 |    0.0 |  0.0 |  4.08134 |   0.0 | 2.836595 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.017318 |   0.0002 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.084102 | 30.287249 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.017318 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |   0.0002 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | -0.25 |  17.88034 | -1.617232 |             15.465744 |                 15.465744 |                  15.465744 |                  2.520953 |                      0.0 |     12.94479 |              0.0 |          12.94479 |             74.333488 |                 100.0 |               192.486718 |              118.223505 |    0.0 |  0.0 | 5.175424 |   0.0 | 3.277008 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.038074 |      0.0 | 0.009173 | 0.000106 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.451334 | 27.048315 |      0.0 |   0.0 |   0.0 |   0.0 | -0.038074 | 0.009241 |   0.0 |   0.0 |   0.0 | -0.000068 |   0.0 | 0.000106 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | -0.25 | 22.437779 |  0.434305 |             15.289825 |                 15.289825 |                  15.289825 |                  2.702703 |                      0.0 |    12.587122 |              0.0 |         12.587122 |             70.616813 |                 100.0 |               204.997694 |              119.967773 |    0.0 |  0.0 | 5.738005 |   0.0 | 3.581497 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.239328 |      0.0 | 0.004366 | 0.000051 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.767029 | 23.706145 |      0.0 |   0.0 |   0.0 |   0.0 | -0.239328 | 0.004931 |   0.0 |   0.0 |   0.0 | -0.000565 |   0.0 | 0.000051 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | -0.25 | 20.156763 | -0.488974 |             13.113987 |                 13.113987 |                  13.113987 |                  2.307856 |                      0.0 |    10.806131 |              0.0 |         10.806131 |             54.320626 |                 100.0 |                221.73546 |              122.070817 |    0.0 |  0.0 | 5.867083 |   0.0 | 3.296906 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.446997 |      0.0 | 0.000839 |  0.00001 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.006003 | 20.856236 |      0.0 |   0.0 |   0.0 |   0.0 | -0.446997 | 0.002631 |   0.0 |   0.0 |   0.0 | -0.001793 |   0.0 |  0.00001 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | -0.25 | 20.953343 | -0.749947 |             15.350398 |                 15.350398 |                  15.350398 |                  2.688476 |                      0.0 |    12.661922 |              0.0 |         12.661922 |             39.231563 |                 100.0 |               241.240755 |              124.200232 |    0.0 |  0.0 | 7.829385 |   0.0 | 3.892052 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.624079 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.50595 | 17.588263 |      0.0 |   0.0 |   0.0 |   0.0 | -0.624079 | 0.001404 |   0.0 |   0.0 |   0.0 | -0.003705 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | -0.25 | 22.619833 |  0.062121 |             16.161308 |                 16.161308 |                  16.161308 |                  3.308913 |                      0.0 |    12.852395 |              0.0 |         12.852395 |             56.493451 |                 100.0 |               274.870616 |              127.357125 |    0.0 |  0.0 |  7.32284 |   0.0 |  4.18324 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.827137 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.193828 | 14.232159 |      0.0 |   0.0 |   0.0 |   0.0 | -0.827137 | 0.000749 |   0.0 |   0.0 |   0.0 | -0.006292 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | -0.25 | 21.579428 |  1.026136 |              5.948207 |                  5.948207 |                   5.948207 |                   0.60663 |                      0.0 |     5.341577 |              0.0 |          5.341577 |             67.254108 |                 100.0 |               337.482269 |              131.866514 |   1.18 |  0.0 |  2.56747 |   0.7 | 1.075772 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   1.18 | 0.013304 |  0.0 |  0.0 | 0.0 | 1.035672 | 0.013304 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.917693 | 15.358755 | 0.013304 |   0.0 |   0.0 |   0.0 | -1.035672 | 0.003825 |   0.0 |   0.0 |   0.0 | -0.009606 |   0.0 |      0.0 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | -0.25 | 17.325643 | -1.541907 |             11.290596 |                 11.290596 |                  11.290596 |                  1.641158 |                      0.0 |     9.649438 |              0.0 |          9.649438 |             94.155751 |                 100.0 |               311.359555 |              129.743335 |    0.0 |  0.0 | 3.374286 | 0.282 | 2.014946 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.96567 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   33.2096 | 14.309479 |      0.0 |   0.0 |   0.0 |   0.0 |  -0.96567 | 0.004821 |   0.0 |   0.0 |   0.0 | -0.013157 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | -0.25 | 21.631999 | -1.279412 |             14.219527 |                 14.219527 |                  14.219527 |                  1.868809 |                      0.0 |    12.350718 |              0.0 |         12.350718 |             64.197103 |                 100.0 |               335.458668 |              131.071625 |    0.0 |  0.0 | 5.573178 |   0.0 |  3.36219 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.030867 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 34.239011 | 11.978156 |      0.0 |   0.0 |   0.0 |   0.0 | -1.030867 | 0.002572 |   0.0 |   0.0 |   0.0 | -0.016686 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | -0.25 | 25.063372 |  0.706772 |             10.960427 |                 10.960427 |                  10.960427 |                  1.401603 |                      0.0 |     9.558824 |              0.0 |          9.558824 |             74.333488 |                 100.0 |               419.640495 |              135.153792 |    0.0 |  0.0 | 4.467819 | 0.094 | 2.719425 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.175726 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.28224 | 10.434457 |      0.0 |   0.0 |   0.0 |   0.0 | -1.175726 | 0.001373 |   0.0 |   0.0 |   0.0 | -0.020579 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | -0.25 | 21.874132 |   0.40131 |              9.585999 |                  9.585999 |                   9.585999 |                  1.261689 |                      0.0 |      8.32431 |              0.0 |           8.32431 |             61.405925 |                 100.0 |               522.198911 |              138.404806 |   9.17 |  0.0 | 3.569973 |   0.7 | 1.597438 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   9.17 | 0.104569 |  0.0 |  0.0 | 0.0 | 1.271645 | 0.104569 | 0.002764 | 0.000032 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.63093 | 19.174095 | 0.104569 |   0.0 |   0.0 |   0.0 | -1.271645 | 0.027653 |   0.0 |   0.0 |   0.0 |  -0.02489 |   0.0 | 0.000032 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | -0.25 | 19.703098 | -0.620474 |             10.870274 |                 10.870274 |                  10.870274 |                  1.498964 |                      0.0 |      9.37131 |              0.0 |           9.37131 |             67.254108 |                 100.0 |               256.494997 |              123.965055 |   5.41 |  0.0 | 3.567075 |   0.7 | 1.722338 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   5.41 | 0.090739 |  0.0 |  0.0 | 0.0 |   0.7286 | 0.090739 | 0.031583 | 0.000366 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.001403 | 23.499618 | 0.090739 |   0.0 |   0.0 |   0.0 |   -0.7286 | 0.059966 |   0.0 |   0.0 |   0.0 | -0.028383 |   0.0 | 0.000366 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | -0.25 | 20.938011 |  0.382585 |              5.758296 |                  5.758296 |                   5.758296 |                  0.589077 |                      0.0 |     5.169219 |              0.0 |          5.169219 |             54.320626 |                 100.0 |                222.95161 |              120.113963 |   13.4 |  0.0 | 2.621374 |   0.7 | 1.072031 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   13.4 | 0.312931 |  0.0 |  0.0 | 0.0 |  0.45983 | 0.312931 | 0.101111 |  0.00117 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.368818 | 35.974485 | 0.312931 |   0.0 |   0.0 |   0.0 |  -0.45983 | 0.131521 |   0.0 |   0.0 |   0.0 |  -0.03041 |   0.0 |  0.00117 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | -0.25 | 18.829395 |  1.414918 |               5.53921 |                   5.53921 |                    5.53921 |                  0.708412 |                      0.0 |     4.830799 |              0.0 |          4.830799 |             54.320626 |                 100.0 |                 175.5055 |              113.153351 |  7.948 |  0.0 | 2.027722 |   0.7 | 0.717235 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  7.948 | 0.249156 |  0.0 |  0.0 | 0.0 |      0.0 | 0.249156 | 0.168586 | 0.001951 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   30.7039 | 42.956093 | 0.249156 |   0.0 |   0.0 |   0.0 |       0.0 |  0.19971 |   0.0 |   0.0 |   0.0 | -0.031124 |   0.0 | 0.001951 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | -0.25 | 13.279667 |  -1.92862 |              9.344591 |                  9.344591 |                   9.344591 |                  1.881434 |                      0.0 |     7.463158 |              0.0 |          7.463158 |             67.254108 |                 100.0 |               159.374248 |              110.061771 |    0.0 |  0.0 | 3.393609 |   0.0 | 2.035968 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.127614 | 0.001477 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.382521 | 40.920125 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.158627 |   0.0 |   0.0 |   0.0 | -0.031013 |   0.0 | 0.001477 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | -0.25 | 18.875069 |  0.349956 |              8.321079 |                  8.321079 |                   8.321079 |                  1.511299 |                      0.0 |      6.80978 |              0.0 |           6.80978 |             67.254108 |                 100.0 |               163.717417 |              110.731287 |    0.0 |  0.0 | 3.322101 |   0.0 | 2.018298 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.053743 | 0.000622 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.782565 | 38.901827 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.084645 |   0.0 |   0.0 |   0.0 | -0.030902 |   0.0 | 0.000622 |

.. _lland_v3_acker_routing_daily:

:ref:`acre (routing) <lland_v1_acker_routing>`
______________________________________________

The following calculation shows the possibility to route inflowing runoff,
discussed in the documentation of |lland_v1|:

.. integration-test::

    >>> inlet.sequences.sim.series = 0.02
    >>> test('lland_v3_acker_routing_daily',
    ...     axis1=(inputs.nied, fluxes.qah),
    ...     axis2=states.bowa)
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |   qz |   qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |     g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes | sbes |     evpo |   evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg |   wada |      qdb | qib1 | qib2 | qbb |     qkap |     qdgz |      qah |       qa | inzp | wats | waes | esnow | taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 |      qbgz |    qdga1 | qdga2 | qiga1 | qiga2 |      qbga | inlet |   outlet |
    --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.02 | 1.728 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | -0.25 | 10.933333 |   -2.6784 |             12.492674 |                 12.492674 |                  12.492674 |                  1.877708 |                      0.0 |    10.614966 |              0.0 |         10.614966 |             58.847345 |                 100.0 |               111.309423 |               99.889466 | 11.802 |  0.0 | 3.623966 |   0.7 | 1.720207 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 11.802 | 0.761343 |  0.0 |  0.0 | 0.0 |      0.0 | 0.761343 | 1.924005 | 0.022269 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   32.4284 |  81.32045 | 0.761343 |   0.0 |   0.0 |   0.0 |       0.0 | 0.196005 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.022269 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.02 | 1.728 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | -0.25 |    19.028 |  0.895795 |              8.382718 |                  8.382718 |                   8.382718 |                   0.96939 |                      0.0 |     7.413327 |              0.0 |          7.413327 |             88.271017 |                 100.0 |                98.302128 |               94.620659 |    0.0 |  0.0 | 2.555713 | 0.094 | 1.697264 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.991667 | 0.023052 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.282605 | 79.623186 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.263667 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.023052 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.02 | 1.728 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | -0.25 | 15.208683 | -1.913579 |             10.795247 |                 10.795247 |                  10.795247 |                  1.386217 |                      0.0 |      9.40903 |              0.0 |           9.40903 |             94.155751 |                 100.0 |               100.629116 |               95.574305 |    0.0 |  0.0 | 3.286521 |   0.0 | 2.329549 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.868696 | 0.021628 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.946184 | 77.293637 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.140696 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.021628 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.02 | 1.728 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | -0.25 | 20.753946 |  1.531325 |             17.664962 |                 17.664962 |                  17.664962 |                  3.108707 |                      0.0 |    14.556255 |              0.0 |         14.556255 |             50.440581 |                 100.0 |               103.850807 |               96.853335 |    0.0 |  0.0 | 5.847436 | 0.094 | 3.314416 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.803077 | 0.020869 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.164858 | 73.979222 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.075077 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020869 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.02 | 1.728 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | -0.25 | 14.816194 | -1.909685 |             18.901295 |                 18.901295 |                  18.901295 |                  3.414868 |                      0.0 |    15.486427 |              0.0 |         15.486427 |             47.077876 |                 100.0 |               108.496391 |               98.617204 |    0.0 |  0.0 | 6.638616 |   0.0 | 3.732358 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.768062 | 0.020464 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.824543 | 70.246864 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.040062 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020464 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.02 | 1.728 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | -0.25 | 20.348477 |  0.232491 |             18.036211 |                 18.036211 |                  18.036211 |                  2.845741 |                      0.0 |     15.19047 |              0.0 |          15.19047 |             40.352465 |                 100.0 |               113.828249 |              100.519671 |    0.0 |  0.0 | 6.829116 |   0.0 | 3.652462 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.749377 | 0.020247 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.342053 | 66.594402 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.021377 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020247 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.02 | 1.728 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | -0.25 | 18.740175 | -1.067813 |             18.651721 |                 18.651721 |                  18.651721 |                  3.051087 |                      0.0 |    15.600634 |              0.0 |         15.600634 |             37.166744 |                 100.0 |               119.168431 |              102.297453 |    0.0 |  0.0 | 7.945284 |   0.0 | 4.128761 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.739407 | 0.020132 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.159866 | 62.465641 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.011407 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020132 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.02 | 1.728 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | -0.25 | 21.466219 |  0.293528 |             17.267745 |                 17.267745 |                  17.267745 |                   2.45312 |                      0.0 |    14.814625 |              0.0 |         14.814625 |             42.798069 |                 100.0 |               125.382697 |              104.231161 |    0.0 |  0.0 | 6.428125 |   0.0 | 3.518477 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.734087 |  0.02007 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.616338 | 58.947163 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.006087 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 |  0.02007 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.02 | 1.728 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | -0.25 | 19.654459 | -1.371448 |             17.877445 |                 17.877445 |                  17.877445 |                  2.688326 |                      0.0 |    15.189119 |              0.0 |         15.189119 |             78.463126 |                 100.0 |               130.861333 |              105.804695 |    0.0 |  0.0 | 6.036736 |   0.0 | 4.091355 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.731248 | 0.020038 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.737786 | 54.855808 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.003248 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020038 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.02 | 1.728 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | -0.25 |  23.39262 |  0.203534 |              17.53355 |                  17.53355 |                   17.53355 |                   2.76572 |                      0.0 |     14.76783 |              0.0 |          14.76783 |             74.333488 |                 100.0 |               137.498449 |               107.59179 |    0.0 |  0.0 | 6.494046 |   0.0 | 4.345849 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.729733 |  0.02002 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.284252 | 50.509959 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.001733 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 |  0.02002 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.02 | 1.728 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | -0.25 | 21.880839 | -0.632013 |             17.704758 |                 17.704758 |                  17.704758 |                  2.799361 |                      0.0 |    14.905396 |              0.0 |         14.905396 |             47.077876 |                 100.0 |               144.953402 |                109.4495 |    0.0 |  0.0 | 7.914762 |   0.0 | 4.548032 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.728925 | 0.020011 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.666265 | 45.961927 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000925 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020011 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.02 | 1.728 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | -0.25 | 23.154216 | -0.023734 |             17.786952 |                 17.786952 |                  17.786952 |                  2.986484 |                      0.0 |    14.800468 |              0.0 |         14.800468 |             56.493451 |                 100.0 |               153.359832 |              111.375646 |    0.0 |  0.0 | 7.566718 | 0.094 | 4.554485 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.728494 | 0.020006 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.439999 | 41.407442 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000494 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020006 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.02 | 1.728 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | -0.25 | 22.399998 | -0.414721 |             17.184083 |                 17.184083 |                  17.184083 |                  2.708378 |                      0.0 |    14.475704 |              0.0 |         14.475704 |             88.271017 |                 100.0 |               162.655293 |                113.3188 |    0.0 |  0.0 | 5.996364 |   0.0 | 4.137143 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.728263 | 0.020003 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.604721 | 37.270299 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000263 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020003 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.02 | 1.728 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | -0.25 | 22.949068 |  0.647517 |             14.968904 |                 14.968904 |                  14.968904 |                  2.239353 |                      0.0 |    12.729551 |              0.0 |         12.729551 |             52.308751 |                 100.0 |               172.213709 |              115.125967 |    0.0 |  0.0 | 5.983925 |   0.0 | 3.428814 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.728141 | 0.020002 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.707203 | 33.841485 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000141 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020002 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.02 | 1.728 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | -0.25 | 19.957345 | -0.436832 |             14.206821 |                 14.206821 |                  14.206821 |                  2.021579 |                      0.0 |    12.185243 |              0.0 |         12.185243 |             44.135508 |                 100.0 |               181.325786 |              116.675553 |  1.838 |  0.0 | 5.470984 |   0.7 |  2.50575 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  1.838 | 0.049891 |  0.0 |  0.0 | 0.0 |      0.0 | 0.049891 | 1.740919 |  0.02015 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.894036 | 33.123844 | 0.049891 |   0.0 |   0.0 |   0.0 |       0.0 | 0.012919 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 |  0.02015 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.02 | 1.728 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | -0.25 |  20.58012 |  0.559934 |             14.881951 |                 14.881951 |                  14.881951 |                  2.487815 |                      0.0 |    12.394137 |              0.0 |         12.394137 |            108.641252 |                 100.0 |               183.415178 |              116.867739 |    0.0 |  0.0 |  4.08134 |   0.0 | 2.836595 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.745318 |   0.0202 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.084102 | 30.287249 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.017318 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 |   0.0202 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.02 | 1.728 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | -0.25 |  17.88034 | -1.617232 |             15.465744 |                 15.465744 |                  15.465744 |                  2.520953 |                      0.0 |     12.94479 |              0.0 |          12.94479 |             74.333488 |                 100.0 |               192.486718 |              118.223505 |    0.0 |  0.0 | 5.175424 |   0.0 | 3.277008 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.038074 |      0.0 | 1.737173 | 0.020106 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.451334 | 27.048315 |      0.0 |   0.0 |   0.0 |   0.0 | -0.038074 | 0.009241 |   0.0 |   0.0 |   0.0 | -0.000068 |  0.02 | 0.020106 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.02 | 1.728 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | -0.25 | 22.437779 |  0.434305 |             15.289825 |                 15.289825 |                  15.289825 |                  2.702703 |                      0.0 |    12.587122 |              0.0 |         12.587122 |             70.616813 |                 100.0 |               204.997694 |              119.967773 |    0.0 |  0.0 | 5.738005 |   0.0 | 3.581497 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.239328 |      0.0 | 1.732366 | 0.020051 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.767029 | 23.706145 |      0.0 |   0.0 |   0.0 |   0.0 | -0.239328 | 0.004931 |   0.0 |   0.0 |   0.0 | -0.000565 |  0.02 | 0.020051 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.02 | 1.728 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | -0.25 | 20.156763 | -0.488974 |             13.113987 |                 13.113987 |                  13.113987 |                  2.307856 |                      0.0 |    10.806131 |              0.0 |         10.806131 |             54.320626 |                 100.0 |                221.73546 |              122.070817 |    0.0 |  0.0 | 5.867083 |   0.0 | 3.296906 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.446997 |      0.0 | 1.728839 |  0.02001 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.006003 | 20.856236 |      0.0 |   0.0 |   0.0 |   0.0 | -0.446997 | 0.002631 |   0.0 |   0.0 |   0.0 | -0.001793 |  0.02 |  0.02001 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.02 | 1.728 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | -0.25 | 20.953343 | -0.749947 |             15.350398 |                 15.350398 |                  15.350398 |                  2.688476 |                      0.0 |    12.661922 |              0.0 |         12.661922 |             39.231563 |                 100.0 |               241.240755 |              124.200232 |    0.0 |  0.0 | 7.829385 |   0.0 | 3.892052 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.624079 |      0.0 | 1.725699 | 0.019973 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.50595 | 17.588263 |      0.0 |   0.0 |   0.0 |   0.0 | -0.624079 | 0.001404 |   0.0 |   0.0 |   0.0 | -0.003705 |  0.02 | 0.019973 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.02 | 1.728 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | -0.25 | 22.619833 |  0.062121 |             16.161308 |                 16.161308 |                  16.161308 |                  3.308913 |                      0.0 |    12.852395 |              0.0 |         12.852395 |             56.493451 |                 100.0 |               274.870616 |              127.357125 |    0.0 |  0.0 |  7.32284 |   0.0 |  4.18324 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.827137 |      0.0 | 1.722458 | 0.019936 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.193828 | 14.232159 |      0.0 |   0.0 |   0.0 |   0.0 | -0.827137 | 0.000749 |   0.0 |   0.0 |   0.0 | -0.006292 |  0.02 | 0.019936 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.02 | 1.728 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | -0.25 | 21.579428 |  1.026136 |              5.948207 |                  5.948207 |                   5.948207 |                   0.60663 |                      0.0 |     5.341577 |              0.0 |          5.341577 |             67.254108 |                 100.0 |               337.482269 |              131.866514 |   1.18 |  0.0 |  2.56747 |   0.7 | 1.075772 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   1.18 | 0.013304 |  0.0 |  0.0 | 0.0 | 1.035672 | 0.013304 | 1.722219 | 0.019933 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.917693 | 15.358755 | 0.013304 |   0.0 |   0.0 |   0.0 | -1.035672 | 0.003825 |   0.0 |   0.0 |   0.0 | -0.009606 |  0.02 | 0.019933 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.02 | 1.728 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | -0.25 | 17.325643 | -1.541907 |             11.290596 |                 11.290596 |                  11.290596 |                  1.641158 |                      0.0 |     9.649438 |              0.0 |          9.649438 |             94.155751 |                 100.0 |               311.359555 |              129.743335 |    0.0 |  0.0 | 3.374286 | 0.282 | 2.014946 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.96567 |      0.0 | 1.719664 | 0.019904 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   33.2096 | 14.309479 |      0.0 |   0.0 |   0.0 |   0.0 |  -0.96567 | 0.004821 |   0.0 |   0.0 |   0.0 | -0.013157 |  0.02 | 0.019904 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.02 | 1.728 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | -0.25 | 21.631999 | -1.279412 |             14.219527 |                 14.219527 |                  14.219527 |                  1.868809 |                      0.0 |    12.350718 |              0.0 |         12.350718 |             64.197103 |                 100.0 |               335.458668 |              131.071625 |    0.0 |  0.0 | 5.573178 |   0.0 |  3.36219 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.030867 |      0.0 | 1.713886 | 0.019837 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 34.239011 | 11.978156 |      0.0 |   0.0 |   0.0 |   0.0 | -1.030867 | 0.002572 |   0.0 |   0.0 |   0.0 | -0.016686 |  0.02 | 0.019837 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.02 | 1.728 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | -0.25 | 25.063372 |  0.706772 |             10.960427 |                 10.960427 |                  10.960427 |                  1.401603 |                      0.0 |     9.558824 |              0.0 |          9.558824 |             74.333488 |                 100.0 |               419.640495 |              135.153792 |    0.0 |  0.0 | 4.467819 | 0.094 | 2.719425 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.175726 |      0.0 | 1.708793 | 0.019778 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.28224 | 10.434457 |      0.0 |   0.0 |   0.0 |   0.0 | -1.175726 | 0.001373 |   0.0 |   0.0 |   0.0 | -0.020579 |  0.02 | 0.019778 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.02 | 1.728 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | -0.25 | 21.874132 |   0.40131 |              9.585999 |                  9.585999 |                   9.585999 |                  1.261689 |                      0.0 |      8.32431 |              0.0 |           8.32431 |             61.405925 |                 100.0 |               522.198911 |              138.404806 |   9.17 |  0.0 | 3.569973 |   0.7 | 1.597438 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   9.17 | 0.104569 |  0.0 |  0.0 | 0.0 | 1.271645 | 0.104569 | 1.730764 | 0.020032 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.63093 | 19.174095 | 0.104569 |   0.0 |   0.0 |   0.0 | -1.271645 | 0.027653 |   0.0 |   0.0 |   0.0 |  -0.02489 |  0.02 | 0.020032 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.02 | 1.728 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | -0.25 | 19.703098 | -0.620474 |             10.870274 |                 10.870274 |                  10.870274 |                  1.498964 |                      0.0 |      9.37131 |              0.0 |           9.37131 |             67.254108 |                 100.0 |               256.494997 |              123.965055 |   5.41 |  0.0 | 3.567075 |   0.7 | 1.722338 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   5.41 | 0.090739 |  0.0 |  0.0 | 0.0 |   0.7286 | 0.090739 | 1.759583 | 0.020366 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.001403 | 23.499618 | 0.090739 |   0.0 |   0.0 |   0.0 |   -0.7286 | 0.059966 |   0.0 |   0.0 |   0.0 | -0.028383 |  0.02 | 0.020366 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.02 | 1.728 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | -0.25 | 20.938011 |  0.382585 |              5.758296 |                  5.758296 |                   5.758296 |                  0.589077 |                      0.0 |     5.169219 |              0.0 |          5.169219 |             54.320626 |                 100.0 |                222.95161 |              120.113963 |   13.4 |  0.0 | 2.621374 |   0.7 | 1.072031 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   13.4 | 0.312931 |  0.0 |  0.0 | 0.0 |  0.45983 | 0.312931 | 1.829111 |  0.02117 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.368818 | 35.974485 | 0.312931 |   0.0 |   0.0 |   0.0 |  -0.45983 | 0.131521 |   0.0 |   0.0 |   0.0 |  -0.03041 |  0.02 |  0.02117 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.02 | 1.728 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | -0.25 | 18.829395 |  1.414918 |               5.53921 |                   5.53921 |                    5.53921 |                  0.708412 |                      0.0 |     4.830799 |              0.0 |          4.830799 |             54.320626 |                 100.0 |                 175.5055 |              113.153351 |  7.948 |  0.0 | 2.027722 |   0.7 | 0.717235 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  7.948 | 0.249156 |  0.0 |  0.0 | 0.0 |      0.0 | 0.249156 | 1.896586 | 0.021951 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   30.7039 | 42.956093 | 0.249156 |   0.0 |   0.0 |   0.0 |       0.0 |  0.19971 |   0.0 |   0.0 |   0.0 | -0.031124 |  0.02 | 0.021951 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.02 | 1.728 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | -0.25 | 13.279667 |  -1.92862 |              9.344591 |                  9.344591 |                   9.344591 |                  1.881434 |                      0.0 |     7.463158 |              0.0 |          7.463158 |             67.254108 |                 100.0 |               159.374248 |              110.061771 |    0.0 |  0.0 | 3.393609 |   0.0 | 2.035968 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.855614 | 0.021477 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.382521 | 40.920125 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.158627 |   0.0 |   0.0 |   0.0 | -0.031013 |  0.02 | 0.021477 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.02 | 1.728 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | -0.25 | 18.875069 |  0.349956 |              8.321079 |                  8.321079 |                   8.321079 |                  1.511299 |                      0.0 |      6.80978 |              0.0 |           6.80978 |             67.254108 |                 100.0 |               163.717417 |              110.731287 |    0.0 |  0.0 | 3.322101 |   0.0 | 2.018298 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.781743 | 0.020622 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.782565 | 38.901827 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.084645 |   0.0 |   0.0 |   0.0 | -0.030902 |  0.02 | 0.020622 |

.. _lland_v3_acker_heavy_rain_daily:

acre (heavy rain)
-----------------

Integration test :ref:`lland_v3_acker_summer_daily` deals with dry summer
conditions.  To show how |lland_v3| works for warm but wet conditions,
we set all values of the precipitation input time series to 20 mm.
Now, the soil water content (|BoWa|) rises from its initial value of
72 mm and nearly reaches its maximum value of 309 mm (|WMax|), resulting
in a relevant response of all runoff components:

.. integration-test::

    >>> lnk(ACKER)
    >>> inlet.sequences.sim.series = 0.0
    >>> nied = inputs.nied.series.copy()
    >>> inputs.nied.series = 20.0
    >>> conditions_acker_heavy_rain = test(
    ...     'lland_v3_acker_heavy_rain_daily',
    ...     axis1=(inputs.nied, fluxes.qah),
    ...     axis2=states.bowa,
    ...     get_conditions='1997-08-03')
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration | nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |     g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |     evpo | evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada |       qdb |     qib1 |     qib2 |      qbb | qkap |      qdgz |       qah |       qa | inzp | wats | waes | esnow | taus |      ebdn |       bowa |     qdgz1 |    qdgz2 |    qigz1 |    qigz2 |     qbgz |     qdga1 |    qdga2 |    qiga1 |    qiga2 |     qbga | inlet |   outlet |
    ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 20.0 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 18.8 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | -0.25 | 10.933333 |   -2.6784 |             12.492674 |                 12.492674 |                  12.492674 |                  1.877708 |                      0.0 |    10.614966 |              0.0 |         10.614966 |             58.847345 |                 100.0 |               111.309423 |               99.889466 | 18.1 |  0.0 | 3.623966 | 0.7 | 1.720207 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  1.217805 |      0.0 |      0.0 |      0.0 |  0.0 |  1.217805 |   0.31352 | 0.003629 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   32.4284 |  87.161988 |  1.217805 |      0.0 |      0.0 |      0.0 |      0.0 |   0.31352 |      0.0 |      0.0 |      0.0 |      0.0 |   0.0 | 0.003629 |
    | 1997-02-08 | 20.0 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 | 18.8 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | -0.25 |    19.028 |  0.895795 |              8.382718 |                  8.382718 |                   8.382718 |                   0.96939 |                      0.0 |     7.413327 |              0.0 |          7.413327 |             88.271017 |                 100.0 |                90.407167 |               91.066405 | 18.1 |  0.0 | 2.555713 | 0.7 | 1.294549 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  1.480868 |      0.0 |      0.0 |      0.0 |  0.0 |  1.480868 |  0.802993 | 0.009294 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.282605 |  102.48657 |  1.480868 |      0.0 |      0.0 |      0.0 |      0.0 |  0.802993 |      0.0 |      0.0 |      0.0 |      0.0 |   0.0 | 0.009294 |
    | 1997-03-08 | 20.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 | 18.8 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | -0.25 | 15.208683 | -1.913579 |             10.795247 |                 10.795247 |                  10.795247 |                  1.386217 |                      0.0 |      9.40903 |              0.0 |           9.40903 |             94.155751 |                 100.0 |                70.327061 |                80.36502 | 18.1 |  0.0 | 3.286521 | 0.7 | 1.922457 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  1.761522 |      0.0 |      0.0 |      0.0 |  0.0 |  1.761522 |    1.1914 | 0.013789 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.946184 | 116.902592 |  1.761522 |      0.0 |      0.0 |      0.0 |      0.0 |    1.1914 |      0.0 |      0.0 |      0.0 |      0.0 |   0.0 | 0.013789 |
    | 1997-04-08 | 20.0 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 | 18.8 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | -0.25 | 20.753946 |  1.531325 |             17.664962 |                 17.664962 |                  17.664962 |                  3.108707 |                      0.0 |    14.556255 |              0.0 |         14.556255 |             50.440581 |                 100.0 |                51.993585 |               67.804653 | 18.1 |  0.0 | 5.847436 | 0.7 | 3.397259 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  2.040913 |      0.0 |      0.0 |      0.0 |  0.0 |  2.040913 |  1.529227 | 0.017699 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.164858 |  129.56442 |  2.040913 |      0.0 |      0.0 |      0.0 |      0.0 |  1.529227 |      0.0 |      0.0 |      0.0 |      0.0 |   0.0 | 0.017699 |
    | 1997-05-08 | 20.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 | 18.8 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | -0.25 | 14.816194 | -1.909685 |             18.901295 |                 18.901295 |                  18.901295 |                  3.414868 |                      0.0 |    15.486427 |              0.0 |         15.486427 |             47.077876 |                 100.0 |                 48.85611 |               65.355461 | 18.1 |  0.0 | 6.638616 | 0.7 | 3.917199 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  2.300279 | 0.101438 |      0.0 | 0.010164 |  0.0 |  2.300279 |  1.849046 | 0.021401 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.824543 |  141.33534 |  2.300279 |      0.0 | 0.101438 |      0.0 | 0.010164 |  1.834644 |      0.0 | 0.014384 |      0.0 | 0.000018 |   0.0 | 0.021401 |
    | 1997-06-08 | 20.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 | 18.8 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | -0.25 | 20.348477 |  0.232491 |             18.036211 |                 18.036211 |                  18.036211 |                  2.845741 |                      0.0 |     15.19047 |              0.0 |          15.19047 |             40.352465 |                 100.0 |                 48.85611 |               65.425338 | 18.1 |  0.0 | 6.829116 | 0.7 | 3.913663 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  2.554684 | 0.110653 |      0.0 | 0.021935 |  0.0 |  2.554684 |  2.156535 |  0.02496 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.342053 | 152.834404 |  2.554684 |      0.0 | 0.110653 |      0.0 | 0.021935 |  2.117307 |      0.0 | 0.039153 |      0.0 | 0.000076 |   0.0 |  0.02496 |
    | 1997-07-08 | 20.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 | 18.8 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | -0.25 | 18.740175 | -1.067813 |             18.651721 |                 18.651721 |                  18.651721 |                  3.051087 |                      0.0 |    15.600634 |              0.0 |         15.600634 |             37.166744 |                 100.0 |                 48.85611 |               65.496024 | 18.1 |  0.0 | 7.945284 | 0.7 | 4.551539 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  2.817223 | 0.119656 |      0.0 | 0.033434 |  0.0 |  2.817223 |  2.448759 | 0.028342 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.159866 | 163.412552 |  2.817223 |      0.0 | 0.119656 |      0.0 | 0.033434 |  2.388884 |      0.0 |   0.0597 |      0.0 | 0.000175 |   0.0 | 0.028342 |
    | 1997-08-08 | 20.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 | 18.8 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | -0.25 | 21.466219 |  0.293528 |             17.267745 |                 17.267745 |                  17.267745 |                   2.45312 |                      0.0 |    14.814625 |              0.0 |         14.814625 |             42.798069 |                 100.0 |                 48.85611 |               65.567492 | 18.1 |  0.0 | 6.428125 | 0.7 | 3.767989 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  3.072608 | 0.127938 |      0.0 | 0.044013 |  0.0 |  3.072608 |   2.73175 | 0.031617 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.616338 | 174.500005 |  3.072608 |      0.0 | 0.127938 |      0.0 | 0.044013 |  2.654405 |      0.0 | 0.077033 |      0.0 | 0.000313 |   0.0 | 0.031617 |
    | 1997-09-08 | 20.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 | 18.8 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | -0.25 | 19.654459 | -1.371448 |             17.877445 |                 17.877445 |                  17.877445 |                  2.688326 |                      0.0 |    15.189119 |              0.0 |         15.189119 |             78.463126 |                 100.0 |                 48.85611 |               65.639719 | 18.1 |  0.0 | 6.036736 | 0.7 | 4.121081 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  3.356591 | 0.136618 |      0.0 |   0.0551 |  0.0 |  3.356591 |  3.015033 | 0.034896 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.737786 | 184.930615 |  3.356591 |      0.0 | 0.136618 |      0.0 |   0.0551 |   2.92256 |      0.0 | 0.091983 |      0.0 | 0.000489 |   0.0 | 0.034896 |
    | 1997-10-08 | 20.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 | 18.8 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | -0.25 |  23.39262 |  0.203534 |              17.53355 |                  17.53355 |                   17.53355 |                   2.76572 |                      0.0 |     14.76783 |              0.0 |          14.76783 |             74.333488 |                 100.0 |                 48.85611 |               65.712681 | 18.1 |  0.0 | 6.494046 | 0.7 | 4.450439 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  3.641259 | 0.144785 |      0.0 | 0.065531 |  0.0 |  3.641259 |  3.304149 | 0.038242 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.284252 | 194.728603 |  3.641259 |      0.0 | 0.144785 |      0.0 | 0.065531 |  3.198274 |      0.0 | 0.105171 |      0.0 | 0.000703 |   0.0 | 0.038242 |
    | 1997-11-08 | 20.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 | 18.8 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | -0.25 | 21.880839 | -0.632013 |             17.704758 |                 17.704758 |                  17.704758 |                  2.799361 |                      0.0 |    14.905396 |              0.0 |         14.905396 |             47.077876 |                 100.0 |                 48.85611 |               65.786357 | 18.1 |  0.0 | 7.914762 | 0.7 | 4.993103 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  3.926571 | 0.152455 |      0.0 | 0.075329 |  0.0 |  3.926571 |  3.596218 | 0.041623 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.666265 | 203.681145 |  3.926571 |      0.0 | 0.152455 |      0.0 | 0.075329 |   3.47833 |      0.0 | 0.116935 |      0.0 | 0.000953 |   0.0 | 0.041623 |
    | 1997-12-08 | 20.0 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 | 18.8 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | -0.25 | 23.154216 | -0.023734 |             17.786952 |                 17.786952 |                  17.786952 |                  2.986484 |                      0.0 |    14.800468 |              0.0 |         14.800468 |             56.493451 |                 100.0 |                 48.85611 |               65.860724 | 18.1 |  0.0 | 7.566718 | 0.7 | 4.979876 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  4.204866 | 0.159465 | 0.009557 | 0.084281 |  0.0 |  4.204866 |  3.889124 | 0.045013 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.439999 |   212.3431 |  4.204866 |      0.0 | 0.159465 | 0.009557 | 0.084281 |  3.759031 |      0.0 | 0.127502 | 0.001355 | 0.001236 |   0.0 | 0.045013 |
    | 1997-13-08 | 20.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 | 18.8 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | -0.25 | 22.399998 | -0.414721 |             17.184083 |                 17.184083 |                  17.184083 |                  2.708378 |                      0.0 |    14.475704 |              0.0 |         14.475704 |             88.271017 |                 100.0 |                 48.85611 |               65.935761 | 18.1 |  0.0 | 5.996364 | 0.7 |  4.19851 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  4.492755 | 0.166246 | 0.045992 | 0.092943 |  0.0 |  4.492755 |  4.188438 | 0.048477 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.604721 | 221.446655 |  4.492755 |      0.0 | 0.166246 | 0.045992 | 0.092943 |  4.041079 |      0.0 | 0.137078 | 0.008732 | 0.001549 |   0.0 | 0.048477 |
    | 1997-14-08 | 20.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 | 18.8 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | -0.25 | 22.949068 |  0.647517 |             14.968904 |                 14.968904 |                  14.968904 |                  2.239353 |                      0.0 |    12.729551 |              0.0 |         12.729551 |             52.308751 |                 100.0 |                 48.85611 |               66.011448 | 18.1 |  0.0 | 5.983925 | 0.7 | 3.702096 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  4.818552 | 0.173373 | 0.100351 | 0.102047 |  0.0 |  4.818552 |  4.509936 | 0.052198 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.707203 | 230.650235 |  4.818552 |      0.0 | 0.173373 | 0.100351 | 0.102047 |  4.335611 |      0.0 |  0.14595 | 0.026482 | 0.001893 |   0.0 | 0.052198 |
    | 1997-15-08 | 20.0 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 | 18.8 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | -0.25 | 19.957345 | -0.436832 |             14.206821 |                 14.206821 |                  14.206821 |                  2.021579 |                      0.0 |    12.185243 |              0.0 |         12.185243 |             44.135508 |                 100.0 |                 48.85611 |               66.087764 | 18.1 |  0.0 | 5.470984 | 0.7 | 3.155298 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  5.176929 | 0.180579 |  0.16802 |  0.11125 |  0.0 |  5.176929 |  4.865729 | 0.056316 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.894036 | 239.958158 |  5.176929 |      0.0 | 0.180579 |  0.16802 |  0.11125 |  4.653113 |      0.0 | 0.154363 | 0.055986 | 0.002268 |   0.0 | 0.056316 |
    | 1997-16-08 | 20.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 | 18.8 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | -0.25 |  20.58012 |  0.559934 |             14.881951 |                 14.881951 |                  14.881951 |                  2.487815 |                      0.0 |    12.394137 |              0.0 |         12.394137 |            108.641252 |                 100.0 |                 48.85611 |                66.16469 | 18.1 |  0.0 |  4.08134 | 0.7 | 2.708461 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  5.575829 | 0.187866 | 0.247347 | 0.120558 |  0.0 |  5.575829 |  5.262676 | 0.060911 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.084102 | 249.218097 |  5.575829 |      0.0 | 0.187866 | 0.247347 | 0.120558 |   5.00011 |      0.0 | 0.162462 | 0.097429 | 0.002675 |   0.0 | 0.060911 |
    | 1997-17-08 | 20.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 | 18.8 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | -0.25 |  17.88034 | -1.617232 |             15.465744 |                 15.465744 |                  15.465744 |                  2.520953 |                      0.0 |     12.94479 |              0.0 |          12.94479 |             74.333488 |                 100.0 |                 48.85611 |               66.242207 | 18.1 |  0.0 | 5.175424 | 0.7 | 3.378706 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  6.018699 | 0.195116 | 0.335804 | 0.129818 |  0.0 |  6.018699 |  5.706462 | 0.066047 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.451334 | 257.259955 |  6.018699 |      0.0 | 0.195116 | 0.335804 | 0.129818 |  5.382634 |      0.0 | 0.170336 | 0.150377 | 0.003114 |   0.0 | 0.066047 |
    | 1997-18-08 | 20.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 | 18.8 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | -0.25 | 22.437779 |  0.434305 |             15.289825 |                 15.289825 |                  15.289825 |                  2.702703 |                      0.0 |    12.587122 |              0.0 |         12.587122 |             70.616813 |                 100.0 |                 48.85611 |               66.320298 | 18.1 |  0.0 | 5.738005 | 0.7 | 3.779831 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  6.451503 | 0.201412 | 0.419614 |  0.13786 |  0.0 |  6.451503 |  6.184438 | 0.071579 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.767029 | 264.369735 |  6.451503 |      0.0 | 0.201412 | 0.419614 |  0.13786 |  5.790712 |      0.0 | 0.177908 | 0.212236 | 0.003582 |   0.0 | 0.071579 |
    | 1997-19-08 | 20.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 | 18.8 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | -0.25 | 20.156763 | -0.488974 |             13.113987 |                 13.113987 |                  13.113987 |                  2.307856 |                      0.0 |    10.806131 |              0.0 |         10.806131 |             54.320626 |                 100.0 |                 48.85611 |               66.398945 | 18.1 |  0.0 | 5.867083 | 0.7 |  3.62847 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  6.883173 | 0.206978 | 0.498724 |  0.14497 |  0.0 |  6.883173 |  6.678482 | 0.077297 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.006003 | 271.107419 |  6.883173 |      0.0 | 0.206978 | 0.498724 |  0.14497 |  6.210029 |      0.0 | 0.185032 | 0.279345 | 0.004076 |   0.0 | 0.077297 |
    | 1997-20-08 | 20.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 | 18.8 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | -0.25 | 20.953343 | -0.749947 |             15.350398 |                 15.350398 |                  15.350398 |                  2.688476 |                      0.0 |    12.661922 |              0.0 |         12.661922 |             39.231563 |                 100.0 |                 48.85611 |                66.47813 | 18.1 |  0.0 | 7.829385 | 0.7 | 4.625035 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  7.348865 | 0.212253 | 0.577784 | 0.151707 |  0.0 |  7.348865 |  7.189836 | 0.083216 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.50595 | 276.291774 |  7.348865 |      0.0 | 0.212253 | 0.577784 | 0.151707 |  6.643867 |      0.0 | 0.191695 | 0.349681 | 0.004593 |   0.0 | 0.083216 |
    | 1997-21-08 | 20.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 | 18.8 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | -0.25 | 22.619833 |  0.062121 |             16.161308 |                 16.161308 |                  16.161308 |                  3.308913 |                      0.0 |    12.852395 |              0.0 |         12.852395 |             56.493451 |                 100.0 |                 48.85611 |               66.557838 | 18.1 |  0.0 |  7.32284 | 0.7 | 4.757012 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  7.757458 | 0.216312 | 0.641202 | 0.156892 |  0.0 |  7.757458 |  7.700952 | 0.089131 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.193828 | 280.862897 |  7.757458 |      0.0 | 0.216312 | 0.641202 | 0.156892 |  7.077861 |      0.0 | 0.197811 |  0.42015 | 0.005129 |   0.0 | 0.089131 |
    | 1997-22-08 | 20.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 | 18.8 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | -0.25 | 21.579428 |  1.026136 |              5.948207 |                  5.948207 |                   5.948207 |                   0.60663 |                      0.0 |     5.341577 |              0.0 |          5.341577 |             67.254108 |                 100.0 |                 48.85611 |               66.638052 | 18.1 |  0.0 |  2.56747 | 0.7 | 1.361228 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  8.166601 | 0.219891 | 0.698918 | 0.161463 |  0.0 |  8.166601 |  8.197047 | 0.094873 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.917693 | 288.354796 |  8.166601 |      0.0 | 0.219891 | 0.698918 | 0.161463 |   7.50015 |      0.0 | 0.203305 | 0.487911 | 0.005681 |   0.0 | 0.094873 |
    | 1997-23-08 | 20.0 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 | 18.8 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | -0.25 | 17.325643 | -1.541907 |             11.290596 |                 11.290596 |                  11.290596 |                  1.641158 |                      0.0 |     9.649438 |              0.0 |          9.649438 |             94.155751 |                 100.0 |                 48.85611 |               66.718758 | 18.1 |  0.0 | 3.374286 | 0.7 | 2.097563 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  8.984192 | 0.225757 | 0.797025 | 0.168955 |  0.0 |  8.984192 |  8.795013 | 0.101794 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   33.2096 | 294.181304 |  8.984192 |      0.0 | 0.225757 | 0.797025 | 0.168955 |  8.021462 |      0.0 | 0.208607 | 0.558691 | 0.006253 |   0.0 | 0.101794 |
    | 1997-24-08 | 20.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 | 18.8 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | -0.25 | 21.631999 | -1.279412 |             14.219527 |                 14.219527 |                  14.219527 |                  1.868809 |                      0.0 |    12.350718 |              0.0 |         12.350718 |             64.197103 |                 100.0 |                 48.85611 |                66.79994 | 18.1 |  0.0 | 5.573178 | 0.7 | 3.649927 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  9.834859 | 0.230318 | 0.876239 | 0.174781 |  0.0 |  9.834859 |   9.54435 | 0.110467 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 34.239011 | 297.515179 |  9.834859 |      0.0 | 0.230318 | 0.876239 | 0.174781 |   8.68947 |      0.0 | 0.213876 | 0.634158 | 0.006846 |   0.0 | 0.110467 |
    | 1997-25-08 | 20.0 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 | 18.8 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | -0.25 | 25.063372 |  0.706772 |             10.960427 |                 10.960427 |                  10.960427 |                  1.401603 |                      0.0 |     9.558824 |              0.0 |          9.558824 |             74.333488 |                 100.0 |                 48.85611 |               66.881585 | 18.1 |  0.0 | 4.467819 | 0.7 | 2.895981 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 10.482817 | 0.232928 | 0.922678 | 0.178115 |  0.0 | 10.482817 | 10.322599 | 0.119475 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.28224 |  300.90266 | 10.482817 |      0.0 | 0.232928 | 0.922678 | 0.178115 |  9.390481 |      0.0 | 0.218677 | 0.705987 | 0.007454 |   0.0 | 0.119475 |
    | 1997-26-08 | 20.0 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 | 18.8 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | -0.25 | 21.874132 |   0.40131 |              9.585999 |                  9.585999 |                   9.585999 |                  1.261689 |                      0.0 |      8.32431 |              0.0 |           8.32431 |             61.405925 |                 100.0 |                 48.85611 |               66.963678 | 18.1 |  0.0 | 3.569973 | 0.7 | 2.071555 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  11.39404 |  0.23558 | 0.970674 | 0.181503 |  0.0 |  11.39404 | 11.136685 | 0.128897 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.63093 | 304.149308 |  11.39404 |      0.0 |  0.23558 | 0.970674 | 0.181503 | 10.134526 |      0.0 | 0.222894 | 0.771194 | 0.008071 |   0.0 | 0.128897 |
    | 1997-27-08 | 20.0 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 | 18.8 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | -0.25 | 19.703098 | -0.620474 |             10.870274 |                 10.870274 |                  10.870274 |                  1.498964 |                      0.0 |      9.37131 |              0.0 |           9.37131 |             67.254108 |                 100.0 |                 48.85611 |               67.046207 | 18.1 |  0.0 | 3.567075 | 0.7 | 2.108966 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 13.249308 | 0.238122 | 1.017431 | 0.184749 |  0.0 | 13.249308 | 12.737269 | 0.147422 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.001403 | 305.450732 |      12.0 | 1.249308 | 0.238122 | 1.017431 | 0.184749 | 10.877952 | 0.792359 | 0.226674 | 0.831586 | 0.008698 |   0.0 | 0.147422 |
    | 1997-28-08 | 20.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 | 18.8 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | -0.25 | 20.938011 |  0.382585 |              5.758296 |                  5.758296 |                   5.758296 |                  0.589077 |                      0.0 |     5.169219 |              0.0 |          5.169219 |             54.320626 |                 100.0 |                 48.85611 |               67.129158 | 18.1 |  0.0 | 2.621374 | 0.7 | 1.331712 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 14.550732 | 0.239141 | 1.036378 | 0.186051 |  0.0 | 14.550732 | 14.562531 | 0.168548 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.368818 | 306.206718 |      12.0 | 2.550732 | 0.239141 | 1.036378 | 0.186051 | 11.401263 | 2.037672 | 0.229904 | 0.884361 | 0.009331 |   0.0 | 0.168548 |
    | 1997-29-08 | 20.0 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 | 18.8 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | -0.25 | 18.829395 |  1.414918 |               5.53921 |                   5.53921 |                    5.53921 |                  0.708412 |                      0.0 |     4.830799 |              0.0 |          4.830799 |             54.320626 |                 100.0 |                 48.85611 |               67.212519 | 18.1 |  0.0 | 2.027722 | 0.7 | 0.881862 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 15.306718 | 0.239733 | 1.047438 | 0.186807 |  0.0 | 15.306718 |  15.83846 | 0.183316 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   30.7039 |  306.64416 |      12.0 | 3.306718 | 0.239733 | 1.047438 | 0.186807 | 11.680507 |  2.98861 | 0.232477 |   0.9269 | 0.009966 |   0.0 | 0.183316 |
    | 1997-30-08 | 20.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 | 18.8 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | -0.25 | 13.279667 |  -1.92862 |              9.344591 |                  9.344591 |                   9.344591 |                  1.881434 |                      0.0 |     7.463158 |              0.0 |          7.463158 |             67.254108 |                 100.0 |                 48.85611 |               67.296277 | 18.1 |  0.0 | 3.393609 | 0.7 | 1.913447 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  15.74416 | 0.240076 | 1.053855 | 0.187244 |  0.0 |  15.74416 | 16.593262 | 0.192052 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.382521 | 305.605378 |      12.0 |  3.74416 | 0.240076 | 1.053855 | 0.187244 | 11.829515 | 3.558369 | 0.234481 | 0.960296 |   0.0106 |   0.0 | 0.192052 |
    | 1997-31-08 | 20.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 | 18.8 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | -0.25 | 18.875069 |  0.349956 |              8.321079 |                  8.321079 |                   8.321079 |                  1.511299 |                      0.0 |      6.80978 |              0.0 |           6.80978 |             67.254108 |                 100.0 |                 48.85611 |               67.380423 | 18.1 |  0.0 | 3.322101 | 0.7 | 1.882222 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 14.705378 | 0.239262 | 1.038637 | 0.186205 |  0.0 | 14.705378 | 16.209748 | 0.187613 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.782565 | 305.653673 |      12.0 | 2.705378 | 0.239262 | 1.038637 | 0.186205 | 11.909027 | 3.070262 | 0.235874 | 0.983354 | 0.011231 |   0.0 | 0.187613 |

>>> inputs.nied.series = nied

.. _lland_v3_wasser_daily:

water
-----

L-Land defines three types of water areas, |WASSER|, |FLUSS| and |SEE|.
|lland_v3| calculates their evaporation via the Penman equation based on
daily input data.  Besides that, there are no structural differences to
|lland_v1|,  so please read the detailed discussions in integration test
:ref:`lland_v1_wasser` for better understanding the following test
results for water type |WASSER|:

.. integration-test::

    >>> lnk(WASSER)
    >>> surfaceresistance.wasser_aug = 0.0
    >>> cropheight.wasser_aug = 0.05
    >>> albedo.wasser_aug = 0.7
    >>> conditions_wasser = test(
    ...     'lland_v3_wasser_daily',
    ...     axis1=(fluxes.nkor, fluxes.evi, fluxes.qah),
    ...     get_conditions='1997-08-03')
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |       qah |       qa | inzp | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz | qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | 0.0 | nan | 0.0 |              4.931319 |                  4.931319 |                   4.931319 |                  1.877708 |                      0.0 |      3.05361 |              0.0 |           3.05361 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077197 | 1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 11.424803 | 0.132232 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.132232 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | 0.0 | nan | 0.0 |              3.308967 |                  3.308967 |                   3.308967 |                   0.96939 |                      0.0 |     2.339577 |              0.0 |          2.339577 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.86545 |    0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | 0.0 | nan | 0.0 |              4.261282 |                  4.261282 |                   4.261282 |                  1.386217 |                      0.0 |     2.875064 |              0.0 |          2.875064 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.092365 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | 0.0 | nan | 0.0 |              6.973011 |                  6.973011 |                   6.973011 |                  3.108707 |                      0.0 |     3.864304 |              0.0 |          3.864304 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.565894 |    0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | 0.0 | nan | 0.0 |              7.461037 |                  7.461037 |                   7.461037 |                  3.414868 |                      0.0 |     4.046169 |              0.0 |          4.046169 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.729733 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | 0.0 | nan | 0.0 |              7.119557 |                  7.119557 |                   7.119557 |                  2.845741 |                      0.0 |     4.273816 |              0.0 |          4.273816 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826087 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | 0.0 | nan | 0.0 |              7.362521 |                  7.362521 |                   7.362521 |                  3.051087 |                      0.0 |     4.311434 |              0.0 |          4.311434 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.041956 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | 0.0 | nan | 0.0 |              6.816215 |                  6.816215 |                   6.816215 |                   2.45312 |                      0.0 |     4.363095 |              0.0 |          4.363095 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.800053 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | 0.0 | nan | 0.0 |              7.056886 |                  7.056886 |                   7.056886 |                  2.688326 |                      0.0 |      4.36856 |              0.0 |           4.36856 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.82186 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | 0.0 | nan | 0.0 |              6.921138 |                  6.921138 |                   6.921138 |                   2.76572 |                      0.0 |     4.155419 |              0.0 |          4.155419 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.891469 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | 0.0 | nan | 0.0 |               6.98872 |                   6.98872 |                    6.98872 |                  2.799361 |                      0.0 |     4.189359 |              0.0 |          4.189359 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.106575 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | 0.0 | nan | 0.0 |              7.021165 |                  7.021165 |                   7.021165 |                  2.986484 |                      0.0 |     4.034682 |              0.0 |          4.034682 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.036929 |    0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | 0.0 | nan | 0.0 |               6.78319 |                   6.78319 |                    6.78319 |                  2.708378 |                      0.0 |     4.074812 |              0.0 |          4.074812 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.808559 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | 0.0 | nan | 0.0 |              5.908778 |                  5.908778 |                   5.908778 |                  2.239353 |                      0.0 |     3.669425 |              0.0 |          3.669425 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672736 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | 0.0 | nan | 0.0 |              5.607956 |                  5.607956 |                   5.607956 |                  2.021579 |                      0.0 |     3.586377 |              0.0 |          3.586377 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.529542 | 1.529542 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  1.008458 | 0.011672 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011672 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | 0.0 | nan | 0.0 |              5.874454 |                  5.874454 |                   5.874454 |                  2.487815 |                      0.0 |      3.38664 |              0.0 |           3.38664 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.283901 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | 0.0 | nan | 0.0 |              6.104899 |                  6.104899 |                   6.104899 |                  2.520953 |                      0.0 |     3.583945 |              0.0 |          3.583945 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.510743 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | 0.0 | nan | 0.0 |              6.035457 |                  6.035457 |                   6.035457 |                  2.702703 |                      0.0 |     3.332754 |              0.0 |          3.332754 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.594572 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | 0.0 | nan | 0.0 |              5.176574 |                  5.176574 |                   5.176574 |                  2.307856 |                      0.0 |     2.868718 |              0.0 |          2.868718 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.541567 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | 0.0 | nan | 0.0 |              6.059368 |                  6.059368 |                   6.059368 |                  2.688476 |                      0.0 |     3.370892 |              0.0 |          3.370892 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.939383 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | 0.0 | nan | 0.0 |              6.379464 |                  6.379464 |                   6.379464 |                  3.308913 |                      0.0 |     3.070551 |              0.0 |          3.070551 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826811 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | 0.0 | nan | 0.0 |              2.347976 |                  2.347976 |                   2.347976 |                   0.60663 |                      0.0 |     1.741346 |              0.0 |          1.741346 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.806188 | 0.806188 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  1.073812 | 0.012428 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012428 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | 0.0 | nan | 0.0 |              4.456814 |                  4.456814 |                   4.456814 |                  1.641158 |                      0.0 |     2.815656 |              0.0 |          2.815656 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.08421 |    0.282 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | 0.0 | nan | 0.0 |              5.612971 |                  5.612971 |                   5.612971 |                  1.868809 |                      0.0 |     3.744162 |              0.0 |          3.744162 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672232 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | 0.0 | nan | 0.0 |              4.326484 |                  4.326484 |                   4.326484 |                  1.401603 |                      0.0 |     2.924881 |              0.0 |          2.924881 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.374102 |    0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | 0.0 | nan | 0.0 |              3.783947 |                  3.783947 |                   3.783947 |                  1.261689 |                      0.0 |     2.522258 |              0.0 |          2.522258 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077096 | 1.077096 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  8.792904 |  0.10177 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.10177 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | 0.0 | nan | 0.0 |              4.290898 |                  4.290898 |                   4.290898 |                  1.498964 |                      0.0 |     2.791934 |              0.0 |          2.791934 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.09756 |  1.09756 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   5.01244 | 0.058014 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.058014 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | 0.0 | nan | 0.0 |              2.273012 |                  2.273012 |                   2.273012 |                  0.589077 |                      0.0 |     1.683934 |              0.0 |          1.683934 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.794859 | 0.794859 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 13.305141 | 0.153995 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.153995 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | 0.0 | nan | 0.0 |               2.18653 |                   2.18653 |                    2.18653 |                  0.708412 |                      0.0 |     1.478119 |              0.0 |          1.478119 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.607776 | 0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  8.040224 | 0.093058 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.093058 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | 0.0 | nan | 0.0 |              3.688655 |                  3.688655 |                   3.688655 |                  1.881434 |                      0.0 |     1.807221 |              0.0 |          1.807221 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.90065 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | 0.0 | nan | 0.0 |              3.284636 |                  3.284636 |                   3.284636 |                  1.511299 |                      0.0 |     1.773338 |              0.0 |          1.773338 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.912078 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |

.. _lland_v3_wasser_routing_daily:

water (routing)
---------------

The following calculation shows the possibility to subtract evaporation
from inflowing runoff, discussed in the integration test
:ref:`lland_v1_wasser_routing` of |lland_v1|:

.. integration-test::

    >>> inlet.sequences.sim.series = 0.02
    >>> test('lland_v3_wasser_routing_daily',
    ...     axis1=(fluxes.nkor, fluxes.evi, fluxes.qah))
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |   qz |   qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |       qah |       qa | inzp | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz | qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.02 | 1.728 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | 0.0 | nan | 0.0 |              4.931319 |                  4.931319 |                   4.931319 |                  1.877708 |                      0.0 |      3.05361 |              0.0 |           3.05361 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077197 | 1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 13.152803 | 0.152232 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.152232 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.02 | 1.728 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | 0.0 | nan | 0.0 |              3.308967 |                  3.308967 |                   3.308967 |                   0.96939 |                      0.0 |     2.339577 |              0.0 |          2.339577 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.86545 |  0.86545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   0.95655 | 0.011071 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.011071 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.02 | 1.728 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | 0.0 | nan | 0.0 |              4.261282 |                  4.261282 |                   4.261282 |                  1.386217 |                      0.0 |     2.875064 |              0.0 |          2.875064 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.092365 | 1.092365 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.635635 | 0.007357 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.007357 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.02 | 1.728 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | 0.0 | nan | 0.0 |              6.973011 |                  6.973011 |                   6.973011 |                  3.108707 |                      0.0 |     3.864304 |              0.0 |          3.864304 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.565894 | 1.565894 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.256106 | 0.002964 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.002964 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.02 | 1.728 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | 0.0 | nan | 0.0 |              7.461037 |                  7.461037 |                   7.461037 |                  3.414868 |                      0.0 |     4.046169 |              0.0 |          4.046169 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.729733 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.02 | 1.728 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | 0.0 | nan | 0.0 |              7.119557 |                  7.119557 |                   7.119557 |                  2.845741 |                      0.0 |     4.273816 |              0.0 |          4.273816 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826087 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.02 | 1.728 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | 0.0 | nan | 0.0 |              7.362521 |                  7.362521 |                   7.362521 |                  3.051087 |                      0.0 |     4.311434 |              0.0 |          4.311434 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.041956 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.02 | 1.728 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | 0.0 | nan | 0.0 |              6.816215 |                  6.816215 |                   6.816215 |                   2.45312 |                      0.0 |     4.363095 |              0.0 |          4.363095 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.800053 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.02 | 1.728 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | 0.0 | nan | 0.0 |              7.056886 |                  7.056886 |                   7.056886 |                  2.688326 |                      0.0 |      4.36856 |              0.0 |           4.36856 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.82186 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.02 | 1.728 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | 0.0 | nan | 0.0 |              6.921138 |                  6.921138 |                   6.921138 |                   2.76572 |                      0.0 |     4.155419 |              0.0 |          4.155419 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.891469 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.02 | 1.728 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | 0.0 | nan | 0.0 |               6.98872 |                   6.98872 |                    6.98872 |                  2.799361 |                      0.0 |     4.189359 |              0.0 |          4.189359 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.106575 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.02 | 1.728 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | 0.0 | nan | 0.0 |              7.021165 |                  7.021165 |                   7.021165 |                  2.986484 |                      0.0 |     4.034682 |              0.0 |          4.034682 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.036929 |    1.822 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.02 | 1.728 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | 0.0 | nan | 0.0 |               6.78319 |                   6.78319 |                    6.78319 |                  2.708378 |                      0.0 |     4.074812 |              0.0 |          4.074812 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.808559 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.02 | 1.728 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | 0.0 | nan | 0.0 |              5.908778 |                  5.908778 |                   5.908778 |                  2.239353 |                      0.0 |     3.669425 |              0.0 |          3.669425 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672736 | 1.672736 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.055264 |  0.00064 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |  0.00064 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.02 | 1.728 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | 0.0 | nan | 0.0 |              5.607956 |                  5.607956 |                   5.607956 |                  2.021579 |                      0.0 |     3.586377 |              0.0 |          3.586377 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.529542 | 1.529542 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  2.736458 | 0.031672 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.031672 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.02 | 1.728 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | 0.0 | nan | 0.0 |              5.874454 |                  5.874454 |                   5.874454 |                  2.487815 |                      0.0 |      3.38664 |              0.0 |           3.38664 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.283901 | 1.283901 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.444099 |  0.00514 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |  0.00514 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.02 | 1.728 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | 0.0 | nan | 0.0 |              6.104899 |                  6.104899 |                   6.104899 |                  2.520953 |                      0.0 |     3.583945 |              0.0 |          3.583945 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.510743 | 1.510743 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.217257 | 0.002515 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.002515 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.02 | 1.728 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | 0.0 | nan | 0.0 |              6.035457 |                  6.035457 |                   6.035457 |                  2.702703 |                      0.0 |     3.332754 |              0.0 |          3.332754 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.594572 | 1.594572 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.133428 | 0.001544 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.001544 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.02 | 1.728 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | 0.0 | nan | 0.0 |              5.176574 |                  5.176574 |                   5.176574 |                  2.307856 |                      0.0 |     2.868718 |              0.0 |          2.868718 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.541567 | 1.541567 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.186433 | 0.002158 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.002158 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.02 | 1.728 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | 0.0 | nan | 0.0 |              6.059368 |                  6.059368 |                   6.059368 |                  2.688476 |                      0.0 |     3.370892 |              0.0 |          3.370892 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.939383 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.02 | 1.728 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | 0.0 | nan | 0.0 |              6.379464 |                  6.379464 |                   6.379464 |                  3.308913 |                      0.0 |     3.070551 |              0.0 |          3.070551 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826811 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.02 | 1.728 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | 0.0 | nan | 0.0 |              2.347976 |                  2.347976 |                   2.347976 |                   0.60663 |                      0.0 |     1.741346 |              0.0 |          1.741346 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.806188 | 0.806188 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  2.801812 | 0.032428 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.032428 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.02 | 1.728 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | 0.0 | nan | 0.0 |              4.456814 |                  4.456814 |                   4.456814 |                  1.641158 |                      0.0 |     2.815656 |              0.0 |          2.815656 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.08421 |  1.08421 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   0.92579 | 0.010715 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.010715 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.02 | 1.728 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | 0.0 | nan | 0.0 |              5.612971 |                  5.612971 |                   5.612971 |                  1.868809 |                      0.0 |     3.744162 |              0.0 |          3.744162 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672232 | 1.672232 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.055768 | 0.000645 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.000645 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.02 | 1.728 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | 0.0 | nan | 0.0 |              4.326484 |                  4.326484 |                   4.326484 |                  1.401603 |                      0.0 |     2.924881 |              0.0 |          2.924881 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.374102 | 1.374102 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.447898 | 0.005184 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.005184 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.02 | 1.728 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | 0.0 | nan | 0.0 |              3.783947 |                  3.783947 |                   3.783947 |                  1.261689 |                      0.0 |     2.522258 |              0.0 |          2.522258 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077096 | 1.077096 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 10.520904 |  0.12177 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |  0.12177 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.02 | 1.728 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | 0.0 | nan | 0.0 |              4.290898 |                  4.290898 |                   4.290898 |                  1.498964 |                      0.0 |     2.791934 |              0.0 |          2.791934 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.09756 |  1.09756 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   6.74044 | 0.078014 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.078014 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.02 | 1.728 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | 0.0 | nan | 0.0 |              2.273012 |                  2.273012 |                   2.273012 |                  0.589077 |                      0.0 |     1.683934 |              0.0 |          1.683934 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.794859 | 0.794859 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 15.033141 | 0.173995 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.173995 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.02 | 1.728 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | 0.0 | nan | 0.0 |               2.18653 |                   2.18653 |                    2.18653 |                  0.708412 |                      0.0 |     1.478119 |              0.0 |          1.478119 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.607776 | 0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  9.768224 | 0.113058 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.113058 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.02 | 1.728 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | 0.0 | nan | 0.0 |              3.688655 |                  3.688655 |                   3.688655 |                  1.881434 |                      0.0 |     1.807221 |              0.0 |          1.807221 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.90065 |  0.90065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   0.82735 | 0.009576 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.009576 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.02 | 1.728 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | 0.0 | nan | 0.0 |              3.284636 |                  3.284636 |                   3.284636 |                  1.511299 |                      0.0 |     1.773338 |              0.0 |          1.773338 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.912078 | 0.912078 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.815922 | 0.009444 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.009444 |

.. _lland_v3_wasser_negq_daily:

water (negative runoff)
-----------------------

The following calculation shows the possibility to calculate negative
discharge values, discussed in the integration test :ref:`lland_v1_wasser_negq`
of |lland_v1|:

.. integration-test::

    >>> negq(True)
    >>> inlet.sequences.sim.series = 0.0
    >>> test('lland_v3_wasser_negq_daily',
    ...      axis1=(fluxes.nkor, fluxes.evi, fluxes.qah))
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |       qah |        qa | inzp | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz | qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |    outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | 0.0 | nan | 0.0 |              4.931319 |                  4.931319 |                   4.931319 |                  1.877708 |                      0.0 |      3.05361 |              0.0 |           3.05361 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077197 | 1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 11.424803 |  0.132232 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.132232 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | 0.0 | nan | 0.0 |              3.308967 |                  3.308967 |                   3.308967 |                   0.96939 |                      0.0 |     2.339577 |              0.0 |          2.339577 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.86545 |  0.86545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  -0.77145 | -0.008929 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.008929 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | 0.0 | nan | 0.0 |              4.261282 |                  4.261282 |                   4.261282 |                  1.386217 |                      0.0 |     2.875064 |              0.0 |          2.875064 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.092365 | 1.092365 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.092365 | -0.012643 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.012643 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | 0.0 | nan | 0.0 |              6.973011 |                  6.973011 |                   6.973011 |                  3.108707 |                      0.0 |     3.864304 |              0.0 |          3.864304 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.565894 | 1.565894 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.471894 | -0.017036 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.017036 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | 0.0 | nan | 0.0 |              7.461037 |                  7.461037 |                   7.461037 |                  3.414868 |                      0.0 |     4.046169 |              0.0 |          4.046169 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.729733 | 1.729733 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.729733 |  -0.02002 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  -0.02002 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | 0.0 | nan | 0.0 |              7.119557 |                  7.119557 |                   7.119557 |                  2.845741 |                      0.0 |     4.273816 |              0.0 |          4.273816 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826087 | 1.826087 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.826087 | -0.021135 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.021135 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | 0.0 | nan | 0.0 |              7.362521 |                  7.362521 |                   7.362521 |                  3.051087 |                      0.0 |     4.311434 |              0.0 |          4.311434 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.041956 | 2.041956 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -2.041956 | -0.023634 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.023634 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | 0.0 | nan | 0.0 |              6.816215 |                  6.816215 |                   6.816215 |                   2.45312 |                      0.0 |     4.363095 |              0.0 |          4.363095 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.800053 | 1.800053 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.800053 | -0.020834 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.020834 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | 0.0 | nan | 0.0 |              7.056886 |                  7.056886 |                   7.056886 |                  2.688326 |                      0.0 |      4.36856 |              0.0 |           4.36856 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.82186 |  1.82186 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  -1.82186 | -0.021086 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.021086 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | 0.0 | nan | 0.0 |              6.921138 |                  6.921138 |                   6.921138 |                   2.76572 |                      0.0 |     4.155419 |              0.0 |          4.155419 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.891469 | 1.891469 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.891469 | -0.021892 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.021892 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | 0.0 | nan | 0.0 |               6.98872 |                   6.98872 |                    6.98872 |                  2.799361 |                      0.0 |     4.189359 |              0.0 |          4.189359 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.106575 | 2.106575 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -2.106575 | -0.024382 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.024382 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | 0.0 | nan | 0.0 |              7.021165 |                  7.021165 |                   7.021165 |                  2.986484 |                      0.0 |     4.034682 |              0.0 |          4.034682 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.036929 | 2.036929 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.942929 | -0.022488 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.022488 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | 0.0 | nan | 0.0 |               6.78319 |                   6.78319 |                    6.78319 |                  2.708378 |                      0.0 |     4.074812 |              0.0 |          4.074812 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.808559 | 1.808559 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.808559 | -0.020932 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.020932 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | 0.0 | nan | 0.0 |              5.908778 |                  5.908778 |                   5.908778 |                  2.239353 |                      0.0 |     3.669425 |              0.0 |          3.669425 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672736 | 1.672736 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.672736 |  -0.01936 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  -0.01936 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | 0.0 | nan | 0.0 |              5.607956 |                  5.607956 |                   5.607956 |                  2.021579 |                      0.0 |     3.586377 |              0.0 |          3.586377 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.529542 | 1.529542 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  1.008458 |  0.011672 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.011672 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | 0.0 | nan | 0.0 |              5.874454 |                  5.874454 |                   5.874454 |                  2.487815 |                      0.0 |      3.38664 |              0.0 |           3.38664 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.283901 | 1.283901 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.283901 |  -0.01486 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  -0.01486 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | 0.0 | nan | 0.0 |              6.104899 |                  6.104899 |                   6.104899 |                  2.520953 |                      0.0 |     3.583945 |              0.0 |          3.583945 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.510743 | 1.510743 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.510743 | -0.017485 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.017485 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | 0.0 | nan | 0.0 |              6.035457 |                  6.035457 |                   6.035457 |                  2.702703 |                      0.0 |     3.332754 |              0.0 |          3.332754 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.594572 | 1.594572 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.594572 | -0.018456 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.018456 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | 0.0 | nan | 0.0 |              5.176574 |                  5.176574 |                   5.176574 |                  2.307856 |                      0.0 |     2.868718 |              0.0 |          2.868718 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.541567 | 1.541567 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.541567 | -0.017842 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.017842 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | 0.0 | nan | 0.0 |              6.059368 |                  6.059368 |                   6.059368 |                  2.688476 |                      0.0 |     3.370892 |              0.0 |          3.370892 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.939383 | 1.939383 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.939383 | -0.022447 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.022447 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | 0.0 | nan | 0.0 |              6.379464 |                  6.379464 |                   6.379464 |                  3.308913 |                      0.0 |     3.070551 |              0.0 |          3.070551 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826811 | 1.826811 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.826811 | -0.021144 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.021144 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | 0.0 | nan | 0.0 |              2.347976 |                  2.347976 |                   2.347976 |                   0.60663 |                      0.0 |     1.741346 |              0.0 |          1.741346 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.806188 | 0.806188 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  1.073812 |  0.012428 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.012428 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | 0.0 | nan | 0.0 |              4.456814 |                  4.456814 |                   4.456814 |                  1.641158 |                      0.0 |     2.815656 |              0.0 |          2.815656 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.08421 |  1.08421 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  -0.80221 | -0.009285 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.009285 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | 0.0 | nan | 0.0 |              5.612971 |                  5.612971 |                   5.612971 |                  1.868809 |                      0.0 |     3.744162 |              0.0 |          3.744162 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672232 | 1.672232 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.672232 | -0.019355 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.019355 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | 0.0 | nan | 0.0 |              4.326484 |                  4.326484 |                   4.326484 |                  1.401603 |                      0.0 |     2.924881 |              0.0 |          2.924881 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.374102 | 1.374102 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.280102 | -0.014816 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.014816 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | 0.0 | nan | 0.0 |              3.783947 |                  3.783947 |                   3.783947 |                  1.261689 |                      0.0 |     2.522258 |              0.0 |          2.522258 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077096 | 1.077096 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  8.792904 |   0.10177 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.10177 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | 0.0 | nan | 0.0 |              4.290898 |                  4.290898 |                   4.290898 |                  1.498964 |                      0.0 |     2.791934 |              0.0 |          2.791934 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.09756 |  1.09756 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   5.01244 |  0.058014 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.058014 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | 0.0 | nan | 0.0 |              2.273012 |                  2.273012 |                   2.273012 |                  0.589077 |                      0.0 |     1.683934 |              0.0 |          1.683934 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.794859 | 0.794859 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 13.305141 |  0.153995 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.153995 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | 0.0 | nan | 0.0 |               2.18653 |                   2.18653 |                    2.18653 |                  0.708412 |                      0.0 |     1.478119 |              0.0 |          1.478119 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.607776 | 0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  8.040224 |  0.093058 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.093058 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | 0.0 | nan | 0.0 |              3.688655 |                  3.688655 |                   3.688655 |                  1.881434 |                      0.0 |     1.807221 |              0.0 |          1.807221 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.90065 |  0.90065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  -0.90065 | -0.010424 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.010424 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | 0.0 | nan | 0.0 |              3.284636 |                  3.284636 |                   3.284636 |                  1.511299 |                      0.0 |     1.773338 |              0.0 |          1.773338 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.912078 | 0.912078 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -0.912078 | -0.010556 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.010556 |

.. _lland_v3_see_daily:

lakes
-----

The following example focusses on water-type |SEE| (for further information,
see integration test :ref:`lland_v1_see` of |lland_v1|):

.. integration-test::

    >>> lnk(SEE)
    >>> surfaceresistance.see_aug = 0.0
    >>> cropheight.see_aug = 0.05
    >>> albedo.see_aug = 0.7
    >>> negq(False)
    >>> test('lland_v3_see_daily',
    ...      axis1=(fluxes.nkor, fluxes.evi, fluxes.qah))
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |      qah |       qa | inzp | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 |      qbgz | qdga1 | qdga2 | qiga1 | qiga2 |      qbga | inlet |   outlet |
    ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | 0.0 | nan | 0.0 |              4.931319 |                  4.931319 |                   4.931319 |                  1.877708 |                      0.0 |      3.05361 |              0.0 |           3.05361 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077197 | 1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.020478 | 0.000237 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | 11.424803 |   0.0 |   0.0 |   0.0 |   0.0 |  0.020478 |   0.0 | 0.000237 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | 0.0 | nan | 0.0 |              3.308967 |                  3.308967 |                   3.308967 |                   0.96939 |                      0.0 |     2.339577 |              0.0 |          2.339577 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.86545 |  0.86545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.039475 | 0.000457 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.77145 |   0.0 |   0.0 |   0.0 |   0.0 |  0.039475 |   0.0 | 0.000457 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | 0.0 | nan | 0.0 |              4.261282 |                  4.261282 |                   4.261282 |                  1.386217 |                      0.0 |     2.875064 |              0.0 |          2.875064 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.092365 | 1.092365 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.035994 | 0.000417 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.092365 |   0.0 |   0.0 |   0.0 |   0.0 |  0.035994 |   0.0 | 0.000417 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | 0.0 | nan | 0.0 |              6.973011 |                  6.973011 |                   6.973011 |                  3.108707 |                      0.0 |     3.864304 |              0.0 |          3.864304 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.565894 | 1.565894 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.031272 | 0.000362 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.471894 |   0.0 |   0.0 |   0.0 |   0.0 |  0.031272 |   0.0 | 0.000362 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | 0.0 | nan | 0.0 |              7.461037 |                  7.461037 |                   7.461037 |                  3.414868 |                      0.0 |     4.046169 |              0.0 |          4.046169 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.729733 | 1.729733 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.025424 | 0.000294 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.729733 |   0.0 |   0.0 |   0.0 |   0.0 |  0.025424 |   0.0 | 0.000294 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | 0.0 | nan | 0.0 |              7.119557 |                  7.119557 |                   7.119557 |                  2.845741 |                      0.0 |     4.273816 |              0.0 |          4.273816 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826087 | 1.826087 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.018963 | 0.000219 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.826087 |   0.0 |   0.0 |   0.0 |   0.0 |  0.018963 |   0.0 | 0.000219 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | 0.0 | nan | 0.0 |              7.362521 |                  7.362521 |                   7.362521 |                  3.051087 |                      0.0 |     4.311434 |              0.0 |          4.311434 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.041956 | 2.041956 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.011966 | 0.000138 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -2.041956 |   0.0 |   0.0 |   0.0 |   0.0 |  0.011966 |   0.0 | 0.000138 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | 0.0 | nan | 0.0 |              6.816215 |                  6.816215 |                   6.816215 |                   2.45312 |                      0.0 |     4.363095 |              0.0 |          4.363095 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.800053 | 1.800053 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.005041 | 0.000058 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.800053 |   0.0 |   0.0 |   0.0 |   0.0 |  0.005041 |   0.0 | 0.000058 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | 0.0 | nan | 0.0 |              7.056886 |                  7.056886 |                   7.056886 |                  2.688326 |                      0.0 |      4.36856 |              0.0 |           4.36856 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.82186 | 1.820396 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  -1.82186 |   0.0 |   0.0 |   0.0 |   0.0 | -0.001465 |   0.0 |      0.0 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | 0.0 | nan | 0.0 |              6.921138 |                  6.921138 |                   6.921138 |                   2.76572 |                      0.0 |     4.155419 |              0.0 |          4.155419 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.891469 | 1.883357 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.891469 |   0.0 |   0.0 |   0.0 |   0.0 | -0.008111 |   0.0 |      0.0 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | 0.0 | nan | 0.0 |               6.98872 |                   6.98872 |                    6.98872 |                  2.799361 |                      0.0 |     4.189359 |              0.0 |          4.189359 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.106575 | 2.091331 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -2.106575 |   0.0 |   0.0 |   0.0 |   0.0 | -0.015244 |   0.0 |      0.0 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | 0.0 | nan | 0.0 |              7.021165 |                  7.021165 |                   7.021165 |                  2.986484 |                      0.0 |     4.034682 |              0.0 |          4.034682 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.036929 | 2.014486 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.942929 |   0.0 |   0.0 |   0.0 |   0.0 | -0.022443 |   0.0 |      0.0 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | 0.0 | nan | 0.0 |               6.78319 |                   6.78319 |                    6.78319 |                  2.708378 |                      0.0 |     4.074812 |              0.0 |          4.074812 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.808559 | 1.779476 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.808559 |   0.0 |   0.0 |   0.0 |   0.0 | -0.029083 |   0.0 |      0.0 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | 0.0 | nan | 0.0 |              5.908778 |                  5.908778 |                   5.908778 |                  2.239353 |                      0.0 |     3.669425 |              0.0 |          3.669425 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672736 | 1.637521 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.672736 |   0.0 |   0.0 |   0.0 |   0.0 | -0.035215 |   0.0 |      0.0 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | 0.0 | nan | 0.0 |              5.607956 |                  5.607956 |                   5.607956 |                  2.021579 |                      0.0 |     3.586377 |              0.0 |          3.586377 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.529542 | 1.493266 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  1.008458 |   0.0 |   0.0 |   0.0 |   0.0 | -0.036276 |   0.0 |      0.0 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | 0.0 | nan | 0.0 |              5.874454 |                  5.874454 |                   5.874454 |                  2.487815 |                      0.0 |      3.38664 |              0.0 |           3.38664 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.283901 | 1.247259 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.283901 |   0.0 |   0.0 |   0.0 |   0.0 | -0.036642 |   0.0 |      0.0 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | 0.0 | nan | 0.0 |              6.104899 |                  6.104899 |                   6.104899 |                  2.520953 |                      0.0 |     3.583945 |              0.0 |          3.583945 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.510743 | 1.469226 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.510743 |   0.0 |   0.0 |   0.0 |   0.0 | -0.041517 |   0.0 |      0.0 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | 0.0 | nan | 0.0 |              6.035457 |                  6.035457 |                   6.035457 |                  2.702703 |                      0.0 |     3.332754 |              0.0 |          3.332754 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.594572 | 1.547641 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.594572 |   0.0 |   0.0 |   0.0 |   0.0 | -0.046931 |   0.0 |      0.0 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | 0.0 | nan | 0.0 |              5.176574 |                  5.176574 |                   5.176574 |                  2.307856 |                      0.0 |     2.868718 |              0.0 |          2.868718 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.541567 | 1.489186 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.541567 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.05238 |   0.0 |      0.0 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | 0.0 | nan | 0.0 |              6.059368 |                  6.059368 |                   6.059368 |                  2.688476 |                      0.0 |     3.370892 |              0.0 |          3.370892 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.939383 | 1.880954 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.939383 |   0.0 |   0.0 |   0.0 |   0.0 | -0.058428 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | 0.0 | nan | 0.0 |              6.379464 |                  6.379464 |                   6.379464 |                  3.308913 |                      0.0 |     3.070551 |              0.0 |          3.070551 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826811 | 1.761846 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.826811 |   0.0 |   0.0 |   0.0 |   0.0 | -0.064965 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | 0.0 | nan | 0.0 |              2.347976 |                  2.347976 |                   2.347976 |                   0.60663 |                      0.0 |     1.741346 |              0.0 |          1.741346 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.806188 | 0.740109 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  1.073812 |   0.0 |   0.0 |   0.0 |   0.0 | -0.066078 |   0.0 |      0.0 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | 0.0 | nan | 0.0 |              4.456814 |                  4.456814 |                   4.456814 |                  1.641158 |                      0.0 |     2.815656 |              0.0 |          2.815656 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.08421 | 1.018853 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.80221 |   0.0 |   0.0 |   0.0 |   0.0 | -0.065357 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | 0.0 | nan | 0.0 |              5.612971 |                  5.612971 |                   5.612971 |                  1.868809 |                      0.0 |     3.744162 |              0.0 |          3.744162 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672232 | 1.602675 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.672232 |   0.0 |   0.0 |   0.0 |   0.0 | -0.069557 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | 0.0 | nan | 0.0 |              4.326484 |                  4.326484 |                   4.326484 |                  1.401603 |                      0.0 |     2.924881 |              0.0 |          2.924881 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.374102 | 1.299507 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.280102 |   0.0 |   0.0 |   0.0 |   0.0 | -0.074595 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | 0.0 | nan | 0.0 |              3.783947 |                  3.783947 |                   3.783947 |                  1.261689 |                      0.0 |     2.522258 |              0.0 |          2.522258 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077096 | 1.016237 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  8.792904 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.06086 |   0.0 |      0.0 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | 0.0 | nan | 0.0 |              4.290898 |                  4.290898 |                   4.290898 |                  1.498964 |                      0.0 |     2.791934 |              0.0 |          2.791934 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.09756 | 1.061645 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |   5.01244 |   0.0 |   0.0 |   0.0 |   0.0 | -0.035916 |   0.0 |      0.0 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | 0.0 | nan | 0.0 |              2.273012 |                  2.273012 |                   2.273012 |                  0.589077 |                      0.0 |     1.683934 |              0.0 |          1.683934 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.794859 | 0.791893 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | 13.305141 |   0.0 |   0.0 |   0.0 |   0.0 | -0.002966 |   0.0 |      0.0 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | 0.0 | nan | 0.0 |               2.18653 |                   2.18653 |                    2.18653 |                  0.708412 |                      0.0 |     1.478119 |              0.0 |          1.478119 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.607776 | 0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.035276 | 0.000408 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  8.040224 |   0.0 |   0.0 |   0.0 |   0.0 |  0.035276 |   0.0 | 0.000408 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | 0.0 | nan | 0.0 |              3.688655 |                  3.688655 |                   3.688655 |                  1.881434 |                      0.0 |     1.807221 |              0.0 |          1.807221 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.90065 |  0.90065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.047929 | 0.000555 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.90065 |   0.0 |   0.0 |   0.0 |   0.0 |  0.047929 |   0.0 | 0.000555 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | 0.0 | nan | 0.0 |              3.284636 |                  3.284636 |                   3.284636 |                  1.511299 |                      0.0 |     1.773338 |              0.0 |          1.773338 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.912078 | 0.912078 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.04451 | 0.000515 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -0.912078 |   0.0 |   0.0 |   0.0 |   0.0 |   0.04451 |   0.0 | 0.000515 |

.. _lland_v3_fluss_daily:

streams
-------

The following example focusses on water-type |FLUSS| (for further information,
see integration test :ref:`lland_v1_fluss` of |lland_v1|):

.. integration-test::

    >>> lnk(FLUSS)
    >>> surfaceresistance.fluss_aug = 0.0
    >>> cropheight.fluss_aug = 0.05
    >>> albedo.fluss_aug = 0.7
    >>> test('lland_v3_fluss_daily',
    ...      axis1=(fluxes.nkor, fluxes.evi, fluxes.qah))
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |     evpo |       evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap |      qdgz |      qah |       qa | inzp | wats | waes | esnow | taus | ebdn | bowa |     qdgz1 |    qdgz2 | qigz1 | qigz2 | qbgz |     qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | 0.0 | nan | 0.0 |              4.931319 |                  4.931319 |                   4.931319 |                  1.877708 |                      0.0 |      3.05361 |              0.0 |           3.05361 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077197 |  1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | 11.424803 | 2.941276 | 0.034043 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | 11.424803 |      0.0 |   0.0 |   0.0 |  0.0 |  2.941276 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034043 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | 0.0 | nan | 0.0 |              3.308967 |                  3.308967 |                   3.308967 |                   0.96939 |                      0.0 |     2.339577 |              0.0 |          2.339577 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.86545 |   0.86545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  -0.77145 | 3.758016 | 0.043496 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  -0.77145 |      0.0 |   0.0 |   0.0 |  0.0 |  3.758016 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043496 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | 0.0 | nan | 0.0 |              4.261282 |                  4.261282 |                   4.261282 |                  1.386217 |                      0.0 |     2.875064 |              0.0 |          2.875064 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.092365 |  1.092365 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.092365 | 1.562905 | 0.018089 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.092365 |      0.0 |   0.0 |   0.0 |  0.0 |  1.562905 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018089 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | 0.0 | nan | 0.0 |              6.973011 |                  6.973011 |                   6.973011 |                  3.108707 |                      0.0 |     3.864304 |              0.0 |          3.864304 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.565894 |  1.565894 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.471894 | 0.226808 | 0.002625 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.471894 |      0.0 |   0.0 |   0.0 |  0.0 |  0.226808 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002625 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | 0.0 | nan | 0.0 |              7.461037 |                  7.461037 |                   7.461037 |                  3.414868 |                      0.0 |     4.046169 |              0.0 |          4.046169 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.729733 |  1.097905 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.729733 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.729733 |      0.0 |   0.0 |   0.0 |  0.0 | -0.631827 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | 0.0 | nan | 0.0 |              7.119557 |                  7.119557 |                   7.119557 |                  2.845741 |                      0.0 |     4.273816 |              0.0 |          4.273816 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826087 |  0.657403 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.826087 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.826087 |      0.0 |   0.0 |   0.0 |  0.0 | -1.168684 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | 0.0 | nan | 0.0 |              7.362521 |                  7.362521 |                   7.362521 |                  3.051087 |                      0.0 |     4.311434 |              0.0 |          4.311434 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.041956 |  0.511091 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -2.041956 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -2.041956 |      0.0 |   0.0 |   0.0 |  0.0 | -1.530864 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | 0.0 | nan | 0.0 |              6.816215 |                  6.816215 |                   6.816215 |                   2.45312 |                      0.0 |     4.363095 |              0.0 |          4.363095 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.800053 |  0.093098 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.800053 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.800053 |      0.0 |   0.0 |   0.0 |  0.0 | -1.706955 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | 0.0 | nan | 0.0 |              7.056886 |                  7.056886 |                   7.056886 |                  2.688326 |                      0.0 |      4.36856 |              0.0 |           4.36856 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.82186 |  0.065871 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  -1.82186 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  -1.82186 |      0.0 |   0.0 |   0.0 |  0.0 | -1.755989 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | 0.0 | nan | 0.0 |              6.921138 |                  6.921138 |                   6.921138 |                   2.76572 |                      0.0 |     4.155419 |              0.0 |          4.155419 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.891469 |  0.086838 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.891469 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.891469 |      0.0 |   0.0 |   0.0 |  0.0 | -1.804631 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | 0.0 | nan | 0.0 |               6.98872 |                   6.98872 |                    6.98872 |                  2.799361 |                      0.0 |     4.189359 |              0.0 |          4.189359 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.106575 |  0.206065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -2.106575 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -2.106575 |      0.0 |   0.0 |   0.0 |  0.0 | -1.900509 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | 0.0 | nan | 0.0 |              7.021165 |                  7.021165 |                   7.021165 |                  2.986484 |                      0.0 |     4.034682 |              0.0 |          4.034682 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 2.036929 |  0.082443 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.942929 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.942929 |      0.0 |   0.0 |   0.0 |  0.0 | -1.954486 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | 0.0 | nan | 0.0 |               6.78319 |                   6.78319 |                    6.78319 |                  2.708378 |                      0.0 |     4.074812 |              0.0 |          4.074812 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.808559 | -0.105944 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.808559 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.808559 |      0.0 |   0.0 |   0.0 |  0.0 | -1.914503 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | 0.0 | nan | 0.0 |              5.908778 |                  5.908778 |                   5.908778 |                  2.239353 |                      0.0 |     3.669425 |              0.0 |          3.669425 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672736 | -0.157388 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.672736 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.672736 |      0.0 |   0.0 |   0.0 |  0.0 | -1.830125 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | 0.0 | nan | 0.0 |              5.607956 |                  5.607956 |                   5.607956 |                  2.021579 |                      0.0 |     3.586377 |              0.0 |          3.586377 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.529542 |  0.463086 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  1.008458 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  1.008458 |      0.0 |   0.0 |   0.0 |  0.0 | -1.066456 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | 0.0 | nan | 0.0 |              5.874454 |                  5.874454 |                   5.874454 |                  2.487815 |                      0.0 |      3.38664 |              0.0 |           3.38664 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.283901 |  0.595002 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.283901 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.283901 |      0.0 |   0.0 |   0.0 |  0.0 | -0.688899 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | 0.0 | nan | 0.0 |              6.104899 |                  6.104899 |                   6.104899 |                  2.520953 |                      0.0 |     3.583945 |              0.0 |          3.583945 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.510743 |  0.485942 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.510743 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.510743 |      0.0 |   0.0 |   0.0 |  0.0 | -1.024801 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | 0.0 | nan | 0.0 |              6.035457 |                  6.035457 |                   6.035457 |                  2.702703 |                      0.0 |     3.332754 |              0.0 |          3.332754 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.594572 |  0.321551 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.594572 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.594572 |      0.0 |   0.0 |   0.0 |  0.0 |  -1.27302 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | 0.0 | nan | 0.0 |              5.176574 |                  5.176574 |                   5.176574 |                  2.307856 |                      0.0 |     2.868718 |              0.0 |          2.868718 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.541567 |  0.132224 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.541567 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.541567 |      0.0 |   0.0 |   0.0 |  0.0 | -1.409342 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | 0.0 | nan | 0.0 |              6.059368 |                  6.059368 |                   6.059368 |                  2.688476 |                      0.0 |     3.370892 |              0.0 |          3.370892 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.939383 |  0.365956 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.939383 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.939383 |      0.0 |   0.0 |   0.0 |  0.0 | -1.573427 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | 0.0 | nan | 0.0 |              6.379464 |                  6.379464 |                   6.379464 |                  3.308913 |                      0.0 |     3.070551 |              0.0 |          3.070551 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.826811 |  0.111688 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.826811 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.826811 |      0.0 |   0.0 |   0.0 |  0.0 | -1.715123 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | 0.0 | nan | 0.0 |              2.347976 |                  2.347976 |                   2.347976 |                   0.60663 |                      0.0 |     1.741346 |              0.0 |          1.741346 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.806188 |  -0.21427 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  1.073812 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  1.073812 |      0.0 |   0.0 |   0.0 |  0.0 | -1.020458 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | 0.0 | nan | 0.0 |              4.456814 |                  4.456814 |                   4.456814 |                  1.641158 |                      0.0 |     2.815656 |              0.0 |          2.815656 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.08421 |  0.557521 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  -0.80221 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  -0.80221 |      0.0 |   0.0 |   0.0 |  0.0 | -0.526689 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | 0.0 | nan | 0.0 |              5.612971 |                  5.612971 |                   5.612971 |                  1.868809 |                      0.0 |     3.744162 |              0.0 |          3.744162 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.672232 |  0.793059 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.672232 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.672232 |      0.0 |   0.0 |   0.0 |  0.0 | -0.879173 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | 0.0 | nan | 0.0 |              4.326484 |                  4.326484 |                   4.326484 |                  1.401603 |                      0.0 |     2.924881 |              0.0 |          2.924881 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.374102 |  0.226008 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.280102 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.280102 |      0.0 |   0.0 |   0.0 |  0.0 | -1.148095 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | 0.0 | nan | 0.0 |              3.783947 |                  3.783947 |                   3.783947 |                  1.261689 |                      0.0 |     2.522258 |              0.0 |          2.522258 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 1.077096 |  1.077096 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  8.792904 |   1.3836 | 0.016014 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  8.792904 |      0.0 |   0.0 |   0.0 |  0.0 |    1.3836 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016014 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | 0.0 | nan | 0.0 |              4.290898 |                  4.290898 |                   4.290898 |                  1.498964 |                      0.0 |     2.791934 |              0.0 |          2.791934 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  1.09756 |   1.09756 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |   5.01244 |  3.86595 | 0.044745 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   5.01244 |      0.0 |   0.0 |   0.0 |  0.0 |   3.86595 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.044745 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | 0.0 | nan | 0.0 |              2.273012 |                  2.273012 |                   2.273012 |                  0.589077 |                      0.0 |     1.683934 |              0.0 |          1.683934 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.794859 |  0.794859 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | 13.305141 | 7.027354 | 0.081335 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |      12.0 | 1.305141 |   0.0 |   0.0 |  0.0 |  6.199583 |  0.82777 |   0.0 |   0.0 |  0.0 |   0.0 | 0.081335 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | 0.0 | nan | 0.0 |               2.18653 |                   2.18653 |                    2.18653 |                  0.708412 |                      0.0 |     1.478119 |              0.0 |          1.478119 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.607776 |  0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  8.040224 |  8.32407 | 0.096343 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  8.040224 |      0.0 |   0.0 |   0.0 |  0.0 |  7.885403 | 0.438667 |   0.0 |   0.0 |  0.0 |   0.0 | 0.096343 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | 0.0 | nan | 0.0 |              3.688655 |                  3.688655 |                   3.688655 |                  1.881434 |                      0.0 |     1.807221 |              0.0 |          1.807221 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.90065 |   0.90065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  -0.90065 | 5.691378 | 0.065872 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  -0.90065 |      0.0 |   0.0 |   0.0 |  0.0 |  5.655812 | 0.035566 |   0.0 |   0.0 |  0.0 |   0.0 | 0.065872 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | 0.0 | nan | 0.0 |              3.284636 |                  3.284636 |                   3.284636 |                  1.511299 |                      0.0 |     1.773338 |              0.0 |          1.773338 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.912078 |  0.912078 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -0.912078 | 2.597892 | 0.030068 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -0.912078 |      0.0 |   0.0 |   0.0 |  0.0 |  2.595008 | 0.002884 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030068 |

.. _lland_v3_vers_daily:

sealed surfaces
---------------

As also shown in the integration test :ref:`lland_v1_vers` of |lland_v1|,
sealed surfaces route the not intercepted water to the linear storages for
direct discharge immediately:

.. integration-test::

    >>> lnk(VERS)
    >>> surfaceresistance.vers_aug = 500
    >>> cropheight.vers_aug = 5.0
    >>> albedo.vers_aug = 0.10
    >>> lai.vers_aug = 10.0
    >>> test('lland_v3_vers_daily',
    ...      axis1=(inputs.nied, fluxes.qah, states.bowa))
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |     g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes | sbes |      evpo |   evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg |   wada |    qdb | qib1 | qib2 | qbb | qkap |   qdgz |      qah |       qa | inzp | wats | waes | esnow | taus |      ebdn | bowa |  qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | -0.25 | 10.933333 |   -2.6784 |             14.793956 |                 14.793956 |                  14.793956 |                  1.877708 |                      0.0 |    12.916248 |              0.0 |         12.916248 |             16.038785 |                   nan |                    500.0 |                   500.0 | 10.502 |  0.0 |  7.049585 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 10.502 | 10.502 |  0.0 |  0.0 | 0.0 |  0.0 | 10.502 | 2.703704 | 0.031293 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   32.4284 |  0.0 | 10.502 |   0.0 |   0.0 |   0.0 |  0.0 | 2.703704 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.031293 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | -0.25 |    19.028 |  0.895795 |              9.926902 |                  9.926902 |                   9.926902 |                   0.96939 |                      0.0 |     8.957512 |              0.0 |          8.957512 |             24.058177 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  5.020895 | 0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 3.637039 | 0.042095 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.282605 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 3.637039 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042095 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | -0.25 | 15.208683 | -1.913579 |             12.783845 |                 12.783845 |                  12.783845 |                  1.386217 |                      0.0 |    11.397628 |              0.0 |         11.397628 |             25.662055 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  6.335547 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 1.940764 | 0.022463 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.946184 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 1.940764 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022463 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | -0.25 | 20.753946 |  1.531325 |             20.919034 |                 20.919034 |                  20.919034 |                  3.108707 |                      0.0 |    17.810327 |              0.0 |         17.810327 |              13.74753 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 12.230738 | 0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 1.035613 | 0.011986 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.164858 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 1.035613 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011986 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | -0.25 | 14.816194 | -1.909685 |             22.383112 |                 22.383112 |                  22.383112 |                  3.414868 |                      0.0 |    18.968244 |              0.0 |         18.968244 |             12.831028 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 14.271962 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.552615 | 0.006396 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.824543 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.552615 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.006396 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | -0.25 | 20.348477 |  0.232491 |             21.358671 |                 21.358671 |                  21.358671 |                  2.845741 |                      0.0 |     18.51293 |              0.0 |          18.51293 |             10.998024 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 14.791696 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.294881 | 0.003413 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.342053 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.294881 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003413 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | -0.25 | 18.740175 | -1.067813 |             22.087564 |                 22.087564 |                  22.087564 |                  3.051087 |                      0.0 |    19.036477 |              0.0 |         19.036477 |             10.129759 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 18.203529 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.157352 | 0.001821 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.159866 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.157352 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001821 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | -0.25 | 21.466219 |  0.293528 |             20.448645 |                 20.448645 |                  20.448645 |                   2.45312 |                      0.0 |    17.995525 |              0.0 |         17.995525 |             11.664571 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 13.456812 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.083965 | 0.000972 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.616338 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.083965 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000972 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | -0.25 | 19.654459 | -1.371448 |             21.170659 |                 21.170659 |                  21.170659 |                  2.688326 |                      0.0 |    18.482332 |              0.0 |         18.482332 |             21.385046 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  11.84053 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.044804 | 0.000519 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.737786 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.044804 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000519 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | -0.25 |  23.39262 |  0.203534 |             20.763414 |                 20.763414 |                  20.763414 |                   2.76572 |                      0.0 |    17.997695 |              0.0 |         17.997695 |             20.259517 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 13.388795 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.023908 | 0.000277 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.284252 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.023908 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000277 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | -0.25 | 21.880839 | -0.632013 |              20.96616 |                  20.96616 |                   20.96616 |                  2.799361 |                      0.0 |    18.166799 |              0.0 |         18.166799 |             12.831028 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 18.009869 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.012758 | 0.000148 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.666265 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.012758 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000148 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | -0.25 | 23.154216 | -0.023734 |             21.063496 |                 21.063496 |                  21.063496 |                  2.986484 |                      0.0 |    18.077012 |              0.0 |         18.077012 |             15.397233 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 16.851837 | 0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.006808 | 0.000079 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.439999 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.006808 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000079 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | -0.25 | 22.399998 | -0.414721 |             20.349571 |                 20.349571 |                  20.349571 |                  2.708378 |                      0.0 |    17.641193 |              0.0 |         17.641193 |             24.058177 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 11.954049 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.003633 | 0.000042 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.604721 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.003633 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000042 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | -0.25 | 22.949068 |  0.647517 |             17.726334 |                 17.726334 |                  17.726334 |                  2.239353 |                      0.0 |    15.486981 |              0.0 |         15.486981 |             14.256697 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 13.005903 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.001938 | 0.000022 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.707203 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001938 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000022 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | -0.25 | 19.957345 | -0.436832 |             16.823868 |                 16.823868 |                  16.823868 |                  2.021579 |                      0.0 |    14.802289 |              0.0 |         14.802289 |             12.029088 |                   nan |                    500.0 |                   500.0 |  0.538 |  0.0 | 11.728768 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.538 |  0.538 |  0.0 |  0.0 | 0.0 |  0.0 |  0.538 | 0.139541 | 0.001615 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.894036 |  0.0 |  0.538 |   0.0 |   0.0 |   0.0 |  0.0 | 0.139541 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001615 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | -0.25 |  20.58012 |  0.559934 |             17.623363 |                 17.623363 |                  17.623363 |                  2.487815 |                      0.0 |    15.135549 |              0.0 |         15.135549 |             29.610064 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |   7.56081 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.186871 | 0.002163 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.084102 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.186871 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002163 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | -0.25 |  17.88034 | -1.617232 |             18.314696 |                 18.314696 |                  18.314696 |                  2.520953 |                      0.0 |    15.793743 |              0.0 |         15.793743 |             20.259517 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  10.39177 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.099717 | 0.001154 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.451334 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.099717 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001154 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | -0.25 | 22.437779 |  0.434305 |             18.106372 |                 18.106372 |                  18.106372 |                  2.702703 |                      0.0 |    15.403668 |              0.0 |         15.403668 |             19.246541 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 12.298568 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 |  0.05321 | 0.000616 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.767029 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.05321 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000616 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | -0.25 | 20.156763 | -0.488974 |             15.529721 |                 15.529721 |                  15.529721 |                  2.307856 |                      0.0 |    13.221865 |              0.0 |         13.221865 |             14.805032 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 13.759102 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.028393 | 0.000329 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.006003 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.028393 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000329 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | -0.25 | 20.953343 | -0.749947 |             18.178103 |                 18.178103 |                  18.178103 |                  2.688476 |                      0.0 |    15.489627 |              0.0 |         15.489627 |             10.692523 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 19.233472 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.015151 | 0.000175 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.50595 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.015151 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000175 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | -0.25 | 22.619833 |  0.062121 |             19.138391 |                 19.138391 |                  19.138391 |                  3.308913 |                      0.0 |    15.829478 |              0.0 |         15.829478 |             15.397233 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  17.26069 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.008085 | 0.000094 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.193828 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.008085 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000094 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | -0.25 | 21.579428 |  1.026136 |              7.043929 |                  7.043929 |                   7.043929 |                   0.60663 |                      0.0 |     6.437299 |              0.0 |          6.437299 |             18.330039 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  5.830469 |  1.88 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.004314 |  0.00005 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.917693 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.004314 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00005 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | -0.25 | 17.325643 | -1.541907 |             13.370443 |                 13.370443 |                  13.370443 |                  1.641158 |                      0.0 |    11.729284 |              0.0 |         11.729284 |             25.662055 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  6.371493 | 0.282 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.002302 | 0.000027 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   33.2096 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.002302 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000027 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | -0.25 | 21.631999 | -1.279412 |             16.838913 |                 16.838913 |                  16.838913 |                  1.868809 |                      0.0 |    14.970104 |              0.0 |         14.970104 |             17.496856 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 | 11.504016 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.001228 | 0.000014 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 34.239011 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001228 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000014 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | -0.25 | 25.063372 |  0.706772 |             12.979453 |                 12.979453 |                  12.979453 |                  1.401603 |                      0.0 |     11.57785 |              0.0 |          11.57785 |             20.259517 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  9.461825 | 0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.000655 | 0.000008 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.28224 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000655 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000008 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | -0.25 | 21.874132 |   0.40131 |             11.351841 |                 11.351841 |                  11.351841 |                  1.261689 |                      0.0 |    10.090152 |              0.0 |         10.090152 |             16.736123 |                   nan |                    500.0 |                   500.0 |   7.87 |  0.0 |  7.472563 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   7.87 |   7.87 |  0.0 |  0.0 | 0.0 |  0.0 |   7.87 | 2.026454 | 0.023454 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.63093 |  0.0 |   7.87 |   0.0 |   0.0 |   0.0 |  0.0 | 2.026454 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023454 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | -0.25 | 19.703098 | -0.620474 |             12.872693 |                 12.872693 |                  12.872693 |                  1.498964 |                      0.0 |    11.373729 |              0.0 |         11.373729 |             18.330039 |                   nan |                    500.0 |                   500.0 |   4.11 |  0.0 |  6.960765 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   4.11 |   4.11 |  0.0 |  0.0 | 0.0 |  0.0 |   4.11 |  3.78382 | 0.043794 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.001403 |  0.0 |   4.11 |   0.0 |   0.0 |   0.0 |  0.0 |  3.78382 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043794 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | -0.25 | 20.938011 |  0.382585 |              6.819035 |                  6.819035 |                   6.819035 |                  0.589077 |                      0.0 |     6.229957 |              0.0 |          6.229957 |             14.805032 |                   nan |                    500.0 |                   500.0 |   12.1 |  0.0 |  6.075158 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   12.1 |   12.1 |  0.0 |  0.0 | 0.0 |  0.0 |   12.1 | 6.030624 | 0.069799 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.368818 |  0.0 |   12.0 |   0.1 |   0.0 |   0.0 |  0.0 | 5.967201 | 0.063424 |   0.0 |   0.0 |  0.0 |   0.0 | 0.069799 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | -0.25 | 18.829395 |  1.414918 |              6.559591 |                  6.559591 |                   6.559591 |                  0.708412 |                      0.0 |     5.851179 |              0.0 |          5.851179 |             14.805032 |                   nan |                    500.0 |                   500.0 |  6.648 |  0.0 |  4.534236 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  6.648 |  6.648 |  0.0 |  0.0 | 0.0 |  0.0 |  6.648 | 7.436588 | 0.086072 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   30.7039 |  0.0 |  6.648 |   0.0 |   0.0 |   0.0 |  0.0 | 7.402977 | 0.033611 |   0.0 |   0.0 |  0.0 |   0.0 | 0.086072 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | -0.25 | 13.279667 |  -1.92862 |             11.065964 |                 11.065964 |                  11.065964 |                  1.881434 |                      0.0 |      9.18453 |              0.0 |           9.18453 |             18.330039 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  7.752367 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 5.342084 |  0.06183 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.382521 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 5.339359 | 0.002725 |   0.0 |   0.0 |  0.0 |   0.0 |  0.06183 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | -0.25 | 18.875069 |  0.349956 |              9.853909 |                  9.853909 |                   9.853909 |                  1.511299 |                      0.0 |      8.34261 |              0.0 |           8.34261 |             18.330039 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |  7.711096 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 2.849362 | 0.032979 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.782565 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 2.849141 | 0.000221 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032979 |

.. _lland_v3_nadelw_summer_daily:

conifers (summer)
-----------------

The following tests focus on functionalities of |lland_v3| not implemented
by application model |lland_v1|.  We start with the land-use type |NADELW|.
For |lland_v1|, there is no structural difference between |NADELW| and,
for example, |ACKER|.  Different simulation results only are based solely
on different parameterisations.  In contrast, |lland_v3| handles |NADELW|
unlike all other land use types (even unlike |LAUBW| and |MISCHW|) regarding
the calculation of the surface resistance, relevant for the Penman-Monteith
equation (see method |Calc_LanduseSurfaceResistance_V1| for more details).
The calculated soil evapotranspiration (|EvB|) is similar to the one
calculated with the same input data for land-use type |ACKER| in the
integration test :ref:`lland_v3_acker_summer_daily`:

.. integration-test::

    >>> lnk(NADELW)
    >>> surfaceresistance.nadelw = 56.0
    >>> cropheight.nadelw = 10.0
    >>> albedo.nadelw = 0.12
    >>> lai.nadelw = 11.0
    >>> conditions_nadelw_summer = test(
    ...     'lland_v3_nadelw_summer_daily',
    ...     axis1=(inputs.nied, fluxes.qah, states.bowa),
    ...     get_conditions='1997-08-03')
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |     g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes | sbes |      evpo |   evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg |   wada |      qdb | qib1 | qib2 | qbb |     qkap |     qdgz |      qah |       qa | inzp | wats | waes | esnow | taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 |      qbgz |    qdga1 | qdga2 | qiga1 | qiga2 |      qbga | inlet |   outlet |
    ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           0.851977 |          2.4 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | -0.25 | 10.933333 |   -2.6784 |             14.465201 |                  2.686395 |                  14.465201 |                  1.877708 |                      0.0 |    12.587493 |              0.0 |         12.587493 |             39.166667 |                 100.0 |               211.364559 |              122.747678 | 10.302 |  0.0 |  4.686428 |   2.2 | 1.110404 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 10.302 | 0.657815 |  0.0 |  0.0 | 0.0 |      0.0 | 0.657815 | 0.169352 |  0.00196 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   32.4284 | 80.533782 | 0.657815 |   0.0 |   0.0 |   0.0 |       0.0 | 0.169352 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.00196 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           0.567985 |          1.6 |                 1.977914 |                      1.977914 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | -0.25 |    19.028 |  0.895795 |              9.706305 |                  1.802599 |                   9.706305 |                   0.96939 |                      0.0 |     8.736914 |              0.0 |          8.736914 |                 58.75 |                 100.0 |               181.457925 |               115.35055 |    0.0 |  0.0 |  3.312002 | 0.094 | 1.809101 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.227814 | 0.002637 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.282605 | 78.724681 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.227814 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.002637 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           0.532486 |          1.5 |                 2.187158 |                      2.187158 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | -0.25 | 15.208683 | -1.913579 |             12.499759 |                  2.321384 |                  12.499759 |                  1.386217 |                      0.0 |    11.113542 |              0.0 |         11.113542 |             62.666667 |                 100.0 |               184.734569 |              115.873944 |    0.0 |  0.0 |   4.24805 |   0.0 | 2.495048 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.121564 | 0.001407 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.946184 | 76.229633 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.121564 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.001407 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           0.993973 |          2.8 |                 2.041337 |                      2.041337 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | -0.25 | 20.753946 |  1.531325 |             20.454167 |                  3.798631 |                  20.454167 |                  3.108707 |                      0.0 |     17.34546 |              0.0 |          17.34546 |             33.571429 |                 100.0 |               209.622626 |              121.261768 |    0.0 |  0.0 |  7.655528 | 0.094 |  3.24309 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.064868 | 0.000751 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.164858 | 72.986543 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.064868 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000751 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           1.064971 |          3.0 |                 2.133112 |                      2.133112 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | -0.25 | 14.816194 | -1.909685 |              21.88571 |                  4.064489 |                   21.88571 |                  3.414868 |                      0.0 |    18.470842 |              0.0 |         18.470842 |             31.333333 |                 100.0 |               223.524712 |              123.686516 |    0.0 |  0.0 |  8.731502 |   0.0 | 3.618105 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.034614 | 0.000401 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.824543 | 69.368437 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.034614 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000401 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           1.242466 |          3.5 |                 2.327535 |                      2.327535 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | -0.25 | 20.348477 |  0.232491 |             20.884034 |                  3.878463 |                  20.884034 |                  2.845741 |                      0.0 |    18.038293 |              0.0 |         18.038293 |             26.857143 |                 100.0 |               217.806222 |              122.172235 |    0.0 |  0.0 |  8.967691 |   0.0 | 3.538589 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.018471 | 0.000214 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.342053 | 65.829848 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.018471 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000214 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           1.348964 |          3.8 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | -0.25 | 18.740175 | -1.067813 |             21.596729 |                  4.010821 |                  21.596729 |                  3.051087 |                      0.0 |    18.545642 |              0.0 |         18.545642 |             24.736842 |                 100.0 |                249.89699 |              127.503607 |    0.0 |  0.0 | 10.546527 |   0.0 | 3.943139 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.009856 | 0.000114 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.159866 | 61.886709 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.009856 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000114 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           1.171468 |          3.3 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | -0.25 | 21.466219 |  0.293528 |             19.994231 |                  3.713214 |                  19.994231 |                   2.45312 |                      0.0 |    17.541111 |              0.0 |         17.541111 |             28.484848 |                 100.0 |               234.560208 |              124.470702 |    0.0 |  0.0 |  8.375029 |   0.0 | 3.447428 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.005259 | 0.000061 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.616338 | 58.439281 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.005259 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000061 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           0.638983 |          1.8 |                 2.696965 |                      2.696965 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | -0.25 | 19.654459 | -1.371448 |               20.7002 |                  3.844323 |                    20.7002 |                  2.688326 |                      0.0 |    18.011873 |              0.0 |         18.011873 |             52.222222 |                 100.0 |                290.49321 |              132.398807 |    0.0 |  0.0 |  7.801371 |   0.0 | 4.238315 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.002806 | 0.000032 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.737786 | 54.200966 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.002806 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000032 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                 2.813915 |                      2.813915 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | -0.25 |  23.39262 |  0.203534 |             20.302005 |                  3.770372 |                  20.302005 |                   2.76572 |                      0.0 |    17.536286 |              0.0 |         17.536286 |             49.473684 |                 100.0 |               348.466656 |              138.318012 |    0.0 |  0.0 |  8.457407 |   0.0 | 4.453782 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.001498 | 0.000017 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.284252 | 49.747184 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.001498 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000017 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           1.064971 |          3.0 |                 2.830978 |                      2.830978 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | -0.25 | 21.880839 | -0.632013 |             20.500246 |                  3.807188 |                  20.500246 |                  2.799361 |                      0.0 |    17.700884 |              0.0 |         17.700884 |             31.333333 |                 100.0 |               368.725681 |              139.570111 |    0.0 |  0.0 | 10.478761 |   0.0 | 4.442334 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000799 | 0.000009 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.666265 |  45.30485 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000799 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000009 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           0.887476 |          2.5 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | -0.25 | 23.154216 | -0.023734 |             20.595418 |                  3.824863 |                  20.595418 |                  2.986484 |                      0.0 |    17.608935 |              0.0 |         17.608935 |                  37.6 |                 100.0 |               419.711234 |              142.937942 |    0.0 |  0.0 |  9.985603 | 0.094 | 4.546047 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000426 | 0.000005 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.439999 | 40.758803 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000426 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000005 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           0.567985 |          1.6 |                  2.84813 |                       2.84813 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | -0.25 | 22.399998 | -0.414721 |             19.897359 |                  3.695224 |                  19.897359 |                  2.708378 |                      0.0 |    17.188981 |              0.0 |         17.188981 |                 58.75 |                 100.0 |               411.253273 |              141.649314 |    0.0 |  0.0 |  7.775599 |   0.0 | 4.347844 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000228 | 0.000003 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.604721 | 36.410959 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000228 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000003 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           0.958474 |          2.7 |                 2.600118 |                      2.600118 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | -0.25 | 22.949068 |  0.647517 |             17.332416 |                  3.218877 |                  17.332416 |                  2.239353 |                      0.0 |    15.093063 |              0.0 |         15.093063 |             34.814815 |                 100.0 |               357.548831 |              136.707247 |    0.0 |  0.0 |  7.857065 |   0.0 | 3.459696 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000121 | 0.000001 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.707203 | 32.951263 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000121 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000001 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           1.135969 |          3.2 |                 2.460535 |                      2.460535 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | -0.25 | 19.957345 | -0.436832 |             16.450004 |                  3.055001 |                  16.450004 |                  2.021579 |                      0.0 |    14.428425 |              0.0 |         14.428425 |                29.375 |                 100.0 |               333.908335 |              133.915514 |  0.338 |  0.0 |  7.162052 |   2.2 | 1.984501 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.338 | 0.008727 |  0.0 |  0.0 | 0.0 |      0.0 | 0.008727 | 0.002312 | 0.000027 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.894036 | 31.296034 | 0.008727 |   0.0 |   0.0 |   0.0 |       0.0 | 0.002312 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000027 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           0.461488 |          1.3 |                 2.270468 |                      2.270468 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | -0.25 |  20.58012 |  0.559934 |             17.231733 |                  3.200179 |                  17.231733 |                  2.487815 |                      0.0 |    14.743918 |              0.0 |         14.743918 |             72.307692 |                 100.0 |               354.695699 |              135.231698 |    0.0 |  0.0 |  5.274274 |   0.0 | 3.081998 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.003057 | 0.000035 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.084102 | 28.214037 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.003057 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000035 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                 2.490975 |                      2.490975 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | -0.25 |  17.88034 | -1.617232 |             17.907703 |                  3.325716 |                  17.907703 |                  2.520953 |                      0.0 |     15.38675 |              0.0 |          15.38675 |             49.473684 |                 100.0 |               400.426126 |               138.20342 |    0.0 |  0.0 |  6.725557 |   0.0 | 3.429544 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.166895 |      0.0 | 0.001332 | 0.000015 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.451334 | 24.951387 |      0.0 |   0.0 |   0.0 |   0.0 | -0.166895 | 0.001631 |   0.0 |   0.0 |   0.0 | -0.000299 |   0.0 | 0.000015 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           0.709981 |          2.0 |                 2.584275 |                      2.584275 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | -0.25 | 22.437779 |  0.434305 |             17.704008 |                  3.287887 |                  17.704008 |                  2.702703 |                      0.0 |    15.001304 |              0.0 |         15.001304 |                  47.0 |                 100.0 |               491.324867 |              142.949964 |    0.0 |  0.0 |   7.53971 |   0.0 | 3.733858 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.369623 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.767029 | 21.587152 |      0.0 |   0.0 |   0.0 |   0.0 | -0.369623 |  0.00087 |   0.0 |   0.0 |   0.0 | -0.001259 |   0.0 |      0.0 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           0.922975 |          2.6 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | -0.25 | 20.156763 | -0.488974 |             15.184616 |                      2.82 |                  15.184616 |                  2.307856 |                      0.0 |     12.87676 |              0.0 |          12.87676 |             36.153846 |                 100.0 |               525.389413 |              143.816091 |    0.0 |  0.0 |  7.832014 |   0.0 |  3.37772 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.578662 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.006003 | 18.788095 |      0.0 |   0.0 |   0.0 |   0.0 | -0.578662 | 0.000464 |   0.0 |   0.0 |   0.0 | -0.002954 |   0.0 |      0.0 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           1.277965 |          3.6 |                 2.713408 |                      2.713408 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | -0.25 | 20.953343 | -0.749947 |             17.774145 |                  3.300913 |                  17.774145 |                  2.688476 |                      0.0 |    15.085669 |              0.0 |         15.085669 |             26.111111 |                 100.0 |               629.540329 |              146.963401 |    0.0 |  0.0 | 10.542903 |   0.0 | 3.855581 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.752584 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.50595 | 15.685099 |      0.0 |   0.0 |   0.0 |   0.0 | -0.752584 | 0.000248 |   0.0 |   0.0 |   0.0 | -0.005328 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           0.887476 |          2.5 |                 2.729939 |                      2.729939 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | -0.25 | 22.619833 |  0.062121 |             18.713094 |                  3.475289 |                  18.713094 |                  3.308913 |                      0.0 |    15.404181 |              0.0 |         15.404181 |                  37.6 |                 100.0 |               884.502325 |              152.120949 |    0.0 |  0.0 |  9.796451 |   0.0 |  4.29127 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.945392 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.193828 | 12.339221 |      0.0 |   0.0 |   0.0 |   0.0 | -0.945392 | 0.000132 |   0.0 |   0.0 |   0.0 | -0.008351 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |            0.74548 |          2.1 |                 2.284619 |                      2.284619 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | -0.25 | 21.579428 |  1.026136 |              6.887397 |                  1.279088 |                   6.887397 |                   0.60663 |                      0.0 |     6.280768 |              0.0 |          6.280768 |             44.761905 |                 100.0 |               696.001395 |              147.247892 |    0.0 |  0.0 |  3.397063 |  1.88 | 0.698098 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.153291 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.917693 | 12.794413 |      0.0 |   0.0 |   0.0 |   0.0 | -1.153291 | 0.000071 |   0.0 |   0.0 |   0.0 | -0.012081 |   0.0 |      0.0 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           0.532486 |          1.5 |                  2.38585 |                       2.38585 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | -0.25 | 17.325643 | -1.541907 |             13.073322 |                  2.427903 |                  13.073322 |                  1.641158 |                      0.0 |    11.432163 |              0.0 |         11.432163 |             62.666667 |                 100.0 |               667.674311 |              145.668066 |    0.0 |  0.0 |  4.351914 | 0.282 | 2.206487 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.125007 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   33.2096 | 11.712934 |      0.0 |   0.0 |   0.0 |   0.0 | -1.125007 | 0.000038 |   0.0 |   0.0 |   0.0 | -0.016118 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           0.780979 |          2.2 |                 3.006626 |                      3.006626 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | -0.25 | 21.631999 | -1.279412 |             16.464715 |                  3.057733 |                  16.464715 |                  1.868809 |                      0.0 |    14.595906 |              0.0 |         14.595906 |             42.727273 |                 100.0 |               938.453811 |              150.290353 |    0.0 |  0.0 |  7.236094 |   0.0 | 3.494179 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.192206 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 34.239011 |  9.410961 |      0.0 |   0.0 |   0.0 |   0.0 | -1.192206 |  0.00002 |   0.0 |   0.0 |   0.0 | -0.020212 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                  2.93526 |                       2.93526 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | -0.25 | 25.063372 |  0.706772 |              12.69102 |                  2.356904 |                   12.69102 |                  1.401603 |                      0.0 |    11.289418 |              0.0 |         11.289418 |             49.473684 |                 100.0 |              1353.717528 |               153.70968 |    0.0 |  0.0 |  5.830325 | 0.094 | 2.901447 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.335241 |      0.0 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  33.28224 |  7.844755 |      0.0 |   0.0 |   0.0 |   0.0 | -1.335241 | 0.000011 |   0.0 |   0.0 |   0.0 | -0.024667 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |           0.816478 |          2.3 |                 2.506318 |                      2.506318 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | -0.25 | 21.874132 |   0.40131 |             11.099578 |                   2.06135 |                  11.099578 |                  1.261689 |                      0.0 |     9.837889 |              0.0 |          9.837889 |             40.869565 |                 100.0 |              1694.004243 |              154.756666 |   7.67 |  0.0 |  4.656548 |   2.2 | 1.080941 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   7.67 | 0.067757 |  0.0 |  0.0 | 0.0 | 1.432559 | 0.067757 |      0.0 |      0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.63093 | 15.798616 | 0.067757 |   0.0 |   0.0 |   0.0 | -1.432559 |  0.01745 |   0.0 |   0.0 |   0.0 | -0.029537 |   0.0 |      0.0 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |            0.74548 |          2.1 |                 2.475714 |                      2.475714 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | -0.25 | 19.703098 | -0.620474 |             12.586633 |                  2.337518 |                  12.586633 |                  1.498964 |                      0.0 |    11.087669 |              0.0 |         11.087669 |             44.761905 |                 100.0 |               514.362016 |              137.508257 |   3.91 |  0.0 |  4.603863 |   2.2 | 1.172642 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   3.91 | 0.052955 |  0.0 |  0.0 | 0.0 | 0.938338 | 0.052955 | 0.003425 |  0.00004 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 33.001403 | 19.421358 | 0.052955 |   0.0 |   0.0 |   0.0 | -0.938338 | 0.037102 |   0.0 |   0.0 |   0.0 | -0.033677 |   0.0 |  0.00004 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           0.922975 |          2.6 |                 2.371153 |                      2.371153 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | -0.25 | 20.938011 |  0.382585 |                6.6675 |                   1.23825 |                     6.6675 |                  0.589077 |                      0.0 |     6.078423 |              0.0 |          6.078423 |             36.153846 |                 100.0 |               423.391318 |               132.52466 |   11.9 |  0.0 |  3.480236 |   2.2 | 0.567552 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   11.9 |   0.2321 |  0.0 |  0.0 | 0.0 | 0.713236 |   0.2321 | 0.054101 | 0.000626 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.368818 | 31.234942 |   0.2321 |   0.0 |   0.0 |   0.0 | -0.713236 | 0.090616 |   0.0 |   0.0 |   0.0 | -0.036515 |   0.0 | 0.000626 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           0.922975 |          2.6 |                 1.832679 |                      1.832679 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | -0.25 | 18.829395 |  1.414918 |              6.413823 |                  1.191138 |                   6.413823 |                  0.708412 |                      0.0 |     5.705411 |              0.0 |          5.705411 |             36.153846 |                 100.0 |               347.072847 |              127.000616 |  6.448 |  0.0 |  2.685024 |   2.2 | 0.203792 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  6.448 | 0.173145 |  0.0 |  0.0 | 0.0 |      0.0 | 0.173145 | 0.103764 | 0.001201 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |   30.7039 | 37.306005 | 0.173145 |   0.0 |   0.0 |   0.0 |       0.0 | 0.141425 |   0.0 |   0.0 |   0.0 | -0.037661 |   0.0 | 0.001201 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |            0.74548 |          2.1 |                 1.940693 |                      1.940693 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | -0.25 | 13.279667 |  -1.92862 |             10.820053 |                  2.009438 |                  10.820053 |                  1.881434 |                      0.0 |      8.93862 |              0.0 |           8.93862 |             44.761905 |                 100.0 |               346.875642 |              126.428035 |    0.0 |  0.0 |  4.533079 |   0.0 | 2.160946 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.074117 | 0.000858 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.382521 | 35.145059 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.111643 |   0.0 |   0.0 |   0.0 | -0.037526 |   0.0 | 0.000858 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |            0.74548 |          2.1 |                 2.093348 |                      2.093348 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | -0.25 | 18.875069 |  0.349956 |              9.634933 |                  1.789345 |                   9.634933 |                  1.511299 |                      0.0 |     8.123635 |              0.0 |          8.123635 |             44.761905 |                 100.0 |               345.936719 |              125.805103 |    0.0 |  0.0 |   4.43981 |   0.0 | 2.165797 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.022183 | 0.000257 |  0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.782565 | 32.979262 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.059574 |   0.0 |   0.0 |   0.0 | -0.037392 |   0.0 | 0.000257 |

.. _lland_v3_acker_winter_daily:

acre (winter)
-------------

None of the examples discussed so far dealt with snow processes.  Thus,
we now switch from a warm August to a snowy December and adapt the initial
conditions and the input time series accordingly:

>>> pub.timegrids = '2010-12-01', '2011-01-01', '1d'
>>> lnk(ACKER)
>>> lai.acker_dec = 0.3
>>> cropheight.acker_dec = 0.05
>>> albedo.acker_dec = 0.1
>>> wg2z.dec = 0.7416
>>> test = IntegrationTest(land)
>>> test.dateformat = '%Y-%d-%m'

>>> test.inits = (
...     (states.inzp, 0.0),
...     (states.wats, 0.0),
...     (states.waes, 0.0),
...     (states.esnow, 0.0),
...     (states.taus, 0.0),
...     (states.ebdn, 0.0),
...     (states.bowa, 72.0),
...     (states.qdgz1, 0.0),
...     (states.qdgz2, 0.0),
...     (states.qigz1, 0.0),
...     (states.qigz2, 0.0),
...     (states.qbgz, 0.0),
...     (states.qdga1, 0.0),
...     (states.qdga2, 0.0),
...     (states.qiga1, 0.0),
...     (states.qiga2, 0.0),
...     (states.qbga, 0.0))

>>> inputs.nied.series = (
...     0.0, 11.1, 0.4, 1.2, 6.4, 0.0, 0.0, 0.0, 0.7, 3.3, 9.4, 0.2, 2.9, 0.0,
...     0.0, 6.7, 0.0, 0.1, 0.0, 0.2, 0.2, 2.4, 7.0, 0.2, 0.0, 1.2, 0.9, 0.0,
...     0.0, 0.2, 0.9)
>>> inputs.teml.series = (
...     -5.8, -3.0, -5.6, -3.8, 0.2, -1.1, -2.9, -2.9, -2.1, -3.1, 5.1, 0.2,
...     -3.5, -3.9, -6.3, -3.5, -7.3, -5.8, -7.9, -5.7, -5.6, -4.8, -1.1,
...     -2.9, -9.1, -8.6, -5.3, -7.7, -9.2, -7.4, 1.7)
>>> inputs.sunshineduration.series = (
...     3.5, 0.0, 0.1, 3.7, 0.0, 0.1, 0.0, 0.0, 4.4, 0.1, 0.2, 0.7, 0.0, 1.2,
...     3.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.8, 0.0, 0.0, 0.0, 3.2, 4.3, 0.0, 3.4,
...     0.0, 0.0, 4.0)
>>> inputs.windspeed.series = (
...     6.9, 5.2, 2.0, 4.3, 4.0, 2.4, 1.9, 1.8, 2.7, 3.0, 7.4, 5.7, 2.2, 3.1,
...     2.8, 3.7, 0.9, 1.3, 1.9, 2.4, 2.3, 4.6, 8.6, 7.6, 3.8, 1.5, 1.3, 3.0,
...     2.7, 1.4, 4.2)
>>> inputs.relativehumidity.series = (
...     70.3, 85.0, 95.9, 90.6, 96.5, 99.0, 99.2, 99.1, 96.9, 95.0, 90.6,
...     78.7, 90.5, 88.1, 92.8, 98.1, 96.6, 97.7, 95.8, 97.5, 95.0, 93.8,
...     94.9, 85.4, 87.2, 92.7, 97.8, 95.4, 94.9, 96.0, 96.5)
>>> inputs.atmosphericpressure.series = (
...     101.5, 100.7, 100.5, 100.8, 99.6, 99.8, 100.4, 100.6, 100.9, 101.8,
...     100.4, 101.2, 102.0, 102.5, 102.3, 100.0, 99.1, 99.7, 99.7, 99.7,
...     100.4, 100.7, 100.4, 100.8, 101.8, 101.9, 101.7, 102.2, 102.4, 102.1,
...     101.6)

When comparing the following results with those of integration test
:ref:`lland_v1_acker_winter` on |lland_v1|, it is not immediately clear what
differences are due to the much higher complexities of the snow module of
|lland_v3|.  At least, one can see that |lland_v3| allows for the (re)freezing
of liquid water within the snow-layer.  Refreezing  occurs around December 7th,
where the frozen water equivalent (|WATS|) grows until it reaches to total
water equivalent (|WAeS|) (a note for `LARSIM`_ users: currently, `LARSIM`_
does not implement such a freezing feature; you can disable it in |lland_v3|
through setting |RefreezeFlag| to |False|):

.. integration-test::

    >>> conditions_acker_winter = test(
    ...     'lland_v3_acker_winter_daily',
    ...     axis1=(inputs.nied, fluxes.wada),
    ...     axis2=(states.waes, states.wats),
    ...     get_conditions='2010-12-10')
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |      g |         tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes |     sbes |      evpo |       evi |     evb |       evs |     wnied | tempssurface | actualalbedo |   schmpot |      schm |  gefrpot |     gefr |  wlatsnow | wsenssnow |     wsurf |      sff |      fvg |      wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz | qah |  qa |     inzp |      wats |      waes |     esnow |     taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet | outlet |
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-01-12 |  0.0 | -5.8 |              3.5 |             70.3 |       6.9 |               101.5 | nan | nan |    -5.8 |                  70.3 |                   3.5 |    0.0 | -5.8 |    -5.8 |     5.53098 |          5.53098 |            5.53098 |          6.9 |                 0.395811 |                      0.395811 |                          0.0 |                      0.030356 |                           0.030356 |             0.278255 |                  0.278255 |     101.221745 |   1.321192 |        -0.389384 | 8.532819 | 16.173848 |         1.028719 |                  5.761386 |                 7.641029 |                      7.641029 |             2.759858 | 100.0 |        2.759858 |                2.759858 | 0.7416 |        0.0 |   3.00672 |              2.483872 |                  2.483872 |                   2.483872 |                   4.01525 |                      0.0 |    -1.531378 |              0.0 |         -1.531378 |             30.870386 |                 100.0 |               111.309423 |               99.512695 |      0.0 |      0.0 |  1.808179 |       0.0 | 0.55374 |       0.0 |       0.0 |          nan |          0.1 |       0.0 |       0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  -2.26512 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-02-12 | 11.1 | -3.0 |              0.0 |             85.0 |       5.2 |               100.7 | nan | nan |    -3.0 |                  85.0 |                   0.0 | 10.434 | -3.0 |    -3.0 |    4.168274 |         4.168274 |           4.168274 |          5.2 |                 0.489349 |                      0.489349 |                     0.381881 |                      0.036626 |                           0.036626 |             0.415947 |                  0.415947 |     100.284053 |   1.296511 |        -0.391536 | 8.548511 | 16.158156 |         1.028994 |                  5.687365 |                 7.609644 |                      7.609644 |             0.853105 | 100.0 |        0.853105 |                0.853105 | 0.7416 |    -7.5504 | -0.665713 |              0.106229 |                  0.106229 |                   0.106229 |                  1.277835 |                 2.406897 |    -1.171606 |        -2.300668 |         -1.171606 |             40.962628 |                 100.0 |               112.102145 |               99.530051 |   10.374 |   10.374 |  0.779906 |      0.06 |     0.0 | -0.168357 | -0.065045 |    -6.266232 |     0.875479 |       0.0 |       0.0 | 0.406773 |      0.0 | -0.449083 |  -2.44648 | -0.594896 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 10.542357 | 10.542357 | -0.135862 | 0.084415 | -0.857807 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-03-12 |  0.4 | -5.6 |              0.1 |             95.9 |       2.0 |               100.5 | nan | nan |    -5.6 |                  95.9 |                   0.1 |  0.376 | -5.6 |    -5.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 0.401924 |                      0.401924 |                     0.357079 |                      0.030771 |                           0.030771 |             0.385445 |                  0.385445 |     100.114555 |   1.306656 |        -0.393572 | 8.563421 | 16.143245 |          1.02926 |                  5.617593 |                 7.579824 |                      7.579824 |             1.441454 | 100.0 |        1.441454 |                1.441454 | 0.7416 |  -2.859357 |  2.216462 |              0.200706 |                  0.200706 |                   0.200706 |                  1.329454 |                  3.24186 |    -1.128748 |        -3.041154 |         -1.128748 |            106.502833 |                 100.0 |               112.102145 |               99.527226 |    0.316 |    0.316 |  0.039575 |  0.039575 |     0.0 | -0.073822 | -0.003698 |    -7.134939 |     0.860761 |       0.0 |       0.0 | 0.484791 |      0.0 | -0.196916 | -0.605417 |  2.238821 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.020425 | 10.932178 | 10.932178 |  -0.16192 | 0.142292 | -2.332669 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-04-12 |  1.2 | -3.8 |              3.7 |             90.6 |       4.3 |               100.8 | nan | nan |    -3.8 |                  90.6 |                   3.7 |  1.128 | -3.8 |    -3.8 |    3.446842 |         3.446842 |           3.446842 |          4.3 |                 0.460812 |                      0.460812 |                     0.323612 |                       0.03473 |                            0.03473 |             0.417496 |                  0.417496 |     100.382504 |   1.301648 |        -0.395492 | 8.577535 | 16.129131 |         1.029518 |                  5.552051 |                 7.551596 |                      7.551596 |             2.748162 | 100.0 |        2.748162 |                2.748162 | 0.7416 |  -7.775563 |  0.321699 |              0.409591 |                  0.409591 |                   0.409591 |                  3.727317 |                 4.828749 |    -3.317726 |        -4.419159 |         -3.317726 |             49.536202 |                 100.0 |               112.102145 |               99.524551 | 1.088425 | 1.088425 |  0.092184 |      0.06 |     0.0 | -0.402205 | -0.008644 |    -8.396125 |     0.850958 |       0.0 |       0.0 | 0.631546 |      0.0 | -1.072863 | -2.984225 |  0.362071 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 12.422809 | 12.422809 | -0.210936 | 0.184396 | -1.912768 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-05-12 |  6.4 |  0.2 |              0.0 |             96.5 |       4.0 |                99.6 | nan | nan |     0.2 |                  96.5 |                   0.0 |  6.016 |  0.2 |     0.2 |    3.206365 |         3.206365 |           3.206365 |          4.0 |                 0.619748 |                      0.619748 |                     0.491269 |                      0.045128 |                           0.045128 |             0.598057 |                  0.598057 |      99.001943 |   1.266438 |        -0.397294 | 8.590839 | 16.115828 |         1.029767 |                  5.490726 |                 7.524989 |                      7.524989 |             0.823609 | 100.0 |        0.823609 |                0.823609 | 0.7416 |  -6.375893 | -1.777198 |              0.126436 |                  0.126436 |                   0.126436 |                  1.140917 |                 1.372579 |    -1.014481 |        -1.246143 |         -1.014481 |             53.251417 |                 100.0 |               112.102145 |               99.522031 |    5.956 |   2.6802 |  0.134809 |      0.06 |     0.0 |  -0.43406 |  0.003863 |    -2.947657 |     0.846485 |       0.0 |       0.0 | 0.399614 | 0.399614 | -1.157835 | -1.939108 |   -1.8508 | 0.977319 | 0.286546 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 15.868755 | 18.812869 |       0.0 | 0.204657 |   0.60603 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-06-12 |  0.0 | -1.1 |              0.1 |             99.0 |       2.4 |                99.8 | nan | nan |    -1.1 |                  99.0 |                   0.1 |    0.0 | -1.1 |    -1.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 0.563475 |                      0.563475 |                     0.515142 |                      0.041489 |                           0.041489 |             0.557841 |                  0.557841 |      99.242159 |   1.275245 |        -0.398979 | 8.603319 | 16.103347 |         1.030007 |                  5.433602 |                 7.500028 |                      7.500028 |             1.394624 | 100.0 |        1.394624 |                1.394624 | 0.7416 |        0.0 |  1.198359 |              0.260442 |                  0.260442 |                   0.260442 |                  1.217715 |                 2.425289 |    -0.957272 |        -2.164847 |         -0.957272 |             88.752361 |                 100.0 |               112.102145 |               99.519666 |      0.0 |      0.0 |  0.000623 |       0.0 |     0.0 | -0.123607 |       0.0 |    -2.311649 |     0.813253 |       0.0 |       0.0 | 0.314856 | 0.314856 | -0.329716 |  -0.53161 |  1.303521 | 0.994414 | 0.296657 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 16.289943 | 18.936477 |       0.0 | 0.380049 |  0.149271 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-07-12 |  0.0 | -2.9 |              0.0 |             99.2 |       1.9 |               100.4 | nan | nan |    -2.9 |                  99.2 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 0.493024 |                      0.493024 |                     0.462335 |                      0.036869 |                           0.036869 |             0.489079 |                  0.489079 |      99.910921 |   1.291808 |        -0.400545 | 8.614964 | 16.091703 |         1.030238 |                  5.380663 |                 7.476739 |                      7.476739 |             0.807099 | 100.0 |        0.807099 |                0.807099 | 0.7416 |        0.0 |  1.947227 |              0.171497 |                  0.171497 |                   0.171497 |                  1.193664 |                 2.796896 |    -1.022168 |          -2.6254 |         -1.022168 |            112.108246 |                 100.0 |               112.102145 |                99.51746 |      0.0 |      0.0 | -0.015896 | -0.015896 |     0.0 | -0.067645 |       0.0 |    -3.756225 |     0.787515 |       0.0 |       0.0 | 0.507501 | 0.507501 |  -0.18044 | -0.328227 |  2.116732 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.015896 | 16.857448 | 19.004122 |       0.0 | 0.555441 | -1.056356 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-08-12 |  0.0 | -2.9 |              0.0 |             99.1 |       1.8 |               100.6 | nan | nan |    -2.9 |                  99.1 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 0.493024 |                      0.493024 |                     0.422308 |                      0.036869 |                           0.036869 |             0.488586 |                  0.488586 |     100.111414 |   1.294388 |        -0.401992 | 8.625761 | 16.080906 |         1.030461 |                  5.331896 |                 7.455145 |                      7.455145 |             0.799784 | 100.0 |        0.799784 |                0.799784 | 0.7416 |  -3.521188 |  0.741715 |              0.186355 |                  0.186355 |                   0.186355 |                  1.194208 |                 2.345156 |    -1.007852 |        -2.158801 |         -1.007852 |            118.336481 |                 100.0 |               112.102145 |               99.515415 |      0.0 |      0.0 |  -0.01263 |  -0.01263 |     0.0 | -0.162793 |       0.0 |    -4.951965 |     0.766993 |       0.0 |       0.0 | 0.655613 | 0.655613 | -0.434243 | -0.763868 |   0.96069 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.028526 | 17.663081 | 19.166915 |       0.0 | 0.730833 | -1.056472 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-09-12 |  0.7 | -2.1 |              4.4 |             96.9 |       2.7 |               100.9 | nan | nan |    -2.1 |                  96.9 |                   4.4 |  0.658 | -2.1 |    -2.1 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 0.523311 |                      0.523311 |                     0.389674 |                      0.038864 |                           0.038864 |             0.507088 |                  0.507088 |     100.392912 |   1.294334 |        -0.403321 | 8.635699 | 16.070967 |         1.030674 |                  5.287288 |                 7.435268 |                      7.435268 |             2.886262 | 100.0 |        2.886262 |                2.886262 | 0.7416 |  -3.521572 |  1.286683 |              0.708095 |                  0.708095 |                   0.708095 |                  4.005891 |                 5.077157 |    -3.297796 |        -4.369062 |         -3.297796 |             78.890988 |                 100.0 |               112.102145 |               99.513532 | 0.626526 | 0.626526 | -0.234885 | -0.234885 |     0.0 | -0.365657 |  -0.00275 |    -6.003598 |     0.754667 |       0.0 |       0.0 | 0.800268 | 0.800268 | -0.975372 | -1.842468 |  1.551222 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.294885 | 19.442535 | 20.159098 |       0.0 | 0.856575 | -1.601554 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-10-12 |  3.3 | -3.1 |              0.1 |             95.0 |       3.0 |               101.8 | nan | nan |    -3.1 |                  95.0 |                   0.1 |  3.102 | -3.1 |    -3.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 0.485698 |                      0.485698 |                     0.402618 |                      0.036384 |                           0.036384 |             0.461413 |                  0.461413 |     101.338587 |   1.310959 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                  7.41713 |                       7.41713 |             1.347076 | 100.0 |        1.347076 |                1.347076 | 0.7416 |  -5.338514 |  0.123879 |              0.312643 |                  0.312643 |                   0.312643 |                  1.286731 |                 2.471706 |    -0.974088 |        -2.159063 |         -0.974088 |             71.001889 |                 100.0 |               112.102145 |               99.511814 | 3.336885 | 3.336885 |  0.136846 |      0.06 |     0.0 | -0.196001 |  -0.02162 |    -5.577477 |      0.76791 |       0.0 |       0.0 | 0.845123 | 0.716563 | -0.522823 |  -1.25171 |   0.38453 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 23.691984 | 23.691984 | -0.042939 | 0.722158 | -0.983833 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-11-12 |  9.4 |  5.1 |              0.2 |             90.6 |       7.4 |               100.4 | nan | nan |     5.1 |                  90.6 |                   0.2 |  8.836 |  5.1 |     5.1 |    5.931775 |         5.931775 |           5.931775 |          7.4 |                 0.879017 |                      0.879017 |                      0.61078 |                      0.061412 |                           0.061412 |              0.79639 |                   0.79639 |       99.60361 |   1.253213 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 7.400749 |                      7.400749 |             1.373029 | 100.0 |        1.373029 |                1.373029 | 0.7416 |  -3.279442 | -1.700063 |              0.343097 |                  0.343097 |                   0.343097 |                   1.16859 |                -0.065197 |    -0.825492 |         0.408294 |         -0.825492 |              28.78455 |                 100.0 |               112.102145 |               99.510262 |    8.776 |      0.0 |  1.032578 |      0.06 |     0.0 | -1.215856 |  0.187391 |          0.0 |     0.750116 | 21.434861 | 21.434861 |      0.0 |      0.0 | -3.243236 | -5.063324 | -8.714854 |  0.94544 | 0.268157 | 30.335284 | 9.737948 |  0.0 |  0.0 | 0.0 |  0.0 | 9.737948 | nan | nan |      0.0 |  2.341648 |  3.348556 |       0.0 | 0.907745 |   1.45783 | 92.043596 | 9.737948 |   0.0 |   0.0 |   0.0 |  0.0 | 2.507002 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 |  0.2 |  0.2 |              0.7 |             78.7 |       5.7 |               101.2 | nan | nan |     0.2 |                  78.7 |                   0.7 |  0.188 |  0.2 |     0.2 |     4.56907 |          4.56907 |            4.56907 |          5.7 |                 0.619748 |                      0.619748 |                      0.51762 |                      0.045128 |                           0.045128 |             0.487741 |                  0.487741 |     100.712259 |    1.28736 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 7.386143 |                      7.386143 |             1.539952 | 100.0 |        1.539952 |                1.539952 | 0.7416 |        0.0 |  1.164919 |              0.405635 |                  0.405635 |                   0.405635 |                  1.733999 |                 3.137107 |    -1.328364 |        -2.731472 |         -1.328364 |             37.369415 |                 100.0 |                83.924464 |               98.591233 |    0.128 |   0.0576 |  1.429059 |      0.06 |     0.0 |  0.158585 |  0.000045 |    -2.247143 |     0.736593 |       0.0 |       0.0 | 0.062798 | 0.062798 |  0.423016 | -1.968551 |  1.185938 | 0.961283 |  0.27722 |  0.045632 | 0.015236 |  0.0 |  0.0 | 0.0 |  0.0 | 0.015236 | nan | nan |      0.0 |  2.348245 |   3.27234 |       0.0 | 1.077936 |  1.034511 | 92.073992 | 0.015236 |   0.0 |   0.0 |   0.0 |  0.0 | 3.376356 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-13-12 |  2.9 | -3.5 |              0.0 |             90.5 |       2.2 |               102.0 | nan | nan |    -3.5 |                  90.5 |                   0.0 |  2.726 | -3.5 |    -3.5 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 0.471336 |                      0.471336 |                     0.432015 |                      0.035431 |                           0.035431 |             0.426559 |                  0.426559 |     101.573441 |   1.315658 |        -0.407436 | 8.666669 | 16.039997 |         1.031437 |                  5.150203 |                 7.373328 |                      7.373328 |              0.77253 | 100.0 |         0.77253 |                 0.77253 | 0.7416 |        0.0 |  2.412148 |              0.192965 |                  0.192965 |                   0.192965 |                  1.254261 |                 3.092729 |    -1.061296 |        -2.899764 |         -1.061296 |             96.820757 |                 100.0 |                83.884383 |               98.589894 |    2.666 |    2.666 |  0.193851 |      0.06 |     0.0 |  0.014995 | -0.019502 |    -4.653064 |     0.750217 |       0.0 |       0.0 | 0.199897 | 0.199897 |  0.039999 | -0.480351 |  2.459412 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |  5.200976 |  5.923345 |       0.0 | 0.906585 | -0.636037 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 1.804844 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-14-12 |  0.0 | -3.9 |              1.2 |             88.1 |       3.1 |               102.5 | nan | nan |    -3.9 |                  88.1 |                   1.2 |    0.0 | -3.9 |    -3.9 |    2.484933 |         2.484933 |           2.484933 |          3.1 |                 0.457351 |                      0.457351 |                     0.383836 |                      0.034499 |                           0.034499 |             0.402926 |                  0.402926 |     102.097074 |   1.324197 |        -0.408164 | 8.672175 | 16.034492 |         1.031605 |                  5.126218 |                 7.362317 |                      7.362317 |             1.699321 | 100.0 |        1.699321 |                1.699321 | 0.7416 |  -2.120123 |  2.114956 |              0.448113 |                  0.448113 |                   0.448113 |                  2.107297 |                  4.01017 |    -1.659183 |        -3.562057 |         -1.659183 |             68.711505 |                 100.0 |                83.884383 |               98.590228 |      0.0 |      0.0 |  0.281575 |       0.0 |     0.0 | -0.065035 |       0.0 |    -6.199899 |     0.736299 |       0.0 |       0.0 | 0.257917 | 0.257917 | -0.173479 | -1.187477 |  2.201101 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |  5.518828 |   5.98838 |       0.0 | 1.081977 | -2.009393 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.963085 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-15-12 |  0.0 | -6.3 |              3.0 |             92.8 |       2.8 |               102.3 | nan | nan |    -6.3 |                  92.8 |                   3.0 |    0.0 | -6.3 |    -6.3 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 0.380889 |                      0.380889 |                     0.282602 |                       0.02934 |                            0.02934 |             0.353465 |                  0.353465 |     101.946535 |    1.33374 |         -0.40877 | 8.676772 | 16.029895 |         1.031763 |                  5.106331 |                 7.353123 |                      7.353123 |             2.318249 | 100.0 |        2.318249 |                2.318249 | 0.7416 |  -6.697977 |  1.768774 |              0.638578 |                  0.638578 |                   0.638578 |                  3.397964 |                 4.986818 |    -2.759386 |        -4.348241 |         -2.759386 |             76.073452 |                 100.0 |                83.884383 |               98.590507 |      0.0 |      0.0 |   -0.0286 |   -0.0286 |     0.0 | -0.225866 |       0.0 |   -10.109965 |     0.724543 |       0.0 |       0.0 | 0.408644 | 0.408644 | -0.602486 | -1.840493 |  1.905262 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |   0.0286 |  6.151042 |  6.214246 |       0.0 | 1.257369 | -3.036568 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.513913 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-16-12 |  6.7 | -3.5 |              0.0 |             98.1 |       3.7 |               100.0 | nan | nan |    -3.5 |                  98.1 |                   0.0 |  6.298 | -3.5 |    -3.5 |    2.965888 |         2.965888 |           2.965888 |          3.7 |                 0.471336 |                      0.471336 |                     0.357863 |                      0.035431 |                           0.035431 |             0.462381 |                  0.462381 |      99.537619 |   1.289645 |        -0.409256 | 8.680456 | 16.026211 |         1.031912 |                  5.090537 |                 7.345755 |                      7.345755 |             0.763581 | 100.0 |        0.763581 |                0.763581 | 0.7416 | -10.121892 | -1.563101 |              0.164206 |                  0.164206 |                   0.164206 |                  1.211777 |                  1.91563 |    -1.047571 |        -1.751423 |         -1.047571 |             57.569099 |                 100.0 |                83.884383 |               98.590731 |   6.2666 |   6.2666 |  0.035323 |  0.035323 |     0.0 |  -0.40191 |  -0.04584 |    -7.106651 |     0.784952 |       0.0 |       0.0 |  0.55784 | 0.063204 | -1.072074 | -2.101972 | -1.422622 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.024677 | 12.882755 | 12.882755 | -0.165208 | 0.575375 | -0.731867 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.274229 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-17-12 |  0.0 | -7.3 |              0.0 |             96.6 |       0.9 |                99.1 | nan | nan |    -7.3 |                  96.6 |                   0.0 |    0.0 | -7.3 |    -7.3 |    0.721432 |         0.721432 |           0.721432 |          0.9 |                 0.352531 |                      0.352531 |                     0.326227 |                      0.027395 |                           0.027395 |             0.340545 |                  0.340545 |      98.759455 |   1.296889 |         -0.40962 | 8.683222 | 16.023444 |         1.032051 |                  5.078832 |                 7.340222 |                      7.340222 |             0.761825 | 100.0 |        0.761825 |                0.761825 | 0.7416 |  -2.439556 |  3.034693 |              0.169472 |                  0.169472 |                   0.169472 |                  1.287913 |                   3.6017 |     -1.11844 |        -3.432228 |          -1.11844 |            236.672963 |                 100.0 |                83.884383 |               98.590899 |      0.0 |      0.0 | -0.001233 | -0.001233 |     0.0 | -0.025746 |       0.0 |    -8.293516 |     0.777544 |       0.0 |       0.0 | 0.668571 |      0.0 | -0.068677 | -0.270764 |  3.092787 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  0.02591 | 12.908502 | 12.908502 | -0.223303 | 0.635987 | -3.024959 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.146332 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-18-12 |  0.1 | -5.8 |              0.0 |             97.7 |       1.3 |                99.7 | nan | nan |    -5.8 |                  97.7 |                   0.0 |  0.094 | -5.8 |    -5.8 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 0.395811 |                      0.395811 |                     0.283779 |                      0.030356 |                           0.030356 |             0.386708 |                  0.386708 |      99.313292 |   1.297204 |        -0.409863 | 8.685068 | 16.021598 |         1.032182 |                  5.071211 |                  7.33653 |                       7.33653 |             0.760682 | 100.0 |        0.760682 |                0.760682 | 0.7416 | -10.083198 | -0.013178 |              0.173382 |                  0.173382 |                   0.173382 |                  1.257034 |                 2.131881 |    -1.083652 |          -1.9585 |         -1.083652 |            163.850513 |                 100.0 |                83.884383 |               98.591011 |  0.05991 |  0.05991 | -0.002775 | -0.002775 |     0.0 | -0.215187 | -0.000726 |   -10.057778 |      0.77207 |       0.0 |       0.0 | 0.816185 |      0.0 |    -0.574 | -1.349101 |  0.035399 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.062775 | 13.183599 | 13.183599 | -0.272606 | 0.683858 | -2.270182 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.078084 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-19-12 |  0.0 | -7.9 |              0.0 |             95.8 |       1.9 |                99.7 | nan | nan |    -7.9 |                  95.8 |                   0.0 |    0.0 | -7.9 |    -7.9 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  0.33643 |                       0.33643 |                     0.272284 |                      0.026283 |                           0.026283 |               0.3223 |                    0.3223 |        99.3777 |   1.307793 |        -0.409985 | 8.685992 | 16.020675 |         1.032302 |                  5.067673 |                 7.334683 |                      7.334683 |             0.760151 | 100.0 |        0.760151 |                0.760151 | 0.7416 |  -7.567272 |  1.559622 |              0.177036 |                  0.177036 |                   0.177036 |                  1.301361 |                 3.118663 |    -1.124325 |        -2.941627 |         -1.124325 |            112.108246 |                 100.0 |                83.884383 |               98.591067 | 0.002775 | 0.002775 |   0.03257 |   0.03257 |     0.0 | -0.126504 | -0.000046 |   -10.575801 |     0.767104 |       0.0 |       0.0 | 0.872647 |      0.0 | -0.337444 | -1.025749 |  1.578434 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  0.02743 | 13.312878 | 13.312878 | -0.291464 | 0.729776 | -3.088203 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.041667 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-20-12 |  0.2 | -5.7 |              0.2 |             97.5 |       2.4 |                99.7 | nan | nan |    -5.7 |                  97.5 |                   0.2 |  0.188 | -5.7 |    -5.7 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 0.398857 |                      0.398857 |                     0.293059 |                      0.030563 |                           0.030563 |             0.388886 |                  0.388886 |      99.311114 |   1.296708 |        -0.409985 | 8.685992 | 16.020675 |         1.032413 |                  5.068219 |                 7.334683 |                      7.334683 |             1.336154 | 100.0 |        1.336154 |                1.336154 | 0.7416 | -10.294011 | -0.332318 |              0.315755 |                  0.315755 |                   0.315755 |                  1.393316 |                 2.436172 |    -1.077561 |        -2.120417 |         -1.077561 |             88.752361 |                 100.0 |                83.884383 |               98.591067 |  0.15543 |  0.15543 |  0.021054 |  0.021054 |     0.0 | -0.277409 | -0.001852 |    -9.652965 |     0.763684 |       0.0 |       0.0 |  0.81353 |      0.0 | -0.739974 | -1.734358 | -0.353915 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.038946 | 13.745716 | 13.745716 | -0.271719 |  0.76289 | -2.014285 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.022234 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-21-12 |  0.2 | -5.6 |              0.8 |             95.0 |       2.3 |               100.4 | nan | nan |    -5.6 |                  95.0 |                   0.8 |  0.188 | -5.6 |    -5.6 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 0.401924 |                      0.401924 |                     0.312873 |                      0.030771 |                           0.030771 |             0.381828 |                  0.381828 |     100.018172 |   1.305372 |        -0.409863 | 8.685068 | 16.021598 |         1.032515 |                  5.072848 |                  7.33653 |                       7.33653 |             1.544792 | 100.0 |        1.544792 |                1.544792 | 0.7416 |  -6.714283 |  1.094405 |              0.370424 |                  0.370424 |                   0.370424 |                  1.820398 |                 3.346196 |    -1.449974 |        -2.975772 |         -1.449974 |             92.611159 |                 100.0 |                83.884383 |               98.591011 | 0.166946 | 0.166946 |  0.035079 |  0.035079 |     0.0 | -0.194575 | -0.001954 |    -8.825403 |     0.760211 |       0.0 |       0.0 | 0.768324 |      0.0 | -0.519019 |   -1.3794 |  1.077352 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.024921 | 14.107237 | 14.107237 |  -0.25662 | 0.797831 |  -2.36709 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.011864 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-22-12 |  2.4 | -4.8 |              0.0 |             93.8 |       4.6 |               100.7 | nan | nan |    -4.8 |                  93.8 |                   0.0 |  2.256 | -4.8 |    -4.8 |     3.68732 |          3.68732 |            3.68732 |          4.6 |                 0.427219 |                      0.427219 |                     0.344247 |                      0.032479 |                           0.032479 |             0.400731 |                  0.400731 |     100.299269 |   1.305282 |         -0.40962 | 8.683222 | 16.023444 |         1.032607 |                  5.081565 |                 7.340222 |                      7.340222 |             0.762235 | 100.0 |        0.762235 |                0.762235 | 0.7416 |  -7.890299 | -0.147589 |              0.172781 |                  0.172781 |                   0.172781 |                  1.259735 |                 2.599284 |    -1.086953 |        -2.426503 |         -1.086953 |              46.30558 |                 100.0 |                83.884383 |               98.590899 | 2.220921 | 2.220921 |  0.238308 |      0.06 |     0.0 | -0.254372 |  -0.02228 |    -7.605598 |     0.773322 |       0.0 |       0.0 | 0.777088 |      0.0 | -0.678525 | -1.914919 | -0.166942 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |  16.58253 |  16.58253 | -0.259548 | 0.672663 | -1.477901 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.006331 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-23-12 |  7.0 | -1.1 |              0.0 |             94.9 |       8.6 |               100.4 | nan | nan |    -1.1 |                  94.9 |                   0.0 |   6.58 | -1.1 |    -1.1 |    6.893685 |         6.893685 |           6.893685 |          8.6 |                 0.563475 |                      0.563475 |                     0.487686 |                      0.041489 |                           0.041489 |             0.534738 |                  0.534738 |      99.865262 |    1.28304 |        -0.409256 | 8.680456 | 16.026211 |         1.032689 |                  5.094372 |                 7.345755 |                      7.345755 |             0.764156 | 100.0 |        0.764156 |                0.764156 | 0.7416 |  -4.926336 | -0.975038 |              0.144842 |                  0.144842 |                   0.144842 |                  1.179808 |                 2.170767 |    -1.034966 |        -2.025925 |         -1.034966 |             24.768101 |                 100.0 |                83.884383 |               98.590731 |     6.52 |    5.053 |  0.461781 |      0.06 |     0.0 | -0.349504 | -0.018373 |    -3.045475 |     0.810455 |       0.0 |       0.0 | 0.468313 | 0.468313 | -0.932284 | -2.190183 | -1.096542 | 0.991065 | 0.294663 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 22.438239 | 23.452034 |       0.0 | 0.397177 |  0.238737 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.003378 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-24-12 |  0.2 | -2.9 |              0.0 |             85.4 |       7.6 |               100.8 | nan | nan |    -2.9 |                  85.4 |                   0.0 |  0.188 | -2.9 |    -2.9 |    6.092093 |         6.092093 |           6.092093 |          7.6 |                 0.493024 |                      0.493024 |                      0.45101 |                      0.036869 |                           0.036869 |             0.421042 |                  0.421042 |     100.378958 |   1.297296 |         -0.40877 | 8.676772 | 16.029895 |         1.032762 |                  5.111276 |                 7.353123 |                      7.353123 |             0.766691 | 100.0 |        0.766691 |                0.766691 | 0.7416 |        0.0 |  2.117638 |              0.164009 |                  0.164009 |                   0.164009 |                  1.273508 |                 3.158235 |    -1.109498 |        -2.994226 |         -1.109498 |             28.027061 |                 100.0 |                83.884383 |               98.590507 |    0.128 |    0.128 |  1.132583 |      0.06 |     0.0 |  0.200688 | -0.000776 |    -4.084949 |     0.786082 |       0.0 |       0.0 |  0.62874 |  0.62874 |  0.535324 | -1.202689 |  2.326861 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 22.997568 | 23.379346 |       0.0 | 0.566529 |   -1.1373 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001803 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-25-12 |  0.0 | -9.1 |              3.2 |             87.2 |       3.8 |               101.8 | nan | nan |    -9.1 |                  87.2 |                   3.2 |    0.0 | -9.1 |    -9.1 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                  0.30617 |                       0.30617 |                     0.260779 |                      0.024175 |                           0.024175 |             0.266981 |                  0.266981 |     101.533019 |   1.341718 |        -0.408164 | 8.672175 | 16.034492 |         1.032825 |                  5.132282 |                 7.362317 |                      7.362317 |             2.398433 | 100.0 |        2.398433 |                2.398433 | 0.7416 |  -3.791001 |  3.796323 |              0.561629 |                  0.561629 |                   0.561629 |                  3.730222 |                 6.170919 |    -3.168592 |         -5.60929 |         -3.168592 |             56.054123 |                 100.0 |                83.884383 |               98.590228 |      0.0 |      0.0 |  0.146367 |       0.0 |     0.0 | -0.024302 |       0.0 |   -11.114155 |     0.765835 |       0.0 |       0.0 | 1.652593 | 0.381779 | -0.064824 | -1.196177 |  4.348289 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 23.403648 | 23.403648 | -0.424452 | 0.741921 | -4.192023 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000962 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-26-12 |  1.2 | -8.6 |              4.3 |             92.7 |       1.5 |               101.9 | nan | nan |    -8.6 |                  92.7 |                   4.3 |  1.128 | -8.6 |    -8.6 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 0.318472 |                      0.318472 |                     0.181886 |                      0.025034 |                           0.025034 |             0.295223 |                  0.295223 |     101.604777 |   1.340358 |        -0.407436 | 8.666669 | 16.039997 |         1.032879 |                  5.157401 |                 7.373328 |                      7.373328 |             2.793204 | 100.0 |        2.793204 |                2.793204 | 0.7416 | -13.973411 |  0.795365 |               0.64551 |                   0.64551 |                    0.64551 |                  4.423241 |                 4.808458 |    -3.777731 |        -4.162948 |         -3.777731 |            142.003778 |                 100.0 |                83.884383 |               98.589894 |    1.068 |    1.068 |  -0.16832 |  -0.16832 |     0.0 |  -0.25352 | -0.019196 |   -15.507681 |       0.7689 |       0.0 |       0.0 | 2.374706 |      0.0 | -0.676251 | -2.341828 |  1.144869 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  0.22832 | 24.725168 | 24.725168 | -0.793152 | 0.712889 | -4.245789 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000513 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-27-12 |  0.9 | -5.3 |              0.0 |             97.8 |       1.3 |               101.7 | nan | nan |    -5.3 |                  97.8 |                   0.0 |  0.846 | -5.3 |    -5.3 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                  0.41125 |                       0.41125 |                     0.253812 |                      0.031402 |                           0.031402 |             0.402202 |                  0.402202 |     101.297798 |   1.320718 |        -0.406588 | 8.660262 | 16.046405 |         1.032922 |                  5.186639 |                 7.386143 |                      7.386143 |             0.777996 | 100.0 |        0.777996 |                0.777996 | 0.7416 | -14.152629 | -1.400795 |              0.176393 |                  0.176393 |                   0.176393 |                  1.247575 |                 1.363527 |    -1.071182 |        -1.187134 |         -1.071182 |            163.850513 |                 100.0 |                83.884383 |               98.589505 |  1.01432 |  1.01432 | -0.002431 | -0.002431 |     0.0 | -0.310229 | -0.011236 |   -11.450477 |     0.773273 |       0.0 |       0.0 | 1.844265 |      0.0 | -0.827519 | -1.948813 | -1.589198 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.062431 | 26.049716 | 26.049716 | -0.615984 | 0.673105 | -2.103393 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000274 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-28-12 |  0.0 | -7.7 |              3.4 |             95.4 |       3.0 |               102.2 | nan | nan |    -7.7 |                  95.4 |                   3.4 |    0.0 | -7.7 |    -7.7 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 0.341723 |                      0.341723 |                      0.25535 |                      0.026649 |                           0.026649 |             0.326003 |                  0.326003 |     101.873997 |   1.339598 |        -0.405619 | 8.652959 | 16.053708 |         1.032956 |                   5.22001 |                 7.400749 |                      7.400749 |             2.504073 | 100.0 |        2.504073 |                2.504073 | 0.7416 |  -7.011311 |  2.262411 |              0.579189 |                  0.579189 |                   0.579189 |                    3.6896 |                 5.330232 |    -3.110411 |        -4.751043 |         -3.110411 |             71.001889 |                 100.0 |                83.884383 |               98.589062 | 0.002431 | 0.002431 | -0.122138 | -0.122138 |     0.0 | -0.235528 | -0.000039 |   -11.375531 |     0.768701 |       0.0 |       0.0 | 1.854449 |      0.0 | -0.628259 |  -1.85701 |  2.265774 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.182138 | 26.287676 | 26.287676 | -0.619386 |  0.71474 | -3.624205 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000146 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-29-12 |  0.0 | -9.2 |              0.0 |             94.9 |       2.7 |               102.4 | nan | nan |    -9.2 |                  94.9 |                   0.0 |    0.0 | -9.2 |    -9.2 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 0.303761 |                      0.303761 |                     0.225287 |                      0.024006 |                           0.024006 |              0.28827 |                   0.28827 |      102.11173 |   1.350039 |         -0.40453 | 8.644768 | 16.061898 |         1.032981 |                  5.257524 |                  7.41713 |                       7.41713 |             0.788629 | 100.0 |        0.788629 |                0.788629 | 0.7416 | -12.080682 |  0.434734 |              0.185161 |                  0.185161 |                   0.185161 |                  1.325064 |                  2.98994 |    -1.139902 |        -2.804778 |         -1.139902 |             78.890988 |                 100.0 |                83.884383 |               98.588564 | 0.122138 | 0.122138 |  0.072845 |      0.06 |     0.0 | -0.196144 | -0.002348 |   -12.919289 |     0.765211 |       0.0 |       0.0 | 2.135029 |      0.0 | -0.523203 | -1.755476 |  0.526099 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 26.605957 | 26.605957 |   -0.7131 | 0.747953 | -3.317338 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000078 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-30-12 |  0.2 | -7.4 |              0.0 |             96.0 |       1.4 |               102.1 | nan | nan |    -7.4 |                  96.0 |                   0.0 |  0.188 | -7.4 |    -7.4 |    1.122228 |         1.122228 |           1.122228 |          1.4 |                 0.349801 |                      0.349801 |                     0.247537 |                      0.027207 |                           0.027207 |             0.335809 |                  0.335809 |     101.764191 |   1.336727 |        -0.403321 | 8.635699 | 16.070967 |         1.032995 |                  5.299194 |                 7.435268 |                      7.435268 |             0.794879 | 100.0 |        0.794879 |                0.794879 | 0.7416 | -11.057795 |  0.364297 |                0.1889 |                    0.1889 |                     0.1889 |                  1.292505 |                 2.434681 |    -1.103604 |        -2.245781 |         -1.103604 |            152.146905 |                 100.0 |                83.884383 |               98.588014 |    0.128 |    0.128 |  0.018911 |  0.018911 |     0.0 | -0.190998 |  -0.00198 |   -11.760528 |     0.762353 |       0.0 |       0.0 | 1.967387 |      0.0 | -0.509478 | -1.429978 |  0.306325 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.041089 | 26.924955 | 26.924955 | -0.657107 | 0.776116 | -2.940036 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000042 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-31-12 |  0.9 |  1.7 |              4.0 |             96.5 |       4.2 |               101.6 | nan | nan |     1.7 |                  96.5 |                   4.0 |  0.846 |  1.7 |     1.7 |    3.366683 |         3.366683 |           3.366683 |          4.2 |                 0.690794 |                      0.690794 |                     0.446978 |                      0.049663 |                           0.049663 |             0.666616 |                  0.666616 |     100.933384 |   1.284547 |        -0.401992 | 8.625761 | 16.080906 |            1.033 |                  5.345035 |                 7.455145 |                      7.455145 |             2.770177 | 100.0 |        2.770177 |                2.770177 | 0.7416 |  -9.800119 | -2.901162 |              0.668643 |                  0.668643 |                   0.668643 |                  3.475519 |                 3.601527 |    -2.806876 |        -2.932884 |         -2.806876 |             50.715635 |                 100.0 |                83.884383 |                98.58741 | 0.827089 | 0.062032 | -0.128392 | -0.128392 |     0.0 | -0.924887 |  0.005666 |    -4.203742 |     0.758628 |       0.0 |       0.0 | 0.750204 | 0.750204 | -2.467091 | -3.767829 | -3.302036 |   0.9737 | 0.284428 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.188392 | 28.661583 | 28.676932 |       0.0 | 0.814224 |  0.702726 | 92.073992 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000022 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |

.. _lland_v3_nadelw_winter_daily:

conifers (winter)
-----------------

Next, we repeat the integration test :ref:`lland_v3_acker_winter_daily` but
select the land use type |NADELW| instead of |ACKER|.  Like for |LAUBW|
and |MISCHW|, |lland_v3| now modifies the global radiation, the net longwave
radiation, and the wind speed to take the shadowing effects of the
tree-canopies into account.  Here, this net-effect of these modifications is
that the peak amounts of the total water equivalent are two to three times
smaller than for land use type |ACKER|:

.. integration-test::

    >>> lnk(NADELW)
    >>> conditions_nadelw_winter = test(
    ...     'lland_v3_nadelw_winter_daily',
    ...     axis1=(inputs.nied, fluxes.wada),
    ...     axis2=(states.waes, states.wats),
    ...     get_conditions='2010-12-10')
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |      g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes |     sbes |      evpo |       evi |      evb |       evs |     wnied | tempssurface | actualalbedo |  schmpot |     schm |  gefrpot |     gefr |  wlatsnow | wsenssnow |     wsurf |      sff |      fvg |      wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz | qah |  qa |     inzp |      wats |      waes |     esnow |     taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet | outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-01-12 |  0.0 | -5.8 |              3.5 |             70.3 |       6.9 |               101.5 | nan | nan |    -5.8 |                  70.3 |                   3.5 |    0.0 | -5.8 |    -5.8 |     5.53098 |          5.53098 |           2.449434 |          6.9 |                 0.395811 |                      0.395811 |                          0.0 |                      0.030356 |                           0.030356 |             0.278255 |                  0.278255 |     101.221745 |   1.321192 |        -0.389384 | 8.532819 | 16.173848 |         1.028719 |                  5.761386 |                 7.641029 |                      7.641029 |             2.759858 | 100.0 |        2.759858 |                2.759858 | 0.7416 |       0.0 |   3.00672 |              2.428675 |                   0.45104 |                   2.428675 |                   4.01525 |                      0.0 |    -1.586575 |              0.0 |         -1.586575 |             13.623188 |                 100.0 |             27827.355778 |              101.116829 |      0.0 |      0.0 |  4.142976 |       0.0 | 0.674992 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  -2.26512 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-02-12 | 11.1 | -3.0 |              0.0 |             85.0 |       5.2 |               100.7 | nan | nan |    -3.0 |                  85.0 |                   0.0 | 10.434 | -3.0 |    -3.0 |    4.168274 |         4.168274 |            1.84595 |          5.2 |                 0.489349 |                      0.489349 |                     0.395271 |                      0.036626 |                           0.036626 |             0.415947 |                  0.415947 |     100.284053 |   1.296511 |        -0.391536 | 8.548511 | 16.158156 |         1.028994 |                  5.687365 |                 7.609644 |                      7.609644 |             0.853105 | 100.0 |        0.853105 |                0.853105 | 0.7416 |   -7.5504 | -0.898177 |              0.106229 |                  0.019728 |                   0.106229 |                  1.277835 |                 0.531756 |    -1.171606 |        -0.512028 |         -1.171606 |             18.076923 |                 100.0 |              2039.690308 |               99.509579 |    8.234 |    8.234 |  1.794481 |  1.794481 |      0.0 | -0.058384 | -0.051627 |    -5.817805 |     0.875479 |      0.0 |      0.0 | 0.299757 |      0.0 | -0.155737 | -1.205976 | -0.849686 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.405519 |  8.292384 |  8.292384 | -0.100119 | 0.084415 | -0.625343 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-03-12 |  0.4 | -5.6 |              0.1 |             95.9 |       2.0 |               100.5 | nan | nan |    -5.6 |                  95.9 |                   0.1 |  0.376 | -5.6 |    -5.6 |    1.603182 |         1.603182 |           0.709981 |          2.0 |                 0.401924 |                      0.401924 |                     0.442417 |                      0.030771 |                           0.030771 |             0.385445 |                  0.385445 |     100.114555 |   1.306656 |        -0.393572 | 8.563421 | 16.143245 |          1.02926 |                  5.617593 |                 7.579824 |                      7.579824 |             1.441454 | 100.0 |        1.441454 |                1.441454 | 0.7416 | -2.084476 |  1.168887 |              0.202282 |                  0.037567 |                   0.202282 |                  1.329454 |                 0.568267 |    -1.127172 |        -0.530701 |         -1.127172 |                  47.0 |                 100.0 |             28069.018521 |              100.749738 |      0.0 |      0.0 |   0.12953 |   0.12953 |      0.0 |   0.10185 |       0.0 |    -4.339273 |     0.859668 |      0.0 |      0.0 | 0.225163 |      0.0 |   0.27168 |  0.341591 |  1.143972 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.651989 |  8.190534 |  8.190534 | -0.075204 |  0.14684 | -1.052629 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-04-12 |  1.2 | -3.8 |              3.7 |             90.6 |       4.3 |               100.8 | nan | nan |    -3.8 |                  90.6 |                   3.7 |  1.128 | -3.8 |    -3.8 |    3.446842 |         3.446842 |           1.526459 |          4.3 |                 0.460812 |                      0.460812 |                     0.427292 |                       0.03473 |                            0.03473 |             0.417496 |                  0.417496 |     100.382504 |   1.301648 |        -0.395492 | 8.577535 | 16.129131 |         1.029518 |                  5.552051 |                 7.551596 |                      7.551596 |             2.748162 | 100.0 |        2.748162 |                2.748162 | 0.7416 | -3.508765 |  0.668199 |              0.429653 |                  0.079793 |                   0.429653 |                  3.727317 |                 1.071685 |    -3.297664 |        -0.991893 |         -3.297664 |             21.860465 |                 100.0 |              3347.213388 |                99.76651 |      0.0 |      0.0 |  0.584703 |  0.584703 |      0.0 |  0.024809 |       0.0 |    -4.797729 |     0.843658 |      0.0 |      0.0 | 0.245894 |      0.0 |  0.066176 | -0.382946 |  0.675123 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.195286 |  8.165725 |  8.165725 | -0.082129 | 0.217823 | -0.979228 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-05-12 |  6.4 |  0.2 |              0.0 |             96.5 |       4.0 |                99.6 | nan | nan |     0.2 |                  96.5 |                   0.0 |  6.016 |  0.2 |     0.2 |    3.206365 |         3.206365 |           1.419962 |          4.0 |                 0.619748 |                      0.619748 |                     0.537181 |                      0.045128 |                           0.045128 |             0.598057 |                  0.598057 |      99.001943 |   1.266438 |        -0.397294 | 8.590839 | 16.115828 |         1.029767 |                  5.490726 |                 7.524989 |                      7.524989 |             0.823609 | 100.0 |        0.823609 |                0.823609 | 0.7416 | -3.264094 |  -0.78634 |               0.13628 |                  0.025309 |                    0.13628 |                  1.140917 |                 0.332476 |    -1.004637 |        -0.307167 |         -1.004637 |                  23.5 |                 100.0 |               763.990266 |               96.624998 | 5.011286 | 2.255079 |  0.342821 |  0.342821 |      0.0 | -0.148252 |  0.003251 |    -1.747234 |     0.834534 |      0.0 |      0.0 | 0.174301 | 0.174301 | -0.395454 | -0.718715 | -0.807002 | 0.979464 | 0.287805 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.857179 | 10.714308 | 13.325263 |       0.0 | 0.262355 |  0.548712 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-06-12 |  0.0 | -1.1 |              0.1 |             99.0 |       2.4 |                99.8 | nan | nan |    -1.1 |                  99.0 |                   0.1 |    0.0 | -1.1 |    -1.1 |    1.923819 |         1.923819 |           0.851977 |          2.4 |                 0.563475 |                      0.563475 |                     0.572963 |                      0.041489 |                           0.041489 |             0.557841 |                  0.557841 |      99.242159 |   1.275245 |        -0.398979 | 8.603319 | 16.103347 |         1.030007 |                  5.433602 |                 7.500028 |                      7.500028 |             1.394624 | 100.0 |        1.394624 |                1.394624 | 0.7416 |       0.0 |  0.452563 |              0.273217 |                   0.05074 |                   0.273217 |                  1.217715 |                 0.389096 |    -0.944497 |        -0.338355 |         -0.944497 |             39.166667 |                 100.0 |              1010.452619 |               97.419093 |      0.0 |      0.0 |   0.03379 |   0.03379 |      0.0 |  0.028994 |       0.0 |       -0.873 |     0.804093 |      0.0 |      0.0 | 0.087103 | 0.087103 |  0.077339 |  0.065961 |  0.481656 | 0.968647 | 0.281483 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.823389 | 10.777908 | 13.296269 |       0.0 | 0.437747 |  0.837749 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-07-12 |  0.0 | -2.9 |              0.0 |             99.2 |       1.9 |               100.4 | nan | nan |    -2.9 |                  99.2 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                 0.493024 |                      0.493024 |                     0.538352 |                      0.036869 |                           0.036869 |             0.489079 |                  0.489079 |      99.910921 |   1.291808 |        -0.400545 | 8.614964 | 16.091703 |         1.030238 |                  5.380663 |                 7.476739 |                      7.476739 |             0.807099 | 100.0 |        0.807099 |                0.807099 | 0.7416 |       0.0 |  0.890518 |              0.177343 |                  0.032935 |                   0.177343 |                  1.193664 |                 0.435045 |    -1.016321 |         -0.40211 |         -1.016321 |             49.473684 |                 100.0 |               1874.96551 |               98.634658 |      0.0 |      0.0 |  0.000946 |  0.000946 |      0.0 |   0.08649 |       0.0 |    -1.717821 |     0.780271 |      0.0 |      0.0 | 0.170083 | 0.170083 |  0.230709 |  0.314507 |  0.947326 |  0.97422 | 0.284732 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.822443 | 10.876776 | 13.209779 |       0.0 | 0.613139 |  0.688831 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-08-12 |  0.0 | -2.9 |              0.0 |             99.1 |       1.8 |               100.6 | nan | nan |    -2.9 |                  99.1 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.442864 |         1.442864 |           0.638983 |          1.8 |                 0.493024 |                      0.493024 |                     0.538702 |                      0.036869 |                           0.036869 |             0.488586 |                  0.488586 |     100.111414 |   1.294388 |        -0.401992 | 8.625761 | 16.080906 |         1.030461 |                  5.331896 |                 7.455145 |                      7.455145 |             0.799784 | 100.0 |        0.799784 |                0.799784 | 0.7416 |       0.0 |  0.885964 |              0.191051 |                  0.035481 |                   0.191051 |                  1.194208 |                 0.435661 |    -1.003157 |         -0.40018 |         -1.003157 |             52.222222 |                 100.0 |              1875.428739 |                98.51545 |      0.0 |      0.0 |  0.005497 |  0.005497 |      0.0 |  0.086348 |       0.0 |    -1.709036 |     0.761122 |      0.0 |      0.0 |   0.1663 |   0.1663 |  0.230328 |     0.311 |  0.941508 | 0.979623 | 0.287899 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.816946 | 10.970891 | 13.123431 |       0.0 | 0.788531 |  0.544466 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-09-12 |  0.7 | -2.1 |              4.4 |             96.9 |       2.7 |               100.9 | nan | nan |    -2.1 |                  96.9 |                   4.4 |  0.658 | -2.1 |    -2.1 |    2.164296 |         2.164296 |           0.958474 |          2.7 |                 0.523311 |                      0.523311 |                     0.528127 |                      0.038864 |                           0.038864 |             0.507088 |                  0.507088 |     100.392912 |   1.294334 |        -0.403321 | 8.635699 | 16.070967 |         1.030674 |                  5.287288 |                 7.435268 |                      7.435268 |             2.886262 | 100.0 |        2.886262 |                2.886262 | 0.7416 |       0.0 |  1.024664 |               0.72913 |                   0.13541 |                    0.72913 |                  4.005891 |                 1.072417 |     -3.27676 |        -0.937007 |          -3.27676 |             34.814815 |                 100.0 |              1366.137242 |               97.824519 | 0.274946 | 0.274946 | -0.145708 | -0.145708 |      0.0 |  0.042379 | -0.001207 |     -1.97659 |     0.747379 |      0.0 |      0.0 | 0.192428 | 0.192428 |  0.113045 |  0.037677 |  1.087729 | 0.990217 | 0.294159 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.345708 | 11.402086 | 13.355998 |       0.0 | 0.939906 |  0.261402 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-10-12 |  3.3 | -3.1 |              0.1 |             95.0 |       3.0 |               101.8 | nan | nan |    -3.1 |                  95.0 |                   0.1 |  3.102 | -3.1 |    -3.1 |    2.404774 |         2.404774 |           1.064971 |          3.0 |                 0.485698 |                      0.485698 |                     0.525121 |                      0.036384 |                           0.036384 |             0.461413 |                  0.461413 |     101.338587 |   1.310959 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                  7.41713 |                       7.41713 |             1.347076 | 100.0 |        1.347076 |                1.347076 | 0.7416 |       0.0 |  1.064535 |              0.320728 |                  0.059564 |                   0.320728 |                  1.286731 |                 0.489004 |    -0.966004 |        -0.429441 |         -0.966004 |             31.333333 |                 100.0 |              2093.665722 |               98.468592 | 3.247708 | 3.247708 |  0.333009 |  0.333009 |      0.0 |  0.134519 | -0.021042 |    -2.053502 |     0.761908 |      0.0 |      0.0 | 0.238543 | 0.238543 |  0.358824 |  0.334902 |  1.123167 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.866991 | 14.767714 | 16.469186 |       0.0 | 0.780583 | -0.061533 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-11-12 |  9.4 |  5.1 |              0.2 |             90.6 |       7.4 |               100.4 | nan | nan |     5.1 |                  90.6 |                   0.2 |  8.836 |  5.1 |     5.1 |    5.931775 |         5.931775 |           2.626929 |          7.4 |                 0.879017 |                      0.879017 |                      0.61078 |                      0.061412 |                           0.061412 |              0.79639 |                   0.79639 |       99.60361 |   1.253213 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 7.400749 |                      7.400749 |             1.373029 | 100.0 |        1.373029 |                1.373029 | 0.7416 | -0.205111 |  -0.10633 |              0.349872 |                  0.064976 |                   0.349872 |                   1.16859 |                 0.213257 |    -0.818717 |        -0.148281 |         -0.818717 |             12.702703 |                 100.0 |               405.842399 |               92.890929 | 8.502991 |      0.0 |  2.300327 |       2.2 |      0.0 | -0.656354 |  0.099261 |          0.0 |     0.745182 | 12.96039 | 12.96039 |      0.0 |      0.0 | -1.750792 | -2.733328 | -4.335839 | 0.970569 | 0.282601 | 22.963731 | 7.623127 |  0.0 |  0.0 | 0.0 |  0.0 | 7.623127 | nan | nan |      0.0 |  1.863497 |  2.664801 |       0.0 |  0.96651 |  0.786396 | 86.665611 | 7.623127 |   0.0 |   0.0 |   0.0 |  0.0 | 1.962548 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 |  0.2 |  0.2 |              0.7 |             78.7 |       5.7 |               101.2 | nan | nan |     0.2 |                  78.7 |                   0.7 |  0.188 |  0.2 |     0.2 |     4.56907 |          4.56907 |           2.023445 |          5.7 |                 0.619748 |                      0.619748 |                     0.564435 |                      0.045128 |                           0.045128 |             0.487741 |                  0.487741 |     100.712259 |    1.28736 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 7.386143 |                      7.386143 |             1.539952 | 100.0 |        1.539952 |                1.539952 | 0.7416 |       0.0 |  0.558256 |              0.412606 |                  0.076627 |                   0.412606 |                  1.733999 |                 0.609687 |    -1.321393 |        -0.533061 |         -1.321393 |             16.491228 |                 100.0 |               656.301186 |               95.339607 |      0.0 |      0.0 |  3.264511 |     0.188 |      0.0 |  0.228989 |       0.0 |    -1.076882 |     0.732066 |      0.0 |      0.0 | 0.023374 | 0.023374 |  0.610818 | -0.577816 |  0.566063 | 0.963707 |  0.27862 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |   1.72473 |  2.435811 |       0.0 | 1.141902 |  0.969741 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 2.640032 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-13-12 |  2.9 | -3.5 |              0.0 |             90.5 |       2.2 |               102.0 | nan | nan |    -3.5 |                  90.5 |                   0.0 |  2.726 | -3.5 |    -3.5 |    1.763501 |         1.763501 |           0.780979 |          2.2 |                 0.471336 |                      0.471336 |                     0.513926 |                      0.035431 |                           0.035431 |             0.426559 |                  0.426559 |     101.573441 |   1.315658 |        -0.407436 | 8.666669 | 16.039997 |         1.031437 |                  5.150203 |                 7.373328 |                      7.373328 |              0.77253 | 100.0 |         0.77253 |                 0.77253 | 0.7416 |       0.0 |  1.214827 |              0.211461 |                  0.039271 |                   0.211461 |                  1.254261 |                 0.511811 |      -1.0428 |         -0.47254 |           -1.0428 |             42.727273 |                 100.0 |              2173.673838 |               98.277258 |    0.526 |    0.526 |  0.463476 |  0.463476 |      0.0 |  0.161846 | -0.003848 |    -2.343417 |     0.726275 |      0.0 |      0.0 | 0.053893 | 0.053893 |  0.431715 |  0.324725 |   1.22898 | 0.981418 | 0.288954 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.736524 |  2.178689 |  2.799966 |       0.0 | 1.229697 |  0.496513 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  1.40875 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-14-12 |  0.0 | -3.9 |              1.2 |             88.1 |       3.1 |               102.5 | nan | nan |    -3.9 |                  88.1 |                   1.2 |    0.0 | -3.9 |    -3.9 |    2.484933 |         2.484933 |            1.10047 |          3.1 |                 0.457351 |                      0.457351 |                     0.493405 |                      0.034499 |                           0.034499 |             0.402926 |                  0.402926 |     102.097074 |   1.324197 |        -0.408164 | 8.672175 | 16.034492 |         1.031605 |                  5.126218 |                 7.362317 |                      7.362317 |             1.699321 | 100.0 |        1.699321 |                1.699321 | 0.7416 |       0.0 |  1.497998 |              0.482654 |                  0.089636 |                   0.482654 |                  2.107297 |                 0.762591 |    -1.624642 |        -0.672956 |         -1.624642 |             30.322581 |                 100.0 |              2978.799239 |               98.578374 |      0.0 |      0.0 |  0.747786 |  0.747786 |      0.0 |  0.193978 |       0.0 |    -2.889656 |     0.715972 |      0.0 |      0.0 | 0.061899 | 0.061899 |  0.517426 |   0.32829 |  1.518672 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.988738 |  2.085363 |  2.605988 |       0.0 | 1.405089 | -0.259884 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.751725 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-15-12 |  0.0 | -6.3 |              3.0 |             92.8 |       2.8 |               102.3 | nan | nan |    -6.3 |                  92.8 |                   3.0 |    0.0 | -6.3 |    -6.3 |    2.244455 |         2.244455 |           0.993973 |          2.8 |                 0.380889 |                      0.380889 |                     0.436373 |                       0.02934 |                            0.02934 |             0.353465 |                  0.353465 |     101.946535 |    1.33374 |         -0.40877 | 8.676772 | 16.029895 |         1.031763 |                  5.106331 |                 7.353123 |                      7.353123 |             2.318249 | 100.0 |        2.318249 |                2.318249 | 0.7416 | -0.866281 |  1.894499 |              0.679084 |                  0.126116 |                   0.679084 |                  3.397964 |                 1.045602 |     -2.71888 |        -0.919486 |          -2.71888 |             33.571429 |                 100.0 |             22767.949382 |               99.384825 |      0.0 |      0.0 |  0.185118 |  0.185118 |      0.0 |  0.169693 |       0.0 |    -4.520792 |      0.70707 |      0.0 |      0.0 | 0.088496 | 0.088496 |  0.452648 |  0.551922 |  1.924056 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  0.80362 |  2.032305 |  2.436295 |       0.0 | 1.580481 | -1.412783 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.401129 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-16-12 |  6.7 | -3.5 |              0.0 |             98.1 |       3.7 |               100.0 | nan | nan |    -3.5 |                  98.1 |                   0.0 |  6.298 | -3.5 |    -3.5 |    2.965888 |         2.965888 |           1.313465 |          3.7 |                 0.471336 |                      0.471336 |                     0.437438 |                      0.035431 |                           0.035431 |             0.462381 |                  0.462381 |      99.537619 |   1.289645 |        -0.409256 | 8.680456 | 16.026211 |         1.031912 |                  5.090537 |                 7.345755 |                      7.345755 |             0.763581 | 100.0 |        0.763581 |                0.763581 | 0.7416 | -4.709277 | -0.114381 |              0.190433 |                  0.035366 |                   0.190433 |                  1.211777 |                 0.463683 |    -1.021345 |        -0.428317 |         -1.021345 |             25.405405 |                 100.0 |              2134.566544 |               98.100564 |  4.90162 |  4.90162 |  0.125363 |  0.125363 |      0.0 |  -0.05832 | -0.035855 |    -4.488635 |     0.750606 |      0.0 |      0.0 | 0.217488 | 0.217488 | -0.155566 | -0.350346 | -0.077595 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.074637 |  7.208251 |  7.396235 |       0.0 | 0.902108 | -0.556802 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.214047 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-17-12 |  0.0 | -7.3 |              0.0 |             96.6 |       0.9 |                99.1 | nan | nan |    -7.3 |                  96.6 |                   0.0 |    0.0 | -7.3 |    -7.3 |    0.721432 |         0.721432 |           0.319491 |          0.9 |                 0.352531 |                      0.352531 |                     0.433413 |                      0.027395 |                           0.027395 |             0.340545 |                  0.340545 |      98.759455 |   1.296889 |         -0.40962 | 8.683222 | 16.023444 |         1.032051 |                  5.078832 |                 7.340222 |                      7.340222 |             0.761825 | 100.0 |        0.761825 |                0.761825 | 0.7416 | -1.856007 |   1.42792 |              0.200646 |                  0.037263 |                   0.200646 |                  1.287913 |                 0.600105 |    -1.087267 |        -0.562843 |         -1.087267 |            104.444444 |                 100.0 |             22767.949382 |               99.309706 |      0.0 |      0.0 |  0.027204 |  0.027204 |      0.0 |  0.132947 |       0.0 |    -4.610482 |     0.736625 |      0.0 |      0.0 | 0.218823 | 0.187984 |  0.354629 |  0.583535 |  1.501006 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.047433 |  7.263288 |  7.263288 |   -0.0103 |   1.0775 | -1.243122 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.114218 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-18-12 |  0.1 | -5.8 |              0.0 |             97.7 |       1.3 |                99.7 | nan | nan |    -5.8 |                  97.7 |                   0.0 |  0.094 | -5.8 |    -5.8 |    1.042069 |         1.042069 |           0.461488 |          1.3 |                 0.395811 |                      0.395811 |                     0.409234 |                      0.030356 |                           0.030356 |             0.386708 |                  0.386708 |      99.313292 |   1.297204 |        -0.409863 | 8.685068 | 16.021598 |         1.032182 |                  5.071211 |                  7.33653 |                       7.33653 |             0.760682 | 100.0 |        0.760682 |                0.760682 | 0.7416 |  -4.14374 |  0.632759 |              0.207622 |                  0.038558 |                   0.207622 |                  1.257034 |                 0.545575 |    -1.049412 |        -0.507016 |         -1.049412 |             72.307692 |                 100.0 |             22767.949382 |               99.288228 |      0.0 |      0.0 |  0.029597 |  0.029597 |      0.0 |  0.035165 |       0.0 |     -5.36434 |     0.727057 |      0.0 |      0.0 | 0.243809 |      0.0 |    0.0938 |  0.103075 |  0.703891 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.111836 |  7.228124 |  7.228124 | -0.081432 | 1.217412 | -1.134281 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.060948 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-19-12 |  0.0 | -7.9 |              0.0 |             95.8 |       1.9 |                99.7 | nan | nan |    -7.9 |                  95.8 |                   0.0 |    0.0 | -7.9 |    -7.9 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                  0.33643 |                       0.33643 |                     0.382007 |                      0.026283 |                           0.026283 |               0.3223 |                    0.3223 |        99.3777 |   1.307793 |        -0.409985 | 8.685992 | 16.020675 |         1.032302 |                  5.067673 |                 7.334683 |                      7.334683 |             0.760151 | 100.0 |        0.760151 |                0.760151 | 0.7416 | -3.780937 |  1.286157 |              0.210531 |                  0.039099 |                   0.210531 |                  1.301361 |                 0.623068 |     -1.09083 |        -0.583969 |          -1.09083 |             49.473684 |                 100.0 |             22767.949382 |               99.277487 |      0.0 |      0.0 |  0.112465 |  0.112465 |      0.0 |  0.104807 |       0.0 |    -6.261951 |      0.72304 |      0.0 |      0.0 | 0.283227 |      0.0 |  0.279567 |  0.435787 |  1.299323 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.999371 |  7.123317 |  7.123317 | -0.094598 | 1.281941 | -1.678839 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.032522 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-20-12 |  0.2 | -5.7 |              0.2 |             97.5 |       2.4 |                99.7 | nan | nan |    -5.7 |                  97.5 |                   0.2 |  0.188 | -5.7 |    -5.7 |    1.923819 |         1.923819 |           0.851977 |          2.4 |                 0.398857 |                      0.398857 |                     0.382911 |                      0.030563 |                           0.030563 |             0.388886 |                  0.388886 |      99.311114 |   1.296708 |        -0.409985 | 8.685992 | 16.020675 |         1.032413 |                  5.068219 |                 7.334683 |                      7.334683 |             1.336154 | 100.0 |        1.336154 |                1.336154 | 0.7416 | -5.596129 |  0.329248 |              0.374753 |                  0.069597 |                   0.374753 |                  1.393316 |                 0.581946 |    -1.018563 |        -0.512349 |         -1.018563 |             39.166667 |                 100.0 |             22767.949382 |               99.277487 |      0.0 |      0.0 |  0.092592 |  0.092592 |      0.0 | -0.011456 |       0.0 |    -6.231253 |     0.719529 |      0.0 |      0.0 | 0.277752 |      0.0 | -0.030559 |  -0.15437 |   0.32742 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.094779 |  7.134773 |  7.134773 | -0.092769 | 1.341508 | -1.266487 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.017354 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-21-12 |  0.2 | -5.6 |              0.8 |             95.0 |       2.3 |               100.4 | nan | nan |    -5.6 |                  95.0 |                   0.8 |  0.188 | -5.6 |    -5.6 |     1.84366 |          1.84366 |           0.816478 |          2.3 |                 0.401924 |                      0.401924 |                     0.401392 |                      0.030771 |                           0.030771 |             0.381828 |                  0.381828 |     100.018172 |   1.305372 |        -0.409863 | 8.685068 | 16.021598 |         1.032515 |                  5.072848 |                  7.33653 |                       7.33653 |             1.544792 | 100.0 |        1.544792 |                1.544792 | 0.7416 | -4.221623 |  0.723518 |              0.437541 |                  0.081258 |                   0.437541 |                  1.820398 |                  0.70235 |    -1.382858 |        -0.621092 |         -1.382858 |             40.869565 |                 100.0 |             22767.949382 |               99.288228 | 0.082779 | 0.082779 |  0.167454 |  0.167454 |      0.0 |  0.036876 | -0.000969 |    -5.617299 |     0.716764 |      0.0 |      0.0 | 0.253698 |      0.0 |  0.098365 | -0.004942 |  0.714515 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.032546 |  7.180676 |  7.180676 | -0.084735 | 1.390636 | -1.248405 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.00926 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-22-12 |  2.4 | -4.8 |              0.0 |             93.8 |       4.6 |               100.7 | nan | nan |    -4.8 |                  93.8 |                   0.0 |  2.256 | -4.8 |    -4.8 |     3.68732 |          3.68732 |           1.632956 |          4.6 |                 0.427219 |                      0.427219 |                      0.41713 |                      0.032479 |                           0.032479 |             0.400731 |                  0.400731 |     100.299269 |   1.305282 |         -0.40962 | 8.683222 | 16.023444 |         1.032607 |                  5.081565 |                 7.340222 |                      7.340222 |             0.762235 | 100.0 |        0.762235 |                0.762235 | 0.7416 |  -4.16135 |  0.493809 |              0.205392 |                  0.038144 |                   0.205392 |                  1.259735 |                 0.535417 |    -1.054343 |        -0.497273 |         -1.054343 |             20.434783 |                 100.0 |             16151.471046 |               99.256536 | 2.088546 | 2.088546 |   0.58112 |   0.58112 |      0.0 |  0.043122 | -0.020952 |    -5.113914 |      0.73054 |      0.0 |      0.0 | 0.296617 |      0.0 |  0.115027 | -0.125107 |  0.487192 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  1.61888 |    9.2261 |    9.2261 |  -0.09907 | 1.164357 | -1.000614 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.004941 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-23-12 |  7.0 | -1.1 |              0.0 |             94.9 |       8.6 |               100.4 | nan | nan |    -1.1 |                  94.9 |                   0.0 |   6.58 | -1.1 |    -1.1 |    6.893685 |         6.893685 |           3.052917 |          8.6 |                 0.563475 |                      0.563475 |                     0.515075 |                      0.041489 |                           0.041489 |             0.534738 |                  0.534738 |      99.865262 |    1.28304 |        -0.409256 | 8.680456 | 16.026211 |         1.032689 |                  5.094372 |                 7.345755 |                      7.345755 |             0.764156 | 100.0 |        0.764156 |                0.764156 | 0.7416 | -3.335381 | -0.529796 |              0.175813 |                  0.032651 |                   0.175813 |                  1.179808 |                 0.394667 |    -1.003995 |        -0.362016 |         -1.003995 |             10.930233 |                 100.0 |               829.225836 |               96.016503 |  5.99888 | 4.649132 |  1.086922 |  1.086922 |      0.0 | -0.077173 | -0.016905 |    -2.313399 |     0.769925 |      0.0 |      0.0 |     0.24 |     0.24 | -0.205854 | -0.721772 |  -0.56561 | 0.989866 |  0.29395 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.113078 |  14.18678 | 15.302153 |       0.0 | 0.703392 |  0.270781 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.002637 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-24-12 |  0.2 | -2.9 |              0.0 |             85.4 |       7.6 |               100.8 | nan | nan |    -2.9 |                  85.4 |                   0.0 |  0.188 | -2.9 |    -2.9 |    6.092093 |         6.092093 |           2.697927 |          7.6 |                 0.493024 |                      0.493024 |                     0.503792 |                      0.036869 |                           0.036869 |             0.421042 |                  0.421042 |     100.378958 |   1.297296 |         -0.40877 | 8.676772 | 16.029895 |         1.032762 |                  5.111276 |                 7.353123 |                      7.353123 |             0.766691 | 100.0 |        0.766691 |                0.766691 | 0.7416 |       0.0 |  1.353398 |              0.189632 |                  0.035217 |                   0.189632 |                  1.273508 |                 0.525497 |    -1.083876 |        -0.490279 |         -1.083876 |             12.368421 |                 100.0 |               1574.53183 |               97.661543 |      0.0 |      0.0 |  2.588786 |  1.301078 |      0.0 |  0.297978 |       0.0 |    -2.610721 |     0.752662 |      0.0 |      0.0 | 0.268265 | 0.268265 |  0.794841 |  0.157877 |  1.442998 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 14.173563 | 15.004175 |       0.0 | 0.878784 | -0.341016 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001407 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-25-12 |  0.0 | -9.1 |              3.2 |             87.2 |       3.8 |               101.8 | nan | nan |    -9.1 |                  87.2 |                   3.2 |    0.0 | -9.1 |    -9.1 |    3.046047 |         3.046047 |           1.348964 |          3.8 |                  0.30617 |                       0.30617 |                     0.384345 |                      0.024175 |                           0.024175 |             0.266981 |                  0.266981 |     101.533019 |   1.341718 |        -0.408164 | 8.672175 | 16.034492 |         1.032825 |                  5.132282 |                 7.362317 |                      7.362317 |             2.398433 | 100.0 |        2.398433 |                2.398433 | 0.7416 | -1.136721 |  2.615827 |              0.627559 |                  0.116547 |                   0.627559 |                  3.730222 |                 1.146755 |    -3.102663 |        -1.030208 |         -3.102663 |             24.736842 |                 100.0 |             22767.949382 |               99.438432 |      0.0 |      0.0 |  0.606006 |       0.0 |      0.0 |  0.278218 |       0.0 |    -6.182684 |     0.738346 |      0.0 |      0.0 | 0.612722 | 0.612722 |  0.742132 |  1.048136 |  2.820476 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 14.512107 | 14.725957 |       0.0 | 1.054176 | -2.215243 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000751 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-26-12 |  1.2 | -8.6 |              4.3 |             92.7 |       1.5 |               101.9 | nan | nan |    -8.6 |                  92.7 |                   4.3 |  1.128 | -8.6 |    -8.6 |    1.202387 |         1.202387 |           0.532486 |          1.5 |                 0.318472 |                      0.318472 |                     0.311154 |                      0.025034 |                           0.025034 |             0.295223 |                  0.295223 |     101.604777 |   1.340358 |        -0.407436 | 8.666669 | 16.039997 |         1.032879 |                  5.157401 |                 7.373328 |                      7.373328 |             2.793204 | 100.0 |        2.793204 |                2.793204 | 0.7416 | -7.384144 |  0.783397 |              0.764548 |                  0.141987 |                   0.764548 |                  4.423241 |                  1.20683 |    -3.658694 |        -1.064843 |         -3.658694 |             62.666667 |                 100.0 |             22767.949382 |               99.502706 |      0.0 |      0.0 |  -0.08636 |  -0.08636 |      0.0 |  0.025901 |       0.0 |    -8.895328 |     0.726283 |      0.0 |      0.0 | 0.831624 |  0.21385 |  0.069089 | -0.072772 |   1.06116 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  1.21436 | 14.700056 | 14.700056 | -0.206336 | 1.229568 | -2.257041 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000401 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-27-12 |  0.9 | -5.3 |              0.0 |             97.8 |       1.3 |               101.7 | nan | nan |    -5.3 |                  97.8 |                   0.0 |  0.846 | -5.3 |    -5.3 |    1.042069 |         1.042069 |           0.461488 |          1.3 |                  0.41125 |                       0.41125 |                     0.355399 |                      0.031402 |                           0.031402 |             0.402202 |                  0.402202 |     101.297798 |   1.320718 |        -0.406588 | 8.660262 | 16.046405 |         1.032922 |                  5.186639 |                 7.386143 |                      7.386143 |             0.777996 | 100.0 |        0.777996 |                0.777996 | 0.7416 | -7.523469 | -0.169932 |              0.215734 |                  0.040065 |                   0.215734 |                  1.247575 |                 0.528268 |    -1.031841 |        -0.488204 |         -1.031841 |             72.307692 |                 100.0 |             22767.949382 |               99.577619 |      0.0 |      0.0 |  0.030808 |  0.030808 |      0.0 | -0.073062 |       0.0 |    -7.195669 |     0.722706 |      0.0 |      0.0 | 0.661896 |      0.0 |  -0.19489 | -0.448508 | -0.155195 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.029552 | 14.773119 | 14.773119 | -0.221073 | 1.287482 | -1.345509 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000214 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-28-12 |  0.0 | -7.7 |              3.4 |             95.4 |       3.0 |               102.2 | nan | nan |    -7.7 |                  95.4 |                   3.4 |    0.0 | -7.7 |    -7.7 |    2.404774 |         2.404774 |           1.064971 |          3.0 |                 0.341723 |                      0.341723 |                     0.358381 |                      0.026649 |                           0.026649 |             0.326003 |                  0.326003 |     101.873997 |   1.339598 |        -0.405619 | 8.652959 | 16.053708 |         1.032956 |                   5.22001 |                 7.400749 |                      7.400749 |             2.504073 | 100.0 |        2.504073 |                2.504073 | 0.7416 |  -4.48503 |  1.349382 |              0.702605 |                  0.130484 |                   0.702605 |                    3.6896 |                 1.099425 |    -2.986995 |        -0.968941 |         -2.986995 |             31.333333 |                 100.0 |             22767.949382 |               99.663136 |      0.0 |      0.0 |  0.016489 |  0.016489 |      0.0 |  0.068366 |       0.0 |    -7.088005 |     0.719415 |      0.0 |      0.0 | 0.655233 |      0.0 |  0.182364 |  0.195852 |  1.347156 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.013063 | 14.704753 | 14.704753 | -0.218848 | 1.343486 | -1.953291 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000114 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-29-12 |  0.0 | -9.2 |              0.0 |             94.9 |       2.7 |               102.4 | nan | nan |    -9.2 |                  94.9 |                   0.0 |    0.0 | -9.2 |    -9.2 |    2.164296 |         2.164296 |           0.958474 |          2.7 |                 0.303761 |                      0.303761 |                     0.322491 |                      0.024006 |                           0.024006 |              0.28827 |                   0.28827 |      102.11173 |   1.350039 |         -0.40453 | 8.644768 | 16.061898 |         1.032981 |                  5.257524 |                  7.41713 |                       7.41713 |             0.788629 | 100.0 |        0.788629 |                0.788629 | 0.7416 |  -6.51097 |  1.000185 |              0.223758 |                  0.041555 |                   0.223758 |                  1.325064 |                 0.666484 |    -1.101305 |        -0.624929 |         -1.101305 |             34.814815 |                 100.0 |             22767.949382 |                99.75922 |      0.0 |      0.0 |  0.206218 |  0.206218 |      0.0 |  0.068935 |       0.0 |     -8.44034 |     0.716269 |      0.0 |      0.0 | 0.776636 |      0.0 |  0.183881 |  0.231924 |  1.040734 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.806845 | 14.635818 | 14.635818 | -0.259396 | 1.399653 | -2.211876 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000061 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-30-12 |  0.2 | -7.4 |              0.0 |             96.0 |       1.4 |               102.1 | nan | nan |    -7.4 |                  96.0 |                   0.0 |  0.188 | -7.4 |    -7.4 |    1.122228 |         1.122228 |           0.496987 |          1.4 |                 0.349801 |                      0.349801 |                      0.33091 |                      0.027207 |                           0.027207 |             0.335809 |                  0.335809 |     101.764191 |   1.336727 |        -0.403321 | 8.635699 | 16.070967 |         1.032995 |                  5.299194 |                 7.435268 |                      7.435268 |             0.794879 | 100.0 |        0.794879 |                0.794879 | 0.7416 | -7.372921 |  0.382915 |              0.227699 |                  0.042287 |                   0.227699 |                  1.292505 |                 0.606599 |    -1.064805 |        -0.564312 |         -1.064805 |             67.142857 |                 100.0 |             22767.949382 |               99.865828 |      0.0 |      0.0 |   0.07775 |   0.07775 |      0.0 | -0.007807 |       0.0 |    -8.111568 |     0.713542 |      0.0 |      0.0 | 0.742885 |      0.0 | -0.020824 | -0.171846 |  0.371642 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 1.917095 | 14.643624 | 14.643624 | -0.248124 | 1.450552 | -1.853191 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000032 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-31-12 |  0.9 |  1.7 |              4.0 |             96.5 |       4.2 |               101.6 | nan | nan |     1.7 |                  96.5 |                   4.0 |  0.846 |  1.7 |     1.7 |    3.366683 |         3.366683 |            1.49096 |          4.2 |                 0.690794 |                      0.690794 |                     0.501108 |                      0.049663 |                           0.049663 |             0.666616 |                  0.666616 |     100.933384 |   1.284547 |        -0.401992 | 8.625761 | 16.080906 |            1.033 |                  5.345035 |                 7.455145 |                      7.455145 |             2.770177 | 100.0 |        2.770177 |                2.770177 | 0.7416 | -6.177303 | -1.811807 |              0.800486 |                  0.148662 |                   0.800486 |                  3.475519 |                 0.945014 |    -2.675033 |        -0.796353 |         -2.675033 |             22.380952 |                 100.0 |               481.569845 |                94.15647 | 0.563095 | 0.042232 |  0.114221 |  0.114221 |      0.0 | -0.413784 |  0.003857 |    -2.682304 |     0.711034 |      0.0 |      0.0 | 0.264008 | 0.264008 | -1.103747 |   -1.6605 | -1.967895 | 0.973794 | 0.284483 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.085779 | 15.356659 | 15.620503 |       0.0 | 1.499321 |  0.700216 | 86.665611 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000017 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |

.. _lland_v3_acker_winter_ktschnee_daily:

acre (snow surface temperature)
-------------------------------

In integration test :ref:`lland_v3_acker_winter_daily`, |lland_v3| assumes
that the snow-layer has a constant temperature over its complete depth.
For less simplified modelling of the processes within the snow layer, one
can enable an iterative adjustment of the snow surface temperature (see
method |Return_TempSSurface_V1|).  Therefore, one must set the thermal
conductivity of the top snow layer (|KTSchnee|) to a value smaller than
|numpy.inf|.  The lower the conductivity, the higher the possible differences
between the bulk and the surface temperature of the snow layer. Setting
zero values (which is not advisable) would prevent any thermal exchange,
setting |numpy.inf| values results in an infinitely fast exchange, avoiding
any temperature differences within the snow layer.

We show what happens when we use the default |KTSchnee| value used by
`LARSIM`_.  In integration test :ref:`lland_v3_acker_winter_daily`, with
:math:`KTSchnee = inf`, nearly the complete frozen water equivalent melts
on December 11th, where the average air temperature is 5 °C.  In our next
test, with :math:`KTSchnee = 0.018 MJ/m³/K/h`, the limited thermal
conductivity of the top snow layer weakens the flux of energy into the
snow layer and fewer snow melts.  The snow surface temperature rises
faster than the bulk temperature, and the higher surface temperature results
in smaller sensible (|WSensSnow|) and latent (|WLatSnow|) heat fluxes
from the atmosphere into the snow layer and to larger longwave radiation
losses (|NetLongwaveRadiationSnow|):

.. integration-test::

    >>> lnk(ACKER)
    >>> ktschnee(0.018)
    >>> conditions_acker_winter_ktschnee = test(
    ...     'lland_v3_acker_winter_ktschnee_daily',
    ...     axis1=(inputs.nied, fluxes.wada),
    ...     axis2=(states.waes, states.wats),
    ...     get_conditions='2010-12-10')
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |      g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes |     sbes |      evpo |       evi |     evb |       evs |     wnied | tempssurface | actualalbedo |  schmpot |     schm |  gefrpot |     gefr |  wlatsnow | wsenssnow |     wsurf |      sff |      fvg |     wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz | qah |  qa |     inzp |      wats |      waes |     esnow |     taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet | outlet |
    -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-01-12 |  0.0 | -5.8 |              3.5 |             70.3 |       6.9 |               101.5 | nan | nan |    -5.8 |                  70.3 |                   3.5 |    0.0 | -5.8 |    -5.8 |     5.53098 |          5.53098 |            5.53098 |          6.9 |                 0.395811 |                      0.395811 |                          0.0 |                      0.030356 |                           0.030356 |             0.278255 |                  0.278255 |     101.221745 |   1.321192 |        -0.389384 | 8.532819 | 16.173848 |         1.028719 |                  5.761386 |                 7.641029 |                      7.641029 |             2.759858 | 100.0 |        2.759858 |                2.759858 | 0.7416 |       0.0 |   3.00672 |              2.483872 |                  2.483872 |                   2.483872 |                   4.01525 |                      0.0 |    -1.531378 |              0.0 |         -1.531378 |             30.870386 |                 100.0 |               111.309423 |               99.512695 |      0.0 |      0.0 |  1.808179 |       0.0 | 0.55374 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  -2.26512 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-02-12 | 11.1 | -3.0 |              0.0 |             85.0 |       5.2 |               100.7 | nan | nan |    -3.0 |                  85.0 |                   0.0 | 10.434 | -3.0 |    -3.0 |    4.168274 |         4.168274 |           4.168274 |          5.2 |                 0.489349 |                      0.489349 |                     0.387464 |                      0.036626 |                           0.036626 |             0.415947 |                  0.415947 |     100.284053 |   1.296511 |        -0.391536 | 8.548511 | 16.158156 |         1.028994 |                  5.687365 |                 7.609644 |                      7.609644 |             0.853105 | 100.0 |        0.853105 |                0.853105 | 0.7416 |   -7.5504 | -0.391921 |              0.106229 |                  0.106229 |                   0.106229 |                  1.277835 |                  2.47724 |    -1.171606 |        -2.371011 |         -1.171606 |             40.962628 |                 100.0 |               112.102145 |               99.530051 |   10.374 |   10.374 |  0.779906 |      0.06 |     0.0 | -0.140764 | -0.065045 |     -6.07759 |     0.875479 |      0.0 |      0.0 | 0.441058 |      0.0 |  -0.37548 | -2.305184 | -0.309653 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 10.514764 | 10.514764 | -0.147313 | 0.084415 | -1.131599 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-03-12 |  0.4 | -5.6 |              0.1 |             95.9 |       2.0 |               100.5 | nan | nan |    -5.6 |                  95.9 |                   0.1 |  0.376 | -5.6 |    -5.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 0.401924 |                      0.401924 |                     0.324339 |                      0.030771 |                           0.030771 |             0.385445 |                  0.385445 |     100.114555 |   1.306656 |        -0.393572 | 8.563421 | 16.143245 |          1.02926 |                  5.617593 |                 7.579824 |                      7.579824 |             1.441454 | 100.0 |        1.441454 |                1.441454 | 0.7416 | -3.771996 |  1.092644 |              0.199858 |                  0.199858 |                   0.199858 |                  1.329454 |                 2.790348 |    -1.129596 |        -2.590489 |         -1.129596 |            106.502833 |                 100.0 |               112.102145 |               99.527226 |    0.316 |    0.316 |  0.039485 |  0.039485 |     0.0 | -0.159024 | -0.003698 |    -8.367514 |     0.861349 |      0.0 |      0.0 | 0.398488 |      0.0 | -0.424188 | -1.091574 |  1.074727 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.020515 | 10.989788 | 10.989788 | -0.133095 | 0.139861 | -1.482643 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-04-12 |  1.2 | -3.8 |              3.7 |             90.6 |       4.3 |               100.8 | nan | nan |    -3.8 |                  90.6 |                   3.7 |  1.128 | -3.8 |    -3.8 |    3.446842 |         3.446842 |           3.446842 |          4.3 |                 0.460812 |                      0.460812 |                     0.331306 |                       0.03473 |                            0.03473 |             0.417496 |                  0.417496 |     100.382504 |   1.301648 |        -0.395492 | 8.577535 | 16.129131 |         1.029518 |                  5.552051 |                 7.551596 |                      7.551596 |             2.748162 | 100.0 |        2.748162 |                2.748162 | 0.7416 | -4.942143 |  0.733022 |               0.41176 |                   0.41176 |                    0.41176 |                  3.727317 |                 4.937974 |    -3.315556 |        -4.526213 |         -3.315556 |             49.536202 |                 100.0 |               112.102145 |               99.524551 | 1.088515 | 1.088515 |  0.092461 |      0.06 |     0.0 | -0.369245 | -0.008645 |    -8.096278 |     0.850169 |      0.0 |      0.0 | 0.480397 |      0.0 | -0.984942 | -2.789537 |  0.751735 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 12.447547 | 12.447547 | -0.160452 | 0.187922 | -1.474065 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-05-12 |  6.4 |  0.2 |              0.0 |             96.5 |       4.0 |                99.6 | nan | nan |     0.2 |                  96.5 |                   0.0 |  6.016 |  0.2 |     0.2 |    3.206365 |         3.206365 |           3.206365 |          4.0 |                 0.619748 |                      0.619748 |                     0.523188 |                      0.045128 |                           0.045128 |             0.598057 |                  0.598057 |      99.001943 |   1.266438 |        -0.397294 | 8.590839 | 16.115828 |         1.029767 |                  5.490726 |                 7.524989 |                      7.524989 |             0.823609 | 100.0 |        0.823609 |                0.823609 | 0.7416 |  -4.91355 | -0.666518 |              0.128581 |                  0.128581 |                   0.128581 |                  1.140917 |                 1.700561 |    -1.012336 |         -1.57198 |         -1.012336 |             53.251417 |                 100.0 |               112.102145 |               99.522031 |    5.956 |   2.6802 |  0.135129 |      0.06 |     0.0 | -0.304322 |  0.003863 |    -2.103176 |     0.843881 |      0.0 |      0.0 | 0.492387 | 0.492387 | -0.811763 | -1.418866 |  -0.65865 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 15.878429 | 18.707869 |       0.0 | 0.216776 | -0.065947 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-06-12 |  0.0 | -1.1 |              0.1 |             99.0 |       2.4 |                99.8 | nan | nan |    -1.1 |                  99.0 |                   0.1 |    0.0 | -1.1 |    -1.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 0.563475 |                      0.563475 |                     0.493463 |                      0.041489 |                           0.041489 |             0.557841 |                  0.557841 |      99.242159 |   1.275245 |        -0.398979 | 8.603319 | 16.103347 |         1.030007 |                  5.433602 |                 7.500028 |                      7.500028 |             1.394624 | 100.0 |        1.394624 |                1.394624 | 0.7416 | -0.219823 |  0.595141 |              0.263213 |                  0.263213 |                   0.263213 |                  1.217715 |                 2.201596 |    -0.954501 |        -1.938383 |         -0.954501 |             88.752361 |                 100.0 |               112.102145 |               99.519666 |      0.0 |      0.0 |  0.000991 |       0.0 |     0.0 | -0.186366 |       0.0 |    -2.888087 |     0.811266 |      0.0 |      0.0 | 0.184424 | 0.184424 | -0.497123 | -0.784521 |  0.656739 | 0.996987 | 0.298195 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |  16.22287 | 18.894235 |       0.0 | 0.392168 |  0.080512 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-07-12 |  0.0 | -2.9 |              0.0 |             99.2 |       1.9 |               100.4 | nan | nan |    -2.9 |                  99.2 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 0.493024 |                      0.493024 |                     0.429685 |                      0.036869 |                           0.036869 |             0.489079 |                  0.489079 |      99.910921 |   1.291808 |        -0.400545 | 8.614964 | 16.091703 |         1.030238 |                  5.380663 |                 7.476739 |                      7.476739 |             0.807099 | 100.0 |        0.807099 |                0.807099 | 0.7416 |       0.0 |  1.062786 |               0.17276 |                   0.17276 |                    0.17276 |                  1.193664 |                 2.428029 |    -1.020904 |        -2.255269 |         -1.020904 |            112.108246 |                 100.0 |               112.102145 |                99.51746 |      0.0 |      0.0 | -0.015747 | -0.015747 |     0.0 | -0.150228 |       0.0 |    -4.724262 |     0.785949 |      0.0 |      0.0 | 0.276769 | 0.276769 | -0.400725 | -0.699317 |  1.155226 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.015747 | 16.630827 | 19.044463 |       0.0 |  0.56756 | -0.240674 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-08-12 |  0.0 | -2.9 |              0.0 |             99.1 |       1.8 |               100.6 | nan | nan |    -2.9 |                  99.1 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 0.493024 |                      0.493024 |                     0.424191 |                      0.036869 |                           0.036869 |             0.488586 |                  0.488586 |     100.111414 |   1.294388 |        -0.401992 | 8.625761 | 16.080906 |         1.030461 |                  5.331896 |                 7.455145 |                      7.455145 |             0.799784 | 100.0 |        0.799784 |                0.799784 | 0.7416 | -0.802246 |  0.901846 |              0.187367 |                  0.187367 |                   0.187367 |                  1.194208 |                 2.367263 |     -1.00684 |        -2.179896 |          -1.00684 |            118.336481 |                 100.0 |               112.102145 |               99.515415 |      0.0 |      0.0 | -0.012511 | -0.012511 |     0.0 | -0.158168 |       0.0 |    -4.893505 |     0.765728 |      0.0 |      0.0 | 0.341438 | 0.341438 | -0.421905 | -0.742105 |  1.015886 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.028258 | 17.113222 | 19.202631 |       0.0 | 0.742952 | -0.400919 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-09-12 |  0.7 | -2.1 |              4.4 |             96.9 |       2.7 |               100.9 | nan | nan |    -2.1 |                  96.9 |                   4.4 |  0.658 | -2.1 |    -2.1 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 0.523311 |                      0.523311 |                     0.381219 |                      0.038864 |                           0.038864 |             0.507088 |                  0.507088 |     100.392912 |   1.294334 |        -0.403321 | 8.635699 | 16.070967 |         1.030674 |                  5.287288 |                 7.435268 |                      7.435268 |             2.886262 | 100.0 |        2.886262 |                2.886262 | 0.7416 | -1.336398 |  1.082838 |              0.711106 |                  0.711106 |                   0.711106 |                  4.005891 |                 4.970801 |    -3.294784 |        -4.259695 |         -3.294784 |             78.890988 |                 100.0 |               112.102145 |               99.513532 | 0.626258 | 0.626258 | -0.234493 | -0.234493 |     0.0 | -0.391985 | -0.002749 |    -6.288734 |     0.753624 |      0.0 |      0.0 | 0.469924 | 0.469924 |   -1.0456 |  -1.97705 |  1.237044 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.294493 | 18.569375 | 20.220873 |       0.0 | 0.868064 | -0.742157 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-10-12 |  3.3 | -3.1 |              0.1 |             95.0 |       3.0 |               101.8 | nan | nan |    -3.1 |                  95.0 |                   0.1 |  3.102 | -3.1 |    -3.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 0.485698 |                      0.485698 |                     0.412084 |                      0.036384 |                           0.036384 |             0.461413 |                  0.461413 |     101.338587 |   1.310959 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                  7.41713 |                       7.41713 |             1.347076 | 100.0 |        1.347076 |                1.347076 | 0.7416 | -2.473858 |  0.568831 |              0.313857 |                  0.313857 |                   0.313857 |                  1.286731 |                 2.586023 |    -0.972874 |        -2.272166 |         -0.972874 |             71.001889 |                 100.0 |               112.102145 |               99.511814 | 3.336493 | 3.336493 |     0.137 |      0.06 |     0.0 | -0.164445 | -0.021617 |    -5.273465 |     0.767009 |      0.0 |      0.0 | 0.563446 | 0.563446 | -0.438649 | -1.098112 |  0.735405 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 22.626164 | 23.721812 |       0.0 | 0.730687 | -0.569388 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-11-12 |  9.4 |  5.1 |              0.2 |             90.6 |       7.4 |               100.4 | nan | nan |     5.1 |                  90.6 |                   0.2 |  8.836 |  5.1 |     5.1 |    5.931775 |         5.931775 |           5.931775 |          7.4 |                 0.879017 |                      0.879017 |                     0.770745 |                      0.061412 |                           0.061412 |              0.79639 |                   0.79639 |       99.60361 |   1.253213 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 7.400749 |                      7.400749 |             1.373029 | 100.0 |        1.373029 |                1.373029 | 0.7416 |  -1.89796 | -0.983902 |              0.344426 |                  0.344426 |                   0.344426 |                   1.16859 |                 1.249118 |    -0.824163 |        -0.904692 |         -0.824163 |              28.78455 |                 100.0 |               112.102145 |               99.510262 |    8.776 |      0.0 |  1.032826 |      0.06 |     0.0 | -0.167986 |  0.184351 |     3.233224 |     0.749149 | 1.788026 | 1.788026 |      0.0 |      0.0 | -0.448093 | -1.853352 | -1.396753 | 0.956732 | 0.274601 | 2.712549 | 0.863018 |  0.0 |  0.0 | 0.0 |  0.0 | 0.863018 | nan | nan |      0.0 | 20.946327 | 29.953248 |       0.0 | 0.918992 |  1.156114 | 73.295792 | 0.863018 |   0.0 |   0.0 |   0.0 |  0.0 | 0.222181 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 |  0.2 |  0.2 |              0.7 |             78.7 |       5.7 |               101.2 | nan | nan |     0.2 |                  78.7 |                   0.7 |  0.188 |  0.2 |     0.2 |     4.56907 |          4.56907 |            4.56907 |          5.7 |                 0.619748 |                      0.619748 |                     0.505709 |                      0.045128 |                           0.045128 |             0.487741 |                  0.487741 |     100.712259 |    1.28736 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 7.386143 |                      7.386143 |             1.539952 | 100.0 |        1.539952 |                1.539952 | 0.7416 |       0.0 |    0.5554 |              0.406881 |                  0.406881 |                   0.406881 |                  1.733999 |                 3.015522 |    -1.327117 |        -2.608641 |         -1.327117 |             37.369415 |                 100.0 |               109.464243 |               99.442447 |    0.128 |   0.0576 |  1.429251 |      0.06 |     0.0 |  0.095368 |  0.000045 |    -2.559774 |     0.735783 |      0.0 |      0.0 | 0.262108 | 0.262108 |  0.254389 |  -2.22004 |  0.642989 | 0.949764 | 0.270615 | 0.045632 | 0.014365 |  0.0 |  0.0 | 0.0 |  0.0 | 0.014365 | nan | nan |      0.0 | 21.198513 | 29.940248 |       0.0 | 1.089094 |  1.342314 | 73.327058 | 0.014365 |   0.0 |   0.0 |   0.0 |  0.0 | 0.302578 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-13-12 |  2.9 | -3.5 |              0.0 |             90.5 |       2.2 |               102.0 | nan | nan |    -3.5 |                  90.5 |                   0.0 |  2.726 | -3.5 |    -3.5 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 0.471336 |                      0.471336 |                     0.399979 |                      0.035431 |                           0.035431 |             0.426559 |                  0.426559 |     101.573441 |   1.315658 |        -0.407436 | 8.666669 | 16.039997 |         1.031437 |                  5.150203 |                 7.373328 |                      7.373328 |              0.77253 | 100.0 |         0.77253 |                 0.77253 | 0.7416 |       0.0 |   1.23294 |              0.196488 |                  0.196488 |                   0.196488 |                  1.254261 |                 2.711698 |    -1.057773 |        -2.515209 |         -1.057773 |             96.820757 |                 100.0 |               109.419886 |               99.440208 |    2.666 |    2.666 |  0.194269 |      0.06 |     0.0 | -0.073059 | -0.019502 |    -5.663335 |     0.745656 |      0.0 |      0.0 | 0.615786 | 0.615786 | -0.194883 | -0.901216 |  1.419111 | 0.968152 | 0.281196 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |  24.53515 | 32.679308 |       0.0 | 0.960708 |  0.850974 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.16446 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-14-12 |  0.0 | -3.9 |              1.2 |             88.1 |       3.1 |               102.5 | nan | nan |    -3.9 |                  88.1 |                   1.2 |    0.0 | -3.9 |    -3.9 |    2.484933 |         2.484933 |           2.484933 |          3.1 |                 0.457351 |                      0.457351 |                     0.370717 |                      0.034499 |                           0.034499 |             0.402926 |                  0.402926 |     102.097074 |   1.324197 |        -0.408164 | 8.672175 | 16.034492 |         1.031605 |                  5.126218 |                 7.362317 |                      7.362317 |             1.699321 | 100.0 |        1.699321 |                1.699321 | 0.7416 |       0.0 |  1.438009 |              0.454628 |                  0.454628 |                   0.454628 |                  2.107297 |                 3.842483 |    -1.652669 |        -3.387855 |         -1.652669 |             68.711505 |                 100.0 |               109.419886 |               99.439268 |      0.0 |      0.0 |  0.282374 |       0.0 |     0.0 | -0.109728 |       0.0 |    -6.650875 |     0.732465 |      0.0 |      0.0 | 0.709068 | 0.709068 | -0.292694 | -1.420324 |  1.674837 | 0.994215 | 0.296539 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 25.328982 | 32.789036 |       0.0 |   1.1361 |  0.154566 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.087758 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-15-12 |  0.0 | -6.3 |              3.0 |             92.8 |       2.8 |               102.3 | nan | nan |    -6.3 |                  92.8 |                   3.0 |    0.0 | -6.3 |    -6.3 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 0.380889 |                      0.380889 |                     0.294106 |                       0.02934 |                            0.02934 |             0.353465 |                  0.353465 |     101.946535 |    1.33374 |         -0.40877 | 8.676772 | 16.029895 |         1.031763 |                  5.106331 |                 7.353123 |                      7.353123 |             2.318249 | 100.0 |        2.318249 |                2.318249 | 0.7416 |       0.0 |  2.079803 |              0.646152 |                  0.646152 |                   0.646152 |                  3.397964 |                 5.166342 |    -2.751811 |        -4.520189 |         -2.751811 |             76.073452 |                 100.0 |               109.419886 |               99.438482 |      0.0 |      0.0 |  -0.02778 |  -0.02778 |     0.0 | -0.189198 |       0.0 |     -9.60802 |     0.721276 |      0.0 |      0.0 | 1.011055 | 1.011055 | -0.504677 | -1.598017 |  2.417496 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  0.02778 | 26.492023 | 32.978234 |       0.0 | 1.311492 | -1.183637 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.046829 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-16-12 |  6.7 | -3.5 |              0.0 |             98.1 |       3.7 |               100.0 | nan | nan |    -3.5 |                  98.1 |                   0.0 |  6.298 | -3.5 |    -3.5 |    2.965888 |         2.965888 |           2.965888 |          3.7 |                 0.471336 |                      0.471336 |                     0.403826 |                      0.035431 |                           0.035431 |             0.462381 |                  0.462381 |      99.537619 |   1.289645 |        -0.409256 | 8.680456 | 16.026211 |         1.031912 |                  5.090537 |                 7.345755 |                      7.345755 |             0.763581 | 100.0 |        0.763581 |                0.763581 | 0.7416 | -3.945458 |   0.17677 |              0.170099 |                  0.170099 |                   0.170099 |                  1.211777 |                 2.499465 |    -1.041678 |        -2.329366 |         -1.041678 |             57.569099 |                 100.0 |               109.419886 |               99.437853 |  6.26578 |  6.26578 |  0.036071 |  0.036071 |     0.0 | -0.225166 | -0.045834 |    -5.538318 |     0.777235 |      0.0 |      0.0 |  1.22716 |  1.22716 | -0.600618 |  -1.18794 |  0.540807 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.023929 | 34.179955 |  39.46918 |       0.0 | 0.638623 | -0.618807 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.024988 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-17-12 |  0.0 | -7.3 |              0.0 |             96.6 |       0.9 |                99.1 | nan | nan |    -7.3 |                  96.6 |                   0.0 |    0.0 | -7.3 |    -7.3 |    0.721432 |         0.721432 |           0.721432 |          0.9 |                 0.352531 |                      0.352531 |                     0.291762 |                      0.027395 |                           0.027395 |             0.340545 |                  0.340545 |      98.759455 |   1.296889 |         -0.40962 | 8.683222 | 16.023444 |         1.032051 |                  5.078832 |                 7.340222 |                      7.340222 |             0.761825 | 100.0 |        0.761825 |                0.761825 | 0.7416 | -2.062691 |  1.544362 |              0.183868 |                  0.183868 |                   0.183868 |                  1.287913 |                 3.090526 |    -1.104045 |        -2.906658 |         -1.104045 |            236.672963 |                 100.0 |               109.419886 |                99.43738 |      0.0 |      0.0 | -0.000097 | -0.000097 |     0.0 | -0.087721 |       0.0 |    -9.708863 |     0.758648 |      0.0 |      0.0 | 1.412621 | 1.412621 | -0.233991 | -0.656489 |  2.016178 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.024026 | 35.671681 |   39.5569 |       0.0 | 0.814015 |  -1.42157 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.013334 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-18-12 |  0.1 | -5.8 |              0.0 |             97.7 |       1.3 |                99.7 | nan | nan |    -5.8 |                  97.7 |                   0.0 |  0.094 | -5.8 |    -5.8 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 0.395811 |                      0.395811 |                     0.317558 |                      0.030356 |                           0.030356 |             0.386708 |                  0.386708 |      99.313292 |   1.297204 |        -0.409863 | 8.685068 | 16.021598 |         1.032182 |                  5.071211 |                  7.33653 |                       7.33653 |             0.760682 | 100.0 |        0.760682 |                0.760682 | 0.7416 | -4.738565 |  0.624118 |              0.194921 |                  0.194921 |                   0.194921 |                  1.257034 |                 2.643162 |    -1.062113 |        -2.448241 |         -1.062113 |            163.850513 |                 100.0 |               109.419886 |               99.437065 | 0.058026 | 0.058026 | -0.000728 | -0.000728 |     0.0 | -0.144568 | -0.000703 |    -8.636558 |     0.743755 |      0.0 |      0.0 | 1.618026 | 1.618026 | -0.385627 | -0.898779 |  1.163835 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.060728 | 37.484027 | 39.759493 |       0.0 | 0.984193 | -1.304087 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.007115 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-19-12 |  0.0 | -7.9 |              0.0 |             95.8 |       1.9 |                99.7 | nan | nan |    -7.9 |                  95.8 |                   0.0 |    0.0 | -7.9 |    -7.9 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  0.33643 |                       0.33643 |                     0.275734 |                      0.026283 |                           0.026283 |               0.3223 |                    0.3223 |        99.3777 |   1.307793 |        -0.409985 | 8.685992 | 16.020675 |         1.032302 |                  5.067673 |                 7.334683 |                      7.334683 |             0.760151 | 100.0 |        0.760151 |                0.760151 | 0.7416 | -4.346957 |  1.118849 |              0.204582 |                  0.204582 |                   0.204582 |                  1.301361 |                  3.17458 |    -1.096779 |        -2.969999 |         -1.096779 |            112.108246 |                 100.0 |               109.419886 |               99.436907 | 0.000728 | 0.000728 |  0.035144 |  0.035144 |     0.0 | -0.117779 | -0.000012 |   -10.418315 |     0.730867 |      0.0 |      0.0 | 1.711422 | 1.711422 | -0.314169 | -0.965378 |  1.690452 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.024856 | 39.312285 |    39.878 |       0.0 | 1.159504 | -1.681336 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.003797 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-20-12 |  0.2 | -5.7 |              0.2 |             97.5 |       2.4 |                99.7 | nan | nan |    -5.7 |                  97.5 |                   0.2 |  0.188 | -5.7 |    -5.7 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 0.398857 |                      0.398857 |                     0.325039 |                      0.030563 |                           0.030563 |             0.388886 |                  0.388886 |      99.311114 |   1.296708 |        -0.409985 | 8.685992 | 16.020675 |         1.032413 |                  5.068219 |                 7.334683 |                      7.334683 |             1.336154 | 100.0 |        1.336154 |                1.336154 | 0.7416 | -5.604455 |  0.354551 |              0.372759 |                  0.372759 |                   0.372759 |                  1.393316 |                 2.910387 |    -1.020557 |        -2.537629 |         -1.020557 |             88.752361 |                 100.0 |               109.419886 |               99.436907 | 0.152856 | 0.152856 |  0.027236 |  0.027236 |     0.0 | -0.184828 | -0.001821 |    -8.340018 |     0.721021 |      0.0 |      0.0 | 1.597527 | 0.565715 |  -0.49302 | -1.158304 |  0.886304 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.032764 | 40.215685 | 40.215685 | -0.344625 | 1.315814 | -1.294287 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.002026 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-21-12 |  0.2 | -5.6 |              0.8 |             95.0 |       2.3 |               100.4 | nan | nan |    -5.6 |                  95.0 |                   0.8 |  0.188 | -5.6 |    -5.6 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 0.401924 |                      0.401924 |                       0.3155 |                      0.030771 |                           0.030771 |             0.381828 |                  0.381828 |     100.018172 |   1.305372 |        -0.409863 | 8.685068 | 16.021598 |         1.032515 |                  5.072848 |                  7.33653 |                       7.33653 |             1.544792 | 100.0 |        1.544792 |                1.544792 | 0.7416 | -4.314291 |  0.944516 |              0.435585 |                  0.435585 |                   0.435585 |                  1.820398 |                 3.384648 |    -1.384814 |        -2.949063 |         -1.384814 |             92.611159 |                 100.0 |               109.419886 |               99.437065 | 0.160764 | 0.160764 |  0.042138 |  0.042138 |     0.0 | -0.187163 | -0.001882 |    -8.719212 |      0.71803 |      0.0 |      0.0 |  1.55036 |      0.0 | -0.499248 | -1.333986 |  1.115829 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.017862 | 40.563612 | 40.563612 |  -0.51782 | 1.367881 | -1.497203 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001081 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-22-12 |  2.4 | -4.8 |              0.0 |             93.8 |       4.6 |               100.7 | nan | nan |    -4.8 |                  93.8 |                   0.0 |  2.256 | -4.8 |    -4.8 |     3.68732 |          3.68732 |            3.68732 |          4.6 |                 0.427219 |                      0.427219 |                     0.357951 |                      0.032479 |                           0.032479 |             0.400731 |                  0.400731 |     100.299269 |   1.305282 |         -0.40962 | 8.683222 | 16.023444 |         1.032607 |                  5.081565 |                 7.340222 |                      7.340222 |             0.762235 | 100.0 |        0.762235 |                0.762235 | 0.7416 | -4.990677 |  0.501831 |              0.203398 |                  0.203398 |                   0.203398 |                  1.259735 |                 2.784042 |    -1.056337 |        -2.580645 |         -1.056337 |              46.30558 |                 100.0 |               109.419886 |                99.43738 | 2.213862 | 2.213862 |  0.242061 |      0.06 |     0.0 | -0.192655 | -0.022209 |    -7.103482 |     0.733156 |      0.0 |      0.0 | 1.595025 |      0.0 | -0.513898 | -1.572207 |   0.49454 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 42.970129 | 42.970129 | -0.532738 | 1.126133 | -1.257434 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000577 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-23-12 |  7.0 | -1.1 |              0.0 |             94.9 |       8.6 |               100.4 | nan | nan |    -1.1 |                  94.9 |                   0.0 |   6.58 | -1.1 |    -1.1 |    6.893685 |         6.893685 |           6.893685 |          8.6 |                 0.563475 |                      0.563475 |                     0.499276 |                      0.041489 |                           0.041489 |             0.534738 |                  0.534738 |      99.865262 |    1.28304 |        -0.409256 | 8.680456 | 16.026211 |         1.032689 |                  5.094372 |                 7.345755 |                      7.345755 |             0.764156 | 100.0 |        0.764156 |                0.764156 | 0.7416 | -4.191446 | -0.258034 |              0.168803 |                  0.168803 |                   0.168803 |                  1.179808 |                 2.292273 |    -1.011005 |         -2.12347 |         -1.011005 |             24.768101 |                 100.0 |               109.419886 |               99.437853 |     6.52 |    5.053 |  0.465371 |      0.06 |     0.0 | -0.263413 | -0.018373 |    -2.731361 |     0.779099 |      0.0 |      0.0 | 1.177895 | 1.177895 |  -0.70264 | -1.836558 | -0.415729 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 49.462898 | 49.753542 |       0.0 | 0.622887 |   -0.2578 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000308 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-24-12 |  0.2 | -2.9 |              0.0 |             85.4 |       7.6 |               100.8 | nan | nan |    -2.9 |                  85.4 |                   0.0 |  0.188 | -2.9 |    -2.9 |    6.092093 |         6.092093 |           6.092093 |          7.6 |                 0.493024 |                      0.493024 |                      0.42951 |                      0.036869 |                           0.036869 |             0.421042 |                  0.421042 |     100.378958 |   1.297296 |         -0.40877 | 8.676772 | 16.029895 |         1.032762 |                  5.111276 |                 7.353123 |                      7.353123 |             0.766691 | 100.0 |        0.766691 |                0.766691 | 0.7416 | -0.859333 |  0.777468 |              0.183244 |                  0.183244 |                   0.183244 |                  1.273508 |                 2.913047 |    -1.090264 |        -2.729803 |         -1.090264 |             28.027061 |                 100.0 |               109.419886 |               99.438482 |    0.128 |    0.128 |  1.135241 |      0.06 |     0.0 |  0.056709 | -0.000776 |    -4.729611 |     0.760994 |      0.0 |      0.0 | 0.740651 | 0.290644 |  0.151269 | -1.857002 |   1.02407 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 49.824833 | 49.824833 | -0.150302 | 0.789836 | -0.293668 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000164 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-25-12 |  0.0 | -9.1 |              3.2 |             87.2 |       3.8 |               101.8 | nan | nan |    -9.1 |                  87.2 |                   3.2 |    0.0 | -9.1 |    -9.1 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                  0.30617 |                       0.30617 |                     0.235908 |                      0.024175 |                           0.024175 |             0.266981 |                  0.266981 |     101.533019 |   1.341718 |        -0.408164 | 8.672175 | 16.034492 |         1.032825 |                  5.132282 |                 7.362317 |                      7.362317 |             2.398433 | 100.0 |        2.398433 |                2.398433 | 0.7416 | -0.978894 |  2.439482 |              0.598445 |                  0.598445 |                   0.598445 |                  3.730222 |                 5.736883 |    -3.131777 |        -5.138438 |         -3.131777 |             56.054123 |                 100.0 |               109.419886 |               99.439268 |      0.0 |      0.0 |  0.149964 |       0.0 |     0.0 | -0.121757 |       0.0 |    -12.35401 |     0.750485 |      0.0 |      0.0 | 1.772361 |      0.0 |  -0.32478 |  -1.93251 |  2.881148 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 49.946589 | 49.946589 | -0.591968 | 0.903496 |  -1.99155 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000088 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-26-12 |  1.2 | -8.6 |              4.3 |             92.7 |       1.5 |               101.9 | nan | nan |    -8.6 |                  92.7 |                   4.3 |  1.128 | -8.6 |    -8.6 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 0.318472 |                      0.318472 |                     0.200516 |                      0.025034 |                           0.025034 |             0.295223 |                  0.295223 |     101.604777 |   1.340358 |        -0.407436 | 8.666669 | 16.039997 |         1.032879 |                  5.157401 |                 7.373328 |                      7.373328 |             2.793204 | 100.0 |        2.793204 |                2.793204 | 0.7416 | -6.638502 |  1.583405 |              0.688582 |                  0.688582 |                   0.688582 |                  4.423241 |                 5.203894 |     -3.73466 |        -4.515313 |          -3.73466 |            142.003778 |                 100.0 |               109.419886 |               99.440208 |    1.068 |    1.068 | -0.164611 | -0.164611 |     0.0 | -0.211847 | -0.019196 |   -14.335739 |      0.75348 |      0.0 |      0.0 |   3.0942 |      0.0 | -0.565091 | -1.944519 |  2.005703 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.224611 | 51.226436 | 51.226436 | -1.033463 | 0.869662 | -2.833355 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000047 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-27-12 |  0.9 | -5.3 |              0.0 |             97.8 |       1.3 |               101.7 | nan | nan |    -5.3 |                  97.8 |                   0.0 |  0.846 | -5.3 |    -5.3 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                  0.41125 |                       0.41125 |                     0.297982 |                      0.031402 |                           0.031402 |             0.402202 |                  0.402202 |     101.297798 |   1.320718 |        -0.406588 | 8.660262 | 16.046405 |         1.032922 |                  5.186639 |                 7.386143 |                      7.386143 |             0.777996 | 100.0 |        0.777996 |                0.777996 | 0.7416 | -9.444517 |  0.006269 |              0.188873 |                  0.188873 |                   0.188873 |                  1.247575 |                  2.07679 |    -1.058702 |        -1.887917 |         -1.058702 |            163.850513 |                 100.0 |               109.419886 |               99.441303 | 1.010611 | 1.010611 | -0.001211 | -0.001211 |     0.0 | -0.217886 | -0.011195 |    -9.442818 |     0.757232 |      0.0 |      0.0 | 3.091106 |      0.0 | -0.581201 | -1.312675 | -0.005958 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.061211 | 52.454934 | 52.454934 |  -1.03243 | 0.828932 | -2.098024 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000025 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-28-12 |  0.0 | -7.7 |              3.4 |             95.4 |       3.0 |               102.2 | nan | nan |    -7.7 |                  95.4 |                   3.4 |    0.0 | -7.7 |    -7.7 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 0.341723 |                      0.341723 |                     0.236612 |                      0.026649 |                           0.026649 |             0.326003 |                  0.326003 |     101.873997 |   1.339598 |        -0.405619 | 8.652959 | 16.053708 |         1.032956 |                   5.22001 |                 7.400749 |                      7.400749 |             2.504073 | 100.0 |        2.504073 |                2.504073 | 0.7416 | -6.993415 |   1.25473 |               0.61884 |                   0.61884 |                    0.61884 |                    3.6896 |                 5.000942 |     -3.07076 |        -4.382102 |          -3.07076 |             71.001889 |                 100.0 |               109.419886 |               99.442551 | 0.001211 | 0.001211 | -0.118092 | -0.118092 |     0.0 | -0.297995 | -0.000019 |   -12.317386 |     0.752867 |      0.0 |      0.0 |  3.09002 |      0.0 | -0.794886 | -2.332869 |  1.254347 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.178092 | 52.754139 | 52.754139 | -1.032067 |  0.87649 | -2.611154 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000013 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-29-12 |  0.0 | -9.2 |              0.0 |             94.9 |       2.7 |               102.4 | nan | nan |    -9.2 |                  94.9 |                   0.0 |    0.0 | -9.2 |    -9.2 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 0.303761 |                      0.303761 |                     0.233442 |                      0.024006 |                           0.024006 |              0.28827 |                   0.28827 |      102.11173 |   1.350039 |         -0.40453 | 8.644768 | 16.061898 |         1.032981 |                  5.257524 |                  7.41713 |                       7.41713 |             0.788629 | 100.0 |        0.788629 |                0.788629 | 0.7416 | -8.703847 |  0.833239 |               0.19749 |                   0.19749 |                    0.19749 |                  1.325064 |                  3.14091 |    -1.127573 |         -2.94342 |         -1.127573 |             78.890988 |                 100.0 |               109.419886 |               99.443951 | 0.118092 | 0.118092 |  0.073994 |      0.06 |     0.0 | -0.170745 | -0.002271 |   -12.483207 |     0.749578 |      0.0 |      0.0 | 3.411424 |      0.0 | -0.455455 | -1.549648 |  0.938317 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 53.042976 | 53.042976 | -1.139415 | 0.913988 | -2.702793 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000007 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-30-12 |  0.2 | -7.4 |              0.0 |             96.0 |       1.4 |               102.1 | nan | nan |    -7.4 |                  96.0 |                   0.0 |  0.188 | -7.4 |    -7.4 |    1.122228 |         1.122228 |           1.122228 |          1.4 |                 0.349801 |                      0.349801 |                     0.254353 |                      0.027207 |                           0.027207 |             0.335809 |                  0.335809 |     101.764191 |   1.336727 |        -0.403321 | 8.635699 | 16.070967 |         1.032995 |                  5.299194 |                 7.435268 |                      7.435268 |             0.794879 | 100.0 |        0.794879 |                0.794879 | 0.7416 | -9.009311 |  0.577914 |              0.201287 |                  0.201287 |                   0.201287 |                  1.292505 |                 2.552653 |    -1.091217 |        -2.351366 |         -1.091217 |            152.146905 |                 100.0 |               109.419886 |               99.445501 |    0.128 |    0.128 |  0.020026 |  0.020026 |     0.0 |  -0.17625 |  -0.00198 |    -11.42408 |      0.74677 |      0.0 |      0.0 | 3.368461 |      0.0 | -0.470138 | -1.319644 |  0.561585 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.039974 | 53.347226 | 53.347226 | -1.125066 | 0.947207 | -2.539107 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000004 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-31-12 |  0.9 |  1.7 |              4.0 |             96.5 |       4.2 |               101.6 | nan | nan |     1.7 |                  96.5 |                   4.0 |  0.846 |  1.7 |     1.7 |    3.366683 |         3.366683 |           3.366683 |          4.2 |                 0.690794 |                      0.690794 |                     0.493144 |                      0.049663 |                           0.049663 |             0.666616 |                  0.666616 |     100.933384 |   1.284547 |        -0.401992 | 8.625761 | 16.080906 |            1.033 |                  5.345035 |                 7.455145 |                      7.455145 |             2.770177 | 100.0 |        2.770177 |                2.770177 | 0.7416 | -8.463692 | -1.098561 |              0.710667 |                  0.710667 |                   0.710667 |                  3.475519 |                 4.103406 |    -2.764852 |        -3.392739 |         -2.764852 |             50.715635 |                 100.0 |               109.419886 |               99.447199 | 0.825974 | 0.061948 | -0.121701 | -0.121701 |     0.0 | -0.730482 |  0.005658 |    -2.896724 |     0.743458 |      0.0 |      0.0 | 2.181159 | 0.764026 | -1.948525 | -2.933677 | -1.489462 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.181701 | 54.903682 | 54.903682 | -0.473322 | 0.987911 | -0.698946 | 73.327058 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000002 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |

>>> ktschnee(inf)

hourly simulation
_________________

Now we change the simulation time step size from one day to one hour.
Consequently, we need to change the values of all time-dependent control
parameters by the factor 1/24 (or, in single cases, 24).  Also, we need
to |FixedParameter.restore| all time-dependent fixed parameter values:

>>> pub.timegrids = '1997-08-03', '1997-08-06', '1h'
>>> for parameter in control:
...     if parameter.TIME == True:
...         parameter.value = parameter.value/24
...     if parameter.TIME == False:
...         parameter.value = parameter.value*24
>>> for parameter in fixed:
...     parameter.restore()
>>> parameterstep('1h')
>>> test = IntegrationTest(land)

.. _lland_v3_acker_summer_hourly:

acre (summer)
-------------

The following input values agree with the ones of the respective sub-period
of the daily integration test :ref:`lland_v3_acker_summer_daily`:

>>> inputs.nied.series = 0.0
>>> inputs.teml.series = (
...     16.9, 16.6, 16.4, 16.3, 16.0, 15.9, 16.0, 16.6, 17.4, 19.0, 20.3,
...     21.4, 21.3, 21.8, 22.9, 22.7, 22.5, 21.9, 21.4, 20.7, 19.4, 17.8,
...     17.0, 16.4, 15.9, 15.4, 15.2, 14.9, 14.3, 14.1, 14.3, 15.4, 16.8,
...     18.2, 19.5, 20.3, 21.0, 21.8, 21.9, 21.8, 22.2, 21.8, 20.9, 19.7,
...     17.9, 15.8, 14.8, 14.1, 13.4, 12.8, 12.3, 12.0, 11.7, 11.6, 12.6,
...     14.8, 17.1, 19.4, 20.6, 21.4, 22.5, 23.1, 23.4, 23.8, 24.0, 24.0,
...     23.7, 22.8, 21.3, 19.3, 18.3, 17.4)
>>> inputs.sunshineduration.series = (
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.5, 0.7, 0.8, 0.5,
...     0.4, 0.5, 0.5, 0.3, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.1, 0.9, 1.0, 1.0, 0.9, 0.8, 0.9, 0.8, 0.9, 0.9, 0.9, 1.0,
...     1.0, 1.0, 0.3, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.6, 0.9,
...     1.0, 0.9, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.5, 0.0,
...     0.0, 0.0)
>>> inputs.windspeed.series = (
...     0.8, 0.8, 0.8, 0.8, 0.8, 0.6, 0.9, 0.9, 0.9, 1.3, 1.5, 1.2, 1.3, 1.5,
...     1.9, 1.9, 2.3, 2.4, 2.5, 2.5, 2.2, 1.7, 1.7, 2.3, 2.3, 2.2, 2.3, 2.0,
...     2.3, 2.5, 2.4, 2.5, 2.5, 2.9, 3.1, 3.3, 3.3, 3.2, 2.5, 2.9, 3.6, 4.2,
...     4.2, 3.6, 3.2, 2.2, 2.2, 2.1, 1.8, 1.8, 1.6, 1.2, 1.9, 1.3, 1.6, 1.9,
...     3.2, 4.0, 4.6, 4.1, 4.1, 4.4, 4.6, 4.5, 4.7, 4.5, 4.1, 3.8, 2.5, 1.9,
...     2.2, 2.7)
>>> inputs.relativehumidity.series = (
...     95.1, 94.9, 95.9, 96.7, 97.2, 97.5, 97.7, 97.4, 96.8, 86.1, 76.8,
...     71.8, 67.5, 66.1, 63.4, 62.4, 61.1, 62.1, 67.0, 74.5, 81.2, 86.9,
...     90.1, 90.9, 88.0, 89.2, 88.8, 89.7, 93.0, 93.6, 94.6, 93.3, 86.9,
...     78.9, 75.8, 73.4, 68.3, 61.3, 60.6, 58.8, 57.2, 57.2, 59.9, 66.1,
...     71.7, 80.6, 85.9, 89.3, 91.4, 93.1, 94.6, 95.6, 96.2, 95.7, 95.1,
...     96.3, 89.9, 79.2, 73.4, 67.3, 62.2, 59.6, 55.4, 52.1, 49.5, 48.1,
...     49.4, 57.0, 65.9, 73.0, 76.7, 81.8)
>>> inputs.atmosphericpressure.series = (
...     101.5, 101.5, 101.5, 101.5, 101.5, 101.5, 101.5, 101.5, 101.6, 101.6,
...     101.6, 101.6, 101.6, 101.6, 101.6, 101.6, 101.6, 101.6, 101.6, 101.6,
...     101.6, 101.6, 101.7, 101.7, 101.7, 101.6, 101.6, 101.6, 101.6, 101.7,
...     101.7, 101.7, 101.7, 101.7, 101.7, 101.8, 101.8, 101.8, 101.8, 101.7,
...     101.7, 101.7, 101.7, 101.7, 101.7, 101.8, 101.8, 101.8, 101.8, 101.8,
...     101.8, 101.8, 101.8, 101.8, 101.8, 101.9, 101.9, 101.9, 101.9, 101.9,
...     101.9, 101.9, 101.9, 101.9, 101.8, 101.8, 101.8, 101.8, 101.8, 101.8,
...     101.8, 101.8)

In this and the following "summer" examples, we use the initial conditions
of August 3rd, calculated by the corresponding daily integration tests.
However, this is possible for the state sequences, but not for the (now
higher resolved) log sequences.  Hence, we need to define them manually:

>>> test.inits = (
...     (logs.loggedsunshineduration,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2,
...        0.1, 0.2, 0.1, 0.2, 0.2, 0.3, 0.0, 0.0,
...        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedglobalradiation,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.2, 0.5,
...       0.8, 1.1, 1.2, 1.4, 1.9, 1.6, 0.9, 0.8,
...       0.6, 0.4, 0.2, 0.1, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedteml,
...      [13.2, 13.2, 13.1, 12.6, 12.7, 13.0, 13.5, 14.8,
...       16.2, 17.7, 18.8, 19.4, 20.4, 21.0, 21.5, 21.2,
...       20.4, 20.7, 20.2, 19.7, 19.0, 18.0, 17.5, 17.1]),
...     (logs.loggedrelativehumidity,
...      [95.1, 94.5, 94.8, 96.4, 96.6, 97.1, 97.1, 96.7,
...       92.2, 88.5, 81.1, 76.5, 75.1, 70.8, 68.9, 69.2,
...       75.0, 74.0, 77.4, 81.4, 85.3, 90.1, 92.3, 93.8]),
...     (logs.loggedwindspeed2m,
...      [0.8, 1.0, 1.2, 1.3, 0.9, 1.1, 1.3, 1.3,
...       1.9, 2.2, 1.8, 2.3, 2.4, 2.5, 2.4, 2.5,
...       2.1, 2.2, 1.7, 1.7, 1.3, 1.3, 0.7, 0.8]))

This integration test deals with a dry situation.  Hence, the soil water
content (|BoWa|) shows a pronounced decline.  This decline is sharpest around
noon, when evapotranspiration out of the soil (|EvB|) reaches its maximum:

.. integration-test::

    >>> test('lland_v3_acker_summer_hourly',
    ...      axis1=(fluxes.evb, fluxes.qah),
    ...      axis2=states.bowa,
    ...      use_conditions=conditions_acker_summer)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration | nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |         g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |      evpo |       evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz | qah |  qa |     inzp | wats | waes | esnow | taus |      ebdn |      bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet | outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-08-03 00:00:00 |  0.0 | 16.9 |              0.0 |             95.1 |       0.8 |               101.5 | nan | nan | 17.279167 |             85.883333 |                   1.3 |  0.0 | 16.9 | 17.279167 |    0.641273 |         1.605886 |           0.641273 |          0.8 |                 1.928423 |                      1.975309 |                          0.0 |                      0.122362 |                           0.124959 |              1.83393 |                  1.696461 |       99.66607 |    1.21073 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 |  0.134093 | 15.208683 | -0.036532 |                   0.0 |                       0.0 |                   7.982856 |                  0.901196 |                      0.0 |     -0.03755 |              0.0 |           7.08166 |            176.542034 |                 100.0 |               100.629116 |               87.719298 |  0.0 |  0.0 |   0.02723 |       0.0 | 0.021938 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.308721 | 79.601249 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.256857 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 01:00:00 |  0.0 | 16.6 |              0.0 |             94.9 |       0.8 |               101.5 | nan | nan | 17.241667 |             85.991667 |                   1.3 |  0.0 | 16.6 | 17.241667 |    0.641273 |         1.603439 |           0.641273 |          0.8 |                 1.892018 |                      1.970628 |                          0.0 |                      0.120339 |                             0.1247 |             1.795525 |                  1.694576 |      99.704475 |   1.212158 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 |  0.134078 | 15.295735 | -0.028172 |                   0.0 |                       0.0 |                   7.982856 |                  0.901844 |                      0.0 |    -0.037577 |              0.0 |          7.081012 |            176.542034 |                 100.0 |               100.659301 |               87.719298 |  0.0 |  0.0 |  0.027289 |       0.0 | 0.021953 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.326476 | 79.579296 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.250222 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 02:00:00 |  0.0 | 16.4 |              0.0 |             95.9 |       0.8 |               101.5 | nan | nan |    17.175 |             86.233333 |                   1.3 |  0.0 | 16.4 |    17.175 |    0.641273 |         1.575992 |           0.641273 |          0.8 |                 1.868084 |                       1.96233 |                          0.0 |                      0.119006 |                           0.124241 |             1.791493 |                  1.692183 |      99.708507 |   1.213014 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 |  0.134066 |  15.35492 | -0.022574 |                   0.0 |                       0.0 |                   7.982856 |                  0.902355 |                      0.0 |    -0.037598 |              0.0 |            7.0805 |            176.542034 |                 100.0 |               100.689509 |               87.719298 |  0.0 |  0.0 |  0.025718 |       0.0 | 0.020669 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.338633 | 79.558627 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.243758 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 03:00:00 |  0.0 | 16.3 |              0.0 |             96.7 |       0.8 |               101.5 | nan | nan |   17.0625 |             86.708333 |                   1.3 |  0.0 | 16.3 |   17.0625 |    0.641273 |         1.548545 |           0.641273 |          0.8 |                 1.856216 |                      1.948396 |                          0.0 |                      0.118345 |                           0.123469 |             1.794961 |                  1.689422 |      99.705039 |   1.213417 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 |  0.134066 | 15.395444 | -0.019538 |                   0.0 |                       0.0 |                   7.982856 |                  0.902364 |                      0.0 |    -0.037598 |              0.0 |          7.080492 |            176.542034 |                 100.0 |               100.717954 |               87.719298 |  0.0 |  0.0 |  0.024521 |       0.0 | 0.019696 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.347755 | 79.538931 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.237462 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 04:00:00 |  0.0 | 16.0 |              0.0 |             97.2 |       0.8 |               101.5 | nan | nan | 16.908333 |             87.366667 |                   1.3 |  0.0 | 16.0 | 16.908333 |    0.641273 |         1.504432 |           0.641273 |          0.8 |                 1.821009 |                      1.929443 |                          0.0 |                      0.116379 |                           0.122418 |              1.77002 |                   1.68569 |       99.72998 |    1.21479 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                 0.429734 |                     15.566134 |            10.503757 |  0.086656 |        0.007799 |                0.006997 |  0.038337 | 15.425849 | -0.012402 |              0.005318 |                  0.005318 |                   7.982856 |                  0.902341 |                      0.0 |     -0.03228 |              0.0 |          7.080514 |            176.542034 |                 100.0 |               100.745063 |                92.80129 |  0.0 |  0.0 |  0.005147 |       0.0 | 0.004081 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  31.34974 | 79.534849 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.231328 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 05:00:00 |  0.0 | 15.9 |              0.0 |             97.5 |       0.6 |               101.5 | nan | nan | 16.729167 |             88.204167 |                   1.3 |  0.0 | 15.9 | 16.729167 |    0.480955 |         1.453638 |           0.480955 |          0.6 |                 1.809403 |                      1.907618 |                          0.0 |                       0.11573 |                           0.121206 |             1.764168 |                  1.682599 |      99.735832 |   1.215237 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  0.968392 |        0.087159 |                0.078957 |  -0.08871 | 15.432466 | -0.010099 |              0.060007 |                  0.060007 |                   7.982856 |                  0.901482 |                      0.0 |     0.022445 |              0.0 |          7.081373 |            235.389378 |                 100.0 |                100.75068 |              100.534104 |  0.0 |  0.0 |  -0.00983 |  -0.00983 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |  0.00983 |  0.0 |  0.0 |   0.0 |  nan | 31.349422 | 79.534849 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.225353 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 06:00:00 |  0.0 | 16.0 |              0.0 |             97.7 |       0.9 |               101.5 | nan | nan | 16.533333 |             89.191667 |                   1.3 |  0.0 | 16.0 | 16.533333 |    0.721432 |         1.392031 |           0.721432 |          0.9 |                 1.821009 |                      1.884011 |                          0.0 |                      0.116379 |                           0.119893 |             1.779125 |                   1.68038 |      99.720875 |   1.214748 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |    2.4897 |        0.224084 |                0.206122 | -0.088731 | 15.431406 | -0.012282 |              0.156653 |                  0.156653 |                   7.982856 |                   0.89977 |                      0.0 |     0.119163 |              0.0 |          7.083086 |            156.926252 |                 100.0 |                100.75068 |              100.534104 |  0.0 |  0.0 |  0.009895 |   0.00983 |  0.00005 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.351287 |   79.5348 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.219531 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 07:00:00 |  0.0 | 16.6 |              0.0 |             97.4 |       0.9 |               101.5 | nan | nan |    16.375 |                90.125 |                   1.3 |  0.0 | 16.6 |    16.375 |    0.721432 |         1.334591 |           0.721432 |          0.9 |                 1.892018 |                      1.865111 |                          0.0 |                      0.120339 |                           0.118841 |             1.842826 |                  1.680931 |      99.657174 |   1.211943 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  4.679733 |        0.421196 |                0.393598 | -0.088768 | 15.437623 | -0.025107 |              0.299135 |                  0.299135 |                   7.982856 |                  0.896826 |                      0.0 |     0.261767 |              0.0 |          7.086029 |            156.926252 |                 100.0 |               100.750749 |              100.534152 |  0.0 |  0.0 |  0.040812 |       0.0 | 0.031261 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.365978 | 79.503539 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.213861 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 08:00:00 |  0.0 | 17.4 |              0.0 |             96.8 |       0.9 |               101.6 | nan | nan | 16.216667 |                91.275 |                   1.3 |  0.0 | 17.4 | 16.216667 |    0.721432 |         1.260484 |           0.721432 |          0.9 |                 1.990459 |                      1.846377 |                          0.0 |                      0.125796 |                           0.117796 |             1.926764 |                  1.685281 |      99.673236 |   1.209425 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  7.219274 |        0.649765 |                0.615417 | -0.088836 | 15.486592 |  -0.04133 |              0.467717 |                  0.467717 |                   7.982856 |                  0.891371 |                      0.0 |     0.430576 |              0.0 |          7.091484 |            156.926252 |                 100.0 |               100.793779 |              100.564699 |  0.0 |  0.0 |  0.079122 |       0.0 | 0.060899 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  31.39689 |  79.44264 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.208337 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 09:00:00 |  0.0 | 19.0 |              0.2 |             86.1 |       1.3 |               101.6 | nan | nan |   16.1125 |             91.991667 |                   1.5 |  0.0 | 19.0 |   16.1125 |    1.042069 |         1.203904 |           1.042069 |          1.3 |                 2.200854 |                      1.834142 |                          0.0 |                      0.137341 |                           0.117113 |             1.894936 |                  1.687258 |      99.705064 |   1.202945 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             10.73504 |  9.442017 |         1.18975 |                1.122345 | -0.090647 | 15.589635 | -0.073664 |              0.852983 |                  0.852983 |                    8.15863 |                  0.922604 |                      0.0 |     0.814541 |              0.0 |          7.236026 |            108.641252 |                 100.0 |               100.877624 |              100.624196 |  0.0 |  0.0 |  0.203678 |       0.0 | 0.147201 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.460138 | 79.295438 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.202955 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 10:00:00 |  0.0 | 20.3 |              0.5 |             76.8 |       1.5 |               101.6 | nan | nan | 16.083333 |             92.241667 |                   1.7 |  0.0 | 20.3 | 16.083333 |    1.202387 |         1.149836 |           1.202387 |          1.5 |                  2.38585 |                      1.830729 |                          0.0 |                      0.147368 |                           0.116922 |             1.832333 |                  1.688695 |      99.767667 |   1.197896 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.966322 | 10.709164 |         1.92774 |                  1.8057 | -0.092439 | 15.800459 |  -0.09719 |              1.372332 |                  1.372332 |                   8.334404 |                  0.955299 |                      0.0 |     1.332528 |              0.0 |          7.379106 |             94.155751 |                 100.0 |               101.080377 |              100.767956 |  0.0 |  0.0 |  0.369534 |       0.0 | 0.261128 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.546911 |  79.03431 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.197713 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 11:00:00 |  0.0 | 21.4 |              0.7 |             71.8 |       1.2 |               101.6 | nan | nan |    16.125 |             92.104167 |                   2.2 |  0.0 | 21.4 |    16.125 |    0.961909 |         1.089916 |           0.961909 |          1.2 |                 2.552842 |                      1.835607 |                          0.0 |                      0.156328 |                           0.117195 |             1.832941 |                   1.69067 |      99.767059 |    1.19342 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            11.544527 | 10.773868 |        2.327265 |                2.214067 | -0.096884 | 16.089704 | -0.114702 |              1.682691 |                  1.682691 |                   8.773841 |                  1.039981 |                      0.0 |     1.639359 |              0.0 |          7.733859 |            117.694689 |                 100.0 |               101.440371 |              101.022794 |  0.0 |  0.0 |  0.444939 |       0.0 |  0.33287 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.651197 |  78.70144 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.192605 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 12:00:00 |  0.0 | 21.3 |              0.8 |             67.5 |       1.3 |               101.6 | nan | nan | 16.204167 |             91.729167 |                   2.8 |  0.0 | 21.3 | 16.204167 |    1.042069 |         1.037502 |           1.042069 |          1.3 |                 2.537251 |                      1.844905 |                          0.0 |                      0.155495 |                           0.117714 |             1.712644 |                  1.692316 |      99.887356 |   1.194363 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.238374 | 10.364714 |        2.425457 |                 2.25556 | -0.102203 | 16.437323 | -0.105034 |              1.714225 |                  1.714225 |                   9.301164 |                  1.142661 |                      0.0 |     1.666614 |              0.0 |          8.158503 |            108.641252 |                 100.0 |               101.899865 |              101.347307 |  0.0 |  0.0 |  0.468211 |       0.0 | 0.344049 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.745814 | 78.357392 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.18763 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 13:00:00 |  0.0 | 21.8 |              0.5 |             66.1 |       1.5 |               101.6 | nan | nan | 16.329167 |             91.104167 |                   3.2 |  0.0 | 21.8 | 16.329167 |    1.202387 |         1.012602 |           1.202387 |          1.5 |                 2.616045 |                      1.859671 |                          0.0 |                      0.159698 |                           0.118538 |             1.729206 |                  1.694238 |      99.870794 |   1.192265 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.700938 | 10.872489 |         1.95714 |                 1.78608 | -0.105739 | 16.752713 | -0.109021 |              1.357421 |                  1.357421 |                   9.652713 |                  1.211985 |                      0.0 |     1.306922 |              0.0 |          8.440727 |             94.155751 |                 100.0 |               102.375502 |              101.682325 |  0.0 |  0.0 |  0.400429 |       0.0 | 0.286018 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.844419 | 78.071374 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.182784 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 14:00:00 |  0.0 | 22.9 |              0.4 |             63.4 |       1.9 |               101.6 | nan | nan | 16.545833 |             90.058333 |                   3.4 |  0.0 | 22.9 | 16.545833 |    1.523023 |         0.984394 |           1.523023 |          1.9 |                 2.796942 |                       1.88551 |                          0.0 |                      0.169283 |                           0.119977 |             1.773261 |                  1.698059 |      99.826739 |   1.187639 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             12.93222 | 10.461143 |        1.694785 |                1.510276 | -0.107495 | 17.081396 | -0.125682 |              1.147809 |                  1.147809 |                   9.828487 |                   1.24753 |                      0.0 |     1.095829 |              0.0 |          8.580957 |             74.333488 |                 100.0 |               102.771471 |              101.960537 |  0.0 |  0.0 |  0.389549 |       0.0 | 0.264958 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.959684 | 77.806416 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.178062 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 15:00:00 |  0.0 | 22.7 |              0.5 |             62.4 |       1.9 |               101.6 | nan | nan | 16.816667 |             88.816667 |                   3.8 |  0.0 | 22.7 | 16.816667 |    1.523023 |         0.968687 |           1.523023 |          1.9 |                 2.763263 |                      1.918249 |                          0.0 |                      0.167505 |                           0.121797 |             1.724276 |                  1.703725 |      99.875724 |    1.18866 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.394785 |  8.865382 |        1.595844 |                1.396394 |  -0.11103 | 17.465613 | -0.113063 |               1.06126 |                   1.06126 |                  10.180036 |                  1.316893 |                      0.0 |     1.006389 |              0.0 |          8.863143 |             74.333488 |                 100.0 |               103.138744 |              102.218027 |  0.0 |  0.0 |  0.366746 |       0.0 | 0.248732 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.06233 | 77.557684 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.173463 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 16:00:00 |  0.0 | 22.5 |              0.5 |             61.1 |       2.3 |               101.6 | nan | nan |   17.1375 |             87.333333 |                   4.1 |  0.0 | 22.5 |   17.1375 |     1.84366 |         0.991339 |            1.84366 |          2.3 |                 2.729939 |                      1.957676 |                          0.0 |                      0.165743 |                           0.123983 |             1.667993 |                  1.709704 |      99.932007 |   1.189715 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.741708 |  6.474205 |        1.165412 |                1.002589 | -0.113666 | 17.807767 | -0.101352 |              0.761967 |                  0.761967 |                  10.443698 |                  1.370145 |                      0.0 |     0.704878 |              0.0 |          9.073553 |             61.405925 |                 100.0 |               103.483932 |              102.459542 |  0.0 |  0.0 |  0.314711 |       0.0 | 0.201449 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.153266 | 77.356234 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.168982 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 17:00:00 |  0.0 | 21.9 |              0.3 |             62.1 |       2.4 |               101.6 | nan | nan |   17.4875 |                85.875 |                   4.4 |  0.0 | 21.9 |   17.4875 |    1.923819 |         1.017332 |           1.923819 |          2.4 |                 2.632058 |                      2.001493 |                          0.0 |                       0.16055 |                           0.126406 |             1.634508 |                  1.718782 |      99.965492 |   1.192283 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.088631 |  3.977658 |         0.57281 |                0.493699 | -0.116332 | 18.110885 | -0.081845 |              0.375211 |                  0.375211 |                   10.70736 |                  1.421011 |                      0.0 |     0.316002 |              0.0 |          9.286348 |             58.847345 |                 100.0 |               103.763792 |              102.655003 |  0.0 |  0.0 |  0.208264 |       0.0 | 0.130498 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.224694 | 77.225736 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.164617 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 18:00:00 |  0.0 | 21.4 |              0.1 |             67.0 |       2.5 |               101.6 | nan | nan |   17.8375 |             84.620833 |                   4.5 |  0.0 | 21.4 |   17.8375 |    2.003978 |         1.054998 |           2.003978 |          2.5 |                 2.552842 |                      2.046164 |                          0.0 |                      0.156328 |                           0.128868 |             1.710404 |                  1.731481 |      99.889596 |   1.193968 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  1.972911 |        0.213084 |                0.183891 | -0.117274 | 18.348979 | -0.065902 |              0.139757 |                  0.139757 |                  10.795247 |                  1.433696 |                      0.0 |     0.080019 |              0.0 |           9.36155 |             56.493451 |                 100.0 |               103.945225 |              102.781554 |  0.0 |  0.0 |  0.128039 |       0.0 |  0.07863 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.280179 | 77.147106 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.160365 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 19:00:00 |  0.0 | 20.7 |              0.0 |             74.5 |       2.5 |               101.6 | nan | nan | 18.170833 |                  83.7 |                   4.5 |  0.0 | 20.7 | 18.170833 |    2.003978 |         1.100997 |           2.003978 |          2.5 |                 2.445437 |                      2.089517 |                          0.0 |                      0.150575 |                           0.131251 |              1.82185 |                  1.748925 |       99.77815 |   1.196313 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  0.634046 |        0.057067 |                0.049078 | -0.117398 | 18.533931 | -0.046787 |              0.037299 |                  0.037299 |                  10.795247 |                  1.423756 |                      0.0 |    -0.022024 |              0.0 |          9.371491 |             56.493451 |                 100.0 |               104.054599 |              102.857781 |  0.0 |  0.0 |  0.068101 |       0.0 | 0.041435 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.31655 | 77.105671 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.156222 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 20:00:00 |  0.0 | 19.4 |              0.0 |             81.2 |       2.2 |               101.6 | nan | nan | 18.454167 |             83.066667 |                   4.5 |  0.0 | 19.4 | 18.454167 |    1.763501 |         1.120309 |           1.763501 |          2.2 |                 2.256393 |                      2.126995 |                          0.0 |                      0.140363 |                           0.133306 |             1.832191 |                  1.766824 |      99.767809 |   1.201582 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                   0.1364 |                     15.566134 |            14.204272 |  0.008648 |        0.000778 |                0.000669 |  0.145725 | 18.655165 | -0.016088 |              0.000509 |                  0.000509 |                  10.795247 |                  1.412078 |                      0.0 |    -0.058328 |              0.0 |          9.383169 |             64.197103 |                 100.0 |                104.11225 |               89.520507 |  0.0 |  0.0 |  0.088999 |       0.0 |  0.05843 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.322221 | 77.047242 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.152187 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 21:00:00 |  0.0 | 17.8 |              0.0 |             86.9 |       1.7 |               101.6 | nan | nan |     18.65 |               82.7375 |                   4.5 |  0.0 | 17.8 |     18.65 |    1.362705 |         1.127089 |           1.362705 |          1.7 |                 2.041337 |                      2.153241 |                          0.0 |                      0.128603 |                           0.134742 |             1.773922 |                  1.781538 |      99.826078 |   1.208454 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 |   0.18755 | 18.674071 |   0.01888 |                   0.0 |                       0.0 |                  10.795247 |                  1.401564 |                      0.0 |    -0.058399 |              0.0 |          9.393683 |             83.078604 |                 100.0 |               104.193568 |               87.719298 |  0.0 |  0.0 |  0.066711 |       0.0 | 0.046565 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.292925 | 77.000676 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.148256 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 22:00:00 |  0.0 | 17.0 |              0.0 |             90.1 |       1.7 |               101.7 | nan | nan | 18.808333 |             82.554167 |                   4.5 |  0.0 | 17.0 | 18.808333 |    1.362705 |         1.142201 |           1.362705 |          1.7 |                 1.940693 |                      2.174668 |                          0.0 |                      0.123042 |                           0.135912 |             1.748564 |                  1.795279 |      99.951436 |   1.213101 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 |   0.18779 | 18.576416 |  0.034051 |                   0.0 |                       0.0 |                  10.795247 |                  1.391158 |                      0.0 |    -0.057965 |              0.0 |          9.404089 |             83.078604 |                 100.0 |               104.258389 |               87.719298 |  0.0 |  0.0 |  0.057008 |       0.0 | 0.039502 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.248458 | 76.961174 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.144426 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 23:00:00 |  0.0 | 16.4 |              0.0 |             90.9 |       2.3 |               101.7 | nan | nan | 18.941667 |             82.379167 |                   4.5 |  0.0 | 16.4 | 18.941667 |     1.84366 |         1.185687 |            1.84366 |          2.3 |                 1.868084 |                      2.192855 |                          0.0 |                      0.119006 |                           0.136905 |             1.698088 |                  1.806456 |     100.001912 |   1.215845 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 |  0.187983 | 18.428192 |  0.043809 |                   0.0 |                       0.0 |                  10.795247 |                   1.38284 |                      0.0 |    -0.057618 |              0.0 |          9.412407 |             61.405925 |                 100.0 |                104.31339 |               87.719298 |  0.0 |  0.0 |  0.062194 |       0.0 |  0.03922 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.194232 | 76.921954 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.140696 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 00:00:00 |  0.0 | 15.9 |              0.0 |             88.0 |       2.3 |               101.7 | nan | nan |      18.9 |             82.083333 |                   4.5 |  0.0 | 15.9 |      18.9 |     1.84366 |         1.235786 |            1.84366 |          2.3 |                 1.809403 |                      2.187158 |                          0.0 |                       0.11573 |                           0.136594 |             1.592275 |                  1.795292 |     100.107725 |    1.21843 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 |  0.184992 |  18.24744 |  0.050705 |                   0.0 |                       0.0 |                  10.739718 |                   1.39654 |                      0.0 |    -0.058189 |              0.0 |          9.343179 |             61.405925 |                 100.0 |               104.368008 |               87.719298 |  0.0 |  0.0 |  0.070298 |       0.0 | 0.044076 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.133111 | 76.877878 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.137061 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 01:00:00 |  0.0 | 15.4 |              0.0 |             89.2 |       2.2 |               101.6 | nan | nan |     18.85 |             81.845833 |                   4.5 |  0.0 | 15.4 |     18.85 |    1.763501 |         1.282546 |           1.763501 |          2.2 |                 1.752342 |                      2.180337 |                          0.0 |                      0.112529 |                           0.136222 |             1.563089 |                  1.784515 |     100.036911 |   1.219467 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 |   0.18476 | 18.043702 |  0.057104 |                   0.0 |                       0.0 |                  10.739718 |                  1.406633 |                      0.0 |     -0.05861 |              0.0 |          9.333085 |             64.197103 |                 100.0 |               104.429401 |               87.719298 |  0.0 |  0.0 |  0.063738 |       0.0 | 0.040313 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.06559 | 76.837564 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.133521 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 02:00:00 |  0.0 | 15.2 |              0.0 |             88.8 |       2.3 |               101.6 | nan | nan |      18.8 |                 81.55 |                   4.5 |  0.0 | 15.2 |      18.8 |     1.84366 |         1.332645 |            1.84366 |          2.3 |                 1.729962 |                      2.173535 |                          0.0 |                      0.111271 |                            0.13585 |             1.536206 |                  1.772518 |     100.063794 |   1.220436 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 |  0.184498 | 17.818633 |  0.056562 |                   0.0 |                       0.0 |                  10.739718 |                  1.418069 |                      0.0 |    -0.059086 |              0.0 |           9.32165 |             61.405925 |                 100.0 |               104.485563 |               87.719298 |  0.0 |  0.0 |  0.066018 |       0.0 | 0.041058 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.998611 | 76.796507 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.130072 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 03:00:00 |  0.0 | 14.9 |              0.0 |             89.7 |       2.0 |               101.6 | nan | nan | 18.741667 |             81.258333 |                   4.5 |  0.0 | 14.9 | 18.741667 |    1.603182 |         1.372725 |           1.603182 |          2.0 |                 1.696861 |                      2.165623 |                          0.0 |                      0.109405 |                           0.135418 |             1.522084 |                  1.759749 |     100.077916 |   1.221771 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 |   0.18422 | 17.595369 |   0.05822 |                   0.0 |                       0.0 |                  10.739718 |                   1.43017 |                      0.0 |     -0.05959 |              0.0 |          9.309548 |             70.616813 |                 100.0 |               104.542774 |               87.719298 |  0.0 |  0.0 |  0.058137 |       0.0 | 0.037684 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.929974 | 76.758823 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.126712 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 04:00:00 |  0.0 | 14.3 |              0.0 |             93.0 |       2.3 |               101.6 | nan | nan | 18.670833 |             81.083333 |                   4.5 |  0.0 | 14.3 | 18.670833 |     1.84366 |         1.422824 |            1.84366 |          2.3 |                 1.632319 |                       2.15605 |                          0.0 |                      0.105752 |                           0.134895 |             1.518057 |                  1.748197 |     100.081943 |    1.22434 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.398044 |                     15.502754 |            14.131208 |   0.07485 |        0.006692 |                0.005726 |  0.064189 |  17.36658 |  0.066238 |              0.004352 |                  0.004352 |                  10.739718 |                  1.440702 |                      0.0 |    -0.055678 |              0.0 |          9.299016 |             61.405925 |                 100.0 |               104.595293 |                93.30058 |  0.0 |  0.0 |  0.024299 |       0.0 | 0.014597 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.853319 | 76.744226 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.123439 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 05:00:00 |  0.0 | 14.1 |              0.0 |             93.6 |       2.5 |               101.7 | nan | nan | 18.595833 |             80.920833 |                   4.5 |  0.0 | 14.1 | 18.595833 |    2.003978 |         1.486284 |           2.003978 |          2.5 |                 1.611289 |                      2.145953 |                          0.0 |                      0.104558 |                           0.134343 |             1.508166 |                  1.736523 |     100.191834 |   1.226451 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.131208 |  0.938714 |        0.083928 |                0.071823 | -0.116833 | 17.111064 |  0.065039 |              0.054586 |                  0.054586 |                  10.739718 |                  1.451305 |                      0.0 |    -0.005885 |              0.0 |          9.288414 |             56.493451 |                 100.0 |                104.61564 |              103.248047 |  0.0 |  0.0 | -0.006212 | -0.006212 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.006212 |  0.0 |  0.0 |   0.0 |  nan | 31.777864 | 76.744226 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.120251 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 06:00:00 |  0.0 | 14.3 |              0.1 |             94.6 |       2.4 |               101.7 | nan | nan |    18.525 |             80.791667 |                   4.6 |  0.0 | 14.3 |    18.525 |    1.923819 |         1.536383 |           1.923819 |          2.4 |                 1.632319 |                      2.136456 |                          0.0 |                      0.105752 |                           0.133824 |             1.544174 |                  1.726078 |     100.155826 |   1.225432 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.246552 |  2.455474 |        0.263445 |                0.226749 | -0.117598 | 16.859546 |  0.055286 |              0.172329 |                  0.172329 |                   10.82738 |                  1.478175 |                      0.0 |     0.110738 |              0.0 |          9.349205 |             58.847345 |                 100.0 |                104.61564 |              103.248047 |  0.0 |  0.0 |   0.01635 |  0.006212 | 0.005748 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.712161 | 76.738477 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.117144 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 07:00:00 |  0.0 | 15.4 |              0.9 |             93.3 |       2.5 |               101.7 | nan | nan |    18.475 |             80.620833 |                   5.5 |  0.0 | 15.4 |    18.475 |    2.003978 |         1.589823 |           2.003978 |          2.5 |                 1.752342 |                      2.129774 |                          0.0 |                      0.112529 |                           0.133458 |             1.634935 |                  1.717041 |     100.065065 |   1.220347 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            15.284647 |  4.650125 |        1.164115 |                1.028795 | -0.125428 | 16.640536 |  0.026796 |              0.781884 |                  0.781884 |                  11.616331 |                   1.64464 |                      0.0 |     0.713358 |              0.0 |          9.971692 |             56.493451 |                 100.0 |               104.623652 |              103.253611 |  0.0 |  0.0 |  0.162578 |       0.0 |  0.09208 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.674949 | 76.646397 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.114118 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 08:00:00 |  0.0 | 16.8 |              1.0 |             86.9 |       2.5 |               101.7 | nan | nan |     18.45 |             80.208333 |                   6.5 |  0.0 | 16.8 |     18.45 |    2.003978 |         1.643262 |           2.003978 |          2.5 |                  1.91622 |                       2.12644 |                          0.0 |                      0.121684 |                           0.133275 |             1.665196 |                  1.705582 |     100.034804 |   1.214317 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            16.438085 |  7.209034 |        1.933622 |                1.710811 | -0.134064 | 16.516495 | -0.006124 |              1.300216 |                  1.300216 |                  12.492945 |                   1.83472 |                      0.0 |      1.22377 |              0.0 |         10.658225 |             56.493451 |                 100.0 |               104.752035 |              103.342738 |  0.0 |  0.0 |  0.313651 |       0.0 | 0.180951 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.670656 | 76.465447 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.111171 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 09:00:00 |  0.0 | 18.2 |              1.0 |             78.9 |       2.9 |               101.7 | nan | nan | 18.416667 |             79.908333 |                   7.3 |  0.0 | 18.2 | 18.416667 |    2.324615 |         1.696701 |           2.324615 |          2.9 |                 2.093348 |                      2.122001 |                          0.0 |                      0.131461 |                           0.133032 |             1.651651 |                  1.695656 |     100.048349 |   1.208543 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.360836 |  9.458797 |        2.537058 |                2.210424 | -0.140938 | 16.502185 | -0.036673 |              1.679922 |                  1.679922 |                  13.194235 |                  1.989515 |                      0.0 |     1.597026 |              0.0 |          11.20472 |             48.701251 |                 100.0 |               105.004493 |               103.51781 |  0.0 |  0.0 |  0.461151 |       0.0 | 0.256056 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.696912 | 76.209391 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.108299 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 10:00:00 |  0.0 | 19.5 |              0.9 |             75.8 |       3.1 |               101.7 | nan | nan | 18.383333 |             79.866667 |                   7.7 |  0.0 | 19.5 | 18.383333 |    2.484933 |         1.750141 |           2.484933 |          3.1 |                 2.270468 |                      2.117571 |                          0.0 |                      0.141127 |                            0.13279 |             1.721014 |                  1.691233 |      99.978986 |   1.202862 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.822211 | 10.747438 |         2.69052 |                2.317703 | -0.144383 | 16.589706 | -0.062862 |              1.761455 |                  1.761455 |                   13.54488 |                   2.06621 |                      0.0 |     1.675363 |              0.0 |          11.47867 |             45.559235 |                 100.0 |               105.362113 |              103.765383 |  0.0 |  0.0 |  0.515542 |       0.0 | 0.283847 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.749357 | 75.925543 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.105502 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 11:00:00 |  0.0 | 20.3 |              0.8 |             73.4 |       3.3 |               101.8 | nan | nan |   18.3375 |             79.933333 |                   7.8 |  0.0 | 20.3 |   18.3375 |    2.645251 |          1.82028 |           2.645251 |          3.3 |                  2.38585 |                      2.111492 |                          0.0 |                      0.147368 |                           0.132457 |             1.751214 |                  1.687786 |     100.048786 |   1.200635 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.937555 | 10.819272 |        2.515039 |                2.160947 | -0.145217 | 16.764525 | -0.076366 |               1.64232 |                   1.64232 |                  13.632542 |                  2.087634 |                      0.0 |     1.555335 |              0.0 |         11.544908 |             42.798069 |                 100.0 |               105.759071 |              104.039602 |  0.0 |  0.0 |  0.512611 |       0.0 |  0.27823 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.815307 | 75.647314 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.102776 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 12:00:00 |  0.0 | 21.0 |              0.9 |             68.3 |       3.3 |               101.8 | nan | nan |    18.325 |             79.966667 |                   7.9 |  0.0 | 21.0 |    18.325 |    2.645251 |         1.887079 |           2.645251 |          3.3 |                 2.490975 |                      2.109837 |                          0.0 |                      0.153018 |                           0.132366 |             1.701336 |                  1.687166 |     100.098664 |   1.198001 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.052899 | 10.408076 |        2.605564 |                2.233851 | -0.146088 | 16.984357 | -0.086738 |              1.697727 |                  1.697727 |                  13.720203 |                  2.106001 |                      0.0 |     1.609977 |              0.0 |         11.614202 |             42.798069 |                 100.0 |               106.148714 |              104.308172 |  0.0 |  0.0 |  0.561107 |       0.0 | 0.307496 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.891628 | 75.339818 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.100122 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 13:00:00 |  0.0 | 21.8 |              0.8 |             61.3 |       3.2 |               101.8 | nan | nan |    18.325 |             79.766667 |                   8.2 |  0.0 | 21.8 |    18.325 |    2.565092 |         1.943859 |           2.565092 |          3.2 |                 2.616045 |                      2.109837 |                          0.0 |                      0.159698 |                           0.132366 |             1.603636 |                  1.682947 |     100.196364 |   1.195188 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             18.39893 | 10.917404 |         2.53785 |                2.157989 | -0.148637 | 17.238761 | -0.098523 |              1.640072 |                  1.640072 |                  13.983187 |                  2.166369 |                      0.0 |     1.549806 |              0.0 |         11.816818 |             44.135508 |                 100.0 |               106.579974 |              104.604739 |  0.0 |  0.0 |  0.584215 |       0.0 | 0.327638 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.979734 |  75.01218 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.097535 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 14:00:00 |  0.0 | 21.9 |              0.9 |             60.6 |       2.5 |               101.8 | nan | nan | 18.283333 |                 79.65 |                   8.7 |  0.0 | 21.9 | 18.283333 |    2.003978 |         1.963899 |           2.003978 |          2.5 |                 2.632058 |                      2.104328 |                          0.0 |                       0.16055 |                           0.132064 |             1.595027 |                  1.676097 |     100.204973 |   1.194821 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.975649 | 10.494383 |         2.62717 |                2.208784 |   -0.1529 | 17.532448 | -0.094339 |              1.678676 |                  1.678676 |                  14.421494 |                  2.265756 |                      0.0 |      1.58427 |              0.0 |         12.155738 |             56.493451 |                 100.0 |               107.040221 |              104.920447 |  0.0 |  0.0 |  0.550969 |       0.0 | 0.337877 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.063657 | 74.674303 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.095016 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 15:00:00 |  0.0 | 21.8 |              0.9 |             58.8 |       2.9 |               101.7 | nan | nan | 18.245833 |                  79.5 |                   9.1 |  0.0 | 21.8 | 18.245833 |    2.324615 |         1.997298 |           2.324615 |          2.9 |                 2.616045 |                      2.099381 |                          0.0 |                      0.159698 |                           0.131792 |             1.538235 |                  1.669008 |     100.161765 |   1.194298 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            19.437025 |  8.874215 |        2.221576 |                1.861585 |  -0.15627 | 17.812189 | -0.086137 |              1.414805 |                  1.414805 |                  14.772139 |                  2.348437 |                      0.0 |     1.316953 |              0.0 |         12.423701 |             48.701251 |                 100.0 |               107.515665 |              105.245721 |  0.0 |  0.0 |  0.511992 |       0.0 | 0.297253 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.139377 |  74.37705 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.092562 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 16:00:00 |  0.0 | 22.2 |              0.9 |             57.2 |       3.6 |               101.7 | nan | nan | 18.233333 |               79.3375 |                   9.5 |  0.0 | 22.2 | 18.233333 |    2.885728 |         2.040718 |           2.885728 |          3.6 |                 2.680609 |                      2.097734 |                          0.0 |                      0.163129 |                           0.131702 |             1.533308 |                   1.66429 |     100.166692 |   1.192703 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |              19.8984 |   6.45697 |        1.616442 |                1.360212 | -0.159672 |  18.06459 | -0.089325 |              1.033761 |                  1.033761 |                  15.122784 |                  2.428674 |                      0.0 |     0.932566 |              0.0 |          12.69411 |             39.231563 |                 100.0 |               107.934634 |               105.53164 |  0.0 |  0.0 |  0.464653 |       0.0 | 0.249062 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.218285 | 74.127988 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.090171 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 17:00:00 |  0.0 | 21.8 |              1.0 |             57.2 |       4.2 |               101.7 | nan | nan | 18.229167 |             79.133333 |                  10.2 |  0.0 | 21.8 | 18.229167 |    3.366683 |         2.100837 |           3.366683 |          4.2 |                 2.616045 |                      2.097185 |                          0.0 |                      0.159698 |                           0.131672 |             1.496378 |                  1.659573 |     100.203622 |   1.194485 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            20.705807 |  3.945159 |        1.058179 |                0.907937 | -0.165683 | 18.327617 | -0.075003 |              0.690032 |                  0.690032 |                  15.736413 |                  2.564427 |                      0.0 |     0.583181 |              0.0 |         13.171986 |             33.627054 |                 100.0 |               108.286185 |              105.771032 |  0.0 |  0.0 |  0.401816 |       0.0 | 0.199809 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.282872 | 73.928179 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.087841 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 18:00:00 |  0.0 | 20.9 |              1.0 |             59.9 |       4.2 |               101.7 | nan | nan | 18.208333 |               78.8375 |                  11.1 |  0.0 | 20.9 | 18.208333 |    3.366683 |         2.157616 |           3.366683 |          4.2 |                 2.475714 |                      2.094444 |                          0.0 |                        0.1522 |                           0.131522 |             1.482953 |                  1.651207 |     100.217047 |   1.198201 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            21.743901 |  1.939642 |        0.520255 |                0.462876 | -0.173346 | 18.542906 | -0.050913 |              0.351786 |                  0.351786 |                  16.525365 |                  2.744116 |                      0.0 |     0.237448 |              0.0 |         13.781249 |             33.627054 |                 100.0 |               108.568552 |               105.96297 |  0.0 |  0.0 |  0.280556 |       0.0 |  0.13719 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.323368 |  73.79099 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.085572 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 19:00:00 |  0.0 | 19.7 |              1.0 |             66.1 |       3.6 |               101.7 | nan | nan | 18.166667 |               78.4875 |                  12.1 |  0.0 | 19.7 | 18.166667 |    2.885728 |         2.194356 |           2.885728 |          3.6 |                 2.298847 |                       2.08897 |                          0.0 |                      0.142666 |                           0.131221 |             1.519538 |                   1.63958 |     100.180462 |   1.202947 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             22.89734 |   0.60531 |        0.162358 |                0.151461 | -0.181782 | 18.677895 | -0.022077 |              0.115111 |                  0.115111 |                  17.401978 |                  2.950104 |                      0.0 |     -0.00781 |              0.0 |         14.451875 |             39.231563 |                 100.0 |               108.762602 |                106.0947 |  0.0 |  0.0 |  0.137894 |       0.0 | 0.070789 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.335029 | 73.720201 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.083362 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 20:00:00 |  0.0 | 17.9 |              0.3 |             71.7 |       3.2 |               101.7 | nan | nan | 18.104167 |             78.091667 |                  12.4 |  0.0 | 17.9 | 18.104167 |    2.565092 |         2.227756 |           2.565092 |          3.2 |                 2.054233 |                      2.080783 |                          0.0 |                      0.129312 |                           0.130772 |             1.472885 |                  1.624918 |     100.227115 |   1.210597 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.104711 |                     15.502754 |            23.243371 |  0.005136 |        0.003091 |                0.002926 |  0.254985 | 18.716764 |  0.017642 |              0.002224 |                  0.002224 |                  17.664962 |                  3.034376 |                      0.0 |    -0.124208 |              0.0 |         14.630586 |             44.135508 |                 100.0 |               108.862786 |               89.344572 |  0.0 |  0.0 |  0.167821 |       0.0 | 0.095224 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.30697 | 73.624977 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.081209 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 21:00:00 |  0.0 | 15.8 |              0.0 |             80.6 |       2.2 |               101.8 | nan | nan | 18.020833 |             77.829167 |                  12.4 |  0.0 | 15.8 | 18.020833 |    1.763501 |         2.244455 |           1.763501 |          2.2 |                 1.797863 |                       2.06991 |                          0.0 |                      0.115083 |                           0.130174 |             1.449077 |                  1.610994 |     100.350923 |    1.22071 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 |  0.305727 | 18.623235 |  0.060982 |                   0.0 |                       0.0 |                  17.664962 |                  3.060678 |                      0.0 |    -0.127528 |              0.0 |         14.604285 |             64.197103 |                 100.0 |               108.997612 |               87.719298 |  0.0 |  0.0 |  0.104453 |       0.0 |  0.06637 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.235572 | 73.558607 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.079111 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 22:00:00 |  0.0 | 14.8 |              0.0 |             85.9 |       2.2 |               101.8 | nan | nan | 17.929167 |             77.654167 |                  12.4 |  0.0 | 14.8 | 17.929167 |    1.763501 |         2.261155 |           1.763501 |          2.2 |                 1.685952 |                      2.058007 |                          0.0 |                      0.108789 |                            0.12952 |             1.448232 |                  1.598128 |     100.351768 |   1.224953 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 |  0.305184 |  18.38524 |  0.077441 |                   0.0 |                       0.0 |                  17.664962 |                  3.084324 |                      0.0 |    -0.128514 |              0.0 |         14.580638 |             64.197103 |                 100.0 |               109.091626 |               87.719298 |  0.0 |  0.0 |  0.084738 |       0.0 |  0.05324 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.147714 | 73.505367 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.077068 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 23:00:00 |  0.0 | 14.1 |              0.0 |             89.3 |       2.1 |               101.8 | nan | nan | 17.833333 |               77.5875 |                  12.4 |  0.0 | 14.1 | 17.833333 |    1.683342 |         2.254475 |           1.683342 |          2.1 |                 1.611289 |                      2.045628 |                          0.0 |                      0.104558 |                           0.128839 |             1.438881 |                  1.587151 |     100.361119 |   1.227981 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 |  0.304741 |  18.09238 |  0.086235 |                   0.0 |                       0.0 |                  17.664962 |                  3.103639 |                      0.0 |    -0.129318 |              0.0 |         14.561323 |             67.254108 |                 100.0 |               109.167066 |               87.719298 |  0.0 |  0.0 |  0.070962 |       0.0 | 0.044913 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.051062 | 73.460455 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.075077 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 00:00:00 |  0.0 | 13.4 |              0.0 |             91.4 |       1.8 |               101.8 | nan | nan | 17.729167 |             77.729167 |                  12.4 |  0.0 | 13.4 | 17.729167 |    1.442864 |         2.237776 |           1.442864 |          1.8 |                 1.539538 |                      2.032245 |                          0.0 |                      0.100468 |                           0.128102 |             1.407137 |                  1.579647 |     100.392863 |   1.231127 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 |  0.300273 | 17.770207 |  0.094396 |                   0.0 |                       0.0 |                   17.58988 |                  3.124611 |                      0.0 |    -0.130192 |              0.0 |         14.465269 |             78.463126 |                 100.0 |               109.230723 |               87.719298 |  0.0 |  0.0 |  0.058672 |       0.0 | 0.038666 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.946249 | 73.421789 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.073138 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 01:00:00 |  0.0 | 12.8 |              0.0 |             93.1 |       1.8 |               101.8 | nan | nan | 17.620833 |             77.891667 |                  12.4 |  0.0 | 12.8 | 17.620833 |    1.442864 |         2.224416 |           1.442864 |          1.8 |                 1.480281 |                      2.018409 |                          0.0 |                      0.097071 |                           0.127339 |             1.378142 |                  1.572172 |     100.421858 |   1.233843 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 |  0.300025 |  17.42083 |   0.09981 |                   0.0 |                       0.0 |                   17.58988 |                  3.135507 |                      0.0 |    -0.130646 |              0.0 |         14.454373 |             78.463126 |                 100.0 |               109.285539 |               87.719298 |  0.0 |  0.0 |  0.053362 |       0.0 | 0.034959 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.836022 | 73.386829 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.071248 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 02:00:00 |  0.0 | 12.3 |              0.0 |             94.6 |       1.6 |               101.8 | nan | nan |      17.5 |             78.133333 |                  12.4 |  0.0 | 12.3 |      17.5 |    1.282546 |         2.201036 |           1.282546 |          1.6 |                 1.432438 |                      2.003073 |                          0.0 |                      0.094315 |                           0.126493 |             1.355086 |                  1.565068 |     100.444914 |   1.236111 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 |  0.299811 | 17.053408 |  0.102674 |                   0.0 |                       0.0 |                   17.58988 |                  3.144898 |                      0.0 |    -0.131037 |              0.0 |         14.444982 |             88.271017 |                 100.0 |                109.33511 |               87.719298 |  0.0 |  0.0 |  0.047083 |       0.0 |  0.03177 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.722932 | 73.355059 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.069408 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 03:00:00 |  0.0 | 12.0 |              0.0 |             95.6 |       1.2 |               101.8 | nan | nan | 17.379167 |             78.379167 |                  12.4 |  0.0 | 12.0 | 17.379167 |    0.961909 |         2.174316 |           0.961909 |          1.2 |                 1.404387 |                       1.98784 |                          0.0 |                      0.092693 |                           0.125652 |             1.342594 |                  1.558052 |     100.457406 |   1.237469 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 |  0.299602 |  16.67644 |  0.101011 |                   0.0 |                       0.0 |                   17.58988 |                   3.15411 |                      0.0 |    -0.131421 |              0.0 |          14.43577 |            117.694689 |                 100.0 |               109.380167 |               87.719298 |  0.0 |  0.0 |  0.040901 |       0.0 | 0.029637 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.611504 | 73.325423 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.067615 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 04:00:00 |  0.0 | 11.7 |              0.0 |             96.2 |       1.9 |               101.8 | nan | nan | 17.270833 |               78.5125 |                  12.4 |  0.0 | 11.7 | 17.270833 |    1.523023 |         2.160956 |           1.523023 |          1.9 |                 1.376819 |                      1.974268 |                          0.0 |                      0.091096 |                           0.124901 |               1.3245 |                  1.550047 |       100.4755 |   1.238856 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.366022 |                     15.438712 |            23.144579 |  0.063733 |         0.00566 |                0.005336 |  0.123083 | 16.305014 |  0.099468 |              0.004056 |                  0.004056 |                   17.58988 |                  3.166293 |                      0.0 |    -0.127873 |              0.0 |         14.423587 |             74.333488 |                 100.0 |               109.422206 |               93.783653 |  0.0 |  0.0 |  0.008433 |       0.0 | 0.005242 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.501619 | 73.320181 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.065868 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 05:00:00 |  0.0 | 11.6 |              0.0 |             95.7 |       1.3 |               101.8 | nan | nan | 17.166667 |                  78.6 |                  12.4 |  0.0 | 11.6 | 17.166667 |    1.042069 |         2.120877 |           1.042069 |          1.3 |                 1.367736 |                      1.961295 |                          0.0 |                      0.090568 |                           0.124183 |             1.308923 |                  1.541578 |     100.491077 |   1.239363 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.144579 |  0.908598 |        0.080684 |                0.076088 | -0.182008 | 15.938731 |  0.093717 |              0.057827 |                  0.057827 |                   17.58988 |                  3.179808 |                      0.0 |    -0.074665 |              0.0 |         14.410072 |            108.641252 |                 100.0 |               109.429642 |              106.546425 |  0.0 |  0.0 |  -0.04351 |  -0.04351 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |  0.04351 |  0.0 |  0.0 |   0.0 |  nan | 31.397486 | 73.320181 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.064167 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 06:00:00 |  0.0 | 12.6 |              0.6 |             95.1 |       1.6 |               101.8 | nan | nan | 17.095833 |             78.620833 |                  12.9 |  0.0 | 12.6 | 17.095833 |    1.282546 |         2.094157 |           1.282546 |          1.6 |                 1.460978 |                      1.952516 |                          0.0 |                       0.09596 |                           0.123697 |              1.38939 |                  1.535084 |      100.41061 |   1.234655 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  2.420659 |        0.472904 |                 0.45318 | -0.186152 |  15.59162 |  0.064619 |              0.344417 |                  0.344417 |                  18.027018 |                  3.288825 |                      0.0 |     0.207383 |              0.0 |         14.738193 |             88.271017 |                 100.0 |               109.429642 |              106.546425 |  0.0 |  0.0 |  0.015278 |  0.015278 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.028231 |  0.0 |  0.0 |   0.0 |  nan |  31.32245 | 73.320181 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.06251 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 07:00:00 |  0.0 | 14.8 |              0.9 |             96.3 |       1.9 |               101.9 | nan | nan | 17.070833 |             78.745833 |                  12.9 |  0.0 | 14.8 | 17.070833 |    1.523023 |         2.074117 |           1.523023 |          1.9 |                 1.685952 |                      1.949426 |                          0.0 |                      0.108789 |                           0.123526 |             1.623571 |                  1.535091 |     100.276429 |   1.225361 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  4.619722 |        1.148656 |                1.101438 | -0.186171 | 15.341501 |  0.011696 |              0.837093 |                  0.837093 |                  18.027018 |                  3.287363 |                      0.0 |     0.700119 |              0.0 |         14.739656 |             74.333488 |                 100.0 |               109.429642 |              106.546425 |  0.0 |  0.0 |  0.126839 |  0.028231 | 0.060415 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.300337 | 73.259766 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.060895 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 08:00:00 |  0.0 | 17.1 |              1.0 |             89.9 |       3.2 |               101.9 | nan | nan | 17.083333 |             78.870833 |                  12.9 |  0.0 | 17.1 | 17.083333 |    2.565092 |         2.097497 |           2.565092 |          3.2 |                 1.953031 |                       1.95097 |                          0.0 |                      0.123726 |                           0.123612 |             1.755775 |                  1.538746 |     100.144225 |   1.215051 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  7.198198 |        1.917614 |                1.839976 |  -0.18627 | 15.267791 | -0.039576 |              1.398382 |                  1.398382 |                  18.027018 |                  3.279458 |                      0.0 |     1.261738 |              0.0 |         14.747561 |             44.135508 |                 100.0 |               109.515362 |              106.604354 |  0.0 |  0.0 |  0.315007 |       0.0 |   0.1629 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.329496 | 73.096866 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.059322 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 09:00:00 |  0.0 | 19.4 |              0.9 |             79.2 |       4.0 |               101.9 | nan | nan | 17.133333 |             78.883333 |                  12.8 |  0.0 | 19.4 | 17.133333 |    3.206365 |         2.134237 |           3.206365 |          4.0 |                 2.256393 |                      1.957159 |                          0.0 |                      0.140363 |                           0.123954 |             1.787063 |                  1.543872 |     100.112937 |   1.205358 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.604725 |  9.475573 |        2.356024 |                2.266269 | -0.185529 | 15.364988 | -0.087156 |              1.722365 |                  1.722365 |                  17.939591 |                  3.250753 |                      0.0 |     1.586917 |              0.0 |         14.688838 |             35.308407 |                 100.0 |               109.746636 |              106.760508 |  0.0 |  0.0 |  0.495578 |       0.0 |  0.24054 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.406236 | 72.856326 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.05779 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 10:00:00 |  0.0 | 20.6 |              1.0 |             73.4 |       4.6 |               101.9 | nan | nan | 17.179167 |             78.783333 |                  12.9 |  0.0 | 20.6 | 17.179167 |     3.68732 |         2.184336 |            3.68732 |          4.6 |                  2.43042 |                      1.962848 |                          0.0 |                      0.149767 |                           0.124269 |             1.783928 |                  1.546397 |     100.116072 |   1.200448 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 | 10.786309 |        2.873495 |                 2.75694 | -0.186428 | 15.620786 | -0.107551 |              2.095274 |                  2.095274 |                  18.027018 |                  3.266966 |                      0.0 |     1.959151 |              0.0 |         14.760052 |             30.702962 |                 100.0 |               110.088522 |              106.990975 |  0.0 |  0.0 |  0.669077 |       0.0 | 0.309949 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  31.50337 | 72.546377 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.056297 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 11:00:00 |  0.0 | 21.4 |              1.0 |             67.3 |       4.1 |               101.9 | nan | nan |    17.225 |             78.529167 |                  13.1 |  0.0 | 21.4 |    17.225 |    3.286524 |         2.211056 |           3.286524 |          4.1 |                 2.552842 |                      1.968551 |                          0.0 |                      0.156328 |                           0.124585 |             1.718063 |                  1.545886 |     100.181937 |   1.197482 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.949834 | 10.865527 |        2.894599 |                2.761725 | -0.188095 | 15.944568 | -0.117837 |              2.098911 |                  2.098911 |                  18.201874 |                  3.309837 |                      0.0 |     1.961001 |              0.0 |         14.892037 |             34.447226 |                 100.0 |               110.529741 |              107.287756 |  0.0 |  0.0 |  0.697445 |       0.0 | 0.345741 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.610791 | 72.200636 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.054843 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 12:00:00 |  0.0 | 22.5 |              1.0 |             62.2 |       4.1 |               101.9 | nan | nan |   17.2875 |                78.275 |                  13.2 |  0.0 | 22.5 |   17.2875 |    3.286524 |         2.237776 |           3.286524 |          4.1 |                 2.729939 |                      1.976351 |                          0.0 |                      0.165743 |                           0.125016 |             1.698022 |                  1.546988 |     100.201978 |   1.193116 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |             24.06487 |  10.45226 |        2.784503 |                 2.65055 |  -0.18894 | 16.302636 | -0.133863 |              2.014418 |                  2.014418 |                  18.289301 |                  3.330376 |                      0.0 |     1.875652 |              0.0 |         14.958925 |             34.447226 |                 100.0 |                111.02283 |              107.618564 |  0.0 |  0.0 |  0.721689 |       0.0 | 0.364125 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.734237 | 71.836511 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.053426 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 13:00:00 |  0.0 | 23.1 |              1.0 |             59.6 |       4.4 |               101.9 | nan | nan | 17.341667 |             78.204167 |                  13.4 |  0.0 | 23.1 | 17.341667 |    3.527001 |         2.277855 |           3.527001 |          4.4 |                 2.830978 |                      1.983133 |                          0.0 |                      0.171077 |                           0.125391 |             1.687263 |                  1.550892 |     100.212737 |   1.190747 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.294943 |  10.96314 |        2.920603 |                2.764822 | -0.190734 | 16.714124 | -0.137935 |              2.101265 |                  2.101265 |                  18.464157 |                  3.363175 |                      0.0 |     1.961133 |              0.0 |         15.100981 |             32.098552 |                 100.0 |               111.543203 |              107.966694 |  0.0 |  0.0 |  0.790142 |       0.0 | 0.389537 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.861756 | 71.446974 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.052046 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 14:00:00 |  0.0 | 23.4 |              1.0 |             55.4 |       4.6 |               101.9 | nan | nan | 17.404167 |               77.9875 |                  13.5 |  0.0 | 23.4 | 17.404167 |     3.68732 |         2.347994 |            3.68732 |          4.6 |                 2.882708 |                      1.990983 |                          0.0 |                      0.173797 |                           0.125825 |              1.59702 |                  1.552718 |      100.30298 |   1.189943 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.409979 | 10.528069 |        2.804699 |                2.649346 | -0.191602 | 17.139185 | -0.135234 |              2.013503 |                  2.013503 |                  18.551584 |                   3.38194 |                      0.0 |     1.872589 |              0.0 |         15.169644 |             30.702962 |                 100.0 |               112.101122 |              108.338827 |  0.0 |  0.0 |  0.813761 |       0.0 | 0.394418 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.986572 | 71.052556 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.050702 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 15:00:00 |  0.0 | 23.8 |              1.0 |             52.1 |       4.5 |               101.9 | nan | nan |   17.4875 |             77.708333 |                  13.6 |  0.0 | 23.8 |   17.4875 |    3.607161 |         2.401434 |           3.607161 |          4.5 |                 2.952961 |                      2.001493 |                          0.0 |                      0.177481 |                           0.126406 |             1.538493 |                  1.555327 |     100.361507 |     1.1886 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.525016 |  8.882849 |        2.366409 |                 2.23335 | -0.192477 | 17.555242 | -0.134887 |              1.697346 |                  1.697346 |                  18.639012 |                  3.400029 |                      0.0 |     1.555678 |              0.0 |         15.238983 |              31.38525 |                 100.0 |               112.667352 |              108.715327 |  0.0 |  0.0 |  0.746481 |       0.0 | 0.367549 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.111043 | 70.685007 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.049392 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 16:00:00 |  0.0 | 24.0 |              1.0 |             49.5 |       4.7 |               101.8 | nan | nan |   17.5625 |               77.3875 |                  13.7 |  0.0 | 24.0 |   17.5625 |    3.767479 |         2.438173 |           3.767479 |          4.7 |                 2.988644 |                      2.010993 |                          0.0 |                      0.179348 |                            0.12693 |             1.479379 |                  1.556257 |     100.320621 |   1.186889 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  6.439028 |         1.71537 |                1.620343 | -0.193307 | 17.970142 | -0.130245 |              1.231461 |                  1.231461 |                   18.72644 |                   3.42173 |                      0.0 |     1.088889 |              0.0 |         15.304709 |             30.049708 |                 100.0 |               113.196228 |              109.065924 |  0.0 |  0.0 |  0.650303 |       0.0 | 0.314382 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.230871 | 70.370625 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.048116 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 17:00:00 |  0.0 | 24.0 |              1.0 |             48.1 |       4.5 |               101.8 | nan | nan | 17.654167 |             77.008333 |                  13.7 |  0.0 | 24.0 | 17.654167 |    3.607161 |         2.448193 |           3.607161 |          4.5 |                 2.988644 |                      2.022657 |                          0.0 |                      0.179348 |                           0.127573 |             1.437538 |                  1.557615 |     100.362462 |   1.187075 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  3.911896 |        1.042137 |                0.985011 |  -0.19328 |  18.36957 | -0.121617 |              0.748608 |                  0.748608 |                   18.72644 |                  3.423912 |                      0.0 |     0.605945 |              0.0 |         15.302528 |              31.38525 |                 100.0 |               113.649549 |               109.36562 |  0.0 |  0.0 |  0.506995 |       0.0 |  0.24977 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.342072 | 70.120855 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.046873 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 18:00:00 |  0.0 | 23.7 |              1.0 |             49.4 |       4.1 |               101.8 | nan | nan | 17.770833 |             76.570833 |                  13.7 |  0.0 | 23.7 | 17.770833 |    3.286524 |         2.444853 |           3.286524 |          4.1 |                  2.93526 |                      2.037589 |                          0.0 |                      0.176554 |                           0.128396 |             1.450018 |                  1.560199 |     100.349982 |   1.188219 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  1.905864 |        0.507726 |                0.480125 | -0.193271 | 18.740238 | -0.107131 |              0.364895 |                  0.364895 |                   18.72644 |                   3.42458 |                      0.0 |     0.222204 |              0.0 |          15.30186 |             34.447226 |                 100.0 |               114.010338 |              109.603606 |  0.0 |  0.0 |  0.354517 |       0.0 | 0.180987 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.438786 | 69.939868 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.045662 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 19:00:00 |  0.0 | 22.8 |              1.0 |             57.0 |       3.8 |               101.8 | nan | nan |      17.9 |             76.191667 |                  13.7 |  0.0 | 22.8 |      17.9 |    3.046047 |         2.451533 |           3.046047 |          3.8 |                 2.780058 |                      2.054233 |                          0.0 |                      0.168392 |                           0.129312 |             1.584633 |                  1.565154 |     100.215367 |   1.191234 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  0.576079 |        0.153469 |                0.145175 | -0.193327 | 19.062619 | -0.080727 |              0.110333 |                  0.110333 |                   18.72644 |                  3.420162 |                      0.0 |    -0.032173 |              0.0 |         15.306277 |             37.166744 |                 100.0 |               114.272126 |              109.775993 |  0.0 |  0.0 |  0.206466 |       0.0 | 0.107221 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.509096 | 69.832647 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.044483 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 20:00:00 |  0.0 | 21.3 |              0.5 |             65.9 |       2.5 |               101.8 | nan | nan | 18.041667 |                 75.95 |                  13.9 |  0.0 | 21.3 | 18.041667 |    2.003978 |         2.428153 |           2.003978 |          2.5 |                 2.537251 |                      2.072623 |                          0.0 |                      0.155495 |                           0.130323 |             1.672048 |                  1.574157 |     100.127952 |   1.196911 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.072689 |                     15.438712 |            24.870125 |  0.002495 |        0.003269 |                0.003122 |  0.285181 | 19.296988 | -0.043265 |              0.002372 |                  0.002372 |                  18.901295 |                  3.445356 |                      0.0 |    -0.141184 |              0.0 |         15.455939 |             56.493451 |                 100.0 |               114.427357 |               89.024306 |  0.0 |  0.0 |   0.18067 |       0.0 | 0.116869 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.541945 | 69.715777 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.043334 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 21:00:00 |  0.0 | 19.3 |              0.0 |             73.0 |       1.9 |               101.8 | nan | nan |   18.1875 |             75.633333 |                  13.9 |  0.0 | 19.3 |   18.1875 |    1.523023 |         2.418134 |           1.523023 |          1.9 |                 2.242395 |                      2.091705 |                          0.0 |                      0.139602 |                           0.131371 |             1.636948 |                  1.582026 |     100.163052 |   1.205255 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 |  0.323081 | 19.406483 |    0.0023 |                   0.0 |                       0.0 |                  18.901295 |                  3.434698 |                      0.0 |    -0.143112 |              0.0 |         15.466597 |             74.333488 |                 100.0 |               114.596678 |               87.719298 |  0.0 |  0.0 |  0.129637 |       0.0 | 0.089184 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.529228 | 69.626593 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.042215 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 22:00:00 |  0.0 | 18.3 |              0.0 |             76.7 |       2.2 |               101.8 | nan | nan | 18.333333 |                 75.25 |                  13.9 |  0.0 | 18.3 | 18.333333 |    1.763501 |         2.418134 |           1.763501 |          2.2 |                  2.10653 |                       2.11094 |                          0.0 |                      0.132185 |                           0.132426 |             1.615709 |                  1.588483 |     100.184291 |   1.209486 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 |  0.323243 | 19.364094 |  0.022984 |                   0.0 |                       0.0 |                  18.901295 |                  3.427556 |                      0.0 |    -0.142815 |              0.0 |         15.473739 |             64.197103 |                 100.0 |               114.725974 |               87.719298 |  0.0 |  0.0 |  0.125384 |       0.0 | 0.081911 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.495827 | 69.544682 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.041124 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 23:00:00 |  0.0 | 17.4 |              0.0 |             81.8 |       2.7 |               101.8 | nan | nan | 18.470833 |               74.9375 |                  13.9 |  0.0 | 17.4 | 18.470833 |    2.164296 |         2.438173 |           2.164296 |          2.7 |                 1.990459 |                      2.129218 |                          0.0 |                      0.125796 |                           0.133428 |             1.628195 |                  1.595583 |     100.171805 |   1.213176 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 |  0.323453 | 19.252757 |   0.04002 |                   0.0 |                       0.0 |                  18.901295 |                  3.418333 |                      0.0 |    -0.142431 |              0.0 |         15.482962 |             52.308751 |                 100.0 |                114.84479 |               87.719298 |  0.0 |  0.0 |  0.118321 |       0.0 | 0.071552 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.445391 |  69.47313 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.040062 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |

.. _lland_v3_acker_heavy_rain_hourly:

acre (heavy rain)
-----------------

When comparing the soil moisture content (|BoWa|) at August 6th, 00:00, we
notice only small differences between the daily
(:ref:`lland_v3_acker_summer_daily`, 70.2 mm) and the hourly
(:ref:`lland_v3_acker_summer_daily`, 69.5) simulation results for a dry
situation.  However, when we compare the differences between our daily
(:ref:`lland_v3_acker_heavy_rain_daily`, 152.8 mm) and our hourly (next
table) test for a wet situation, we see much more pronounced differences,
which are mainly due to the prioritisation of throughfall (|NBes|) over
interception evaporation (|EvI|):

.. integration-test::

    >>> inputs.nied.series = 20.0/24.0
    >>> test('lland_v3_acker_heavy_rain_hourly',
    ...      axis1=(inputs.nied, fluxes.qah),
    ...      axis2=states.bowa,
    ...      use_conditions=conditions_acker_heavy_rain)
    |                date |     nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration |     nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |         g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes | sbes |      evpo |       evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg |     wada |      qdb |     qib1 | qib2 |      qbb | qkap |     qdgz | qah |  qa |     inzp | wats | waes | esnow | taus |      ebdn |       bowa |    qdgz1 | qdgz2 |    qigz1 | qigz2 |     qbgz |    qdga1 | qdga2 |    qiga1 | qiga2 |     qbga | inlet | outlet |
    -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-08-03 00:00:00 | 0.833333 | 16.9 |              0.0 |             95.1 |       0.8 |               101.5 | nan | nan | 17.279167 |             85.883333 |                   1.3 | 0.783333 | 16.9 | 17.279167 |    0.641273 |         1.605886 |           0.641273 |          0.8 |                 1.928423 |                      1.975309 |                          0.0 |                      0.122362 |                           0.124959 |              1.83393 |                  1.696461 |       99.66607 |    1.21073 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 |  0.134093 | 15.208683 | -0.036532 |                   0.0 |                       0.0 |                   7.982856 |                  0.901196 |                      0.0 |     -0.03755 |              0.0 |           7.08166 |            176.542034 |                 100.0 |                70.327061 |               87.719298 | 0.083333 |  0.0 |   0.02723 |   0.02723 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.007403 |      0.0 |  0.0 |      0.0 |  0.0 | 0.007403 | nan | nan |  0.67277 |  0.0 |  0.0 |   0.0 |  nan | 31.308721 |   102.5625 | 0.007403 |   0.0 |      0.0 |   0.0 |      0.0 |  0.80139 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 01:00:00 | 0.833333 | 16.6 |              0.0 |             94.9 |       0.8 |               101.5 | nan | nan | 17.241667 |             85.991667 |                   1.3 | 0.783333 | 16.6 | 17.241667 |    0.641273 |         1.603439 |           0.641273 |          0.8 |                 1.892018 |                      1.970628 |                          0.0 |                      0.120339 |                             0.1247 |             1.795525 |                  1.694576 |      99.704475 |   1.212158 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 |  0.134078 | 15.295735 | -0.028172 |                   0.0 |                       0.0 |                   7.982856 |                  0.901844 |                      0.0 |    -0.037577 |              0.0 |          7.081012 |            176.542034 |                 100.0 |                70.229321 |               87.719298 | 0.756104 |  0.0 |  0.027289 |  0.027289 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.756104 | 0.067467 |      0.0 |  0.0 |      0.0 |  0.0 | 0.067467 | nan | nan | 0.672711 |  0.0 |  0.0 |   0.0 |  nan | 31.326476 | 103.251137 | 0.067467 |   0.0 |      0.0 |   0.0 |      0.0 |  0.78166 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 02:00:00 | 0.833333 | 16.4 |              0.0 |             95.9 |       0.8 |               101.5 | nan | nan |    17.175 |             86.233333 |                   1.3 | 0.783333 | 16.4 |    17.175 |    0.641273 |         1.575992 |           0.641273 |          0.8 |                 1.868084 |                       1.96233 |                          0.0 |                      0.119006 |                           0.124241 |             1.791493 |                  1.692183 |      99.708507 |   1.213014 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 |  0.134066 |  15.35492 | -0.022574 |                   0.0 |                       0.0 |                   7.982856 |                  0.902355 |                      0.0 |    -0.037598 |              0.0 |            7.0805 |            176.542034 |                 100.0 |                69.343535 |               87.719298 | 0.756045 |  0.0 |  0.025718 |  0.025718 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.756045 | 0.067993 |      0.0 |  0.0 |      0.0 |  0.0 | 0.067993 | nan | nan | 0.674282 |  0.0 |  0.0 |   0.0 |  nan | 31.338633 | 103.939189 | 0.067993 |   0.0 |      0.0 |   0.0 |      0.0 | 0.763218 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 03:00:00 | 0.833333 | 16.3 |              0.0 |             96.7 |       0.8 |               101.5 | nan | nan |   17.0625 |             86.708333 |                   1.3 | 0.783333 | 16.3 |   17.0625 |    0.641273 |         1.548545 |           0.641273 |          0.8 |                 1.856216 |                      1.948396 |                          0.0 |                      0.118345 |                           0.123469 |             1.794961 |                  1.689422 |      99.705039 |   1.213417 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 |  0.134066 | 15.395444 | -0.019538 |                   0.0 |                       0.0 |                   7.982856 |                  0.902364 |                      0.0 |    -0.037598 |              0.0 |          7.080492 |            176.542034 |                 100.0 |                68.459657 |               87.719298 | 0.757615 |  0.0 |  0.024521 |  0.024521 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.757615 | 0.068668 |      0.0 |  0.0 |      0.0 |  0.0 | 0.068668 | nan | nan | 0.675479 |  0.0 |  0.0 |   0.0 |  nan | 31.347755 | 104.628135 | 0.068668 |   0.0 |      0.0 |   0.0 |      0.0 | 0.745269 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 04:00:00 | 0.833333 | 16.0 |              0.0 |             97.2 |       0.8 |               101.5 | nan | nan | 16.908333 |             87.366667 |                   1.3 | 0.783333 | 16.0 | 16.908333 |    0.641273 |         1.504432 |           0.641273 |          0.8 |                 1.821009 |                      1.929443 |                          0.0 |                      0.116379 |                           0.122418 |              1.77002 |                   1.68569 |       99.72998 |    1.21479 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                 0.429734 |                     15.566134 |            10.503757 |  0.086656 |        0.007799 |                0.006997 |  0.038337 | 15.425849 | -0.012402 |              0.005318 |                  0.005318 |                   7.982856 |                  0.902341 |                      0.0 |     -0.03228 |              0.0 |          7.080514 |            176.542034 |                 100.0 |                67.575761 |               81.508918 | 0.758813 |  0.0 |  0.005147 |  0.005147 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.758813 | 0.069314 |      0.0 |  0.0 |      0.0 |  0.0 | 0.069314 | nan | nan | 0.694853 |  0.0 |  0.0 |   0.0 |  nan |  31.34974 | 105.317634 | 0.069314 |   0.0 |      0.0 |   0.0 |      0.0 |   0.7278 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 05:00:00 | 0.833333 | 15.9 |              0.0 |             97.5 |       0.6 |               101.5 | nan | nan | 16.729167 |             88.204167 |                   1.3 | 0.783333 | 15.9 | 16.729167 |    0.480955 |         1.453638 |           0.480955 |          0.6 |                 1.809403 |                      1.907618 |                          0.0 |                       0.11573 |                           0.121206 |             1.764168 |                  1.682599 |      99.735832 |   1.215237 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  0.968392 |        0.087159 |                0.078957 |  -0.08871 | 15.432466 | -0.010099 |              0.060007 |                  0.060007 |                   7.982856 |                  0.901482 |                      0.0 |     0.022445 |              0.0 |          7.081373 |            235.389378 |                 100.0 |                66.692266 |               73.740751 | 0.778186 |  0.0 |  -0.00983 |  -0.00983 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.778186 | 0.071642 |      0.0 |  0.0 |      0.0 |  0.0 | 0.071642 | nan | nan |  0.70983 |  0.0 |  0.0 |   0.0 |  nan | 31.349422 | 106.024178 | 0.071642 |   0.0 |      0.0 |   0.0 |      0.0 | 0.710821 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 06:00:00 | 0.833333 | 16.0 |              0.0 |             97.7 |       0.9 |               101.5 | nan | nan | 16.533333 |             89.191667 |                   1.3 | 0.783333 | 16.0 | 16.533333 |    0.721432 |         1.392031 |           0.721432 |          0.9 |                 1.821009 |                      1.884011 |                          0.0 |                      0.116379 |                           0.119893 |             1.779125 |                   1.68038 |      99.720875 |   1.214748 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |    2.4897 |        0.224084 |                0.206122 | -0.088731 | 15.431406 | -0.012282 |              0.156653 |                  0.156653 |                   7.982856 |                   0.89977 |                      0.0 |     0.119163 |              0.0 |          7.083086 |            156.926252 |                 100.0 |                65.788057 |                72.95028 | 0.793163 |  0.0 |  0.009895 |  0.009895 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.793163 | 0.073605 |      0.0 |  0.0 |      0.0 |  0.0 | 0.073605 | nan | nan | 0.690105 |  0.0 |  0.0 |   0.0 |  nan | 31.351287 | 106.743736 | 0.073605 |   0.0 |      0.0 |   0.0 |      0.0 | 0.694336 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 07:00:00 | 0.833333 | 16.6 |              0.0 |             97.4 |       0.9 |               101.5 | nan | nan |    16.375 |                90.125 |                   1.3 | 0.783333 | 16.6 |    16.375 |    0.721432 |         1.334591 |           0.721432 |          0.9 |                 1.892018 |                      1.865111 |                          0.0 |                      0.120339 |                           0.118841 |             1.842826 |                  1.680931 |      99.657174 |   1.211943 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  4.679733 |        0.421196 |                0.393598 | -0.088768 | 15.437623 | -0.025107 |              0.299135 |                  0.299135 |                   7.982856 |                  0.896826 |                      0.0 |     0.261767 |              0.0 |          7.086029 |            156.926252 |                 100.0 |                64.868339 |                72.14157 | 0.773439 |  0.0 |  0.040812 |  0.040812 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.773439 | 0.072343 |      0.0 |  0.0 |      0.0 |  0.0 | 0.072343 | nan | nan | 0.659188 |  0.0 |  0.0 |   0.0 |  nan | 31.365978 | 107.444832 | 0.072343 |   0.0 |      0.0 |   0.0 |      0.0 | 0.678285 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 08:00:00 | 0.833333 | 17.4 |              0.0 |             96.8 |       0.9 |               101.6 | nan | nan | 16.216667 |                91.275 |                   1.3 | 0.783333 | 17.4 | 16.216667 |    0.721432 |         1.260484 |           0.721432 |          0.9 |                 1.990459 |                      1.846377 |                          0.0 |                      0.125796 |                           0.117796 |             1.926764 |                  1.685281 |      99.673236 |   1.209425 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  7.219274 |        0.649765 |                0.615417 | -0.088836 | 15.486592 |  -0.04133 |              0.467717 |                  0.467717 |                   7.982856 |                  0.891371 |                      0.0 |     0.430576 |              0.0 |          7.091484 |            156.926252 |                 100.0 |                63.973307 |               71.349996 | 0.742521 |  0.0 |  0.079122 |  0.079122 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.742521 |  0.06998 |      0.0 |  0.0 |      0.0 |  0.0 |  0.06998 | nan | nan | 0.620878 |  0.0 |  0.0 |   0.0 |  nan |  31.39689 | 108.117373 |  0.06998 |   0.0 |      0.0 |   0.0 |      0.0 | 0.662603 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 09:00:00 | 0.833333 | 19.0 |              0.2 |             86.1 |       1.3 |               101.6 | nan | nan |   16.1125 |             91.991667 |                   1.5 | 0.783333 | 19.0 |   16.1125 |    1.042069 |         1.203904 |           1.042069 |          1.3 |                 2.200854 |                      1.834142 |                          0.0 |                      0.137341 |                           0.117113 |             1.894936 |                  1.687258 |      99.705064 |   1.202945 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             10.73504 |  9.442017 |         1.18975 |                1.122345 | -0.090647 | 15.589635 | -0.073664 |              0.852983 |                  0.852983 |                    8.15863 |                  0.922604 |                      0.0 |     0.814541 |              0.0 |          7.236026 |            108.641252 |                 100.0 |                63.115716 |               70.587269 | 0.704211 |  0.0 |  0.203678 |  0.203678 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.704211 | 0.066849 |      0.0 |  0.0 |      0.0 |  0.0 | 0.066849 | nan | nan | 0.496322 |  0.0 |  0.0 |   0.0 |  nan | 31.460138 | 108.754735 | 0.066849 |   0.0 |      0.0 |   0.0 |      0.0 | 0.647254 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 10:00:00 | 0.833333 | 20.3 |              0.5 |             76.8 |       1.5 |               101.6 | nan | nan | 16.083333 |             92.241667 |                   1.7 | 0.783333 | 20.3 | 16.083333 |    1.202387 |         1.149836 |           1.202387 |          1.5 |                  2.38585 |                      1.830729 |                          0.0 |                      0.147368 |                           0.116922 |             1.832333 |                  1.688695 |      99.767667 |   1.197896 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.966322 | 10.709164 |         1.92774 |                  1.8057 | -0.092439 | 15.800459 |  -0.09719 |              1.372332 |                  1.372332 |                   8.334404 |                  0.955299 |                      0.0 |     1.332528 |              0.0 |          7.379106 |             94.155751 |                 100.0 |                62.303857 |               69.861336 | 0.579655 |  0.0 |  0.369534 |  0.369534 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.579655 | 0.055376 |      0.0 |  0.0 |      0.0 |  0.0 | 0.055376 | nan | nan | 0.330466 |  0.0 |  0.0 |   0.0 |  nan | 31.546911 | 109.279014 | 0.055376 |   0.0 |      0.0 |   0.0 |      0.0 | 0.632113 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 11:00:00 | 0.833333 | 21.4 |              0.7 |             71.8 |       1.2 |               101.6 | nan | nan |    16.125 |             92.104167 |                   2.2 | 0.783333 | 21.4 |    16.125 |    0.961909 |         1.089916 |           0.961909 |          1.2 |                 2.552842 |                      1.835607 |                          0.0 |                      0.156328 |                           0.117195 |             1.832941 |                   1.69067 |      99.767059 |    1.19342 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            11.544527 | 10.773868 |        2.327265 |                2.214067 | -0.096884 | 16.089704 | -0.114702 |              1.682691 |                  1.682691 |                   8.773841 |                  1.039981 |                      0.0 |     1.639359 |              0.0 |          7.733859 |            117.694689 |                 100.0 |                61.636668 |               69.261916 | 0.413799 |  0.0 |  0.444939 |  0.444939 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.413799 | 0.039726 |      0.0 |  0.0 |      0.0 |  0.0 | 0.039726 | nan | nan | 0.255061 |  0.0 |  0.0 |   0.0 |  nan | 31.651197 | 109.653088 | 0.039726 |   0.0 |      0.0 |   0.0 |      0.0 | 0.617012 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 12:00:00 | 0.833333 | 21.3 |              0.8 |             67.5 |       1.3 |               101.6 | nan | nan | 16.204167 |             91.729167 |                   2.8 | 0.783333 | 21.3 | 16.204167 |    1.042069 |         1.037502 |           1.042069 |          1.3 |                 2.537251 |                      1.844905 |                          0.0 |                      0.155495 |                           0.117714 |             1.712644 |                  1.692316 |      99.887356 |   1.194363 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.238374 | 10.364714 |        2.425457 |                 2.25556 | -0.102203 | 16.437323 | -0.105034 |              1.714225 |                  1.714225 |                   9.301164 |                  1.142661 |                      0.0 |     1.666614 |              0.0 |          8.158503 |            108.641252 |                 100.0 |                61.160968 |               68.832957 | 0.338395 |  0.0 |  0.468211 |  0.468211 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.338395 | 0.032607 |      0.0 |  0.0 |      0.0 |  0.0 | 0.032607 | nan | nan | 0.231789 |  0.0 |  0.0 |   0.0 |  nan | 31.745814 | 109.958875 | 0.032607 |   0.0 |      0.0 |   0.0 |      0.0 | 0.602008 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 13:00:00 | 0.833333 | 21.8 |              0.5 |             66.1 |       1.5 |               101.6 | nan | nan | 16.329167 |             91.104167 |                   3.2 | 0.783333 | 21.8 | 16.329167 |    1.202387 |         1.012602 |           1.202387 |          1.5 |                 2.616045 |                      1.859671 |                          0.0 |                      0.159698 |                           0.118538 |             1.729206 |                  1.694238 |      99.870794 |   1.192265 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.700938 | 10.872489 |         1.95714 |                 1.78608 | -0.105739 | 16.752713 | -0.109021 |              1.357421 |                  1.357421 |                   9.652713 |                  1.211985 |                      0.0 |     1.306922 |              0.0 |          8.440727 |             94.155751 |                 100.0 |                60.772313 |                68.48151 | 0.315122 |  0.0 |  0.400429 |  0.400429 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.315122 | 0.030462 |      0.0 |  0.0 |      0.0 |  0.0 | 0.030462 | nan | nan | 0.299571 |  0.0 |  0.0 |   0.0 |  nan | 31.844419 | 110.243536 | 0.030462 |   0.0 |      0.0 |   0.0 |      0.0 | 0.587272 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 14:00:00 | 0.833333 | 22.9 |              0.4 |             63.4 |       1.9 |               101.6 | nan | nan | 16.545833 |             90.058333 |                   3.4 | 0.783333 | 22.9 | 16.545833 |    1.523023 |         0.984394 |           1.523023 |          1.9 |                 2.796942 |                       1.88551 |                          0.0 |                      0.169283 |                           0.119977 |             1.773261 |                  1.698059 |      99.826739 |   1.187639 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             12.93222 | 10.461143 |        1.694785 |                1.510276 | -0.107495 | 17.081396 | -0.125682 |              1.147809 |                  1.147809 |                   9.828487 |                   1.24753 |                      0.0 |     1.095829 |              0.0 |          8.580957 |             74.333488 |                 100.0 |                60.410677 |               68.153702 | 0.382904 |  0.0 |  0.389549 |  0.389549 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.382904 | 0.037141 |      0.0 |  0.0 |      0.0 |  0.0 | 0.037141 | nan | nan | 0.310451 |  0.0 |  0.0 |   0.0 |  nan | 31.959684 | 110.589299 | 0.037141 |   0.0 |      0.0 |   0.0 |      0.0 | 0.572976 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 15:00:00 | 0.833333 | 22.7 |              0.5 |             62.4 |       1.9 |               101.6 | nan | nan | 16.816667 |             88.816667 |                   3.8 | 0.783333 | 22.7 | 16.816667 |    1.523023 |         0.968687 |           1.523023 |          1.9 |                 2.763263 |                      1.918249 |                          0.0 |                      0.167505 |                           0.121797 |             1.724276 |                  1.703725 |      99.875724 |    1.18866 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.394785 |  8.865382 |        1.595844 |                1.396394 |  -0.11103 | 17.465613 | -0.113063 |               1.06126 |                   1.06126 |                  10.180036 |                  1.316893 |                      0.0 |     1.006389 |              0.0 |          8.863143 |             74.333488 |                 100.0 |                59.971629 |                67.75469 | 0.393784 |  0.0 |  0.366746 |  0.366746 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.393784 | 0.038341 |      0.0 |  0.0 |      0.0 |  0.0 | 0.038341 | nan | nan | 0.333254 |  0.0 |  0.0 |   0.0 |  nan |  32.06233 | 110.944742 | 0.038341 |   0.0 |      0.0 |   0.0 |      0.0 | 0.559151 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 16:00:00 | 0.833333 | 22.5 |              0.5 |             61.1 |       2.3 |               101.6 | nan | nan |   17.1375 |             87.333333 |                   4.1 | 0.783333 | 22.5 |   17.1375 |     1.84366 |         0.991339 |            1.84366 |          2.3 |                 2.729939 |                      1.957676 |                          0.0 |                      0.165743 |                           0.123983 |             1.667993 |                  1.709704 |      99.932007 |   1.189715 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.741708 |  6.474205 |        1.165412 |                1.002589 | -0.113666 | 17.807767 | -0.101352 |              0.761967 |                  0.761967 |                  10.443698 |                  1.370145 |                      0.0 |     0.704878 |              0.0 |          9.073553 |             61.405925 |                 100.0 |                59.520531 |               67.343541 | 0.416588 |  0.0 |  0.314711 |  0.314711 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.416588 | 0.040722 |      0.0 |  0.0 |      0.0 |  0.0 | 0.040722 | nan | nan | 0.385289 |  0.0 |  0.0 |   0.0 |  nan | 32.153266 | 111.320608 | 0.040722 |   0.0 |      0.0 |   0.0 |      0.0 | 0.545728 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 17:00:00 | 0.833333 | 21.9 |              0.3 |             62.1 |       2.4 |               101.6 | nan | nan |   17.4875 |                85.875 |                   4.4 | 0.783333 | 21.9 |   17.4875 |    1.923819 |         1.017332 |           1.923819 |          2.4 |                 2.632058 |                      2.001493 |                          0.0 |                       0.16055 |                           0.126406 |             1.634508 |                  1.718782 |      99.965492 |   1.192283 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.088631 |  3.977658 |         0.57281 |                0.493699 | -0.116332 | 18.110885 | -0.081845 |              0.375211 |                  0.375211 |                   10.70736 |                  1.421011 |                      0.0 |     0.316002 |              0.0 |          9.286348 |             58.847345 |                 100.0 |                59.043779 |               66.907699 | 0.468622 |  0.0 |  0.208264 |  0.208264 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.468622 | 0.046005 |      0.0 |  0.0 |      0.0 |  0.0 | 0.046005 | nan | nan | 0.491736 |  0.0 |  0.0 |   0.0 |  nan | 32.224694 | 111.743225 | 0.046005 |   0.0 |      0.0 |   0.0 |      0.0 | 0.532752 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 18:00:00 | 0.833333 | 21.4 |              0.1 |             67.0 |       2.5 |               101.6 | nan | nan |   17.8375 |             84.620833 |                   4.5 | 0.783333 | 21.4 |   17.8375 |    2.003978 |         1.054998 |           2.003978 |          2.5 |                 2.552842 |                      2.046164 |                          0.0 |                      0.156328 |                           0.128868 |             1.710404 |                  1.731481 |      99.889596 |   1.193968 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  1.972911 |        0.213084 |                0.183891 | -0.117274 | 18.348979 | -0.065902 |              0.139757 |                  0.139757 |                  10.795247 |                  1.433696 |                      0.0 |     0.080019 |              0.0 |           9.36155 |             56.493451 |                 100.0 |                58.508047 |               66.416324 | 0.575069 |  0.0 |  0.128039 |  0.128039 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.575069 | 0.056741 |      0.0 |  0.0 |      0.0 |  0.0 | 0.056741 | nan | nan | 0.571961 |  0.0 |  0.0 |   0.0 |  nan | 32.280179 | 112.261554 | 0.056741 |   0.0 |      0.0 |   0.0 |      0.0 | 0.520318 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 19:00:00 | 0.833333 | 20.7 |              0.0 |             74.5 |       2.5 |               101.6 | nan | nan | 18.170833 |                  83.7 |                   4.5 | 0.783333 | 20.7 | 18.170833 |    2.003978 |         1.100997 |           2.003978 |          2.5 |                 2.445437 |                      2.089517 |                          0.0 |                      0.150575 |                           0.131251 |              1.82185 |                  1.748925 |       99.77815 |   1.196313 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  0.634046 |        0.057067 |                0.049078 | -0.117398 | 18.533931 | -0.046787 |              0.037299 |                  0.037299 |                  10.795247 |                  1.423756 |                      0.0 |    -0.022024 |              0.0 |          9.371491 |             56.493451 |                 100.0 |                57.851444 |                65.81174 | 0.655295 |  0.0 |  0.068101 |  0.068101 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.655295 |  0.06504 |      0.0 |  0.0 |      0.0 |  0.0 |  0.06504 | nan | nan | 0.631899 |  0.0 |  0.0 |   0.0 |  nan |  32.31655 | 112.851809 |  0.06504 |   0.0 |      0.0 |   0.0 |      0.0 | 0.508451 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 20:00:00 | 0.833333 | 19.4 |              0.0 |             81.2 |       2.2 |               101.6 | nan | nan | 18.454167 |             83.066667 |                   4.5 | 0.783333 | 19.4 | 18.454167 |    1.763501 |         1.120309 |           1.763501 |          2.2 |                 2.256393 |                      2.126995 |                          0.0 |                      0.140363 |                           0.133306 |             1.832191 |                  1.766824 |      99.767809 |   1.201582 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                   0.1364 |                     15.566134 |            14.204272 |  0.008648 |        0.000778 |                0.000669 |  0.145725 | 18.655165 | -0.016088 |              0.000509 |                  0.000509 |                  10.795247 |                  1.412078 |                      0.0 |    -0.058328 |              0.0 |          9.383169 |             64.197103 |                 100.0 |                57.104331 |               83.754798 | 0.715232 |  0.0 |  0.088999 |  0.088999 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.715232 | 0.071457 |      0.0 |  0.0 |      0.0 |  0.0 | 0.071457 | nan | nan | 0.611001 |  0.0 |  0.0 |   0.0 |  nan | 32.322221 | 113.495584 | 0.071457 |   0.0 |      0.0 |   0.0 |      0.0 | 0.497081 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 21:00:00 | 0.833333 | 17.8 |              0.0 |             86.9 |       1.7 |               101.6 | nan | nan |     18.65 |               82.7375 |                   4.5 | 0.783333 | 17.8 |     18.65 |    1.362705 |         1.127089 |           1.362705 |          1.7 |                 2.041337 |                      2.153241 |                          0.0 |                      0.128603 |                           0.134742 |             1.773922 |                  1.781538 |      99.826078 |   1.208454 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 |   0.18755 | 18.674071 |   0.01888 |                   0.0 |                       0.0 |                  10.795247 |                  1.401564 |                      0.0 |    -0.058399 |              0.0 |          9.393683 |             83.078604 |                 100.0 |                56.290196 |               87.719298 | 0.694334 |  0.0 |  0.066711 |  0.066711 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.694334 | 0.069836 |      0.0 |  0.0 |      0.0 |  0.0 | 0.069836 | nan | nan | 0.633289 |  0.0 |  0.0 |   0.0 |  nan | 32.292925 | 114.120082 | 0.069836 |   0.0 |      0.0 |   0.0 |      0.0 | 0.486066 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 22:00:00 | 0.833333 | 17.0 |              0.0 |             90.1 |       1.7 |               101.7 | nan | nan | 18.808333 |             82.554167 |                   4.5 | 0.783333 | 17.0 | 18.808333 |    1.362705 |         1.142201 |           1.362705 |          1.7 |                 1.940693 |                      2.174668 |                          0.0 |                      0.123042 |                           0.135912 |             1.748564 |                  1.795279 |      99.951436 |   1.213101 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 |   0.18779 | 18.576416 |  0.034051 |                   0.0 |                       0.0 |                  10.795247 |                  1.391158 |                      0.0 |    -0.057965 |              0.0 |          9.404089 |             83.078604 |                 100.0 |                55.501146 |               87.719298 | 0.716622 |  0.0 |  0.057008 |  0.057008 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.716622 | 0.072562 |      0.0 |  0.0 |      0.0 |  0.0 | 0.072562 | nan | nan | 0.642992 |  0.0 |  0.0 |   0.0 |  nan | 32.248458 | 114.764142 | 0.072562 |   0.0 |      0.0 |   0.0 |      0.0 | 0.475349 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-03 23:00:00 | 0.833333 | 16.4 |              0.0 |             90.9 |       2.3 |               101.7 | nan | nan | 18.941667 |             82.379167 |                   4.5 | 0.783333 | 16.4 | 18.941667 |     1.84366 |         1.185687 |            1.84366 |          2.3 |                 1.868084 |                      2.192855 |                          0.0 |                      0.119006 |                           0.136905 |             1.698088 |                  1.806456 |     100.001912 |   1.215845 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 |  0.187983 | 18.428192 |  0.043809 |                   0.0 |                       0.0 |                  10.795247 |                   1.38284 |                      0.0 |    -0.057618 |              0.0 |          9.412407 |             61.405925 |                 100.0 |                54.688094 |               87.719298 | 0.726326 |  0.0 |  0.062194 |  0.062194 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.726326 | 0.074047 |      0.0 |  0.0 |      0.0 |  0.0 | 0.074047 | nan | nan | 0.637806 |  0.0 |  0.0 |   0.0 |  nan | 32.194232 |  115.41642 | 0.074047 |   0.0 |      0.0 |   0.0 |      0.0 | 0.464964 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 00:00:00 | 0.833333 | 15.9 |              0.0 |             88.0 |       2.3 |               101.7 | nan | nan |      18.9 |             82.083333 |                   4.5 | 0.783333 | 15.9 |      18.9 |     1.84366 |         1.235786 |            1.84366 |          2.3 |                 1.809403 |                      2.187158 |                          0.0 |                       0.11573 |                           0.136594 |             1.592275 |                  1.795292 |     100.107725 |    1.21843 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 |  0.184992 |  18.24744 |  0.050705 |                   0.0 |                       0.0 |                  10.739718 |                   1.39654 |                      0.0 |    -0.058189 |              0.0 |          9.343179 |             61.405925 |                 100.0 |                53.865393 |               87.719298 |  0.72114 |  0.0 |  0.070298 |  0.070298 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.72114 |  0.07402 |      0.0 |  0.0 |      0.0 |  0.0 |  0.07402 | nan | nan | 0.629702 |  0.0 |  0.0 |   0.0 |  nan | 32.133111 |  116.06354 |  0.07402 |   0.0 |      0.0 |   0.0 |      0.0 | 0.454866 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 01:00:00 | 0.833333 | 15.4 |              0.0 |             89.2 |       2.2 |               101.6 | nan | nan |     18.85 |             81.845833 |                   4.5 | 0.783333 | 15.4 |     18.85 |    1.763501 |         1.282546 |           1.763501 |          2.2 |                 1.752342 |                      2.180337 |                          0.0 |                      0.112529 |                           0.136222 |             1.563089 |                  1.784515 |     100.036911 |   1.219467 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 |   0.18476 | 18.043702 |  0.057104 |                   0.0 |                       0.0 |                  10.739718 |                  1.406633 |                      0.0 |     -0.05861 |              0.0 |          9.333085 |             64.197103 |                 100.0 |                53.049908 |               87.719298 | 0.713036 |  0.0 |  0.063738 |  0.063738 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.713036 |  0.07368 |      0.0 |  0.0 |      0.0 |  0.0 |  0.07368 | nan | nan | 0.636262 |  0.0 |  0.0 |   0.0 |  nan |  32.06559 | 116.702896 |  0.07368 |   0.0 |      0.0 |   0.0 |      0.0 | 0.445024 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 02:00:00 | 0.833333 | 15.2 |              0.0 |             88.8 |       2.3 |               101.6 | nan | nan |      18.8 |                 81.55 |                   4.5 | 0.783333 | 15.2 |      18.8 |     1.84366 |         1.332645 |            1.84366 |          2.3 |                 1.729962 |                      2.173535 |                          0.0 |                      0.111271 |                            0.13585 |             1.536206 |                  1.772518 |     100.063794 |   1.220436 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 |  0.184498 | 17.818633 |  0.056562 |                   0.0 |                       0.0 |                  10.739718 |                  1.418069 |                      0.0 |    -0.059086 |              0.0 |           9.32165 |             61.405925 |                 100.0 |                52.244888 |               87.719298 | 0.719595 |  0.0 |  0.066018 |  0.066018 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.719595 | 0.074855 |      0.0 |  0.0 |      0.0 |  0.0 | 0.074855 | nan | nan | 0.633982 |  0.0 |  0.0 |   0.0 |  nan | 31.998611 | 117.347636 | 0.074855 |   0.0 |      0.0 |   0.0 |      0.0 | 0.435447 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 03:00:00 | 0.833333 | 14.9 |              0.0 |             89.7 |       2.0 |               101.6 | nan | nan | 18.741667 |             81.258333 |                   4.5 | 0.783333 | 14.9 | 18.741667 |    1.603182 |         1.372725 |           1.603182 |          2.0 |                 1.696861 |                      2.165623 |                          0.0 |                      0.109405 |                           0.135418 |             1.522084 |                  1.759749 |     100.077916 |   1.221771 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 |   0.18422 | 17.595369 |   0.05822 |                   0.0 |                       0.0 |                  10.739718 |                   1.43017 |                      0.0 |     -0.05959 |              0.0 |          9.309548 |             70.616813 |                 100.0 |                51.433761 |               87.719298 | 0.717315 |  0.0 |  0.058137 |  0.058137 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.717315 | 0.075115 |      0.0 |  0.0 |      0.0 |  0.0 | 0.075115 | nan | nan | 0.641863 |  0.0 |  0.0 |   0.0 |  nan | 31.929974 | 117.989836 | 0.075115 |   0.0 |      0.0 |   0.0 |      0.0 | 0.426136 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 04:00:00 | 0.833333 | 14.3 |              0.0 |             93.0 |       2.3 |               101.6 | nan | nan | 18.670833 |             81.083333 |                   4.5 | 0.783333 | 14.3 | 18.670833 |     1.84366 |         1.422824 |            1.84366 |          2.3 |                 1.632319 |                       2.15605 |                          0.0 |                      0.105752 |                           0.134895 |             1.518057 |                  1.748197 |     100.081943 |    1.22434 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.398044 |                     15.502754 |            14.131208 |   0.07485 |        0.006692 |                0.005726 |  0.064189 |  17.36658 |  0.066238 |              0.004352 |                  0.004352 |                  10.739718 |                  1.440702 |                      0.0 |    -0.055678 |              0.0 |          9.299016 |             61.405925 |                 100.0 |                50.626491 |               73.471875 | 0.725197 |  0.0 |  0.024299 |  0.024299 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.725197 | 0.076447 |      0.0 |  0.0 |      0.0 |  0.0 | 0.076447 | nan | nan | 0.675701 |  0.0 |  0.0 |   0.0 |  nan | 31.853319 | 118.638585 | 0.076447 |   0.0 |      0.0 |   0.0 |      0.0 | 0.417086 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 05:00:00 | 0.833333 | 14.1 |              0.0 |             93.6 |       2.5 |               101.7 | nan | nan | 18.595833 |             80.920833 |                   4.5 | 0.783333 | 14.1 | 18.595833 |    2.003978 |         1.486284 |           2.003978 |          2.5 |                 1.611289 |                      2.145953 |                          0.0 |                      0.104558 |                           0.134343 |             1.508166 |                  1.736523 |     100.191834 |   1.226451 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.131208 |  0.938714 |        0.083928 |                0.071823 | -0.116833 | 17.111064 |  0.065039 |              0.054586 |                  0.054586 |                  10.739718 |                  1.451305 |                      0.0 |    -0.005885 |              0.0 |          9.288414 |             56.493451 |                 100.0 |                49.811643 |               58.193057 | 0.759034 |  0.0 | -0.006212 | -0.006212 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.759034 |  0.08056 |      0.0 |  0.0 |      0.0 |  0.0 |  0.08056 | nan | nan | 0.706212 |  0.0 |  0.0 |   0.0 |  nan | 31.777864 |  119.31706 |  0.08056 |   0.0 |      0.0 |   0.0 |      0.0 | 0.408341 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 06:00:00 | 0.833333 | 14.3 |              0.1 |             94.6 |       2.4 |               101.7 | nan | nan |    18.525 |             80.791667 |                   4.6 | 0.783333 | 14.3 |    18.525 |    1.923819 |         1.536383 |           1.923819 |          2.4 |                 1.632319 |                      2.136456 |                          0.0 |                      0.105752 |                           0.133824 |             1.544174 |                  1.726078 |     100.155826 |   1.225432 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.246552 |  2.455474 |        0.263445 |                0.226749 | -0.117598 | 16.859546 |  0.055286 |              0.172329 |                  0.172329 |                   10.82738 |                  1.478175 |                      0.0 |     0.110738 |              0.0 |          9.349205 |             58.847345 |                 100.0 |                48.960153 |               57.362046 | 0.789545 |  0.0 |   0.01635 |   0.01635 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.789545 | 0.084392 |      0.0 |  0.0 |      0.0 |  0.0 | 0.084392 | nan | nan |  0.68365 |  0.0 |  0.0 |   0.0 |  nan | 31.712161 | 120.022213 | 0.084392 |   0.0 |      0.0 |   0.0 |      0.0 | 0.399924 |   0.0 |      0.0 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 07:00:00 | 0.833333 | 15.4 |              0.9 |             93.3 |       2.5 |               101.7 | nan | nan |    18.475 |             80.620833 |                   5.5 | 0.783333 | 15.4 |    18.475 |    2.003978 |         1.589823 |           2.003978 |          2.5 |                 1.752342 |                      2.129774 |                          0.0 |                      0.112529 |                           0.133458 |             1.634935 |                  1.717041 |     100.065065 |   1.220347 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            15.284647 |  4.650125 |        1.164115 |                1.028795 | -0.125428 | 16.640536 |  0.026796 |              0.781884 |                  0.781884 |                  11.616331 |                   1.64464 |                      0.0 |     0.713358 |              0.0 |          9.971692 |             56.493451 |                 100.0 |                 48.85611 |               57.260179 | 0.766983 |  0.0 |  0.162578 |  0.162578 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.766983 | 0.082561 | 0.003915 |  0.0 | 0.000026 |  0.0 | 0.082561 | nan | nan | 0.537422 |  0.0 |  0.0 |   0.0 |  nan | 31.674949 | 120.702693 | 0.082561 |   0.0 | 0.003915 |   0.0 | 0.000026 | 0.391749 |   0.0 | 0.000026 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 08:00:00 | 0.833333 | 16.8 |              1.0 |             86.9 |       2.5 |               101.7 | nan | nan |     18.45 |             80.208333 |                   6.5 | 0.783333 | 16.8 |     18.45 |    2.003978 |         1.643262 |           2.003978 |          2.5 |                  1.91622 |                       2.12644 |                          0.0 |                      0.121684 |                           0.133275 |             1.665196 |                  1.705582 |     100.034804 |   1.214317 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            16.438085 |  7.209034 |        1.933622 |                1.710811 | -0.134064 | 16.516495 | -0.006124 |              1.300216 |                  1.300216 |                  12.492945 |                   1.83472 |                      0.0 |      1.22377 |              0.0 |         10.658225 |             56.493451 |                 100.0 |                 48.85611 |               57.260179 | 0.620755 |  0.0 |  0.313651 |  0.313651 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.620755 | 0.067238 | 0.003937 |  0.0 | 0.000054 |  0.0 | 0.067238 | nan | nan | 0.386349 |  0.0 |  0.0 |   0.0 |  nan | 31.670656 | 121.252219 | 0.067238 |   0.0 | 0.003937 |   0.0 | 0.000054 | 0.383564 |   0.0 | 0.000076 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 09:00:00 | 0.833333 | 18.2 |              1.0 |             78.9 |       2.9 |               101.7 | nan | nan | 18.416667 |             79.908333 |                   7.3 | 0.783333 | 18.2 | 18.416667 |    2.324615 |         1.696701 |           2.324615 |          2.9 |                 2.093348 |                      2.122001 |                          0.0 |                      0.131461 |                           0.133032 |             1.651651 |                  1.695656 |     100.048349 |   1.208543 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.360836 |  9.458797 |        2.537058 |                2.210424 | -0.140938 | 16.502185 | -0.036673 |              1.679922 |                  1.679922 |                  13.194235 |                  1.989515 |                      0.0 |     1.597026 |              0.0 |          11.20472 |             48.701251 |                 100.0 |                 48.85611 |               57.260179 | 0.469682 |  0.0 |  0.461151 |  0.461151 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.469682 | 0.051122 | 0.003955 |  0.0 | 0.000077 |  0.0 | 0.051122 | nan | nan | 0.238849 |  0.0 |  0.0 |   0.0 |  nan | 31.696912 | 121.666747 | 0.051122 |   0.0 | 0.003955 |   0.0 | 0.000077 | 0.375184 |   0.0 | 0.000127 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 10:00:00 | 0.833333 | 19.5 |              0.9 |             75.8 |       3.1 |               101.7 | nan | nan | 18.383333 |             79.866667 |                   7.7 | 0.783333 | 19.5 | 18.383333 |    2.484933 |         1.750141 |           2.484933 |          3.1 |                 2.270468 |                      2.117571 |                          0.0 |                      0.141127 |                            0.13279 |             1.721014 |                  1.691233 |      99.978986 |   1.202862 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.822211 | 10.747438 |         2.69052 |                2.317703 | -0.144383 | 16.589706 | -0.062862 |              1.761455 |                  1.761455 |                   13.54488 |                   2.06621 |                      0.0 |     1.675363 |              0.0 |          11.47867 |             45.559235 |                 100.0 |                 48.85611 |               57.260179 | 0.322182 |  0.0 |  0.515542 |  0.515542 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.322182 | 0.035191 | 0.003969 |  0.0 | 0.000094 |  0.0 | 0.035191 | nan | nan | 0.184458 |  0.0 |  0.0 |   0.0 |  nan | 31.749357 | 121.949674 | 0.035191 |   0.0 | 0.003969 |   0.0 | 0.000094 | 0.366607 |   0.0 | 0.000176 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 11:00:00 | 0.833333 | 20.3 |              0.8 |             73.4 |       3.3 |               101.8 | nan | nan |   18.3375 |             79.933333 |                   7.8 | 0.783333 | 20.3 |   18.3375 |    2.645251 |          1.82028 |           2.645251 |          3.3 |                  2.38585 |                      2.111492 |                          0.0 |                      0.147368 |                           0.132457 |             1.751214 |                  1.687786 |     100.048786 |   1.200635 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.937555 | 10.819272 |        2.515039 |                2.160947 | -0.145217 | 16.764525 | -0.076366 |               1.64232 |                   1.64232 |                  13.632542 |                  2.087634 |                      0.0 |     1.555335 |              0.0 |         11.544908 |             42.798069 |                 100.0 |                 48.85611 |               57.260179 | 0.267791 |  0.0 |  0.512611 |  0.512611 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.267791 | 0.029326 | 0.003978 |  0.0 | 0.000106 |  0.0 | 0.029326 | nan | nan | 0.187389 |  0.0 |  0.0 |   0.0 |  nan | 31.815307 | 122.184055 | 0.029326 |   0.0 | 0.003978 |   0.0 | 0.000106 |  0.35797 |   0.0 | 0.000226 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 12:00:00 | 0.833333 | 21.0 |              0.9 |             68.3 |       3.3 |               101.8 | nan | nan |    18.325 |             79.966667 |                   7.9 | 0.783333 | 21.0 |    18.325 |    2.645251 |         1.887079 |           2.645251 |          3.3 |                 2.490975 |                      2.109837 |                          0.0 |                      0.153018 |                           0.132366 |             1.701336 |                  1.687166 |     100.098664 |   1.198001 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.052899 | 10.408076 |        2.605564 |                2.233851 | -0.146088 | 16.984357 | -0.086738 |              1.697727 |                  1.697727 |                  13.720203 |                  2.106001 |                      0.0 |     1.609977 |              0.0 |         11.614202 |             42.798069 |                 100.0 |                 48.85611 |               57.260179 | 0.270722 |  0.0 |  0.561107 |  0.561107 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.270722 | 0.029717 | 0.003986 |  0.0 | 0.000116 |  0.0 | 0.029717 | nan | nan | 0.138893 |  0.0 |  0.0 |   0.0 |  nan | 31.891628 | 122.420958 | 0.029717 |   0.0 | 0.003986 |   0.0 | 0.000116 | 0.349486 |   0.0 | 0.000275 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 13:00:00 | 0.833333 | 21.8 |              0.8 |             61.3 |       3.2 |               101.8 | nan | nan |    18.325 |             79.766667 |                   8.2 | 0.783333 | 21.8 |    18.325 |    2.565092 |         1.943859 |           2.565092 |          3.2 |                 2.616045 |                      2.109837 |                          0.0 |                      0.159698 |                           0.132366 |             1.603636 |                  1.682947 |     100.196364 |   1.195188 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             18.39893 | 10.917404 |         2.53785 |                2.157989 | -0.148637 | 17.238761 | -0.098523 |              1.640072 |                  1.640072 |                  13.983187 |                  2.166369 |                      0.0 |     1.549806 |              0.0 |         11.816818 |             44.135508 |                 100.0 |                 48.85611 |               57.260179 | 0.222227 |  0.0 |  0.584215 |  0.584215 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.222227 | 0.024447 | 0.003994 |  0.0 | 0.000126 |  0.0 | 0.024447 | nan | nan | 0.115785 |  0.0 |  0.0 |   0.0 |  nan | 31.979734 | 122.614619 | 0.024447 |   0.0 | 0.003994 |   0.0 | 0.000126 | 0.341157 |   0.0 | 0.000323 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 14:00:00 | 0.833333 | 21.9 |              0.9 |             60.6 |       2.5 |               101.8 | nan | nan | 18.283333 |                 79.65 |                   8.7 | 0.783333 | 21.9 | 18.283333 |    2.003978 |         1.963899 |           2.003978 |          2.5 |                 2.632058 |                      2.104328 |                          0.0 |                       0.16055 |                           0.132064 |             1.595027 |                  1.676097 |     100.204973 |   1.194821 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.975649 | 10.494383 |         2.62717 |                2.208784 |   -0.1529 | 17.532448 | -0.094339 |              1.678676 |                  1.678676 |                  14.421494 |                  2.265756 |                      0.0 |      1.58427 |              0.0 |         12.155738 |             56.493451 |                 100.0 |                 48.85611 |               57.260179 | 0.199118 |  0.0 |  0.550969 |  0.550969 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.199118 | 0.021945 |    0.004 |  0.0 | 0.000134 |  0.0 | 0.021945 | nan | nan | 0.149031 |  0.0 |  0.0 |   0.0 |  nan | 32.063657 | 122.787658 | 0.021945 |   0.0 |    0.004 |   0.0 | 0.000134 | 0.332944 |   0.0 | 0.000371 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 15:00:00 | 0.833333 | 21.8 |              0.9 |             58.8 |       2.9 |               101.7 | nan | nan | 18.245833 |                  79.5 |                   9.1 | 0.783333 | 21.8 | 18.245833 |    2.324615 |         1.997298 |           2.324615 |          2.9 |                 2.616045 |                      2.099381 |                          0.0 |                      0.159698 |                           0.131792 |             1.538235 |                  1.669008 |     100.161765 |   1.194298 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            19.437025 |  8.874215 |        2.221576 |                1.861585 |  -0.15627 | 17.812189 | -0.086137 |              1.414805 |                  1.414805 |                  14.772139 |                  2.348437 |                      0.0 |     1.316953 |              0.0 |         12.423701 |             48.701251 |                 100.0 |                 48.85611 |               57.260179 | 0.232364 |  0.0 |  0.511992 |  0.511992 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.232364 | 0.025657 | 0.004006 |  0.0 | 0.000141 |  0.0 | 0.025657 | nan | nan | 0.188008 |  0.0 |  0.0 |   0.0 |  nan | 32.139377 | 122.990219 | 0.025657 |   0.0 | 0.004006 |   0.0 | 0.000141 | 0.324959 |   0.0 | 0.000418 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 16:00:00 | 0.833333 | 22.2 |              0.9 |             57.2 |       3.6 |               101.7 | nan | nan | 18.233333 |               79.3375 |                   9.5 | 0.783333 | 22.2 | 18.233333 |    2.885728 |         2.040718 |           2.885728 |          3.6 |                 2.680609 |                      2.097734 |                          0.0 |                      0.163129 |                           0.131702 |             1.533308 |                   1.66429 |     100.166692 |   1.192703 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |              19.8984 |   6.45697 |        1.616442 |                1.360212 | -0.159672 |  18.06459 | -0.089325 |              1.033761 |                  1.033761 |                  15.122784 |                  2.428674 |                      0.0 |     0.932566 |              0.0 |          12.69411 |             39.231563 |                 100.0 |                 48.85611 |               57.260179 | 0.271341 |  0.0 |  0.464653 |  0.464653 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.271341 | 0.030026 | 0.004012 |  0.0 |  0.00015 |  0.0 | 0.030026 | nan | nan | 0.235347 |  0.0 |  0.0 |   0.0 |  nan | 32.218285 | 123.227372 | 0.030026 |   0.0 | 0.004012 |   0.0 |  0.00015 | 0.317284 |   0.0 | 0.000465 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 17:00:00 | 0.833333 | 21.8 |              1.0 |             57.2 |       4.2 |               101.7 | nan | nan | 18.229167 |             79.133333 |                  10.2 | 0.783333 | 21.8 | 18.229167 |    3.366683 |         2.100837 |           3.366683 |          4.2 |                 2.616045 |                      2.097185 |                          0.0 |                      0.159698 |                           0.131672 |             1.496378 |                  1.659573 |     100.203622 |   1.194485 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            20.705807 |  3.945159 |        1.058179 |                0.907937 | -0.165683 | 18.327617 | -0.075003 |              0.690032 |                  0.690032 |                  15.736413 |                  2.564427 |                      0.0 |     0.583181 |              0.0 |         13.171986 |             33.627054 |                 100.0 |                 48.85611 |               57.260179 | 0.318681 |  0.0 |  0.401816 |  0.401816 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.318681 | 0.035356 |  0.00402 |  0.0 | 0.000159 |  0.0 | 0.035356 | nan | nan | 0.298184 |  0.0 |  0.0 |   0.0 |  nan | 32.282872 | 123.506518 | 0.035356 |   0.0 |  0.00402 |   0.0 | 0.000159 | 0.309933 |   0.0 | 0.000511 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 18:00:00 | 0.833333 | 20.9 |              1.0 |             59.9 |       4.2 |               101.7 | nan | nan | 18.208333 |               78.8375 |                  11.1 | 0.783333 | 20.9 | 18.208333 |    3.366683 |         2.157616 |           3.366683 |          4.2 |                 2.475714 |                      2.094444 |                          0.0 |                        0.1522 |                           0.131522 |             1.482953 |                  1.651207 |     100.217047 |   1.198201 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            21.743901 |  1.939642 |        0.520255 |                0.462876 | -0.173346 | 18.542906 | -0.050913 |              0.351786 |                  0.351786 |                  16.525365 |                  2.744116 |                      0.0 |     0.237448 |              0.0 |         13.781249 |             33.627054 |                 100.0 |                 48.85611 |               57.260179 | 0.381518 |  0.0 |  0.280556 |  0.280556 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.381518 | 0.042457 | 0.004029 |  0.0 | 0.000171 |  0.0 | 0.042457 | nan | nan | 0.419444 |  0.0 |  0.0 |   0.0 |  nan | 32.323368 | 123.841378 | 0.042457 |   0.0 | 0.004029 |   0.0 | 0.000171 | 0.302933 |   0.0 | 0.000556 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 19:00:00 | 0.833333 | 19.7 |              1.0 |             66.1 |       3.6 |               101.7 | nan | nan | 18.166667 |               78.4875 |                  12.1 | 0.783333 | 19.7 | 18.166667 |    2.885728 |         2.194356 |           2.885728 |          3.6 |                 2.298847 |                       2.08897 |                          0.0 |                      0.142666 |                           0.131221 |             1.519538 |                   1.63958 |     100.180462 |   1.202947 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             22.89734 |   0.60531 |        0.162358 |                0.151461 | -0.181782 | 18.677895 | -0.022077 |              0.115111 |                  0.115111 |                  17.401978 |                  2.950104 |                      0.0 |     -0.00781 |              0.0 |         14.451875 |             39.231563 |                 100.0 |                 48.85611 |               57.260179 | 0.502777 |  0.0 |  0.137894 |  0.137894 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.502777 | 0.056167 |  0.00404 |  0.0 | 0.000185 |  0.0 | 0.056167 | nan | nan | 0.562106 |  0.0 |  0.0 |   0.0 |  nan | 32.335029 | 124.283763 | 0.056167 |   0.0 |  0.00404 |   0.0 | 0.000185 | 0.296382 |   0.0 | 0.000602 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 20:00:00 | 0.833333 | 17.9 |              0.3 |             71.7 |       3.2 |               101.7 | nan | nan | 18.104167 |             78.091667 |                  12.4 | 0.783333 | 17.9 | 18.104167 |    2.565092 |         2.227756 |           2.565092 |          3.2 |                 2.054233 |                      2.080783 |                          0.0 |                      0.129312 |                           0.130772 |             1.472885 |                  1.624918 |     100.227115 |   1.210597 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.104711 |                     15.502754 |            23.243371 |  0.005136 |        0.003091 |                0.002926 |  0.254985 | 18.716764 |  0.017642 |              0.002224 |                  0.002224 |                  17.664962 |                  3.034376 |                      0.0 |    -0.124208 |              0.0 |         14.630586 |             44.135508 |                 100.0 |                 48.85611 |               83.091125 |  0.64544 |  0.0 |  0.167821 |  0.167821 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.64544 | 0.072467 | 0.004054 |  0.0 | 0.000203 |  0.0 | 0.072467 | nan | nan | 0.532179 |  0.0 |  0.0 |   0.0 |  nan |  32.30697 | 124.852478 | 0.072467 |   0.0 | 0.004054 |   0.0 | 0.000203 | 0.290389 |   0.0 | 0.000646 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 21:00:00 | 0.833333 | 15.8 |              0.0 |             80.6 |       2.2 |               101.8 | nan | nan | 18.020833 |             77.829167 |                  12.4 | 0.783333 | 15.8 | 18.020833 |    1.763501 |         2.244455 |           1.763501 |          2.2 |                 1.797863 |                       2.06991 |                          0.0 |                      0.115083 |                           0.130174 |             1.449077 |                  1.610994 |     100.350923 |    1.22071 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 |  0.305727 | 18.623235 |  0.060982 |                   0.0 |                       0.0 |                  17.664962 |                  3.060678 |                      0.0 |    -0.127528 |              0.0 |         14.604285 |             64.197103 |                 100.0 |                 48.85611 |               87.719298 | 0.615513 |  0.0 |  0.104453 |  0.104453 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.615513 | 0.069487 | 0.004073 |  0.0 | 0.000227 |  0.0 | 0.069487 | nan | nan | 0.595547 |  0.0 |  0.0 |   0.0 |  nan | 32.235572 | 125.394204 | 0.069487 |   0.0 | 0.004073 |   0.0 | 0.000227 | 0.284721 |   0.0 | 0.000691 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 22:00:00 | 0.833333 | 14.8 |              0.0 |             85.9 |       2.2 |               101.8 | nan | nan | 17.929167 |             77.654167 |                  12.4 | 0.783333 | 14.8 | 17.929167 |    1.763501 |         2.261155 |           1.763501 |          2.2 |                 1.685952 |                      2.058007 |                          0.0 |                      0.108789 |                            0.12952 |             1.448232 |                  1.598128 |     100.351768 |   1.224953 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 |  0.305184 |  18.38524 |  0.077441 |                   0.0 |                       0.0 |                  17.664962 |                  3.084324 |                      0.0 |    -0.128514 |              0.0 |         14.580638 |             64.197103 |                 100.0 |                 48.85611 |               87.719298 |  0.67888 |  0.0 |  0.084738 |  0.084738 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.67888 | 0.077072 | 0.004091 |  0.0 |  0.00025 |  0.0 | 0.077072 | nan | nan | 0.615262 |  0.0 |  0.0 |   0.0 |  nan | 32.147714 | 125.991671 | 0.077072 |   0.0 | 0.004091 |   0.0 |  0.00025 |  0.27926 |   0.0 | 0.000735 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-04 23:00:00 | 0.833333 | 14.1 |              0.0 |             89.3 |       2.1 |               101.8 | nan | nan | 17.833333 |               77.5875 |                  12.4 | 0.783333 | 14.1 | 17.833333 |    1.683342 |         2.254475 |           1.683342 |          2.1 |                 1.611289 |                      2.045628 |                          0.0 |                      0.104558 |                           0.128839 |             1.438881 |                  1.587151 |     100.361119 |   1.227981 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 |  0.304741 |  18.09238 |  0.086235 |                   0.0 |                       0.0 |                  17.664962 |                  3.103639 |                      0.0 |    -0.129318 |              0.0 |         14.561323 |             67.254108 |                 100.0 |                 48.85611 |               87.719298 | 0.698595 |  0.0 |  0.070962 |  0.070962 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.698595 | 0.079783 |  0.00411 |  0.0 | 0.000275 |  0.0 | 0.079783 | nan | nan | 0.629038 |  0.0 |  0.0 |   0.0 |  nan | 32.051062 | 126.606099 | 0.079783 |   0.0 |  0.00411 |   0.0 | 0.000275 | 0.274072 |   0.0 | 0.000779 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-05 00:00:00 | 0.833333 | 13.4 |              0.0 |             91.4 |       1.8 |               101.8 | nan | nan | 17.729167 |             77.729167 |                  12.4 | 0.783333 | 13.4 | 17.729167 |    1.442864 |         2.237776 |           1.442864 |          1.8 |                 1.539538 |                      2.032245 |                          0.0 |                      0.100468 |                           0.128102 |             1.407137 |                  1.579647 |     100.392863 |   1.231127 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 |  0.300273 | 17.770207 |  0.094396 |                   0.0 |                       0.0 |                   17.58988 |                  3.124611 |                      0.0 |    -0.130192 |              0.0 |         14.465269 |             78.463126 |                 100.0 |                 48.85611 |               87.719298 | 0.712371 |  0.0 |  0.058672 |  0.058672 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.712371 | 0.081851 |  0.00413 |  0.0 |   0.0003 |  0.0 | 0.081851 | nan | nan | 0.641328 |  0.0 |  0.0 |   0.0 |  nan | 31.946249 | 127.232188 | 0.081851 |   0.0 |  0.00413 |   0.0 |   0.0003 |  0.26908 |   0.0 | 0.000822 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-05 01:00:00 | 0.833333 | 12.8 |              0.0 |             93.1 |       1.8 |               101.8 | nan | nan | 17.620833 |             77.891667 |                  12.4 | 0.783333 | 12.8 | 17.620833 |    1.442864 |         2.224416 |           1.442864 |          1.8 |                 1.480281 |                      2.018409 |                          0.0 |                      0.097071 |                           0.127339 |             1.378142 |                  1.572172 |     100.421858 |   1.233843 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 |  0.300025 |  17.42083 |   0.09981 |                   0.0 |                       0.0 |                   17.58988 |                  3.135507 |                      0.0 |    -0.130646 |              0.0 |         14.454373 |             78.463126 |                 100.0 |                 48.85611 |               87.719298 | 0.724662 |  0.0 |  0.053362 |  0.053362 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.724662 | 0.083777 |  0.00415 |  0.0 | 0.000326 |  0.0 | 0.083777 | nan | nan | 0.646638 |  0.0 |  0.0 |   0.0 |  nan | 31.836022 | 127.868595 | 0.083777 |   0.0 |  0.00415 |   0.0 | 0.000326 | 0.264269 |   0.0 | 0.000865 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-05 02:00:00 | 0.833333 | 12.3 |              0.0 |             94.6 |       1.6 |               101.8 | nan | nan |      17.5 |             78.133333 |                  12.4 | 0.783333 | 12.3 |      17.5 |    1.282546 |         2.201036 |           1.282546 |          1.6 |                 1.432438 |                      2.003073 |                          0.0 |                      0.094315 |                           0.126493 |             1.355086 |                  1.565068 |     100.444914 |   1.236111 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 |  0.299811 | 17.053408 |  0.102674 |                   0.0 |                       0.0 |                   17.58988 |                  3.144898 |                      0.0 |    -0.131037 |              0.0 |         14.444982 |             88.271017 |                 100.0 |                 48.85611 |               87.719298 | 0.729971 |  0.0 |  0.047083 |  0.047083 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.729971 | 0.084916 | 0.004171 |  0.0 | 0.000353 |  0.0 | 0.084916 | nan | nan | 0.652917 |  0.0 |  0.0 |   0.0 |  nan | 31.722932 | 128.509126 | 0.084916 |   0.0 | 0.004171 |   0.0 | 0.000353 | 0.259622 |   0.0 | 0.000908 |   0.0 |      0.0 |   nan |    nan |
    | 1997-08-05 03:00:00 | 0.833333 | 12.0 |              0.0 |             95.6 |       1.2 |               101.8 | nan | nan | 17.379167 |             78.379167 |                  12.4 | 0.783333 | 12.0 | 17.379167 |    0.961909 |         2.174316 |           0.961909 |          1.2 |                 1.404387 |                       1.98784 |                          0.0 |                      0.092693 |                           0.125652 |             1.342594 |                  1.558052 |     100.457406 |   1.237469 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 |  0.299602 |  16.67644 |  0.101011 |                   0.0 |                       0.0 |                   17.58988 |                   3.15411 |                      0.0 |    -0.131421 |              0.0 |          14.43577 |            117.694689 |                 100.0 |                 48.85611 |               87.719298 | 0.736251 |  0.0 |  0.040901 |  0.040901 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.736251 | 0.086181 | 0.004192 |  0.0 |  0.00038 |  0.0 | 0.086181 | nan | nan | 0.659099 |  0.0 |  0.0 |   0.0 |  nan | 31.611504 | 129.154624 | 0.086181 |   0.0 | 0.004192 |   0.0 |  0.00038 | 0.255125 |   0.0 | 0.000951 |   0.0 | 0.000001 |   nan |    nan |
    | 1997-08-05 04:00:00 | 0.833333 | 11.7 |              0.0 |             96.2 |       1.9 |               101.8 | nan | nan | 17.270833 |               78.5125 |                  12.4 | 0.783333 | 11.7 | 17.270833 |    1.523023 |         2.160956 |           1.523023 |          1.9 |                 1.376819 |                      1.974268 |                          0.0 |                      0.091096 |                           0.124901 |               1.3245 |                  1.550047 |       100.4755 |   1.238856 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.366022 |                     15.438712 |            23.144579 |  0.063733 |         0.00566 |                0.005336 |  0.123083 | 16.305014 |  0.099468 |              0.004056 |                  0.004056 |                   17.58988 |                  3.166293 |                      0.0 |    -0.127873 |              0.0 |         14.423587 |             74.333488 |                 100.0 |                 48.85611 |               73.423526 | 0.742432 |  0.0 |  0.008433 |  0.008433 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.742432 |  0.08745 | 0.004213 |  0.0 | 0.000406 |  0.0 |  0.08745 | nan | nan | 0.691567 |  0.0 |  0.0 |   0.0 |  nan | 31.501619 | 129.804986 |  0.08745 |   0.0 | 0.004213 |   0.0 | 0.000406 | 0.250778 |   0.0 | 0.000993 |   0.0 | 0.000001 |   nan |    nan |
    | 1997-08-05 05:00:00 | 0.833333 | 11.6 |              0.0 |             95.7 |       1.3 |               101.8 | nan | nan | 17.166667 |                  78.6 |                  12.4 | 0.783333 | 11.6 | 17.166667 |    1.042069 |         2.120877 |           1.042069 |          1.3 |                 1.367736 |                      1.961295 |                          0.0 |                      0.090568 |                           0.124183 |             1.308923 |                  1.541578 |     100.491077 |   1.239363 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.144579 |  0.908598 |        0.080684 |                0.076088 | -0.182008 | 15.938731 |  0.093717 |              0.057827 |                  0.057827 |                   17.58988 |                  3.179808 |                      0.0 |    -0.074665 |              0.0 |         14.410072 |            108.641252 |                 100.0 |                 48.85611 |               57.260179 |   0.7749 |  0.0 |  -0.04351 |  -0.04351 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |   0.7749 | 0.091859 | 0.004234 |  0.0 | 0.000434 |  0.0 | 0.091859 | nan | nan |  0.74351 |  0.0 |  0.0 |   0.0 |  nan | 31.397486 |  130.48336 | 0.091859 |   0.0 | 0.004234 |   0.0 | 0.000434 | 0.246616 |   0.0 | 0.001035 |   0.0 | 0.000001 |   nan |    nan |
    | 1997-08-05 06:00:00 | 0.833333 | 12.6 |              0.6 |             95.1 |       1.6 |               101.8 | nan | nan | 17.095833 |             78.620833 |                  12.9 | 0.783333 | 12.6 | 17.095833 |    1.282546 |         2.094157 |           1.282546 |          1.6 |                 1.460978 |                      1.952516 |                          0.0 |                       0.09596 |                           0.123697 |              1.38939 |                  1.535084 |      100.41061 |   1.234655 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  2.420659 |        0.472904 |                 0.45318 | -0.186152 |  15.59162 |  0.064619 |              0.344417 |                  0.344417 |                  18.027018 |                  3.288825 |                      0.0 |     0.207383 |              0.0 |         14.738193 |             88.271017 |                 100.0 |                 48.85611 |               57.260179 | 0.826843 |  0.0 |  0.015278 |  0.015278 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.826843 | 0.098676 | 0.004257 |  0.0 | 0.000462 |  0.0 | 0.098676 | nan | nan | 0.684722 |  0.0 |  0.0 |   0.0 |  nan |  31.32245 | 131.206808 | 0.098676 |   0.0 | 0.004257 |   0.0 | 0.000462 | 0.242707 |   0.0 | 0.001077 |   0.0 | 0.000001 |   nan |    nan |
    | 1997-08-05 07:00:00 | 0.833333 | 14.8 |              0.9 |             96.3 |       1.9 |               101.9 | nan | nan | 17.070833 |             78.745833 |                  12.9 | 0.783333 | 14.8 | 17.070833 |    1.523023 |         2.074117 |           1.523023 |          1.9 |                 1.685952 |                      1.949426 |                          0.0 |                      0.108789 |                           0.123526 |             1.623571 |                  1.535091 |     100.276429 |   1.225361 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  4.619722 |        1.148656 |                1.101438 | -0.186171 | 15.341501 |  0.011696 |              0.837093 |                  0.837093 |                  18.027018 |                  3.287363 |                      0.0 |     0.700119 |              0.0 |         14.739656 |             74.333488 |                 100.0 |                 48.85611 |               57.260179 | 0.768055 |  0.0 |  0.126839 |  0.126839 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.768055 | 0.092272 |  0.00428 |  0.0 | 0.000492 |  0.0 | 0.092272 | nan | nan | 0.573161 |  0.0 |  0.0 |   0.0 |  nan | 31.300337 | 131.877818 | 0.092272 |   0.0 |  0.00428 |   0.0 | 0.000492 | 0.238903 |   0.0 | 0.001118 |   0.0 | 0.000001 |   nan |    nan |
    | 1997-08-05 08:00:00 | 0.833333 | 17.1 |              1.0 |             89.9 |       3.2 |               101.9 | nan | nan | 17.083333 |             78.870833 |                  12.9 | 0.783333 | 17.1 | 17.083333 |    2.565092 |         2.097497 |           2.565092 |          3.2 |                 1.953031 |                       1.95097 |                          0.0 |                      0.123726 |                           0.123612 |             1.755775 |                  1.538746 |     100.144225 |   1.215051 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  7.198198 |        1.917614 |                1.839976 |  -0.18627 | 15.267791 | -0.039576 |              1.398382 |                  1.398382 |                  18.027018 |                  3.279458 |                      0.0 |     1.261738 |              0.0 |         14.747561 |             44.135508 |                 100.0 |                 48.85611 |               57.260179 | 0.656494 |  0.0 |  0.315007 |  0.315007 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.656494 | 0.079337 | 0.004302 |  0.0 |  0.00052 |  0.0 | 0.079337 | nan | nan | 0.384993 |  0.0 |  0.0 |   0.0 |  nan | 31.329496 | 132.450153 | 0.079337 |   0.0 | 0.004302 |   0.0 |  0.00052 | 0.234948 |   0.0 | 0.001159 |   0.0 | 0.000001 |   nan |    nan |
    | 1997-08-05 09:00:00 | 0.833333 | 19.4 |              0.9 |             79.2 |       4.0 |               101.9 | nan | nan | 17.133333 |             78.883333 |                  12.8 | 0.783333 | 19.4 | 17.133333 |    3.206365 |         2.134237 |           3.206365 |          4.0 |                 2.256393 |                      1.957159 |                          0.0 |                      0.140363 |                           0.123954 |             1.787063 |                  1.543872 |     100.112937 |   1.205358 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.604725 |  9.475573 |        2.356024 |                2.266269 | -0.185529 | 15.364988 | -0.087156 |              1.722365 |                  1.722365 |                  17.939591 |                  3.250753 |                      0.0 |     1.586917 |              0.0 |         14.688838 |             35.308407 |                 100.0 |                 48.85611 |               57.260179 | 0.468327 |  0.0 |  0.495578 |  0.495578 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.468327 | 0.056861 | 0.004321 |  0.0 | 0.000544 |  0.0 | 0.056861 | nan | nan | 0.204422 |  0.0 |  0.0 |   0.0 |  nan | 31.406236 | 132.856755 | 0.056861 |   0.0 | 0.004321 |   0.0 | 0.000544 | 0.230637 |   0.0 |   0.0012 |   0.0 | 0.000001 |   nan |    nan |
    | 1997-08-05 10:00:00 | 0.833333 | 20.6 |              1.0 |             73.4 |       4.6 |               101.9 | nan | nan | 17.179167 |             78.783333 |                  12.9 | 0.783333 | 20.6 | 17.179167 |     3.68732 |         2.184336 |            3.68732 |          4.6 |                  2.43042 |                      1.962848 |                          0.0 |                      0.149767 |                           0.124269 |             1.783928 |                  1.546397 |     100.116072 |   1.200448 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 | 10.786309 |        2.873495 |                 2.75694 | -0.186428 | 15.620786 | -0.107551 |              2.095274 |                  2.095274 |                  18.027018 |                  3.266966 |                      0.0 |     1.959151 |              0.0 |         14.760052 |             30.702962 |                 100.0 |                 48.85611 |               57.260179 | 0.287755 |  0.0 |  0.669077 |  0.669077 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.287755 | 0.035045 | 0.004334 |  0.0 | 0.000561 |  0.0 | 0.035045 | nan | nan | 0.030923 |  0.0 |  0.0 |   0.0 |  nan |  31.50337 |  133.10457 | 0.035045 |   0.0 | 0.004334 |   0.0 | 0.000561 | 0.225865 |   0.0 | 0.001241 |   0.0 | 0.000001 |   nan |    nan |
    | 1997-08-05 11:00:00 | 0.833333 | 21.4 |              1.0 |             67.3 |       4.1 |               101.9 | nan | nan |    17.225 |             78.529167 |                  13.1 | 0.783333 | 21.4 |    17.225 |    3.286524 |         2.211056 |           3.286524 |          4.1 |                 2.552842 |                      1.968551 |                          0.0 |                      0.156328 |                           0.124585 |             1.718063 |                  1.545886 |     100.181937 |   1.197482 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.949834 | 10.865527 |        2.894599 |                2.761725 | -0.188095 | 15.944568 | -0.117837 |              2.098911 |                  2.098911 |                  18.201874 |                  3.309837 |                      0.0 |     1.961001 |              0.0 |         14.892037 |             34.447226 |                 100.0 |                 48.85611 |               57.260179 | 0.114257 |  0.0 |  0.697445 |  0.697445 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.114257 | 0.013938 | 0.004342 |  0.0 | 0.000571 |  0.0 | 0.013938 | nan | nan | 0.002555 |  0.0 |  0.0 |   0.0 |  nan | 31.610791 | 133.199976 | 0.013938 |   0.0 | 0.004342 |   0.0 | 0.000571 | 0.220662 |   0.0 | 0.001281 |   0.0 | 0.000001 |   nan |    nan |
    | 1997-08-05 12:00:00 | 0.833333 | 22.5 |              1.0 |             62.2 |       4.1 |               101.9 | nan | nan |   17.2875 |                78.275 |                  13.2 | 0.783333 | 22.5 |   17.2875 |    3.286524 |         2.237776 |           3.286524 |          4.1 |                 2.729939 |                      1.976351 |                          0.0 |                      0.165743 |                           0.125016 |             1.698022 |                  1.546988 |     100.201978 |   1.193116 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |             24.06487 |  10.45226 |        2.784503 |                 2.65055 |  -0.18894 | 16.302636 | -0.133863 |              2.014418 |                  2.014418 |                  18.289301 |                  3.330376 |                      0.0 |     1.875652 |              0.0 |         14.958925 |             34.447226 |                 100.0 |                 48.85611 |               57.260179 | 0.085888 |  0.0 |  0.721689 |       0.7 | 0.014246 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.085888 | 0.010485 | 0.004345 |  0.0 | 0.000575 |  0.0 | 0.010485 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.734237 | 133.256212 | 0.010485 |   0.0 | 0.004345 |   0.0 | 0.000575 | 0.215278 |   0.0 | 0.001321 |   0.0 | 0.000001 |   nan |    nan |
    | 1997-08-05 13:00:00 | 0.833333 | 23.1 |              1.0 |             59.6 |       4.4 |               101.9 | nan | nan | 17.341667 |             78.204167 |                  13.4 | 0.783333 | 23.1 | 17.341667 |    3.527001 |         2.277855 |           3.527001 |          4.4 |                 2.830978 |                      1.983133 |                          0.0 |                      0.171077 |                           0.125391 |             1.687263 |                  1.550892 |     100.212737 |   1.190747 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.294943 |  10.96314 |        2.920603 |                2.764822 | -0.190734 | 16.714124 | -0.137935 |              2.101265 |                  2.101265 |                  18.464157 |                  3.363175 |                      0.0 |     1.961133 |              0.0 |         15.100981 |             32.098552 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |  0.790142 |       0.7 | 0.058329 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010179 | 0.004347 |  0.0 | 0.000577 |  0.0 | 0.010179 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.861756 | 133.266114 | 0.010179 |   0.0 | 0.004347 |   0.0 | 0.000577 | 0.209984 |   0.0 |  0.00136 |   0.0 | 0.000001 |   nan |    nan |
    | 1997-08-05 14:00:00 | 0.833333 | 23.4 |              1.0 |             55.4 |       4.6 |               101.9 | nan | nan | 17.404167 |               77.9875 |                  13.5 | 0.783333 | 23.4 | 17.404167 |     3.68732 |         2.347994 |            3.68732 |          4.6 |                 2.882708 |                      1.990983 |                          0.0 |                      0.173797 |                           0.125825 |              1.59702 |                  1.552718 |      100.30298 |   1.189943 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.409979 | 10.528069 |        2.804699 |                2.649346 | -0.191602 | 17.139185 | -0.135234 |              2.013503 |                  2.013503 |                  18.551584 |                   3.38194 |                      0.0 |     1.872589 |              0.0 |         15.169644 |             30.702962 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |  0.813761 |       0.7 | 0.072834 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.083333 |  0.01018 | 0.004347 |  0.0 | 0.000578 |  0.0 |  0.01018 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.986572 | 133.261509 |  0.01018 |   0.0 | 0.004347 |   0.0 | 0.000578 | 0.204823 |   0.0 | 0.001399 |   0.0 | 0.000001 |   nan |    nan |
    | 1997-08-05 15:00:00 | 0.833333 | 23.8 |              1.0 |             52.1 |       4.5 |               101.9 | nan | nan |   17.4875 |             77.708333 |                  13.6 | 0.783333 | 23.8 |   17.4875 |    3.607161 |         2.401434 |           3.607161 |          4.5 |                 2.952961 |                      2.001493 |                          0.0 |                      0.177481 |                           0.126406 |             1.538493 |                  1.555327 |     100.361507 |     1.1886 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.525016 |  8.882849 |        2.366409 |                 2.23335 | -0.192477 | 17.555242 | -0.134887 |              1.697346 |                  1.697346 |                  18.639012 |                  3.400029 |                      0.0 |     1.555678 |              0.0 |         15.238983 |              31.38525 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |  0.746481 |       0.7 | 0.030124 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010179 | 0.004347 |  0.0 | 0.000578 |  0.0 | 0.010179 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.111043 | 133.299614 | 0.010179 |   0.0 | 0.004347 |   0.0 | 0.000578 | 0.199795 |   0.0 | 0.001438 |   0.0 | 0.000001 |   nan |    nan |
    | 1997-08-05 16:00:00 | 0.833333 | 24.0 |              1.0 |             49.5 |       4.7 |               101.8 | nan | nan |   17.5625 |               77.3875 |                  13.7 | 0.783333 | 24.0 |   17.5625 |    3.767479 |         2.438173 |           3.767479 |          4.7 |                 2.988644 |                      2.010993 |                          0.0 |                      0.179348 |                            0.12693 |             1.479379 |                  1.556257 |     100.320621 |   1.186889 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  6.439028 |         1.71537 |                1.620343 | -0.193307 | 17.970142 | -0.130245 |              1.231461 |                  1.231461 |                   18.72644 |                   3.42173 |                      0.0 |     1.088889 |              0.0 |         15.304709 |             30.049708 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |  0.650303 |  0.650303 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010183 | 0.004348 |  0.0 | 0.000579 |  0.0 | 0.010183 | nan | nan | 0.049697 |  0.0 |  0.0 |   0.0 |  nan | 32.230871 | 133.367837 | 0.010183 |   0.0 | 0.004348 |   0.0 | 0.000579 | 0.194897 |   0.0 | 0.001475 |   0.0 | 0.000002 |   nan |    nan |
    | 1997-08-05 17:00:00 | 0.833333 | 24.0 |              1.0 |             48.1 |       4.5 |               101.8 | nan | nan | 17.654167 |             77.008333 |                  13.7 | 0.783333 | 24.0 | 17.654167 |    3.607161 |         2.448193 |           3.607161 |          4.5 |                 2.988644 |                      2.022657 |                          0.0 |                      0.179348 |                           0.127573 |             1.437538 |                  1.557615 |     100.362462 |   1.187075 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  3.911896 |        1.042137 |                0.985011 |  -0.19328 |  18.36957 | -0.121617 |              0.748608 |                  0.748608 |                   18.72644 |                  3.423912 |                      0.0 |     0.605945 |              0.0 |         15.302528 |              31.38525 |                 100.0 |                 48.85611 |               57.260179 |  0.13303 |  0.0 |  0.506995 |  0.506995 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.13303 |  0.01627 | 0.004351 |  0.0 | 0.000582 |  0.0 |  0.01627 | nan | nan | 0.193005 |  0.0 |  0.0 |   0.0 |  nan | 32.342072 | 133.479665 |  0.01627 |   0.0 | 0.004351 |   0.0 | 0.000582 | 0.190205 |   0.0 | 0.001513 |   0.0 | 0.000002 |   nan |    nan |
    | 1997-08-05 18:00:00 | 0.833333 | 23.7 |              1.0 |             49.4 |       4.1 |               101.8 | nan | nan | 17.770833 |             76.570833 |                  13.7 | 0.783333 | 23.7 | 17.770833 |    3.286524 |         2.444853 |           3.286524 |          4.1 |                  2.93526 |                      2.037589 |                          0.0 |                      0.176554 |                           0.128396 |             1.450018 |                  1.560199 |     100.349982 |   1.188219 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  1.905864 |        0.507726 |                0.480125 | -0.193271 | 18.740238 | -0.107131 |              0.364895 |                  0.364895 |                   18.72644 |                   3.42458 |                      0.0 |     0.222204 |              0.0 |          15.30186 |             34.447226 |                 100.0 |                 48.85611 |               57.260179 | 0.276338 |  0.0 |  0.354517 |  0.354517 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.276338 | 0.033852 | 0.004354 |  0.0 | 0.000587 |  0.0 | 0.033852 | nan | nan | 0.345483 |  0.0 |  0.0 |   0.0 |  nan | 32.438786 | 133.717211 | 0.033852 |   0.0 | 0.004354 |   0.0 | 0.000587 |  0.18594 |   0.0 |  0.00155 |   0.0 | 0.000002 |   nan |    nan |
    | 1997-08-05 19:00:00 | 0.833333 | 22.8 |              1.0 |             57.0 |       3.8 |               101.8 | nan | nan |      17.9 |             76.191667 |                  13.7 | 0.783333 | 22.8 |      17.9 |    3.046047 |         2.451533 |           3.046047 |          3.8 |                 2.780058 |                      2.054233 |                          0.0 |                      0.168392 |                           0.129312 |             1.584633 |                  1.565154 |     100.215367 |   1.191234 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  0.576079 |        0.153469 |                0.145175 | -0.193327 | 19.062619 | -0.080727 |              0.110333 |                  0.110333 |                   18.72644 |                  3.420162 |                      0.0 |    -0.032173 |              0.0 |         15.306277 |             37.166744 |                 100.0 |                 48.85611 |               57.260179 | 0.428817 |  0.0 |  0.206466 |  0.206466 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.428817 | 0.052681 | 0.004362 |  0.0 | 0.000597 |  0.0 | 0.052681 | nan | nan | 0.493534 |  0.0 |  0.0 |   0.0 |  nan | 32.509096 | 134.088388 | 0.052681 |   0.0 | 0.004362 |   0.0 | 0.000597 | 0.182255 |   0.0 | 0.001586 |   0.0 | 0.000002 |   nan |    nan |
    | 1997-08-05 20:00:00 | 0.833333 | 21.3 |              0.5 |             65.9 |       2.5 |               101.8 | nan | nan | 18.041667 |                 75.95 |                  13.9 | 0.783333 | 21.3 | 18.041667 |    2.003978 |         2.428153 |           2.003978 |          2.5 |                 2.537251 |                      2.072623 |                          0.0 |                      0.155495 |                           0.130323 |             1.672048 |                  1.574157 |     100.127952 |   1.196911 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.072689 |                     15.438712 |            24.870125 |  0.002495 |        0.003269 |                0.003122 |  0.285181 | 19.296988 | -0.043265 |              0.002372 |                  0.002372 |                  18.901295 |                  3.445356 |                      0.0 |    -0.141184 |              0.0 |         15.455939 |             56.493451 |                 100.0 |                 48.85611 |               84.453771 | 0.576867 |  0.0 |   0.18067 |   0.18067 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.576867 | 0.071161 | 0.004374 |  0.0 | 0.000612 |  0.0 | 0.071161 | nan | nan |  0.51933 |  0.0 |  0.0 |   0.0 |  nan | 32.541945 | 134.589108 | 0.071161 |   0.0 | 0.004374 |   0.0 | 0.000612 | 0.179148 |   0.0 | 0.001622 |   0.0 | 0.000002 |   nan |    nan |
    | 1997-08-05 21:00:00 | 0.833333 | 19.3 |              0.0 |             73.0 |       1.9 |               101.8 | nan | nan |   18.1875 |             75.633333 |                  13.9 | 0.783333 | 19.3 |   18.1875 |    1.523023 |         2.418134 |           1.523023 |          1.9 |                 2.242395 |                      2.091705 |                          0.0 |                      0.139602 |                           0.131371 |             1.636948 |                  1.582026 |     100.163052 |   1.205255 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 |  0.323081 | 19.406483 |    0.0023 |                   0.0 |                       0.0 |                  18.901295 |                  3.434698 |                      0.0 |    -0.143112 |              0.0 |         15.466597 |             74.333488 |                 100.0 |                 48.85611 |               87.719298 | 0.602663 |  0.0 |  0.129637 |  0.129637 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.602663 | 0.074701 |  0.00439 |  0.0 | 0.000633 |  0.0 | 0.074701 | nan | nan | 0.570363 |  0.0 |  0.0 |   0.0 |  nan | 32.529228 | 135.112048 | 0.074701 |   0.0 |  0.00439 |   0.0 | 0.000633 | 0.176405 |   0.0 | 0.001658 |   0.0 | 0.000002 |   nan |    nan |
    | 1997-08-05 22:00:00 | 0.833333 | 18.3 |              0.0 |             76.7 |       2.2 |               101.8 | nan | nan | 18.333333 |                 75.25 |                  13.9 | 0.783333 | 18.3 | 18.333333 |    1.763501 |         2.418134 |           1.763501 |          2.2 |                  2.10653 |                       2.11094 |                          0.0 |                      0.132185 |                           0.132426 |             1.615709 |                  1.588483 |     100.184291 |   1.209486 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 |  0.323243 | 19.364094 |  0.022984 |                   0.0 |                       0.0 |                  18.901295 |                  3.427556 |                      0.0 |    -0.142815 |              0.0 |         15.473739 |             64.197103 |                 100.0 |                 48.85611 |               87.719298 | 0.653696 |  0.0 |  0.125384 |  0.125384 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.653696 |  0.08144 | 0.004408 |  0.0 | 0.000655 |  0.0 |  0.08144 | nan | nan | 0.574616 |  0.0 |  0.0 |   0.0 |  nan | 32.495827 | 135.679241 |  0.08144 |   0.0 | 0.004408 |   0.0 | 0.000655 | 0.173865 |   0.0 | 0.001694 |   0.0 | 0.000002 |   nan |    nan |
    | 1997-08-05 23:00:00 | 0.833333 | 17.4 |              0.0 |             81.8 |       2.7 |               101.8 | nan | nan | 18.470833 |               74.9375 |                  13.9 | 0.783333 | 17.4 | 18.470833 |    2.164296 |         2.438173 |           2.164296 |          2.7 |                 1.990459 |                      2.129218 |                          0.0 |                      0.125796 |                           0.133428 |             1.628195 |                  1.595583 |     100.171805 |   1.213176 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 |  0.323453 | 19.252757 |   0.04002 |                   0.0 |                       0.0 |                  18.901295 |                  3.418333 |                      0.0 |    -0.142431 |              0.0 |         15.482962 |             52.308751 |                 100.0 |                 48.85611 |               87.719298 | 0.657949 |  0.0 |  0.118321 |  0.118321 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 | 0.657949 | 0.082406 | 0.004426 |  0.0 | 0.000678 |  0.0 | 0.082406 | nan | nan | 0.581679 |  0.0 |  0.0 |   0.0 |  nan | 32.445391 |  136.24968 | 0.082406 |   0.0 | 0.004426 |   0.0 | 0.000678 |  0.17149 |   0.0 | 0.001729 |   0.0 | 0.000002 |   nan |    nan |

.. _lland_v3_water_hourly:

water
-----

For our hourly integration test on water type |WASSER|, we deviate from the
input data of the daily simulation (:ref:`lland_v3_wasser_daily`) in setting
precipitation to the constant value of 0.05 mm/h.  Neither the actual (|EvI|)
or the potential evaporation (|EvPo|) does show a diurnal pattern due to
method |Return_Penman_V1| relying on daily aggregated input data (e.g.
|DailyNetRadiation|) only (the same holds for water types |FLUSS| and |SEE|):

.. integration-test::

    >>> lnk(WASSER)
    >>> inputs.nied.series = 0.05
    >>> test('lland_v3_water_hourly',
    ...      axis1=(fluxes.nkor, fluxes.evpo, fluxes.evi, fluxes.qah),
    ...      use_conditions=conditions_wasser)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration |  nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |     evpo | evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz | qah |  qa | inzp | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz | qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet | outlet |
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-08-03 00:00:00 | 0.05 | 16.9 |              0.0 |             95.1 |       0.8 |               101.5 | nan | nan | 17.279167 |             85.883333 |                   1.3 | 0.047 | 16.9 | 17.279167 |    0.641273 |         1.605886 |           0.641273 |          0.8 |                 1.928423 |                      1.975309 |                          0.0 |                      0.122362 |                           0.124959 |              1.83393 |                  1.696461 |       99.66607 |    1.21073 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   3.151127 |                  0.901196 |                      0.0 |     -0.03755 |              0.0 |          2.249931 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.036205 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 01:00:00 | 0.05 | 16.6 |              0.0 |             94.9 |       0.8 |               101.5 | nan | nan | 17.241667 |             85.991667 |                   1.3 | 0.047 | 16.6 | 17.241667 |    0.641273 |         1.603439 |           0.641273 |          0.8 |                 1.892018 |                      1.970628 |                          0.0 |                      0.120339 |                             0.1247 |             1.795525 |                  1.694576 |      99.704475 |   1.212158 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   3.151127 |                  0.901844 |                      0.0 |    -0.037577 |              0.0 |          2.249283 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.036073 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 02:00:00 | 0.05 | 16.4 |              0.0 |             95.9 |       0.8 |               101.5 | nan | nan |    17.175 |             86.233333 |                   1.3 | 0.047 | 16.4 |    17.175 |    0.641273 |         1.575992 |           0.641273 |          0.8 |                 1.868084 |                       1.96233 |                          0.0 |                      0.119006 |                           0.124241 |             1.791493 |                  1.692183 |      99.708507 |   1.213014 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   3.151127 |                  0.902355 |                      0.0 |    -0.037598 |              0.0 |          2.248772 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.035724 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 03:00:00 | 0.05 | 16.3 |              0.0 |             96.7 |       0.8 |               101.5 | nan | nan |   17.0625 |             86.708333 |                   1.3 | 0.047 | 16.3 |   17.0625 |    0.641273 |         1.548545 |           0.641273 |          0.8 |                 1.856216 |                      1.948396 |                          0.0 |                      0.118345 |                           0.123469 |             1.794961 |                  1.689422 |      99.705039 |   1.213417 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   3.151127 |                  0.902364 |                      0.0 |    -0.037598 |              0.0 |          2.248763 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.035169 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 04:00:00 | 0.05 | 16.0 |              0.0 |             97.2 |       0.8 |               101.5 | nan | nan | 16.908333 |             87.366667 |                   1.3 | 0.047 | 16.0 | 16.908333 |    0.641273 |         1.504432 |           0.641273 |          0.8 |                 1.821009 |                      1.929443 |                          0.0 |                      0.116379 |                           0.122418 |              1.77002 |                   1.68569 |       99.72998 |    1.21479 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                 0.429734 |                     15.566134 |            10.503757 |  0.086656 |        0.007799 |                0.006997 | 0.0 | nan | 0.0 |              0.002099 |                  0.002099 |                   3.151127 |                  0.902341 |                      0.0 |    -0.035498 |              0.0 |          2.248786 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.034397 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 05:00:00 | 0.05 | 15.9 |              0.0 |             97.5 |       0.6 |               101.5 | nan | nan | 16.729167 |             88.204167 |                   1.3 | 0.047 | 15.9 | 16.729167 |    0.480955 |         1.453638 |           0.480955 |          0.6 |                 1.809403 |                      1.907618 |                          0.0 |                       0.11573 |                           0.121206 |             1.764168 |                  1.682599 |      99.735832 |   1.215237 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  0.968392 |        0.087159 |                0.078957 | 0.0 | nan | 0.0 |              0.023687 |                  0.023687 |                   3.151127 |                  0.901482 |                      0.0 |    -0.013875 |              0.0 |          2.249645 |            355.009444 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.033483 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 06:00:00 | 0.05 | 16.0 |              0.0 |             97.7 |       0.9 |               101.5 | nan | nan | 16.533333 |             89.191667 |                   1.3 | 0.047 | 16.0 | 16.533333 |    0.721432 |         1.392031 |           0.721432 |          0.9 |                 1.821009 |                      1.884011 |                          0.0 |                      0.116379 |                           0.119893 |             1.779125 |                   1.68038 |      99.720875 |   1.214748 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |    2.4897 |        0.224084 |                0.206122 | 0.0 | nan | 0.0 |              0.061837 |                  0.061837 |                   3.151127 |                   0.89977 |                      0.0 |     0.024346 |              0.0 |          2.251357 |            236.672963 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.032456 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 07:00:00 | 0.05 | 16.6 |              0.0 |             97.4 |       0.9 |               101.5 | nan | nan |    16.375 |                90.125 |                   1.3 | 0.047 | 16.6 |    16.375 |    0.721432 |         1.334591 |           0.721432 |          0.9 |                 1.892018 |                      1.865111 |                          0.0 |                      0.120339 |                           0.118841 |             1.842826 |                  1.680931 |      99.657174 |   1.211943 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  4.679733 |        0.421196 |                0.393598 | 0.0 | nan | 0.0 |              0.118079 |                  0.118079 |                   3.151127 |                  0.896826 |                      0.0 |     0.080712 |              0.0 |          2.254301 |            236.672963 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.031557 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 08:00:00 | 0.05 | 17.4 |              0.0 |             96.8 |       0.9 |               101.6 | nan | nan | 16.216667 |                91.275 |                   1.3 | 0.047 | 17.4 | 16.216667 |    0.721432 |         1.260484 |           0.721432 |          0.9 |                 1.990459 |                      1.846377 |                          0.0 |                      0.125796 |                           0.117796 |             1.926764 |                  1.685281 |      99.673236 |   1.209425 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  7.219274 |        0.649765 |                0.615417 | 0.0 | nan | 0.0 |              0.184625 |                  0.184625 |                   3.151127 |                  0.891371 |                      0.0 |     0.147485 |              0.0 |          2.259756 |            236.672963 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.030534 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 09:00:00 | 0.05 | 19.0 |              0.2 |             86.1 |       1.3 |               101.6 | nan | nan |   16.1125 |             91.991667 |                   1.5 | 0.047 | 19.0 |   16.1125 |    1.042069 |         1.203904 |           1.042069 |          1.3 |                 2.200854 |                      1.834142 |                          0.0 |                      0.137341 |                           0.117113 |             1.894936 |                  1.687258 |      99.705064 |   1.202945 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             10.73504 |  9.442017 |         1.18975 |                1.122345 | 0.0 | nan | 0.0 |              0.336704 |                  0.336704 |                   3.220512 |                  0.922604 |                      0.0 |     0.298262 |              0.0 |          2.297908 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.030275 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 10:00:00 | 0.05 | 20.3 |              0.5 |             76.8 |       1.5 |               101.6 | nan | nan | 16.083333 |             92.241667 |                   1.7 | 0.047 | 20.3 | 16.083333 |    1.202387 |         1.149836 |           1.202387 |          1.5 |                  2.38585 |                      1.830729 |                          0.0 |                      0.147368 |                           0.116922 |             1.832333 |                  1.688695 |      99.767667 |   1.197896 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.966322 | 10.709164 |         1.92774 |                  1.8057 | 0.0 | nan | 0.0 |               0.54171 |                   0.54171 |                   3.289897 |                  0.955299 |                      0.0 |     0.501906 |              0.0 |          2.334598 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.030379 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 11:00:00 | 0.05 | 21.4 |              0.7 |             71.8 |       1.2 |               101.6 | nan | nan |    16.125 |             92.104167 |                   2.2 | 0.047 | 21.4 |    16.125 |    0.961909 |         1.089916 |           0.961909 |          1.2 |                 2.552842 |                      1.835607 |                          0.0 |                      0.156328 |                           0.117195 |             1.832941 |                   1.69067 |      99.767059 |    1.19342 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            11.544527 | 10.773868 |        2.327265 |                2.214067 | 0.0 | nan | 0.0 |               0.66422 |                   0.66422 |                   3.463358 |                  1.039981 |                      0.0 |     0.620888 |              0.0 |          2.423377 |            177.504722 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.031338 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 12:00:00 | 0.05 | 21.3 |              0.8 |             67.5 |       1.3 |               101.6 | nan | nan | 16.204167 |             91.729167 |                   2.8 | 0.047 | 21.3 | 16.204167 |    1.042069 |         1.037502 |           1.042069 |          1.3 |                 2.537251 |                      1.844905 |                          0.0 |                      0.155495 |                           0.117714 |             1.712644 |                  1.692316 |      99.887356 |   1.194363 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.238374 | 10.364714 |        2.425457 |                 2.25556 | 0.0 | nan | 0.0 |              0.676668 |                  0.676668 |                   3.671512 |                  1.142661 |                      0.0 |     0.629057 |              0.0 |          2.528851 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.032665 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 13:00:00 | 0.05 | 21.8 |              0.5 |             66.1 |       1.5 |               101.6 | nan | nan | 16.329167 |             91.104167 |                   3.2 | 0.047 | 21.8 | 16.329167 |    1.202387 |         1.012602 |           1.202387 |          1.5 |                 2.616045 |                      1.859671 |                          0.0 |                      0.159698 |                           0.118538 |             1.729206 |                  1.694238 |      99.870794 |   1.192265 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.700938 | 10.872489 |         1.95714 |                 1.78608 | 0.0 | nan | 0.0 |              0.535824 |                  0.535824 |                   3.810281 |                  1.211985 |                      0.0 |     0.485325 |              0.0 |          2.598296 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.033843 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 14:00:00 | 0.05 | 22.9 |              0.4 |             63.4 |       1.9 |               101.6 | nan | nan | 16.545833 |             90.058333 |                   3.4 | 0.047 | 22.9 | 16.545833 |    1.523023 |         0.984394 |           1.523023 |          1.9 |                 2.796942 |                       1.88551 |                          0.0 |                      0.169283 |                           0.119977 |             1.773261 |                  1.698059 |      99.826739 |   1.187639 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             12.93222 | 10.461143 |        1.694785 |                1.510276 | 0.0 | nan | 0.0 |              0.453083 |                  0.453083 |                   3.879666 |                   1.24753 |                      0.0 |     0.401102 |              0.0 |          2.632136 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.034948 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 15:00:00 | 0.05 | 22.7 |              0.5 |             62.4 |       1.9 |               101.6 | nan | nan | 16.816667 |             88.816667 |                   3.8 | 0.047 | 22.7 | 16.816667 |    1.523023 |         0.968687 |           1.523023 |          1.9 |                 2.763263 |                      1.918249 |                          0.0 |                      0.167505 |                           0.121797 |             1.724276 |                  1.703725 |      99.875724 |    1.18866 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.394785 |  8.865382 |        1.595844 |                1.396394 | 0.0 | nan | 0.0 |              0.418918 |                  0.418918 |                   4.018435 |                  1.316893 |                      0.0 |     0.364048 |              0.0 |          2.701542 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.036637 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 16:00:00 | 0.05 | 22.5 |              0.5 |             61.1 |       2.3 |               101.6 | nan | nan |   17.1375 |             87.333333 |                   4.1 | 0.047 | 22.5 |   17.1375 |     1.84366 |         0.991339 |            1.84366 |          2.3 |                 2.729939 |                      1.957676 |                          0.0 |                      0.165743 |                           0.123983 |             1.667993 |                  1.709704 |      99.932007 |   1.189715 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.741708 |  6.474205 |        1.165412 |                1.002589 | 0.0 | nan | 0.0 |              0.300777 |                  0.300777 |                   4.122512 |                  1.370145 |                      0.0 |     0.243687 |              0.0 |          2.752367 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.038445 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 17:00:00 | 0.05 | 21.9 |              0.3 |             62.1 |       2.4 |               101.6 | nan | nan |   17.4875 |                85.875 |                   4.4 | 0.047 | 21.9 |   17.4875 |    1.923819 |         1.017332 |           1.923819 |          2.4 |                 2.632058 |                      2.001493 |                          0.0 |                       0.16055 |                           0.126406 |             1.634508 |                  1.718782 |      99.965492 |   1.192283 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.088631 |  3.977658 |         0.57281 |                0.493699 | 0.0 | nan | 0.0 |               0.14811 |                   0.14811 |                   4.226589 |                  1.421011 |                      0.0 |     0.088901 |              0.0 |          2.805578 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.040342 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 18:00:00 | 0.05 | 21.4 |              0.1 |             67.0 |       2.5 |               101.6 | nan | nan |   17.8375 |             84.620833 |                   4.5 | 0.047 | 21.4 |   17.8375 |    2.003978 |         1.054998 |           2.003978 |          2.5 |                 2.552842 |                      2.046164 |                          0.0 |                      0.156328 |                           0.128868 |             1.710404 |                  1.731481 |      99.889596 |   1.193968 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  1.972911 |        0.213084 |                0.183891 | 0.0 | nan | 0.0 |              0.055167 |                  0.055167 |                   4.261282 |                  1.433696 |                      0.0 |     -0.00457 |              0.0 |          2.827585 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.041848 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 19:00:00 | 0.05 | 20.7 |              0.0 |             74.5 |       2.5 |               101.6 | nan | nan | 18.170833 |                  83.7 |                   4.5 | 0.047 | 20.7 | 18.170833 |    2.003978 |         1.100997 |           2.003978 |          2.5 |                 2.445437 |                      2.089517 |                          0.0 |                      0.150575 |                           0.131251 |              1.82185 |                  1.748925 |       99.77815 |   1.196313 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  0.634046 |        0.057067 |                0.049078 | 0.0 | nan | 0.0 |              0.014723 |                  0.014723 |                   4.261282 |                  1.423756 |                      0.0 |      -0.0446 |              0.0 |          2.837526 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.043061 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 20:00:00 | 0.05 | 19.4 |              0.0 |             81.2 |       2.2 |               101.6 | nan | nan | 18.454167 |             83.066667 |                   4.5 | 0.047 | 19.4 | 18.454167 |    1.763501 |         1.120309 |           1.763501 |          2.2 |                 2.256393 |                      2.126995 |                          0.0 |                      0.140363 |                           0.133306 |             1.832191 |                  1.766824 |      99.767809 |   1.201582 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                   0.1364 |                     15.566134 |            14.204272 |  0.008648 |        0.000778 |                0.000669 | 0.0 | nan | 0.0 |              0.000201 |                  0.000201 |                   4.261282 |                  1.412078 |                      0.0 |    -0.058636 |              0.0 |          2.849203 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.043962 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 21:00:00 | 0.05 | 17.8 |              0.0 |             86.9 |       1.7 |               101.6 | nan | nan |     18.65 |               82.7375 |                   4.5 | 0.047 | 17.8 |     18.65 |    1.362705 |         1.127089 |           1.362705 |          1.7 |                 2.041337 |                      2.153241 |                          0.0 |                      0.128603 |                           0.134742 |             1.773922 |                  1.781538 |      99.826078 |   1.208454 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   4.261282 |                  1.401564 |                      0.0 |    -0.058399 |              0.0 |          2.859717 |            125.297451 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.044514 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 22:00:00 | 0.05 | 17.0 |              0.0 |             90.1 |       1.7 |               101.7 | nan | nan | 18.808333 |             82.554167 |                   4.5 | 0.047 | 17.0 | 18.808333 |    1.362705 |         1.142201 |           1.362705 |          1.7 |                 1.940693 |                      2.174668 |                          0.0 |                      0.123042 |                           0.135912 |             1.748564 |                  1.795279 |      99.951436 |   1.213101 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   4.261282 |                  1.391158 |                      0.0 |    -0.057965 |              0.0 |          2.870124 |            125.297451 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.044975 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-03 23:00:00 | 0.05 | 16.4 |              0.0 |             90.9 |       2.3 |               101.7 | nan | nan | 18.941667 |             82.379167 |                   4.5 | 0.047 | 16.4 | 18.941667 |     1.84366 |         1.185687 |            1.84366 |          2.3 |                 1.868084 |                      2.192855 |                          0.0 |                      0.119006 |                           0.136905 |             1.698088 |                  1.806456 |     100.001912 |   1.215845 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   4.261282 |                   1.38284 |                      0.0 |    -0.057618 |              0.0 |          2.878441 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.045525 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 00:00:00 | 0.05 | 15.9 |              0.0 |             88.0 |       2.3 |               101.7 | nan | nan |      18.9 |             82.083333 |                   4.5 | 0.047 | 15.9 |      18.9 |     1.84366 |         1.235786 |            1.84366 |          2.3 |                 1.809403 |                      2.187158 |                          0.0 |                       0.11573 |                           0.136594 |             1.592275 |                  1.795292 |     100.107725 |    1.21843 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   4.239363 |                   1.39654 |                      0.0 |    -0.058189 |              0.0 |          2.842823 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.04554 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 01:00:00 | 0.05 | 15.4 |              0.0 |             89.2 |       2.2 |               101.6 | nan | nan |     18.85 |             81.845833 |                   4.5 | 0.047 | 15.4 |     18.85 |    1.763501 |         1.282546 |           1.763501 |          2.2 |                 1.752342 |                      2.180337 |                          0.0 |                      0.112529 |                           0.136222 |             1.563089 |                  1.784515 |     100.036911 |   1.219467 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   4.239363 |                  1.406633 |                      0.0 |     -0.05861 |              0.0 |           2.83273 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.045787 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 02:00:00 | 0.05 | 15.2 |              0.0 |             88.8 |       2.3 |               101.6 | nan | nan |      18.8 |                 81.55 |                   4.5 | 0.047 | 15.2 |      18.8 |     1.84366 |         1.332645 |            1.84366 |          2.3 |                 1.729962 |                      2.173535 |                          0.0 |                      0.111271 |                            0.13585 |             1.536206 |                  1.772518 |     100.063794 |   1.220436 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   4.239363 |                  1.418069 |                      0.0 |    -0.059086 |              0.0 |          2.821294 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.046085 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 03:00:00 | 0.05 | 14.9 |              0.0 |             89.7 |       2.0 |               101.6 | nan | nan | 18.741667 |             81.258333 |                   4.5 | 0.047 | 14.9 | 18.741667 |    1.603182 |         1.372725 |           1.603182 |          2.0 |                 1.696861 |                      2.165623 |                          0.0 |                      0.109405 |                           0.135418 |             1.522084 |                  1.759749 |     100.077916 |   1.221771 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   4.239363 |                   1.43017 |                      0.0 |     -0.05959 |              0.0 |          2.809192 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.046319 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 04:00:00 | 0.05 | 14.3 |              0.0 |             93.0 |       2.3 |               101.6 | nan | nan | 18.670833 |             81.083333 |                   4.5 | 0.047 | 14.3 | 18.670833 |     1.84366 |         1.422824 |            1.84366 |          2.3 |                 1.632319 |                       2.15605 |                          0.0 |                      0.105752 |                           0.134895 |             1.518057 |                  1.748197 |     100.081943 |    1.22434 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.398044 |                     15.502754 |            14.131208 |   0.07485 |        0.006692 |                0.005726 | 0.0 | nan | 0.0 |              0.001718 |                  0.001718 |                   4.239363 |                  1.440702 |                      0.0 |    -0.058312 |              0.0 |           2.79866 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.046528 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 05:00:00 | 0.05 | 14.1 |              0.0 |             93.6 |       2.5 |               101.7 | nan | nan | 18.595833 |             80.920833 |                   4.5 | 0.047 | 14.1 | 18.595833 |    2.003978 |         1.486284 |           2.003978 |          2.5 |                 1.611289 |                      2.145953 |                          0.0 |                      0.104558 |                           0.134343 |             1.508166 |                  1.736523 |     100.191834 |   1.226451 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.131208 |  0.938714 |        0.083928 |                0.071823 | 0.0 | nan | 0.0 |              0.021547 |                  0.021547 |                   4.239363 |                  1.451305 |                      0.0 |    -0.038924 |              0.0 |          2.788058 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.046795 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 06:00:00 | 0.05 | 14.3 |              0.1 |             94.6 |       2.4 |               101.7 | nan | nan |    18.525 |             80.791667 |                   4.6 | 0.047 | 14.3 |    18.525 |    1.923819 |         1.536383 |           1.923819 |          2.4 |                 1.632319 |                      2.136456 |                          0.0 |                      0.105752 |                           0.133824 |             1.544174 |                  1.726078 |     100.155826 |   1.225432 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.246552 |  2.455474 |        0.263445 |                0.226749 | 0.0 | nan | 0.0 |              0.068025 |                  0.068025 |                   4.273966 |                  1.478175 |                      0.0 |     0.006434 |              0.0 |          2.795791 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.047182 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 07:00:00 | 0.05 | 15.4 |              0.9 |             93.3 |       2.5 |               101.7 | nan | nan |    18.475 |             80.620833 |                   5.5 | 0.047 | 15.4 |    18.475 |    2.003978 |         1.589823 |           2.003978 |          2.5 |                 1.752342 |                      2.129774 |                          0.0 |                      0.112529 |                           0.133458 |             1.634935 |                  1.717041 |     100.065065 |   1.220347 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            15.284647 |  4.650125 |        1.164115 |                1.028795 | 0.0 | nan | 0.0 |              0.308638 |                  0.308638 |                   4.585394 |                   1.64464 |                      0.0 |     0.240112 |              0.0 |          2.940754 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |   0.0492 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 08:00:00 | 0.05 | 16.8 |              1.0 |             86.9 |       2.5 |               101.7 | nan | nan |     18.45 |             80.208333 |                   6.5 | 0.047 | 16.8 |     18.45 |    2.003978 |         1.643262 |           2.003978 |          2.5 |                  1.91622 |                       2.12644 |                          0.0 |                      0.121684 |                           0.133275 |             1.665196 |                  1.705582 |     100.034804 |   1.214317 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            16.438085 |  7.209034 |        1.933622 |                1.710811 | 0.0 | nan | 0.0 |              0.513243 |                  0.513243 |                   4.931426 |                   1.83472 |                      0.0 |     0.436797 |              0.0 |          3.096706 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.051571 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 09:00:00 | 0.05 | 18.2 |              1.0 |             78.9 |       2.9 |               101.7 | nan | nan | 18.416667 |             79.908333 |                   7.3 | 0.047 | 18.2 | 18.416667 |    2.324615 |         1.696701 |           2.324615 |          2.9 |                 2.093348 |                      2.122001 |                          0.0 |                      0.131461 |                           0.133032 |             1.651651 |                  1.695656 |     100.048349 |   1.208543 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.360836 |  9.458797 |        2.537058 |                2.210424 | 0.0 | nan | 0.0 |              0.663127 |                  0.663127 |                   5.208251 |                  1.989515 |                      0.0 |     0.580231 |              0.0 |          3.218735 |              73.45023 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.053462 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 10:00:00 | 0.05 | 19.5 |              0.9 |             75.8 |       3.1 |               101.7 | nan | nan | 18.383333 |             79.866667 |                   7.7 | 0.047 | 19.5 | 18.383333 |    2.484933 |         1.750141 |           2.484933 |          3.1 |                 2.270468 |                      2.117571 |                          0.0 |                      0.141127 |                            0.13279 |             1.721014 |                  1.691233 |      99.978986 |   1.202862 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.822211 | 10.747438 |         2.69052 |                2.317703 | 0.0 | nan | 0.0 |              0.695311 |                  0.695311 |                   5.346663 |                   2.06621 |                      0.0 |     0.609219 |              0.0 |          3.280453 |             68.711505 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.054455 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 11:00:00 | 0.05 | 20.3 |              0.8 |             73.4 |       3.3 |               101.8 | nan | nan |   18.3375 |             79.933333 |                   7.8 | 0.047 | 20.3 |   18.3375 |    2.645251 |          1.82028 |           2.645251 |          3.3 |                  2.38585 |                      2.111492 |                          0.0 |                      0.147368 |                           0.132457 |             1.751214 |                  1.687786 |     100.048786 |   1.200635 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.937555 | 10.819272 |        2.515039 |                2.160947 | 0.0 | nan | 0.0 |              0.648284 |                  0.648284 |                   5.381266 |                  2.087634 |                      0.0 |     0.561299 |              0.0 |          3.293633 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.054881 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 12:00:00 | 0.05 | 21.0 |              0.9 |             68.3 |       3.3 |               101.8 | nan | nan |    18.325 |             79.966667 |                   7.9 | 0.047 | 21.0 |    18.325 |    2.645251 |         1.887079 |           2.645251 |          3.3 |                 2.490975 |                      2.109837 |                          0.0 |                      0.153018 |                           0.132366 |             1.701336 |                  1.687166 |     100.098664 |   1.198001 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.052899 | 10.408076 |        2.605564 |                2.233851 | 0.0 | nan | 0.0 |              0.670155 |                  0.670155 |                    5.41587 |                  2.106001 |                      0.0 |     0.582405 |              0.0 |          3.309868 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.055387 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 13:00:00 | 0.05 | 21.8 |              0.8 |             61.3 |       3.2 |               101.8 | nan | nan |    18.325 |             79.766667 |                   8.2 | 0.047 | 21.8 |    18.325 |    2.565092 |         1.943859 |           2.565092 |          3.2 |                 2.616045 |                      2.109837 |                          0.0 |                      0.159698 |                           0.132366 |             1.603636 |                  1.682947 |     100.196364 |   1.195188 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             18.39893 | 10.917404 |         2.53785 |                2.157989 | 0.0 | nan | 0.0 |              0.647397 |                  0.647397 |                   5.519679 |                  2.166369 |                      0.0 |     0.557131 |              0.0 |           3.35331 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.056372 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 14:00:00 | 0.05 | 21.9 |              0.9 |             60.6 |       2.5 |               101.8 | nan | nan | 18.283333 |                 79.65 |                   8.7 | 0.047 | 21.9 | 18.283333 |    2.003978 |         1.963899 |           2.003978 |          2.5 |                 2.632058 |                      2.104328 |                          0.0 |                       0.16055 |                           0.132064 |             1.595027 |                  1.676097 |     100.204973 |   1.194821 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.975649 | 10.494383 |         2.62717 |                2.208784 | 0.0 | nan | 0.0 |              0.662635 |                  0.662635 |                   5.692695 |                  2.265756 |                      0.0 |     0.568229 |              0.0 |          3.426939 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.057374 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 15:00:00 | 0.05 | 21.8 |              0.9 |             58.8 |       2.9 |               101.7 | nan | nan | 18.245833 |                  79.5 |                   9.1 | 0.047 | 21.8 | 18.245833 |    2.324615 |         1.997298 |           2.324615 |          2.9 |                 2.616045 |                      2.099381 |                          0.0 |                      0.159698 |                           0.131792 |             1.538235 |                  1.669008 |     100.161765 |   1.194298 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            19.437025 |  8.874215 |        2.221576 |                1.861585 | 0.0 | nan | 0.0 |              0.558476 |                  0.558476 |                   5.831107 |                  2.348437 |                      0.0 |     0.460624 |              0.0 |           3.48267 |              73.45023 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.058283 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 16:00:00 | 0.05 | 22.2 |              0.9 |             57.2 |       3.6 |               101.7 | nan | nan | 18.233333 |               79.3375 |                   9.5 | 0.047 | 22.2 | 18.233333 |    2.885728 |         2.040718 |           2.885728 |          3.6 |                 2.680609 |                      2.097734 |                          0.0 |                      0.163129 |                           0.131702 |             1.533308 |                   1.66429 |     100.166692 |   1.192703 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |              19.8984 |   6.45697 |        1.616442 |                1.360212 | 0.0 | nan | 0.0 |              0.408064 |                  0.408064 |                    5.96952 |                  2.428674 |                      0.0 |     0.306869 |              0.0 |          3.540846 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.05932 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 17:00:00 | 0.05 | 21.8 |              1.0 |             57.2 |       4.2 |               101.7 | nan | nan | 18.229167 |             79.133333 |                  10.2 | 0.047 | 21.8 | 18.229167 |    3.366683 |         2.100837 |           3.366683 |          4.2 |                 2.616045 |                      2.097185 |                          0.0 |                      0.159698 |                           0.131672 |             1.496378 |                  1.659573 |     100.203622 |   1.194485 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            20.705807 |  3.945159 |        1.058179 |                0.907937 | 0.0 | nan | 0.0 |              0.272381 |                  0.272381 |                   6.211742 |                  2.564427 |                      0.0 |      0.16553 |              0.0 |          3.647315 |             50.715635 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.061051 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 18:00:00 | 0.05 | 20.9 |              1.0 |             59.9 |       4.2 |               101.7 | nan | nan | 18.208333 |               78.8375 |                  11.1 | 0.047 | 20.9 | 18.208333 |    3.366683 |         2.157616 |           3.366683 |          4.2 |                 2.475714 |                      2.094444 |                          0.0 |                        0.1522 |                           0.131522 |             1.482953 |                  1.651207 |     100.217047 |   1.198201 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            21.743901 |  1.939642 |        0.520255 |                0.462876 | 0.0 | nan | 0.0 |              0.138863 |                  0.138863 |                    6.52317 |                  2.744116 |                      0.0 |     0.024525 |              0.0 |          3.779055 |             50.715635 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.063121 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 19:00:00 | 0.05 | 19.7 |              1.0 |             66.1 |       3.6 |               101.7 | nan | nan | 18.166667 |               78.4875 |                  12.1 | 0.047 | 19.7 | 18.166667 |    2.885728 |         2.194356 |           2.885728 |          3.6 |                 2.298847 |                       2.08897 |                          0.0 |                      0.142666 |                           0.131221 |             1.519538 |                   1.63958 |     100.180462 |   1.202947 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             22.89734 |   0.60531 |        0.162358 |                0.151461 | 0.0 | nan | 0.0 |              0.045438 |                  0.045438 |                   6.869202 |                  2.950104 |                      0.0 |    -0.077483 |              0.0 |          3.919098 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |   0.0652 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 20:00:00 | 0.05 | 17.9 |              0.3 |             71.7 |       3.2 |               101.7 | nan | nan | 18.104167 |             78.091667 |                  12.4 | 0.047 | 17.9 | 18.104167 |    2.565092 |         2.227756 |           2.565092 |          3.2 |                 2.054233 |                      2.080783 |                          0.0 |                      0.129312 |                           0.130772 |             1.472885 |                  1.624918 |     100.227115 |   1.210597 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.104711 |                     15.502754 |            23.243371 |  0.005136 |        0.003091 |                0.002926 | 0.0 | nan | 0.0 |              0.000878 |                  0.000878 |                   6.973011 |                  3.034376 |                      0.0 |    -0.125554 |              0.0 |          3.938635 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.06592 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 21:00:00 | 0.05 | 15.8 |              0.0 |             80.6 |       2.2 |               101.8 | nan | nan | 18.020833 |             77.829167 |                  12.4 | 0.047 | 15.8 | 18.020833 |    1.763501 |         2.244455 |           1.763501 |          2.2 |                 1.797863 |                       2.06991 |                          0.0 |                      0.115083 |                           0.130174 |             1.449077 |                  1.610994 |     100.350923 |    1.22071 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   6.973011 |                  3.060678 |                      0.0 |    -0.127528 |              0.0 |          3.912334 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.065866 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 22:00:00 | 0.05 | 14.8 |              0.0 |             85.9 |       2.2 |               101.8 | nan | nan | 17.929167 |             77.654167 |                  12.4 | 0.047 | 14.8 | 17.929167 |    1.763501 |         2.261155 |           1.763501 |          2.2 |                 1.685952 |                      2.058007 |                          0.0 |                      0.108789 |                            0.12952 |             1.448232 |                  1.598128 |     100.351768 |   1.224953 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   6.973011 |                  3.084324 |                      0.0 |    -0.128514 |              0.0 |          3.888687 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.065746 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-04 23:00:00 | 0.05 | 14.1 |              0.0 |             89.3 |       2.1 |               101.8 | nan | nan | 17.833333 |               77.5875 |                  12.4 | 0.047 | 14.1 | 17.833333 |    1.683342 |         2.254475 |           1.683342 |          2.1 |                 1.611289 |                      2.045628 |                          0.0 |                      0.104558 |                           0.128839 |             1.438881 |                  1.587151 |     100.361119 |   1.227981 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   6.973011 |                  3.103639 |                      0.0 |    -0.129318 |              0.0 |          3.869372 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.065422 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 00:00:00 | 0.05 | 13.4 |              0.0 |             91.4 |       1.8 |               101.8 | nan | nan | 17.729167 |             77.729167 |                  12.4 | 0.047 | 13.4 | 17.729167 |    1.442864 |         2.237776 |           1.442864 |          1.8 |                 1.539538 |                      2.032245 |                          0.0 |                      0.100468 |                           0.128102 |             1.407137 |                  1.579647 |     100.392863 |   1.231127 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   6.943374 |                  3.124611 |                      0.0 |    -0.130192 |              0.0 |          3.818763 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.064471 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 01:00:00 | 0.05 | 12.8 |              0.0 |             93.1 |       1.8 |               101.8 | nan | nan | 17.620833 |             77.891667 |                  12.4 | 0.047 | 12.8 | 17.620833 |    1.442864 |         2.224416 |           1.442864 |          1.8 |                 1.480281 |                      2.018409 |                          0.0 |                      0.097071 |                           0.127339 |             1.378142 |                  1.572172 |     100.421858 |   1.233843 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   6.943374 |                  3.135507 |                      0.0 |    -0.130646 |              0.0 |          3.807867 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.063963 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 02:00:00 | 0.05 | 12.3 |              0.0 |             94.6 |       1.6 |               101.8 | nan | nan |      17.5 |             78.133333 |                  12.4 | 0.047 | 12.3 |      17.5 |    1.282546 |         2.201036 |           1.282546 |          1.6 |                 1.432438 |                      2.003073 |                          0.0 |                      0.094315 |                           0.126493 |             1.355086 |                  1.565068 |     100.444914 |   1.236111 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   6.943374 |                  3.144898 |                      0.0 |    -0.131037 |              0.0 |          3.798476 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.063323 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 03:00:00 | 0.05 | 12.0 |              0.0 |             95.6 |       1.2 |               101.8 | nan | nan | 17.379167 |             78.379167 |                  12.4 | 0.047 | 12.0 | 17.379167 |    0.961909 |         2.174316 |           0.961909 |          1.2 |                 1.404387 |                       1.98784 |                          0.0 |                      0.092693 |                           0.125652 |             1.342594 |                  1.558052 |     100.457406 |   1.237469 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   6.943374 |                   3.15411 |                      0.0 |    -0.131421 |              0.0 |          3.789264 |            177.504722 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.062667 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 04:00:00 | 0.05 | 11.7 |              0.0 |             96.2 |       1.9 |               101.8 | nan | nan | 17.270833 |               78.5125 |                  12.4 | 0.047 | 11.7 | 17.270833 |    1.523023 |         2.160956 |           1.523023 |          1.9 |                 1.376819 |                      1.974268 |                          0.0 |                      0.091096 |                           0.124901 |               1.3245 |                  1.550047 |       100.4755 |   1.238856 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.366022 |                     15.438712 |            23.144579 |  0.063733 |         0.00566 |                0.005336 | 0.0 | nan | 0.0 |              0.001601 |                  0.001601 |                   6.943374 |                  3.166293 |                      0.0 |    -0.130328 |              0.0 |          3.777081 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.062183 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 05:00:00 | 0.05 | 11.6 |              0.0 |             95.7 |       1.3 |               101.8 | nan | nan | 17.166667 |                  78.6 |                  12.4 | 0.047 | 11.6 | 17.166667 |    1.042069 |         2.120877 |           1.042069 |          1.3 |                 1.367736 |                      1.961295 |                          0.0 |                      0.090568 |                           0.124183 |             1.308923 |                  1.541578 |     100.491077 |   1.239363 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.144579 |  0.908598 |        0.080684 |                0.076088 | 0.0 | nan | 0.0 |              0.022826 |                  0.022826 |                   6.943374 |                  3.179808 |                      0.0 |    -0.109666 |              0.0 |          3.763566 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.061583 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 06:00:00 | 0.05 | 12.6 |              0.6 |             95.1 |       1.6 |               101.8 | nan | nan | 17.095833 |             78.620833 |                  12.9 | 0.047 | 12.6 | 17.095833 |    1.282546 |         2.094157 |           1.282546 |          1.6 |                 1.460978 |                      1.952516 |                          0.0 |                       0.09596 |                           0.123697 |              1.38939 |                  1.535084 |      100.41061 |   1.234655 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  2.420659 |        0.472904 |                 0.45318 | 0.0 | nan | 0.0 |              0.135954 |                  0.135954 |                   7.115928 |                  3.288825 |                      0.0 |     -0.00108 |              0.0 |          3.827103 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.062021 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 07:00:00 | 0.05 | 14.8 |              0.9 |             96.3 |       1.9 |               101.9 | nan | nan | 17.070833 |             78.745833 |                  12.9 | 0.047 | 14.8 | 17.070833 |    1.523023 |         2.074117 |           1.523023 |          1.9 |                 1.685952 |                      1.949426 |                          0.0 |                      0.108789 |                           0.123526 |             1.623571 |                  1.535091 |     100.276429 |   1.225361 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  4.619722 |        1.148656 |                1.101438 | 0.0 | nan | 0.0 |              0.330431 |                  0.330431 |                   7.115928 |                  3.287363 |                      0.0 |     0.193458 |              0.0 |          3.828566 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.061775 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 08:00:00 | 0.05 | 17.1 |              1.0 |             89.9 |       3.2 |               101.9 | nan | nan | 17.083333 |             78.870833 |                  12.9 | 0.047 | 17.1 | 17.083333 |    2.565092 |         2.097497 |           2.565092 |          3.2 |                 1.953031 |                       1.95097 |                          0.0 |                      0.123726 |                           0.123612 |             1.755775 |                  1.538746 |     100.144225 |   1.215051 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  7.198198 |        1.917614 |                1.839976 | 0.0 | nan | 0.0 |              0.551993 |                  0.551993 |                   7.115928 |                  3.279458 |                      0.0 |     0.415349 |              0.0 |          3.836471 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.061896 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 09:00:00 | 0.05 | 19.4 |              0.9 |             79.2 |       4.0 |               101.9 | nan | nan | 17.133333 |             78.883333 |                  12.8 | 0.047 | 19.4 | 17.133333 |    3.206365 |         2.134237 |           3.206365 |          4.0 |                 2.256393 |                      1.957159 |                          0.0 |                      0.140363 |                           0.123954 |             1.787063 |                  1.543872 |     100.112937 |   1.205358 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.604725 |  9.475573 |        2.356024 |                2.266269 | 0.0 | nan | 0.0 |              0.679881 |                  0.679881 |                   7.081417 |                  3.250753 |                      0.0 |     0.544433 |              0.0 |          3.830665 |             53.251417 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.062093 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 10:00:00 | 0.05 | 20.6 |              1.0 |             73.4 |       4.6 |               101.9 | nan | nan | 17.179167 |             78.783333 |                  12.9 | 0.047 | 20.6 | 17.179167 |     3.68732 |         2.184336 |            3.68732 |          4.6 |                  2.43042 |                      1.962848 |                          0.0 |                      0.149767 |                           0.124269 |             1.783928 |                  1.546397 |     100.116072 |   1.200448 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 | 10.786309 |        2.873495 |                 2.75694 | 0.0 | nan | 0.0 |              0.827082 |                  0.827082 |                   7.115928 |                  3.266966 |                      0.0 |     0.690958 |              0.0 |          3.848962 |              46.30558 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.062733 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 11:00:00 | 0.05 | 21.4 |              1.0 |             67.3 |       4.1 |               101.9 | nan | nan |    17.225 |             78.529167 |                  13.1 | 0.047 | 21.4 |    17.225 |    3.286524 |         2.211056 |           3.286524 |          4.1 |                 2.552842 |                      1.968551 |                          0.0 |                      0.156328 |                           0.124585 |             1.718063 |                  1.545886 |     100.181937 |   1.197482 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.949834 | 10.865527 |        2.894599 |                2.761725 | 0.0 | nan | 0.0 |              0.828518 |                  0.828518 |                    7.18495 |                  3.309837 |                      0.0 |     0.690608 |              0.0 |          3.875113 |             51.952602 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.063478 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 12:00:00 | 0.05 | 22.5 |              1.0 |             62.2 |       4.1 |               101.9 | nan | nan |   17.2875 |                78.275 |                  13.2 | 0.047 | 22.5 |   17.2875 |    3.286524 |         2.237776 |           3.286524 |          4.1 |                 2.729939 |                      1.976351 |                          0.0 |                      0.165743 |                           0.125016 |             1.698022 |                  1.546988 |     100.201978 |   1.193116 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |             24.06487 |  10.45226 |        2.784503 |                 2.65055 | 0.0 | nan | 0.0 |              0.795165 |                  0.795165 |                   7.219461 |                  3.330376 |                      0.0 |     0.656399 |              0.0 |          3.889085 |             51.952602 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.064117 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 13:00:00 | 0.05 | 23.1 |              1.0 |             59.6 |       4.4 |               101.9 | nan | nan | 17.341667 |             78.204167 |                  13.4 | 0.047 | 23.1 | 17.341667 |    3.527001 |         2.277855 |           3.527001 |          4.4 |                 2.830978 |                      1.983133 |                          0.0 |                      0.171077 |                           0.125391 |             1.687263 |                  1.550892 |     100.212737 |   1.190747 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.294943 |  10.96314 |        2.920603 |                2.764822 | 0.0 | nan | 0.0 |              0.829447 |                  0.829447 |                   7.288483 |                  3.363175 |                      0.0 |     0.689314 |              0.0 |          3.925307 |             48.410379 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.064896 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 14:00:00 | 0.05 | 23.4 |              1.0 |             55.4 |       4.6 |               101.9 | nan | nan | 17.404167 |               77.9875 |                  13.5 | 0.047 | 23.4 | 17.404167 |     3.68732 |         2.347994 |            3.68732 |          4.6 |                 2.882708 |                      1.990983 |                          0.0 |                      0.173797 |                           0.125825 |              1.59702 |                  1.552718 |      100.30298 |   1.189943 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.409979 | 10.528069 |        2.804699 |                2.649346 | 0.0 | nan | 0.0 |              0.794804 |                  0.794804 |                   7.322994 |                   3.38194 |                      0.0 |      0.65389 |              0.0 |          3.941053 |              46.30558 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.065782 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 15:00:00 | 0.05 | 23.8 |              1.0 |             52.1 |       4.5 |               101.9 | nan | nan |   17.4875 |             77.708333 |                  13.6 | 0.047 | 23.8 |   17.4875 |    3.607161 |         2.401434 |           3.607161 |          4.5 |                 2.952961 |                      2.001493 |                          0.0 |                      0.177481 |                           0.126406 |             1.538493 |                  1.555327 |     100.361507 |     1.1886 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.525016 |  8.882849 |        2.366409 |                 2.23335 | 0.0 | nan | 0.0 |              0.670005 |                  0.670005 |                   7.357505 |                  3.400029 |                      0.0 |     0.528337 |              0.0 |          3.957476 |             47.334593 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.066681 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 16:00:00 | 0.05 | 24.0 |              1.0 |             49.5 |       4.7 |               101.8 | nan | nan |   17.5625 |               77.3875 |                  13.7 | 0.047 | 24.0 |   17.5625 |    3.767479 |         2.438173 |           3.767479 |          4.7 |                 2.988644 |                      2.010993 |                          0.0 |                      0.179348 |                            0.12693 |             1.479379 |                  1.556257 |     100.320621 |   1.186889 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  6.439028 |         1.71537 |                1.620343 | 0.0 | nan | 0.0 |              0.486103 |                  0.486103 |                   7.392016 |                   3.42173 |                      0.0 |     0.343531 |              0.0 |          3.970285 |             45.320355 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.06748 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 17:00:00 | 0.05 | 24.0 |              1.0 |             48.1 |       4.5 |               101.8 | nan | nan | 17.654167 |             77.008333 |                  13.7 | 0.047 | 24.0 | 17.654167 |    3.607161 |         2.448193 |           3.607161 |          4.5 |                 2.988644 |                      2.022657 |                          0.0 |                      0.179348 |                           0.127573 |             1.437538 |                  1.557615 |     100.362462 |   1.187075 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  3.911896 |        1.042137 |                0.985011 | 0.0 | nan | 0.0 |              0.295503 |                  0.295503 |                   7.392016 |                  3.423912 |                      0.0 |      0.15284 |              0.0 |          3.968104 |             47.334593 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.068039 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 18:00:00 | 0.05 | 23.7 |              1.0 |             49.4 |       4.1 |               101.8 | nan | nan | 17.770833 |             76.570833 |                  13.7 | 0.047 | 23.7 | 17.770833 |    3.286524 |         2.444853 |           3.286524 |          4.1 |                  2.93526 |                      2.037589 |                          0.0 |                      0.176554 |                           0.128396 |             1.450018 |                  1.560199 |     100.349982 |   1.188219 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  1.905864 |        0.507726 |                0.480125 | 0.0 | nan | 0.0 |              0.144038 |                  0.144038 |                   7.392016 |                   3.42458 |                      0.0 |     0.001347 |              0.0 |          3.967436 |             51.952602 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.068633 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 19:00:00 | 0.05 | 22.8 |              1.0 |             57.0 |       3.8 |               101.8 | nan | nan |      17.9 |             76.191667 |                  13.7 | 0.047 | 22.8 |      17.9 |    3.046047 |         2.451533 |           3.046047 |          3.8 |                 2.780058 |                      2.054233 |                          0.0 |                      0.168392 |                           0.129312 |             1.584633 |                  1.565154 |     100.215367 |   1.191234 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  0.576079 |        0.153469 |                0.145175 | 0.0 | nan | 0.0 |              0.043553 |                  0.043553 |                   7.392016 |                  3.420162 |                      0.0 |    -0.098954 |              0.0 |          3.971854 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.069308 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 20:00:00 | 0.05 | 21.3 |              0.5 |             65.9 |       2.5 |               101.8 | nan | nan | 18.041667 |                 75.95 |                  13.9 | 0.047 | 21.3 | 18.041667 |    2.003978 |         2.428153 |           2.003978 |          2.5 |                 2.537251 |                      2.072623 |                          0.0 |                      0.155495 |                           0.130323 |             1.672048 |                  1.574157 |     100.127952 |   1.196911 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.072689 |                     15.438712 |            24.870125 |  0.002495 |        0.003269 |                0.003122 | 0.0 | nan | 0.0 |              0.000936 |                  0.000936 |                   7.461037 |                  3.445356 |                      0.0 |     -0.14262 |              0.0 |          4.015681 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.070109 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 21:00:00 | 0.05 | 19.3 |              0.0 |             73.0 |       1.9 |               101.8 | nan | nan |   18.1875 |             75.633333 |                  13.9 | 0.047 | 19.3 |   18.1875 |    1.523023 |         2.418134 |           1.523023 |          1.9 |                 2.242395 |                      2.091705 |                          0.0 |                      0.139602 |                           0.131371 |             1.636948 |                  1.582026 |     100.163052 |   1.205255 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   7.461037 |                  3.434698 |                      0.0 |    -0.143112 |              0.0 |          4.026339 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.070708 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 22:00:00 | 0.05 | 18.3 |              0.0 |             76.7 |       2.2 |               101.8 | nan | nan | 18.333333 |                 75.25 |                  13.9 | 0.047 | 18.3 | 18.333333 |    1.763501 |         2.418134 |           1.763501 |          2.2 |                  2.10653 |                       2.11094 |                          0.0 |                      0.132185 |                           0.132426 |             1.615709 |                  1.588483 |     100.184291 |   1.209486 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   7.461037 |                  3.427556 |                      0.0 |    -0.142815 |              0.0 |          4.033482 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 | 0.071405 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |
    | 1997-08-05 23:00:00 | 0.05 | 17.4 |              0.0 |             81.8 |       2.7 |               101.8 | nan | nan | 18.470833 |               74.9375 |                  13.9 | 0.047 | 17.4 | 18.470833 |    2.164296 |         2.438173 |           2.164296 |          2.7 |                 1.990459 |                      2.129218 |                          0.0 |                      0.125796 |                           0.133428 |             1.628195 |                  1.595583 |     100.171805 |   1.213176 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                       0.0 |                   7.461037 |                  3.418333 |                      0.0 |    -0.142431 |              0.0 |          4.042705 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |  0.07218 | nan | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 | 0.0 |  0.0 |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    0.0 |

    >>> inputs.nied.series = 0.0

.. _lland_v3_nadelw_summer_hourly:

conifers (summer)
-----------------

Now we repeat the daily integration test :ref:`lland_v3_nadelw_summer_daily`
on an hourly basis. In contrast to the calculation for water areas (see
:ref:`lland_v3_water_hourly`), both the interception evaporation (|EvI|)
and the soil evapotranspiration (|EvB|) rely on hourly input data (see method
|Return_PenmanMonteith_V1|) and thus show a diurnal pattern.  |lland_v3|
always adds condensed water to the interception storage (|Inzp|).  Hence,
|EvI| is sometimes negative during the nighttime, but |Inzp| dries out as
soon as the sun starts to rise:

.. integration-test::

    >>> lnk(NADELW)
    >>> test('lland_v3_nadelw_summer_hourly',
    ...      axis1=(fluxes.evi, fluxes.evb),
    ...      axis2=states.inzp,
    ...      use_conditions=conditions_nadelw_summer)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration | nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |         g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes |      evpo |       evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatsnow | wsenssnow | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz | qah |  qa |     inzp | wats | waes | esnow | taus |      ebdn |      bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet | outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-08-03 00:00:00 |  0.0 | 16.9 |              0.0 |             95.1 |       0.8 |               101.5 | nan | nan | 17.279167 |             85.883333 |                   1.3 |  0.0 | 16.9 | 17.279167 |    0.641273 |         1.605886 |           0.283992 |          0.8 |                 1.928423 |                      1.975309 |                          0.0 |                      0.122362 |                           0.124959 |              1.83393 |                  1.696461 |       99.66607 |    1.21073 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.059316 | 15.208683 | -0.036532 |                   0.0 |                       0.0 |                   9.243306 |                  0.901196 |                      0.0 |     -0.03755 |              0.0 |          8.342111 |                 117.5 |                 100.0 |               170.875609 |               69.444444 |  0.0 |  0.0 | -0.012014 | -0.012014 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.012014 |  0.0 |  0.0 |   0.0 |  nan | 31.308721 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.221929 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 01:00:00 |  0.0 | 16.6 |              0.0 |             94.9 |       0.8 |               101.5 | nan | nan | 17.241667 |             85.991667 |                   1.3 |  0.0 | 16.6 | 17.241667 |    0.641273 |         1.603439 |           0.283992 |          0.8 |                 1.892018 |                      1.970628 |                          0.0 |                      0.120339 |                             0.1247 |             1.795525 |                  1.694576 |      99.704475 |   1.212158 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.059314 | 15.295735 | -0.028172 |                   0.0 |                       0.0 |                   9.243306 |                  0.901844 |                      0.0 |    -0.037577 |              0.0 |          8.341463 |                 117.5 |                 100.0 |               173.430933 |               69.444444 |  0.0 |  0.0 | -0.011569 | -0.011569 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.023583 |  0.0 |  0.0 |   0.0 |  nan | 31.326476 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.216196 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 02:00:00 |  0.0 | 16.4 |              0.0 |             95.9 |       0.8 |               101.5 | nan | nan |    17.175 |             86.233333 |                   1.3 |  0.0 | 16.4 |    17.175 |    0.641273 |         1.575992 |           0.283992 |          0.8 |                 1.868084 |                       1.96233 |                          0.0 |                      0.119006 |                           0.124241 |             1.791493 |                  1.692183 |      99.708507 |   1.213014 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.059312 |  15.35492 | -0.022574 |                   0.0 |                       0.0 |                   9.243306 |                  0.902355 |                      0.0 |    -0.037598 |              0.0 |          8.340951 |                 117.5 |                 100.0 |               173.240523 |               69.444444 |  0.0 |  0.0 | -0.013498 | -0.013498 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.037081 |  0.0 |  0.0 |   0.0 |  nan | 31.338633 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.210612 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 03:00:00 |  0.0 | 16.3 |              0.0 |             96.7 |       0.8 |               101.5 | nan | nan |   17.0625 |             86.708333 |                   1.3 |  0.0 | 16.3 |   17.0625 |    0.641273 |         1.548545 |           0.283992 |          0.8 |                 1.856216 |                      1.948396 |                          0.0 |                      0.118345 |                           0.123469 |             1.794961 |                  1.689422 |      99.705039 |   1.213417 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.059312 | 15.395444 | -0.019538 |                   0.0 |                       0.0 |                   9.243306 |                  0.902364 |                      0.0 |    -0.037598 |              0.0 |          8.340943 |                 117.5 |                 100.0 |               172.677019 |               69.444444 |  0.0 |  0.0 | -0.015036 | -0.015036 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.052118 |  0.0 |  0.0 |   0.0 |  nan | 31.347755 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.205172 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 04:00:00 |  0.0 | 16.0 |              0.0 |             97.2 |       0.8 |               101.5 | nan | nan | 16.908333 |             87.366667 |                   1.3 |  0.0 | 16.0 | 16.908333 |    0.641273 |         1.504432 |           0.283992 |          0.8 |                 1.821009 |                      1.929443 |                          0.0 |                      0.116379 |                           0.122418 |              1.77002 |                   1.68569 |       99.72998 |    1.21479 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                 0.429734 |                     15.566134 |            10.503757 |  0.086656 |        0.007799 |                0.006997 | -0.026916 | 15.425849 | -0.012402 |              0.006158 |                  0.001144 |                   9.243306 |                  0.902341 |                      0.0 |     -0.03144 |              0.0 |          8.340965 |                 117.5 |                 100.0 |                174.22122 |               93.330305 |  0.0 |  0.0 | -0.007429 | -0.007429 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.059547 |  0.0 |  0.0 |   0.0 |  nan |  31.34974 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.199872 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 05:00:00 |  0.0 | 15.9 |              0.0 |             97.5 |       0.6 |               101.5 | nan | nan | 16.729167 |             88.204167 |                   1.3 |  0.0 | 15.9 | 16.729167 |    0.480955 |         1.453638 |           0.212994 |          0.6 |                 1.809403 |                      1.907618 |                          0.0 |                       0.11573 |                           0.121206 |             1.764168 |                  1.682599 |      99.735832 |   1.215237 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  0.968392 |        0.087159 |                0.078957 |  0.016077 | 15.432466 | -0.010099 |              0.069482 |                  0.012904 |                   9.243306 |                  0.901482 |                      0.0 |      0.03192 |              0.0 |          8.341824 |            156.666667 |                 100.0 |               174.539602 |              172.004443 |  0.0 |  0.0 |   0.01385 |   0.01385 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.045698 |  0.0 |  0.0 |   0.0 |  nan | 31.349422 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.194709 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 06:00:00 |  0.0 | 16.0 |              0.0 |             97.7 |       0.9 |               101.5 | nan | nan | 16.533333 |             89.191667 |                   1.3 |  0.0 | 16.0 | 16.533333 |    0.721432 |         1.392031 |           0.319491 |          0.9 |                 1.821009 |                      1.884011 |                          0.0 |                      0.116379 |                           0.119893 |             1.779125 |                   1.68038 |      99.720875 |   1.214748 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |    2.4897 |        0.224084 |                0.206122 |   0.01608 | 15.431406 | -0.012282 |              0.181388 |                  0.033686 |                   9.243306 |                   0.89977 |                      0.0 |     0.143897 |              0.0 |          8.343537 |            104.444444 |                 100.0 |               173.411109 |              170.929929 |  0.0 |  0.0 |  0.040724 |  0.040724 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.004973 |  0.0 |  0.0 |   0.0 |  nan | 31.351287 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.189679 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 07:00:00 |  0.0 | 16.6 |              0.0 |             97.4 |       0.9 |               101.5 | nan | nan |    16.375 |                90.125 |                   1.3 |  0.0 | 16.6 |    16.375 |    0.721432 |         1.334591 |           0.319491 |          0.9 |                 1.892018 |                      1.865111 |                          0.0 |                      0.120339 |                           0.118841 |             1.842826 |                  1.680931 |      99.657174 |   1.211943 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  4.679733 |        0.421196 |                0.393598 |  0.016086 | 15.437623 | -0.025107 |              0.346366 |                  0.064325 |                   9.243306 |                  0.896826 |                      0.0 |     0.308999 |              0.0 |           8.34648 |            104.444444 |                 100.0 |               169.225822 |              166.940703 |  0.0 |  0.0 |  0.079445 |  0.004973 | 0.043681 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.365978 | 78.680999 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.18478 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 08:00:00 |  0.0 | 17.4 |              0.0 |             96.8 |       0.9 |               101.6 | nan | nan | 16.216667 |                91.275 |                   1.3 |  0.0 | 17.4 | 16.216667 |    0.721432 |         1.260484 |           0.319491 |          0.9 |                 1.990459 |                      1.846377 |                          0.0 |                      0.125796 |                           0.117796 |             1.926764 |                  1.685281 |      99.673236 |   1.209425 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  7.219274 |        0.649765 |                0.615417 |  0.016096 | 15.486592 |  -0.04133 |              0.541567 |                  0.100577 |                   9.243306 |                  0.891371 |                      0.0 |     0.504426 |              0.0 |          8.351935 |            104.444444 |                 100.0 |                 164.5016 |              162.429957 |  0.0 |  0.0 |  0.127432 |       0.0 | 0.076301 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  31.39689 | 78.604698 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.180007 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 09:00:00 |  0.0 | 19.0 |              0.2 |             86.1 |       1.3 |               101.6 | nan | nan |   16.1125 |             91.991667 |                   1.5 |  0.0 | 19.0 |   16.1125 |    1.042069 |         1.203904 |           0.461488 |          1.3 |                 2.200854 |                      1.834142 |                          0.0 |                      0.137341 |                           0.117113 |             1.894936 |                  1.687258 |      99.705064 |   1.202945 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             10.73504 |  9.442017 |         1.18975 |                1.122345 |  0.016428 | 15.589635 | -0.073664 |              0.987664 |                  0.183423 |                   9.446835 |                  0.922604 |                      0.0 |     0.949222 |              0.0 |          8.524231 |             72.307692 |                 100.0 |               175.669147 |              173.079486 |  0.0 |  0.0 |  0.284696 |       0.0 | 0.147871 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.460138 | 78.456827 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.175357 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 10:00:00 |  0.0 | 20.3 |              0.5 |             76.8 |       1.5 |               101.6 | nan | nan | 16.083333 |             92.241667 |                   1.7 |  0.0 | 20.3 | 16.083333 |    1.202387 |         1.149836 |           0.532486 |          1.5 |                  2.38585 |                      1.830729 |                          0.0 |                      0.147368 |                           0.116922 |             1.832333 |                  1.688695 |      99.767667 |   1.197896 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.966322 | 10.709164 |         1.92774 |                  1.8057 |  0.016758 | 15.800459 |  -0.09719 |              1.589016 |                  0.295103 |                   9.650363 |                  0.955299 |                      0.0 |     1.549212 |              0.0 |          8.695064 |             62.666667 |                 100.0 |               197.905084 |              194.146593 |  0.0 |  0.0 |  0.491165 |       0.0 | 0.231101 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.546911 | 78.225726 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.170827 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 11:00:00 |  0.0 | 21.4 |              0.7 |             71.8 |       1.2 |               101.6 | nan | nan |    16.125 |             92.104167 |                   2.2 |  0.0 | 21.4 |    16.125 |    0.961909 |         1.089916 |           0.425988 |          1.2 |                 2.552842 |                      1.835607 |                          0.0 |                      0.156328 |                           0.117195 |             1.832941 |                   1.69067 |      99.767059 |    1.19342 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            11.544527 | 10.773868 |        2.327265 |                2.214067 |  0.017575 | 16.089704 | -0.114702 |              1.948379 |                  0.361842 |                  10.159184 |                  1.039981 |                      0.0 |     1.905047 |              0.0 |          9.119203 |             78.333333 |                 100.0 |               224.327744 |              218.945268 |  0.0 |  0.0 |  0.586716 |       0.0 | 0.291264 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.651197 | 77.934463 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.166415 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 12:00:00 |  0.0 | 21.3 |              0.8 |             67.5 |       1.3 |               101.6 | nan | nan | 16.204167 |             91.729167 |                   2.8 |  0.0 | 21.3 | 16.204167 |    1.042069 |         1.037502 |           0.461488 |          1.3 |                 2.537251 |                      1.844905 |                          0.0 |                      0.155495 |                           0.117714 |             1.712644 |                  1.692316 |      99.887356 |   1.194363 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.238374 | 10.364714 |        2.425457 |                 2.25556 |  0.018554 | 16.437323 | -0.105034 |              1.984892 |                  0.368623 |                  10.769769 |                  1.142661 |                      0.0 |     1.937282 |              0.0 |          9.627108 |             72.307692 |                 100.0 |               245.272215 |              238.423486 |  0.0 |  0.0 |  0.618672 |       0.0 | 0.283176 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.745814 | 77.651286 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.162116 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 13:00:00 |  0.0 | 21.8 |              0.5 |             66.1 |       1.5 |               101.6 | nan | nan | 16.329167 |             91.104167 |                   3.2 |  0.0 | 21.8 | 16.329167 |    1.202387 |         1.012602 |           0.532486 |          1.5 |                 2.616045 |                      1.859671 |                          0.0 |                      0.159698 |                           0.118538 |             1.729206 |                  1.694238 |      99.870794 |   1.192265 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.700938 | 10.872489 |         1.95714 |                 1.78608 |  0.019205 | 16.752713 | -0.109021 |              1.571751 |                  0.291897 |                  11.176825 |                  1.211985 |                      0.0 |     1.521251 |              0.0 |           9.96484 |             62.666667 |                 100.0 |               259.972504 |              252.001253 |  0.0 |  0.0 |  0.540126 |       0.0 | 0.225916 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.844419 | 77.425371 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.157929 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 14:00:00 |  0.0 | 22.9 |              0.4 |             63.4 |       1.9 |               101.6 | nan | nan | 16.545833 |             90.058333 |                   3.4 |  0.0 | 22.9 | 16.545833 |    1.523023 |         0.984394 |           0.674482 |          1.9 |                 2.796942 |                       1.88551 |                          0.0 |                      0.169283 |                           0.119977 |             1.773261 |                  1.698059 |      99.826739 |   1.187639 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             12.93222 | 10.461143 |        1.694785 |                1.510276 |  0.019529 | 17.081396 | -0.125682 |              1.329043 |                  0.246822 |                  11.380354 |                   1.24753 |                      0.0 |     1.277062 |              0.0 |         10.132824 |             49.473684 |                 100.0 |               297.310123 |              286.146226 |  0.0 |  0.0 |  0.534524 |       0.0 | 0.186453 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.959684 | 77.238918 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.153849 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 15:00:00 |  0.0 | 22.7 |              0.5 |             62.4 |       1.9 |               101.6 | nan | nan | 16.816667 |             88.816667 |                   3.8 |  0.0 | 22.7 | 16.816667 |    1.523023 |         0.968687 |           0.674482 |          1.9 |                 2.763263 |                      1.918249 |                          0.0 |                      0.167505 |                           0.121797 |             1.724276 |                  1.703725 |      99.875724 |    1.18866 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.394785 |  8.865382 |        1.595844 |                1.396394 |  0.020179 | 17.465613 | -0.113063 |              1.228827 |                  0.228211 |                   11.78741 |                  1.316893 |                      0.0 |     1.173957 |              0.0 |         10.470517 |             49.473684 |                 100.0 |               302.800589 |              291.126335 |  0.0 |  0.0 |  0.509326 |       0.0 | 0.174894 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.06233 | 77.064024 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.149875 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 16:00:00 |  0.0 | 22.5 |              0.5 |             61.1 |       2.3 |               101.6 | nan | nan |   17.1375 |             87.333333 |                   4.1 |  0.0 | 22.5 |   17.1375 |     1.84366 |         0.991339 |           0.816478 |          2.3 |                 2.729939 |                      1.957676 |                          0.0 |                      0.165743 |                           0.123983 |             1.667993 |                  1.709704 |      99.932007 |   1.189715 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.741708 |  6.474205 |        1.165412 |                1.002589 |  0.020665 | 17.807767 | -0.101352 |              0.882278 |                  0.163852 |                  12.092703 |                  1.370145 |                      0.0 |     0.825189 |              0.0 |         10.722558 |             40.869565 |                 100.0 |               310.938004 |              298.488257 |  0.0 |  0.0 |  0.455459 |       0.0 | 0.136509 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.153266 | 76.927515 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.146004 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 17:00:00 |  0.0 | 21.9 |              0.3 |             62.1 |       2.4 |               101.6 | nan | nan |   17.4875 |                85.875 |                   4.4 |  0.0 | 21.9 |   17.4875 |    1.923819 |         1.017332 |           0.851977 |          2.4 |                 2.632058 |                      2.001493 |                          0.0 |                       0.16055 |                           0.126406 |             1.634508 |                  1.718782 |      99.965492 |   1.192283 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.088631 |  3.977658 |         0.57281 |                0.493699 |  0.021156 | 18.110885 | -0.081845 |              0.434455 |                  0.080685 |                  12.397995 |                  1.421011 |                      0.0 |     0.375246 |              0.0 |         10.976984 |             39.166667 |                 100.0 |               291.494538 |              280.859856 |  0.0 |  0.0 |  0.330117 |       0.0 |  0.09909 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.224694 | 76.828425 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.142232 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 18:00:00 |  0.0 | 21.4 |              0.1 |             67.0 |       2.5 |               101.6 | nan | nan |   17.8375 |             84.620833 |                   4.5 |  0.0 | 21.4 |   17.8375 |    2.003978 |         1.054998 |           0.887476 |          2.5 |                 2.552842 |                      2.046164 |                          0.0 |                      0.156328 |                           0.128868 |             1.710404 |                  1.731481 |      99.889596 |   1.193968 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  1.972911 |        0.213084 |                0.183891 |  0.021327 | 18.348979 | -0.065902 |              0.161824 |                  0.030053 |                  12.499759 |                  1.433696 |                      0.0 |     0.102086 |              0.0 |         11.066063 |                  37.6 |                 100.0 |               252.768497 |              245.356921 |  0.0 |  0.0 |  0.231297 |       0.0 | 0.073434 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.280179 | 76.754991 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.138558 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 19:00:00 |  0.0 | 20.7 |              0.0 |             74.5 |       2.5 |               101.6 | nan | nan | 18.170833 |                  83.7 |                   4.5 |  0.0 | 20.7 | 18.170833 |    2.003978 |         1.100997 |           0.887476 |          2.5 |                 2.445437 |                      2.089517 |                          0.0 |                      0.150575 |                           0.131251 |              1.82185 |                  1.748925 |       99.77815 |   1.196313 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  0.634046 |        0.057067 |                0.049078 |  0.021346 | 18.533931 | -0.046787 |              0.043189 |                  0.008021 |                  12.499759 |                  1.423756 |                      0.0 |    -0.016134 |              0.0 |         11.076003 |                  37.6 |                 100.0 |               212.786167 |              208.144236 |  0.0 |  0.0 |  0.151599 |       0.0 | 0.052856 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.31655 | 76.702135 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.134979 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 20:00:00 |  0.0 | 19.4 |              0.0 |             81.2 |       2.2 |               101.6 | nan | nan | 18.454167 |             83.066667 |                   4.5 |  0.0 | 19.4 | 18.454167 |    1.763501 |         1.120309 |           0.780979 |          2.2 |                 2.256393 |                      2.126995 |                          0.0 |                      0.140363 |                           0.133306 |             1.832191 |                  1.766824 |      99.767809 |   1.201582 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                   0.1364 |                     15.566134 |            14.204272 |  0.008648 |        0.000778 |                0.000669 | -0.056745 | 18.655165 | -0.016088 |              0.000589 |                  0.000109 |                  12.499759 |                  1.412078 |                      0.0 |    -0.058248 |              0.0 |         11.087681 |             42.727273 |                 100.0 |               190.567035 |               75.962362 |  0.0 |  0.0 |  0.065573 |       0.0 | 0.040104 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.322221 | 76.662031 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.131493 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 21:00:00 |  0.0 | 17.8 |              0.0 |             86.9 |       1.7 |               101.6 | nan | nan |     18.65 |               82.7375 |                   4.5 |  0.0 | 17.8 |     18.65 |    1.362705 |         1.127089 |           0.603484 |          1.7 |                 2.041337 |                      2.153241 |                          0.0 |                      0.128603 |                           0.134742 |             1.773922 |                  1.781538 |      99.826078 |   1.208454 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 |  -0.06912 | 18.674071 |   0.01888 |                   0.0 |                       0.0 |                  12.499759 |                  1.401564 |                      0.0 |    -0.058399 |              0.0 |         11.098195 |             55.294118 |                 100.0 |               185.584095 |               69.444444 |  0.0 |  0.0 |  0.019064 |       0.0 | 0.012824 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.292925 | 76.649207 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.128096 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 22:00:00 |  0.0 | 17.0 |              0.0 |             90.1 |       1.7 |               101.7 | nan | nan | 18.808333 |             82.554167 |                   4.5 |  0.0 | 17.0 | 18.808333 |    1.362705 |         1.142201 |           0.603484 |          1.7 |                 1.940693 |                      2.174668 |                          0.0 |                      0.123042 |                           0.135912 |             1.748564 |                  1.795279 |      99.951436 |   1.213101 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.069157 | 18.576416 |  0.034051 |                   0.0 |                       0.0 |                  12.499759 |                  1.391158 |                      0.0 |    -0.057965 |              0.0 |         11.108602 |             55.294118 |                 100.0 |               184.354631 |               69.444444 |  0.0 |  0.0 |  0.006463 |       0.0 | 0.004311 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.248458 | 76.644895 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.124787 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-03 23:00:00 |  0.0 | 16.4 |              0.0 |             90.9 |       2.3 |               101.7 | nan | nan | 18.941667 |             82.379167 |                   4.5 |  0.0 | 16.4 | 18.941667 |     1.84366 |         1.185687 |           0.816478 |          2.3 |                 1.868084 |                      2.192855 |                          0.0 |                      0.119006 |                           0.136905 |             1.698088 |                  1.806456 |     100.001912 |   1.215845 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.069186 | 18.428192 |  0.043809 |                   0.0 |                       0.0 |                  12.499759 |                   1.38284 |                      0.0 |    -0.057618 |              0.0 |         11.116919 |             40.869565 |                 100.0 |               187.242012 |               69.444444 |  0.0 |  0.0 |  0.013279 |       0.0 | 0.007964 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.194232 | 76.636931 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.121564 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 00:00:00 |  0.0 | 15.9 |              0.0 |             88.0 |       2.3 |               101.7 | nan | nan |      18.9 |             82.083333 |                   4.5 |  0.0 | 15.9 |      18.9 |     1.84366 |         1.235786 |           0.816478 |          2.3 |                 1.809403 |                      2.187158 |                          0.0 |                       0.11573 |                           0.136594 |             1.592275 |                  1.795292 |     100.107725 |    1.21843 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.068395 |  18.24744 |  0.050705 |                   0.0 |                       0.0 |                  12.435463 |                   1.39654 |                      0.0 |    -0.058189 |              0.0 |         11.038924 |             40.869565 |                 100.0 |                196.81078 |               69.444444 |  0.0 |  0.0 |  0.027068 |       0.0 | 0.016129 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.133111 | 76.620802 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.118424 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 01:00:00 |  0.0 | 15.4 |              0.0 |             89.2 |       2.2 |               101.6 | nan | nan |     18.85 |             81.845833 |                   4.5 |  0.0 | 15.4 |     18.85 |    1.763501 |         1.282546 |           0.780979 |          2.2 |                 1.752342 |                      2.180337 |                          0.0 |                      0.112529 |                           0.136222 |             1.563089 |                  1.784515 |     100.036911 |   1.219467 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.068359 | 18.043702 |  0.057104 |                   0.0 |                       0.0 |                  12.435463 |                  1.406633 |                      0.0 |     -0.05861 |              0.0 |          11.02883 |             42.727273 |                 100.0 |               198.573145 |               69.444444 |  0.0 |  0.0 |  0.018784 |       0.0 | 0.011303 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.06559 | 76.609499 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.115365 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 02:00:00 |  0.0 | 15.2 |              0.0 |             88.8 |       2.3 |               101.6 | nan | nan |      18.8 |                 81.55 |                   4.5 |  0.0 | 15.2 |      18.8 |     1.84366 |         1.332645 |           0.816478 |          2.3 |                 1.729962 |                      2.173535 |                          0.0 |                      0.111271 |                            0.13585 |             1.536206 |                  1.772518 |     100.063794 |   1.220436 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.068319 | 17.818633 |  0.056562 |                   0.0 |                       0.0 |                  12.435463 |                  1.418069 |                      0.0 |    -0.059086 |              0.0 |         11.017395 |             40.869565 |                 100.0 |                201.06939 |               69.444444 |  0.0 |  0.0 |  0.022374 |       0.0 |  0.01321 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.998611 | 76.596289 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.112385 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 03:00:00 |  0.0 | 14.9 |              0.0 |             89.7 |       2.0 |               101.6 | nan | nan | 18.741667 |             81.258333 |                   4.5 |  0.0 | 14.9 | 18.741667 |    1.603182 |         1.372725 |           0.709981 |          2.0 |                 1.696861 |                      2.165623 |                          0.0 |                      0.109405 |                           0.135418 |             1.522084 |                  1.759749 |     100.077916 |   1.221771 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.068276 | 17.595369 |   0.05822 |                   0.0 |                       0.0 |                  12.435463 |                   1.43017 |                      0.0 |     -0.05959 |              0.0 |         11.005293 |                  47.0 |                 100.0 |                202.01382 |               69.444444 |  0.0 |  0.0 |  0.012301 |       0.0 | 0.007608 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.929974 | 76.588682 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.109482 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 04:00:00 |  0.0 | 14.3 |              0.0 |             93.0 |       2.3 |               101.6 | nan | nan | 18.670833 |             81.083333 |                   4.5 |  0.0 | 14.3 | 18.670833 |     1.84366 |         1.422824 |           0.816478 |          2.3 |                 1.632319 |                       2.15605 |                          0.0 |                      0.105752 |                           0.134895 |             1.518057 |                  1.748197 |     100.081943 |    1.22434 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.398044 |                     15.502754 |            14.131208 |   0.07485 |        0.006692 |                0.005726 | -0.032608 |  17.36658 |  0.066238 |              0.005039 |                  0.000936 |                  12.435463 |                  1.440702 |                      0.0 |    -0.054991 |              0.0 |         10.994761 |             40.869565 |                 100.0 |               201.630185 |               93.615707 |  0.0 |  0.0 |  0.011636 |       0.0 | 0.005929 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.853319 | 76.582753 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.106654 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 05:00:00 |  0.0 | 14.1 |              0.0 |             93.6 |       2.5 |               101.7 | nan | nan | 18.595833 |             80.920833 |                   4.5 |  0.0 | 14.1 | 18.595833 |    2.003978 |         1.486284 |           0.887476 |          2.5 |                 1.611289 |                      2.145953 |                          0.0 |                      0.104558 |                           0.134343 |             1.508166 |                  1.736523 |     100.191834 |   1.226451 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.131208 |  0.938714 |        0.083928 |                0.071823 |  0.021256 | 17.111064 |  0.065039 |              0.063204 |                  0.011738 |                  12.435463 |                  1.451305 |                      0.0 |     0.002733 |              0.0 |         10.984159 |                  37.6 |                 100.0 |               202.560953 |              198.534737 |  0.0 |  0.0 |  0.037339 |       0.0 | 0.011605 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.777864 | 76.571147 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.103899 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 06:00:00 |  0.0 | 14.3 |              0.1 |             94.6 |       2.4 |               101.7 | nan | nan |    18.525 |             80.791667 |                   4.6 |  0.0 | 14.3 |    18.525 |    1.923819 |         1.536383 |           0.851977 |          2.4 |                 1.632319 |                      2.136456 |                          0.0 |                      0.105752 |                           0.133824 |             1.544174 |                  1.726078 |     100.155826 |   1.225432 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.246552 |  2.455474 |        0.263445 |                0.226749 |    0.0214 | 16.859546 |  0.055286 |              0.199539 |                  0.037057 |                  12.536966 |                  1.478175 |                      0.0 |     0.137948 |              0.0 |         11.058791 |             39.166667 |                 100.0 |               198.922196 |              195.105895 |  0.0 |  0.0 |   0.06352 |       0.0 | 0.020586 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.712161 | 76.550561 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.101215 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 07:00:00 |  0.0 | 15.4 |              0.9 |             93.3 |       2.5 |               101.7 | nan | nan |    18.475 |             80.620833 |                   5.5 |  0.0 | 15.4 |    18.475 |    2.003978 |         1.589823 |           0.887476 |          2.5 |                 1.752342 |                      2.129774 |                          0.0 |                      0.112529 |                           0.133458 |             1.634935 |                  1.717041 |     100.065065 |   1.220347 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            15.284647 |  4.650125 |        1.164115 |                1.028795 |  0.022846 | 16.640536 |  0.026796 |               0.90534 |                  0.168134 |                  13.450489 |                   1.64464 |                      0.0 |     0.836813 |              0.0 |         11.805849 |                  37.6 |                 100.0 |               191.173528 |              187.788087 |  0.0 |  0.0 |  0.242925 |       0.0 | 0.080662 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.674949 | 76.469899 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.098601 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 08:00:00 |  0.0 | 16.8 |              1.0 |             86.9 |       2.5 |               101.7 | nan | nan |     18.45 |             80.208333 |                   6.5 |  0.0 | 16.8 |     18.45 |    2.003978 |         1.643262 |           0.887476 |          2.5 |                  1.91622 |                       2.12644 |                          0.0 |                      0.121684 |                           0.133275 |             1.665196 |                  1.705582 |     100.034804 |   1.214317 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            16.438085 |  7.209034 |        1.933622 |                1.710811 |  0.024442 | 16.516495 | -0.006124 |              1.505514 |                  0.279595 |                  14.465515 |                   1.83472 |                      0.0 |     1.429067 |              0.0 |         12.630795 |                  37.6 |                 100.0 |               192.770285 |              189.297859 |  0.0 |  0.0 |  0.433136 |       0.0 | 0.147472 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.670656 | 76.322428 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.096054 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 09:00:00 |  0.0 | 18.2 |              1.0 |             78.9 |       2.9 |               101.7 | nan | nan | 18.416667 |             79.908333 |                   7.3 |  0.0 | 18.2 | 18.416667 |    2.324615 |         1.696701 |           1.029472 |          2.9 |                 2.093348 |                      2.122001 |                          0.0 |                      0.131461 |                           0.133032 |             1.651651 |                  1.695656 |     100.048349 |   1.208543 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.360836 |  9.458797 |        2.537058 |                2.210424 |  0.025714 | 16.502185 | -0.036673 |              1.945173 |                  0.361246 |                  15.277535 |                  1.989515 |                      0.0 |     1.862277 |              0.0 |          13.28802 |             32.413793 |                 100.0 |               203.699986 |              199.607068 |  0.0 |  0.0 |  0.622629 |       0.0 | 0.192557 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.696912 | 76.129871 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.093572 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 10:00:00 |  0.0 | 19.5 |              0.9 |             75.8 |       3.1 |               101.7 | nan | nan | 18.383333 |             79.866667 |                   7.7 |  0.0 | 19.5 | 18.383333 |    2.484933 |         1.750141 |            1.10047 |          3.1 |                 2.270468 |                      2.117571 |                          0.0 |                      0.141127 |                            0.13279 |             1.721014 |                  1.691233 |      99.978986 |   1.202862 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.822211 | 10.747438 |         2.69052 |                2.317703 |  0.026351 | 16.589706 | -0.062862 |              2.039579 |                  0.378779 |                  15.683546 |                   2.06621 |                      0.0 |     1.953487 |              0.0 |         13.617336 |             30.322581 |                 100.0 |               207.750743 |              203.416777 |  0.0 |  0.0 |  0.695283 |       0.0 | 0.209727 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.749357 | 75.920144 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.091155 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 11:00:00 |  0.0 | 20.3 |              0.8 |             73.4 |       3.3 |               101.8 | nan | nan |   18.3375 |             79.933333 |                   7.8 |  0.0 | 20.3 |   18.3375 |    2.645251 |          1.82028 |           1.171468 |          3.3 |                  2.38585 |                      2.111492 |                          0.0 |                      0.147368 |                           0.132457 |             1.751214 |                  1.687786 |     100.048786 |   1.200635 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.937555 | 10.819272 |        2.515039 |                2.160947 |  0.026506 | 16.764525 | -0.076366 |              1.901633 |                   0.35316 |                  15.785048 |                  2.087634 |                      0.0 |     1.814648 |              0.0 |         13.697414 |             28.484848 |                 100.0 |                216.89937 |              211.999049 |  0.0 |  0.0 |  0.698385 |       0.0 | 0.200334 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.815307 |  75.71981 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.088801 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 12:00:00 |  0.0 | 21.0 |              0.9 |             68.3 |       3.3 |               101.8 | nan | nan |    18.325 |             79.966667 |                   7.9 |  0.0 | 21.0 |    18.325 |    2.645251 |         1.887079 |           1.171468 |          3.3 |                 2.490975 |                      2.109837 |                          0.0 |                      0.153018 |                           0.132366 |             1.701336 |                  1.687166 |     100.098664 |   1.198001 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.052899 | 10.408076 |        2.605564 |                2.233851 |  0.026667 | 16.984357 | -0.086738 |              1.965789 |                  0.365075 |                  15.886551 |                  2.106001 |                      0.0 |     1.878039 |              0.0 |          13.78055 |             28.484848 |                 100.0 |               245.325144 |               238.47251 |  0.0 |  0.0 |  0.764212 |       0.0 | 0.204943 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.891628 | 75.514867 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.086507 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 13:00:00 |  0.0 | 21.8 |              0.8 |             61.3 |       3.2 |               101.8 | nan | nan |    18.325 |             79.766667 |                   8.2 |  0.0 | 21.8 |    18.325 |    2.565092 |         1.943859 |           1.135969 |          3.2 |                 2.616045 |                      2.109837 |                          0.0 |                      0.159698 |                           0.132366 |             1.603636 |                  1.682947 |     100.196364 |   1.195188 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             18.39893 | 10.917404 |         2.53785 |                2.157989 |   0.02714 | 17.238761 | -0.098523 |               1.89903 |                  0.352677 |                  16.191059 |                  2.166369 |                      0.0 |     1.808765 |              0.0 |          14.02469 |                29.375 |                 100.0 |               301.477482 |              289.927166 |  0.0 |  0.0 |  0.802266 |       0.0 | 0.193994 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.979734 | 75.320873 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.084272 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 14:00:00 |  0.0 | 21.9 |              0.9 |             60.6 |       2.5 |               101.8 | nan | nan | 18.283333 |                 79.65 |                   8.7 |  0.0 | 21.9 | 18.283333 |    2.003978 |         1.963899 |           0.887476 |          2.5 |                 2.632058 |                      2.104328 |                          0.0 |                       0.16055 |                           0.132064 |             1.595027 |                  1.676097 |     100.204973 |   1.194821 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.975649 | 10.494383 |         2.62717 |                2.208784 |   0.02793 | 17.532448 | -0.094339 |               1.94373 |                  0.360978 |                  16.698571 |                  2.265756 |                      0.0 |     1.849324 |              0.0 |         14.432816 |                  37.6 |                 100.0 |               309.977063 |              297.620077 |  0.0 |  0.0 |  0.754875 |       0.0 | 0.211903 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.063657 |  75.10897 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.082096 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 15:00:00 |  0.0 | 21.8 |              0.9 |             58.8 |       2.9 |               101.7 | nan | nan | 18.245833 |                  79.5 |                   9.1 |  0.0 | 21.8 | 18.245833 |    2.324615 |         1.997298 |           1.029472 |          2.9 |                 2.616045 |                      2.099381 |                          0.0 |                      0.159698 |                           0.131792 |             1.538235 |                  1.669008 |     100.161765 |   1.194298 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            19.437025 |  8.874215 |        2.221576 |                1.861585 |  0.028555 | 17.812189 | -0.086137 |              1.638195 |                  0.304236 |                  17.104582 |                  2.348437 |                      0.0 |     1.540344 |              0.0 |         14.756144 |             32.413793 |                 100.0 |               324.588116 |              310.786578 |  0.0 |  0.0 |  0.718415 |       0.0 |  0.17644 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.139377 |  74.93253 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.079975 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 16:00:00 |  0.0 | 22.2 |              0.9 |             57.2 |       3.6 |               101.7 | nan | nan | 18.233333 |               79.3375 |                   9.5 |  0.0 | 22.2 | 18.233333 |    2.885728 |         2.040718 |           1.277965 |          3.6 |                 2.680609 |                      2.097734 |                          0.0 |                      0.163129 |                           0.131702 |             1.533308 |                   1.66429 |     100.166692 |   1.192703 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |              19.8984 |   6.45697 |        1.616442 |                1.360212 |  0.029186 |  18.06459 | -0.089325 |              1.196986 |                  0.222297 |                  17.510592 |                  2.428674 |                      0.0 |     1.095792 |              0.0 |         15.081918 |             26.111111 |                 100.0 |               351.854614 |              335.163509 |  0.0 |  0.0 |  0.679939 |       0.0 | 0.136423 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.218285 | 74.796107 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.077909 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 17:00:00 |  0.0 | 21.8 |              1.0 |             57.2 |       4.2 |               101.7 | nan | nan | 18.229167 |             79.133333 |                  10.2 |  0.0 | 21.8 | 18.229167 |    3.366683 |         2.100837 |            1.49096 |          4.2 |                 2.616045 |                      2.097185 |                          0.0 |                      0.159698 |                           0.131672 |             1.496378 |                  1.659573 |     100.203622 |   1.194485 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            20.705807 |  3.945159 |        1.058179 |                0.907937 |  0.030298 | 18.327617 | -0.075003 |              0.798984 |                  0.148383 |                   18.22111 |                  2.564427 |                      0.0 |     0.692133 |              0.0 |         15.656683 |             22.380952 |                 100.0 |                 341.4206 |              325.864873 |  0.0 |  0.0 |   0.61955 |       0.0 | 0.111926 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.282872 | 74.684181 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.075897 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 18:00:00 |  0.0 | 20.9 |              1.0 |             59.9 |       4.2 |               101.7 | nan | nan | 18.208333 |               78.8375 |                  11.1 |  0.0 | 20.9 | 18.208333 |    3.366683 |         2.157616 |            1.49096 |          4.2 |                 2.475714 |                      2.094444 |                          0.0 |                        0.1522 |                           0.131522 |             1.482953 |                  1.651207 |     100.217047 |   1.198201 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            21.743901 |  1.939642 |        0.520255 |                0.462876 |  0.031718 | 18.542906 | -0.050913 |              0.407331 |                  0.075647 |                  19.134633 |                  2.744116 |                      0.0 |     0.292993 |              0.0 |         16.390518 |             22.380952 |                 100.0 |               298.841688 |              287.536473 |  0.0 |  0.0 |  0.472941 |       0.0 | 0.092148 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.323368 | 74.592033 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.073936 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 19:00:00 |  0.0 | 19.7 |              1.0 |             66.1 |       3.6 |               101.7 | nan | nan | 18.166667 |               78.4875 |                  12.1 |  0.0 | 19.7 | 18.166667 |    2.885728 |         2.194356 |           1.277965 |          3.6 |                 2.298847 |                       2.08897 |                          0.0 |                      0.142666 |                           0.131221 |             1.519538 |                   1.63958 |     100.180462 |   1.202947 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             22.89734 |   0.60531 |        0.162358 |                0.151461 |  0.033284 | 18.677895 | -0.022077 |              0.133286 |                  0.024753 |                  20.149659 |                  2.950104 |                      0.0 |     0.010365 |              0.0 |         17.199555 |             26.111111 |                 100.0 |               249.881996 |              242.689512 |  0.0 |  0.0 |   0.28471 |       0.0 | 0.068597 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.335029 | 74.523436 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.072026 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 20:00:00 |  0.0 | 17.9 |              0.3 |             71.7 |       3.2 |               101.7 | nan | nan | 18.104167 |             78.091667 |                  12.4 |  0.0 | 17.9 | 18.104167 |    2.565092 |         2.227756 |           1.135969 |          3.2 |                 2.054233 |                      2.080783 |                          0.0 |                      0.129312 |                           0.130772 |             1.472885 |                  1.624918 |     100.227115 |   1.210597 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.104711 |                     15.502754 |            23.243371 |  0.005136 |        0.003091 |                0.002926 | -0.077873 | 18.716764 |  0.017642 |              0.002575 |                  0.000478 |                  20.454167 |                  3.034376 |                      0.0 |    -0.123857 |              0.0 |         17.419791 |                29.375 |                 100.0 |               232.121911 |               74.878015 |  0.0 |  0.0 |  0.143314 |       0.0 | 0.074372 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  32.30697 | 74.449064 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.070166 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 21:00:00 |  0.0 | 15.8 |              0.0 |             80.6 |       2.2 |               101.8 | nan | nan | 18.020833 |             77.829167 |                  12.4 |  0.0 | 15.8 | 18.020833 |    1.763501 |         2.244455 |           0.780979 |          2.2 |                 1.797863 |                       2.06991 |                          0.0 |                      0.115083 |                           0.130174 |             1.449077 |                  1.610994 |     100.350923 |    1.22071 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 |  -0.09083 | 18.623235 |  0.060982 |                   0.0 |                       0.0 |                  20.454167 |                  3.060678 |                      0.0 |    -0.127528 |              0.0 |         17.393489 |             42.727273 |                 100.0 |                219.77746 |               69.444444 |  0.0 |  0.0 |  0.036824 |       0.0 | 0.022273 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.235572 | 74.426792 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.068353 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 22:00:00 |  0.0 | 14.8 |              0.0 |             85.9 |       2.2 |               101.8 | nan | nan | 17.929167 |             77.654167 |                  12.4 |  0.0 | 14.8 | 17.929167 |    1.763501 |         2.261155 |           0.780979 |          2.2 |                 1.685952 |                      2.058007 |                          0.0 |                      0.108789 |                            0.12952 |             1.448232 |                  1.598128 |     100.351768 |   1.224953 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.090746 |  18.38524 |  0.077441 |                   0.0 |                       0.0 |                  20.454167 |                  3.084324 |                      0.0 |    -0.128514 |              0.0 |         17.369842 |             42.727273 |                 100.0 |                 216.3895 |               69.444444 |  0.0 |  0.0 |  0.011369 |       0.0 | 0.006789 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.147714 | 74.420002 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.066588 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-04 23:00:00 |  0.0 | 14.1 |              0.0 |             89.3 |       2.1 |               101.8 | nan | nan | 17.833333 |               77.5875 |                  12.4 |  0.0 | 14.1 | 17.833333 |    1.683342 |         2.254475 |            0.74548 |          2.1 |                 1.611289 |                      2.045628 |                          0.0 |                      0.104558 |                           0.128839 |             1.438881 |                  1.587151 |     100.361119 |   1.227981 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.090678 |  18.09238 |  0.086235 |                   0.0 |                       0.0 |                  20.454167 |                  3.103639 |                      0.0 |    -0.129318 |              0.0 |         17.350527 |             44.761905 |                 100.0 |               216.322849 |               69.444444 |  0.0 |  0.0 | -0.006064 | -0.006064 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.006064 |  0.0 |  0.0 |   0.0 |  nan | 32.051062 | 74.420002 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.064868 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 00:00:00 |  0.0 | 13.4 |              0.0 |             91.4 |       1.8 |               101.8 | nan | nan | 17.729167 |             77.729167 |                  12.4 |  0.0 | 13.4 | 17.729167 |    1.442864 |         2.237776 |           0.638983 |          1.8 |                 1.539538 |                      2.032245 |                          0.0 |                      0.100468 |                           0.128102 |             1.407137 |                  1.579647 |     100.392863 |   1.231127 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.089622 | 17.770207 |  0.094396 |                   0.0 |                       0.0 |                   20.36723 |                  3.124611 |                      0.0 |    -0.130192 |              0.0 |         17.242619 |             52.222222 |                 100.0 |               219.742169 |               69.444444 |  0.0 |  0.0 | -0.018897 | -0.018897 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.024961 |  0.0 |  0.0 |   0.0 |  nan | 31.946249 | 74.420002 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.063192 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 01:00:00 |  0.0 | 12.8 |              0.0 |             93.1 |       1.8 |               101.8 | nan | nan | 17.620833 |             77.891667 |                  12.4 |  0.0 | 12.8 | 17.620833 |    1.442864 |         2.224416 |           0.638983 |          1.8 |                 1.480281 |                      2.018409 |                          0.0 |                      0.097071 |                           0.127339 |             1.378142 |                  1.572172 |     100.421858 |   1.233843 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.089584 |  17.42083 |   0.09981 |                   0.0 |                       0.0 |                   20.36723 |                  3.135507 |                      0.0 |    -0.130646 |              0.0 |         17.231723 |             52.222222 |                 100.0 |                223.52738 |               69.444444 |  0.0 |  0.0 | -0.024841 | -0.024841 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.049802 |  0.0 |  0.0 |   0.0 |  nan | 31.836022 | 74.420002 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.06156 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 02:00:00 |  0.0 | 12.3 |              0.0 |             94.6 |       1.6 |               101.8 | nan | nan |      17.5 |             78.133333 |                  12.4 |  0.0 | 12.3 |      17.5 |    1.282546 |         2.201036 |           0.567985 |          1.6 |                 1.432438 |                      2.003073 |                          0.0 |                      0.094315 |                           0.126493 |             1.355086 |                  1.565068 |     100.444914 |   1.236111 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.089551 | 17.053408 |  0.102674 |                   0.0 |                       0.0 |                   20.36723 |                  3.144898 |                      0.0 |    -0.131037 |              0.0 |         17.222331 |                 58.75 |                 100.0 |               227.022589 |               69.444444 |  0.0 |  0.0 | -0.031316 | -0.031316 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.081119 |  0.0 |  0.0 |   0.0 |  nan | 31.722932 | 74.420002 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.05997 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 03:00:00 |  0.0 | 12.0 |              0.0 |             95.6 |       1.2 |               101.8 | nan | nan | 17.379167 |             78.379167 |                  12.4 |  0.0 | 12.0 | 17.379167 |    0.961909 |         2.174316 |           0.425988 |          1.2 |                 1.404387 |                       1.98784 |                          0.0 |                      0.092693 |                           0.125652 |             1.342594 |                  1.558052 |     100.457406 |   1.237469 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.089518 |  16.67644 |  0.101011 |                   0.0 |                       0.0 |                   20.36723 |                   3.15411 |                      0.0 |    -0.131421 |              0.0 |          17.21312 |             78.333333 |                 100.0 |               229.174328 |               69.444444 |  0.0 |  0.0 | -0.035969 | -0.035969 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.117088 |  0.0 |  0.0 |   0.0 |  nan | 31.611504 | 74.420002 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.058421 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 04:00:00 |  0.0 | 11.7 |              0.0 |             96.2 |       1.9 |               101.8 | nan | nan | 17.270833 |               78.5125 |                  12.4 |  0.0 | 11.7 | 17.270833 |    1.523023 |         2.160956 |           0.674482 |          1.9 |                 1.376819 |                      1.974268 |                          0.0 |                      0.091096 |                           0.124901 |               1.3245 |                  1.550047 |       100.4755 |   1.238856 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.366022 |                     15.438712 |            23.144579 |  0.063733 |         0.00566 |                0.005336 | -0.044492 | 16.305014 |  0.099468 |              0.004696 |                  0.000872 |                   20.36723 |                  3.166293 |                      0.0 |    -0.127233 |              0.0 |         17.200937 |             49.473684 |                 100.0 |               232.156458 |               93.048394 |  0.0 |  0.0 | -0.023868 | -0.023868 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.140956 |  0.0 |  0.0 |   0.0 |  nan | 31.501619 | 74.420002 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.056912 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 05:00:00 |  0.0 | 11.6 |              0.0 |             95.7 |       1.3 |               101.8 | nan | nan | 17.166667 |                  78.6 |                  12.4 |  0.0 | 11.6 | 17.166667 |    1.042069 |         2.120877 |           0.461488 |          1.3 |                 1.367736 |                      1.961295 |                          0.0 |                      0.090568 |                           0.124183 |             1.308923 |                  1.541578 |     100.491077 |   1.239363 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.144579 |  0.908598 |        0.080684 |                0.076088 |  0.033398 | 15.938731 |  0.093717 |              0.066957 |                  0.012435 |                   20.36723 |                  3.179808 |                      0.0 |    -0.065535 |              0.0 |         17.187422 |             72.307692 |                 100.0 |               234.336263 |               228.27273 |  0.0 |  0.0 |  0.004154 |  0.004154 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.136802 |  0.0 |  0.0 |   0.0 |  nan | 31.397486 | 74.420002 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.055442 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 06:00:00 |  0.0 | 12.6 |              0.6 |             95.1 |       1.6 |               101.8 | nan | nan | 17.095833 |             78.620833 |                  12.9 |  0.0 | 12.6 | 17.095833 |    1.282546 |         2.094157 |           0.567985 |          1.6 |                 1.460978 |                      1.952516 |                          0.0 |                       0.09596 |                           0.123697 |              1.38939 |                  1.535084 |      100.41061 |   1.234655 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  2.420659 |        0.472904 |                 0.45318 |   0.03417 |  15.59162 |  0.064619 |              0.398799 |                  0.074063 |                   20.87339 |                  3.288825 |                      0.0 |     0.261764 |              0.0 |         17.584565 |                 58.75 |                 100.0 |                222.48592 |               217.22485 |  0.0 |  0.0 |  0.080594 |  0.080594 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan | 0.056208 |  0.0 |  0.0 |   0.0 |  nan |  31.32245 | 74.420002 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.054009 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 07:00:00 |  0.0 | 14.8 |              0.9 |             96.3 |       1.9 |               101.9 | nan | nan | 17.070833 |             78.745833 |                  12.9 |  0.0 | 14.8 | 17.070833 |    1.523023 |         2.074117 |           0.674482 |          1.9 |                 1.685952 |                      1.949426 |                          0.0 |                      0.108789 |                           0.123526 |             1.623571 |                  1.535091 |     100.276429 |   1.225361 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  4.619722 |        1.148656 |                1.101438 |  0.034173 | 15.341501 |  0.011696 |              0.969265 |                  0.180006 |                   20.87339 |                  3.287363 |                      0.0 |     0.832292 |              0.0 |         17.586027 |             49.473684 |                 100.0 |               196.825466 |              193.127924 |  0.0 |  0.0 |  0.218392 |  0.056208 | 0.061372 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.300337 |  74.35863 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.052614 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 08:00:00 |  0.0 | 17.1 |              1.0 |             89.9 |       3.2 |               101.9 | nan | nan | 17.083333 |             78.870833 |                  12.9 |  0.0 | 17.1 | 17.083333 |    2.565092 |         2.097497 |           1.135969 |          3.2 |                 1.953031 |                       1.95097 |                          0.0 |                      0.123726 |                           0.123612 |             1.755775 |                  1.538746 |     100.144225 |   1.215051 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  7.198198 |        1.917614 |                1.839976 |  0.034188 | 15.267791 | -0.039576 |              1.619179 |                  0.300705 |                   20.87339 |                  3.279458 |                      0.0 |     1.482535 |              0.0 |         17.593932 |                29.375 |                 100.0 |               189.686799 |              186.381507 |  0.0 |  0.0 |  0.454201 |       0.0 | 0.134753 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.329496 | 74.223877 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.051255 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 09:00:00 |  0.0 | 19.4 |              0.9 |             79.2 |       4.0 |               101.9 | nan | nan | 17.133333 |             78.883333 |                  12.8 |  0.0 | 19.4 | 17.133333 |    3.206365 |         2.134237 |           1.419962 |          4.0 |                 2.256393 |                      1.957159 |                          0.0 |                      0.140363 |                           0.123954 |             1.787063 |                  1.543872 |     100.112937 |   1.205358 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.604725 |  9.475573 |        2.356024 |                2.266269 |  0.034047 | 15.364988 | -0.087156 |              1.994317 |                  0.370373 |                  20.772158 |                  3.250753 |                      0.0 |     1.858869 |              0.0 |         17.521405 |                  23.5 |                 100.0 |               202.701165 |              198.666764 |  0.0 |  0.0 |  0.690928 |       0.0 | 0.178278 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.406236 | 74.045599 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.049931 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 10:00:00 |  0.0 | 20.6 |              1.0 |             73.4 |       4.6 |               101.9 | nan | nan | 17.179167 |             78.783333 |                  12.9 |  0.0 | 20.6 | 17.179167 |     3.68732 |         2.184336 |           1.632956 |          4.6 |                  2.43042 |                      1.962848 |                          0.0 |                      0.149767 |                           0.124269 |             1.783928 |                  1.546397 |     100.116072 |   1.200448 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 | 10.786309 |        2.873495 |                 2.75694 |  0.034212 | 15.620786 | -0.107551 |              2.426107 |                  0.450563 |                   20.87339 |                  3.266966 |                      0.0 |     2.289983 |              0.0 |         17.606424 |             20.434783 |                 100.0 |               224.252198 |              218.874727 |  0.0 |  0.0 |  0.914147 |       0.0 | 0.204826 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan |  31.50337 | 73.840772 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.048642 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 11:00:00 |  0.0 | 21.4 |              1.0 |             67.3 |       4.1 |               101.9 | nan | nan |    17.225 |             78.529167 |                  13.1 |  0.0 | 21.4 |    17.225 |    3.286524 |         2.211056 |           1.455461 |          4.1 |                 2.552842 |                      1.968551 |                          0.0 |                      0.156328 |                           0.124585 |             1.718063 |                  1.545886 |     100.181937 |   1.197482 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.949834 | 10.865527 |        2.894599 |                2.761725 |  0.034522 | 15.944568 | -0.117837 |              2.430318 |                  0.451345 |                  21.075854 |                  3.309837 |                      0.0 |     2.292408 |              0.0 |         17.766017 |             22.926829 |                 100.0 |               261.184935 |              253.117687 |  0.0 |  0.0 |  0.955451 |       0.0 | 0.212623 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.610791 | 73.628149 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.047385 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 12:00:00 |  0.0 | 22.5 |              1.0 |             62.2 |       4.1 |               101.9 | nan | nan |   17.2875 |                78.275 |                  13.2 |  0.0 | 22.5 |   17.2875 |    3.286524 |         2.237776 |           1.455461 |          4.1 |                 2.729939 |                      1.976351 |                          0.0 |                      0.165743 |                           0.125016 |             1.698022 |                  1.546988 |     100.201978 |   1.193116 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |             24.06487 |  10.45226 |        2.784503 |                 2.65055 |  0.034679 | 16.302636 | -0.133863 |              2.332484 |                  0.433176 |                  21.177086 |                  3.330376 |                      0.0 |     2.193718 |              0.0 |         17.846709 |             22.926829 |                 100.0 |               315.242275 |              302.373144 |  0.0 |  0.0 |  0.996135 |       0.0 | 0.198734 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.734237 | 73.429415 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.046161 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 13:00:00 |  0.0 | 23.1 |              1.0 |             59.6 |       4.4 |               101.9 | nan | nan | 17.341667 |             78.204167 |                  13.4 |  0.0 | 23.1 | 17.341667 |    3.527001 |         2.277855 |           1.561958 |          4.4 |                 2.830978 |                      1.983133 |                          0.0 |                      0.171077 |                           0.125391 |             1.687263 |                  1.550892 |     100.212737 |   1.190747 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.294943 |  10.96314 |        2.920603 |                2.764822 |  0.035009 | 16.714124 | -0.137935 |              2.433044 |                  0.451851 |                   21.37955 |                  3.363175 |                      0.0 |     2.292911 |              0.0 |         18.016374 |             21.363636 |                 100.0 |               357.321701 |              340.021118 |  0.0 |  0.0 |  1.088996 |       0.0 | 0.190495 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.861756 |  73.23892 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.044969 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 14:00:00 |  0.0 | 23.4 |              1.0 |             55.4 |       4.6 |               101.9 | nan | nan | 17.404167 |               77.9875 |                  13.5 |  0.0 | 23.4 | 17.404167 |     3.68732 |         2.347994 |           1.632956 |          4.6 |                 2.882708 |                      1.990983 |                          0.0 |                      0.173797 |                           0.125825 |              1.59702 |                  1.552718 |      100.30298 |   1.189943 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.409979 | 10.528069 |        2.804699 |                2.649346 |  0.035169 | 17.139185 | -0.135234 |              2.331424 |                  0.432979 |                  21.480782 |                   3.38194 |                      0.0 |      2.19051 |              0.0 |         18.098841 |             20.434783 |                 100.0 |               429.400188 |              403.142251 |  0.0 |  0.0 |  1.130574 |       0.0 | 0.167074 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 31.986572 | 73.071846 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.043807 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 15:00:00 |  0.0 | 23.8 |              1.0 |             52.1 |       4.5 |               101.9 | nan | nan |   17.4875 |             77.708333 |                  13.6 |  0.0 | 23.8 |   17.4875 |    3.607161 |         2.401434 |           1.597457 |          4.5 |                 2.952961 |                      2.001493 |                          0.0 |                      0.177481 |                           0.126406 |             1.538493 |                  1.555327 |     100.361507 |     1.1886 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.525016 |  8.882849 |        2.366409 |                 2.23335 |  0.035331 | 17.555242 | -0.134887 |              1.965348 |                  0.364993 |                  21.582014 |                  3.400029 |                      0.0 |      1.82368 |              0.0 |         18.181985 |             20.888889 |                 100.0 |               524.976153 |              484.281194 |  0.0 |  0.0 |  1.058012 |       0.0 | 0.137614 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.111043 | 72.934232 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.042676 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 16:00:00 |  0.0 | 24.0 |              1.0 |             49.5 |       4.7 |               101.8 | nan | nan |   17.5625 |               77.3875 |                  13.7 |  0.0 | 24.0 |   17.5625 |    3.767479 |         2.438173 |           1.668455 |          4.7 |                 2.988644 |                      2.010993 |                          0.0 |                      0.179348 |                            0.12693 |             1.479379 |                  1.556257 |     100.320621 |   1.186889 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  6.439028 |         1.71537 |                1.620343 |  0.035485 | 17.970142 | -0.130245 |              1.425902 |                   0.26481 |                  21.683246 |                   3.42173 |                      0.0 |      1.28333 |              0.0 |         18.261515 |                  20.0 |                 100.0 |               627.503196 |              568.233788 |  0.0 |  0.0 |  0.956476 |       0.0 | 0.104932 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.230871 |   72.8293 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.041573 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 17:00:00 |  0.0 | 24.0 |              1.0 |             48.1 |       4.5 |               101.8 | nan | nan | 17.654167 |             77.008333 |                  13.7 |  0.0 | 24.0 | 17.654167 |    3.607161 |         2.448193 |           1.597457 |          4.5 |                 2.988644 |                      2.022657 |                          0.0 |                      0.179348 |                           0.127573 |             1.437538 |                  1.557615 |     100.362462 |   1.187075 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  3.911896 |        1.042137 |                0.985011 |  0.035481 |  18.36957 | -0.121617 |               0.86681 |                  0.160979 |                  21.683246 |                  3.423912 |                      0.0 |     0.724147 |              0.0 |         18.259334 |             20.888889 |                 100.0 |               686.923021 |              615.492716 |  0.0 |  0.0 |  0.786453 |       0.0 | 0.083319 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.342072 | 72.745981 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.040499 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 18:00:00 |  0.0 | 23.7 |              1.0 |             49.4 |       4.1 |               101.8 | nan | nan | 17.770833 |             76.570833 |                  13.7 |  0.0 | 23.7 | 17.770833 |    3.286524 |         2.444853 |           1.455461 |          4.1 |                  2.93526 |                      2.037589 |                          0.0 |                      0.176554 |                           0.128396 |             1.450018 |                  1.560199 |     100.349982 |   1.188219 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  1.905864 |        0.507726 |                0.480125 |   0.03548 | 18.740238 | -0.107131 |               0.42251 |                  0.078466 |                  21.683246 |                   3.42458 |                      0.0 |     0.279819 |              0.0 |         18.258666 |             22.926829 |                 100.0 |               599.674099 |              545.752698 |  0.0 |  0.0 |  0.593816 |       0.0 | 0.074832 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.438786 | 72.671149 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.039453 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 19:00:00 |  0.0 | 22.8 |              1.0 |             57.0 |       3.8 |               101.8 | nan | nan |      17.9 |             76.191667 |                  13.7 |  0.0 | 22.8 |      17.9 |    3.046047 |         2.451533 |           1.348964 |          3.8 |                 2.780058 |                      2.054233 |                          0.0 |                      0.168392 |                           0.129312 |             1.584633 |                  1.565154 |     100.215367 |   1.191234 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  0.576079 |        0.153469 |                0.145175 |  0.035488 | 19.062619 | -0.080727 |              0.127754 |                  0.023726 |                  21.683246 |                  3.420162 |                      0.0 |    -0.014752 |              0.0 |         18.263084 |             24.736842 |                 100.0 |                384.03591 |              363.614181 |  0.0 |  0.0 |  0.396612 |       0.0 | 0.072803 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.509096 | 72.598346 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.038434 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 20:00:00 |  0.0 | 21.3 |              0.5 |             65.9 |       2.5 |               101.8 | nan | nan | 18.041667 |                 75.95 |                  13.9 |  0.0 | 21.3 | 18.041667 |    2.003978 |         2.428153 |           0.887476 |          2.5 |                 2.537251 |                      2.072623 |                          0.0 |                      0.155495 |                           0.130323 |             1.672048 |                  1.574157 |     100.127952 |   1.196911 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.072689 |                     15.438712 |            24.870125 |  0.002495 |        0.003269 |                0.003122 | -0.084395 | 19.296988 | -0.043265 |              0.002747 |                   0.00051 |                   21.88571 |                  3.445356 |                      0.0 |    -0.140809 |              0.0 |         18.440354 |                  37.6 |                 100.0 |               272.538499 |               73.372486 |  0.0 |  0.0 |  0.142961 |       0.0 | 0.086909 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.541945 | 72.511436 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.037441 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 21:00:00 |  0.0 | 19.3 |              0.0 |             73.0 |       1.9 |               101.8 | nan | nan |   18.1875 |             75.633333 |                  13.9 |  0.0 | 19.3 |   18.1875 |    1.523023 |         2.418134 |           0.674482 |          1.9 |                 2.242395 |                      2.091705 |                          0.0 |                      0.139602 |                           0.131371 |             1.636948 |                  1.582026 |     100.163052 |   1.205255 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.093856 | 19.406483 |    0.0023 |                   0.0 |                       0.0 |                   21.88571 |                  3.434698 |                      0.0 |    -0.143112 |              0.0 |         18.451012 |             49.473684 |                 100.0 |               228.418544 |               69.444444 |  0.0 |  0.0 |  0.059253 |       0.0 | 0.039214 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.529228 | 72.472222 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.036474 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 22:00:00 |  0.0 | 18.3 |              0.0 |             76.7 |       2.2 |               101.8 | nan | nan | 18.333333 |                 75.25 |                  13.9 |  0.0 | 18.3 | 18.333333 |    1.763501 |         2.418134 |           0.780979 |          2.2 |                  2.10653 |                       2.11094 |                          0.0 |                      0.132185 |                           0.132426 |             1.615709 |                  1.588483 |     100.184291 |   1.209486 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.093881 | 19.364094 |  0.022984 |                   0.0 |                       0.0 |                   21.88571 |                  3.427556 |                      0.0 |    -0.142815 |              0.0 |         18.458154 |             42.727273 |                 100.0 |               220.239745 |               69.444444 |  0.0 |  0.0 |   0.05494 |       0.0 | 0.034293 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.495827 | 72.437929 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.035532 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 1997-08-05 23:00:00 |  0.0 | 17.4 |              0.0 |             81.8 |       2.7 |               101.8 | nan | nan | 18.470833 |               74.9375 |                  13.9 |  0.0 | 17.4 | 18.470833 |    2.164296 |         2.438173 |           0.958474 |          2.7 |                 1.990459 |                      2.129218 |                          0.0 |                      0.125796 |                           0.133428 |             1.628195 |                  1.595583 |     100.171805 |   1.213176 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.093914 | 19.252757 |   0.04002 |                   0.0 |                       0.0 |                   21.88571 |                  3.418333 |                      0.0 |    -0.142431 |              0.0 |         18.467377 |             34.814815 |                 100.0 |               211.199082 |               69.444444 |  0.0 |  0.0 |  0.046092 |       0.0 | 0.026377 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |      0.0 |       0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | nan | nan |      0.0 |  0.0 |  0.0 |   0.0 |  nan | 32.445391 | 72.411552 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.034614 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |

.. _lland_v3_acker_winter_hourly:

acre (winter)
-------------

In our hourly integration tests covering snow processes, we focus on the
subperiod from December 10th to December 12th, where, according to the
daily simulations, some melting occurs:

>>> pub.timegrids = '2010-12-10', '2010-12-13', '1h'

The following input values agree with the ones of the respective sub-period
of the daily integration test :ref:`lland_v3_acker_winter_daily`:

>>> inputs.nied.series = (
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.1, 0.0, 0.3, 0.2, 0.4, 0.7, 0.5, 1.0, 2.0, 2.1, 1.8, 1.3,
...     0.5, 0.2, 0.5, 0.2, 0.1, 0.3, 0.1, 0.1, 0.0, 0.1, 0.0, 0.1, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.1, 0.1, 0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0)
>>> inputs.teml.series = (
...     -3.7, -4.8, -5.6, -6.2, -7.3, -8.3, -8.6, -8.5, -9.5, -9.2, -6.9,
...     -4.0, -2.8, -1.4, -0.4, 0.5, 1.0, 1.1, 1.0, 1.4, 1.6, 2.0, 2.4, 2.5,
...     2.7, 3.2, 3.9, 4.4, 4.8, 4.8, 5.1, 5.1, 5.4, 5.3, 5.5, 6.1, 6.4, 6.4,
...     6.5, 6.3, 5.9, 5.5, 5.1, 4.8, 4.9, 4.9, 4.7, 3.8, 3.9, 3.7, 3.5, 3.3,
...     3.3, 3.0, 1.9, 1.2, 0.8, 0.8, 0.5, 0.3, -0.1, -0.4, -0.6, -0.8, -1.3,
...     -1.6, -2.0, -2.1, -2.2, -2.6, -3.1, -3.9)
>>> inputs.sunshineduration.series = (
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.2, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0)
>>> inputs.windspeed.series = (
...     3.4, 2.9, 2.5, 3.0, 2.5, 2.3, 2.1, 1.4, 1.1, 1.0, 1.4, 1.3, 1.3, 1.4,
...     2.3, 2.8, 3.1, 3.5, 4.4, 3.8, 5.7, 5.6, 5.8, 6.2, 5.6, 5.2, 5.7, 6.6,
...     7.0, 7.3, 7.2, 7.6, 7.3, 7.1, 7.1, 8.9, 9.5, 7.9, 9.6, 9.1, 8.7, 8.9,
...     7.9, 7.1, 6.4, 7.2, 7.0, 6.8, 6.4, 6.3, 5.7, 5.2, 4.6, 4.9, 6.3, 5.8,
...     6.1, 5.9, 6.4, 6.8, 7.0, 6.3, 7.3, 7.5, 7.0, 6.0, 4.8, 4.2, 4.9, 4.0,
...     3.4, 3.4)
>>> inputs.relativehumidity.series = (
...     87.5, 90.7, 89.7, 91.3, 94.2, 93.2, 96.5, 95.9, 96.8, 94.3, 96.7,
...     97.4, 93.9, 93.4, 93.3, 94.1, 94.6, 94.4, 96.9, 98.3, 98.7, 99.6,
...     99.4, 99.2, 99.2, 99.6, 99.8, 99.8, 99.0, 97.6, 97.4, 96.3, 95.2,
...     96.3, 95.5, 91.6, 90.0, 85.7, 82.3, 82.5, 82.2, 76.5, 81.7, 83.4,
...     82.8, 85.8, 87.6, 85.6, 85.3, 86.2, 89.3, 91.4, 90.9, 89.2, 85.6,
...     86.3, 91.2, 87.5, 84.5, 76.7, 74.8, 69.7, 66.9, 64.4, 65.4, 67.6,
...     70.2, 70.4, 69.5, 71.6, 74.5, 80.2)
>>> inputs.atmosphericpressure.series = (
...     101.7, 101.8, 101.8, 101.9, 102.0, 102.0, 102.1, 102.1, 102.2, 102.2,
...     102.2, 102.2, 102.2, 102.1, 102.0, 102.0, 101.9, 101.8, 101.6, 101.5,
...     101.4, 101.3, 101.1, 101.0, 100.8, 100.7, 100.6, 100.5, 100.5, 100.4,
...     100.4, 100.4, 100.3, 100.3, 100.3, 100.2, 100.2, 100.2, 100.2, 100.3,
...     100.3, 100.3, 100.3, 100.3, 100.3, 100.3, 100.3, 100.2, 100.2, 100.2,
...     100.2, 100.3, 100.4, 100.5, 100.6, 100.8, 100.9, 101.0, 101.1, 101.2,
...     101.3, 101.4, 101.5, 101.5, 101.7, 101.7, 101.8, 101.9, 101.9, 102.0,
...     102.0, 102.0)

In this and the following "winter" examples, we use the initial conditions
of December 10th, calculated by the corresponding daily integration tests.
However, this is possible for the state sequences, but not for the (now
higher resolved) log sequences.  Hence, we need to define them manually:

>>> test.inits = (
...     (logs.loggedsunshineduration,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...       0.0, 0.0, 0.04, 0.25, 0.59, 0.91, 0.97, 1.0,
...       0.65, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedglobalradiation,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...       0.0, 0.1, 0.2, 0.4, 0.6, 0.5, 0.2, 0.0,
...       0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedteml,
...      [-2.8, -2.6, -2.2, -1.8, -1.8, -2.4, -1.5, -1.4,
...       -0.8, -0.8, -0.9, 0.9, -0.8, -1.0, -1.2, -1.8,
...       -2.8, -3.8, -4.3, -4, -3.9, -1.6, -1.6, 2.6]),
...     (logs.loggedrelativehumidity,
...      [99.4, 99.0, 99.6, 99.6, 99.3, 99.3, 99.8, 99.7,
...       100.0, 99.6, 99.3, 98.9, 96.6, 93.0, 91.4, 92.0,
...       94.5, 96.5, 98.2, 96.9, 98.0, 95.0, 91.8, 88.5]),
...     (logs.loggedwindspeed2m,
...      [2.2, 2.0, 2.5, 2.6, 2.3, 1.7, 2.8, 1.9,
...       2.6, 2.9, 3.5, 3.4, 3.5, 3.0, 3.9, 3.6,
...       2.2, 2.4, 2.5, 1.5, 2.5, 3.0, 3.2, 3.1]))

Compared to the daily simulation :ref:`lland_v3_acker_winter_daily`, the
hourly simulation estimates a slightly more pronounced melting rate for
land-use type |ACKER|, resulting in a complete depletion of the snow-cover:

.. integration-test::

    >>> lnk(ACKER)
    >>> test('lland_v3_acker_winter_hourly',
    ...      axis1=(inputs.nied, fluxes.wada),
    ...      axis2=(states.waes, states.wats),
    ...      use_conditions=conditions_acker_winter)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration |  nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |         g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes |     sbes |      evpo |       evi |      evb |       evs |     wnied | tempssurface | actualalbedo |  schmpot |     schm |  gefrpot |     gefr |  wlatsnow | wsenssnow |     wsurf |      sff |      fvg |     wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz | qah |  qa |     inzp |      wats |      waes |     esnow |     taus |      ebdn |      bowa |    qdgz1 |    qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet | outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-12-10 00:00:00 |  0.0 | -3.7 |              0.0 |             87.5 |       3.4 |               101.7 | nan | nan |    -2.025 |             96.870833 |                  4.41 |   0.0 | -3.7 |    -2.025 |     2.72541 |         2.684392 |            2.72541 |          3.4 |                 0.464297 |                      0.526233 |                      0.41098 |                      0.034962 |                           0.039056 |              0.40626 |                  0.509766 |      101.29374 |   1.312855 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | -0.013422 | -5.338514 | -0.000646 |                   0.0 |                       0.0 |                   0.701061 |                  4.014396 |                 0.266716 |    -0.167266 |        -0.266716 |         -3.313335 |             62.648725 |                 100.0 |               112.102145 |               98.814229 | 0.234885 | 0.234885 | -0.003706 | -0.003706 |      0.0 |  0.000713 | -0.001816 |    -5.308595 |     0.755856 |      0.0 |      0.0 | 0.701337 | 0.701337 |  0.001902 | -0.036834 |  0.231784 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.063706 | 20.378044 |  20.39327 |       0.0 | 0.843656 | -1.570008 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 01:00:00 |  0.0 | -4.8 |              0.0 |             90.7 |       2.9 |               101.8 | nan | nan | -2.158333 |                96.825 |                  4.41 |   0.0 | -4.8 | -2.158333 |    2.324615 |         2.647918 |           2.324615 |          2.9 |                 0.427219 |                      0.521048 |                     0.396007 |                      0.032479 |                           0.038716 |             0.387487 |                  0.504505 |     101.412513 |   1.319627 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | -0.013572 | -5.233359 |    0.0121 |                   0.0 |                       0.0 |                   0.702915 |                  4.024777 |                 0.276595 |    -0.167699 |        -0.276595 |         -3.321862 |              73.45023 |                 100.0 |               112.102145 |               98.814229 | 0.003706 | 0.003706 |  -0.00973 |  -0.00973 |      0.0 |  0.001157 | -0.000037 |    -5.793556 |     0.755211 |      0.0 |      0.0 | 0.740007 | 0.015226 |  0.003087 | -0.020457 |  0.259225 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  0.06973 | 20.395818 | 20.395818 | -0.242077 | 0.850649 | -1.551208 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 02:00:00 |  0.0 | -5.6 |              0.0 |             89.7 |       2.5 |               101.8 | nan | nan |    -2.325 |             96.604167 |                  4.41 |   0.0 | -5.6 |    -2.325 |    2.003978 |         2.606417 |           2.003978 |          2.5 |                 0.401924 |                      0.514631 |                     0.322903 |                      0.030771 |                           0.038294 |             0.360526 |                  0.497155 |     101.439474 |   1.323706 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | -0.013844 | -5.170694 |  0.070273 |                   0.0 |                       0.0 |                   0.703388 |                  4.040774 |                 0.252691 |    -0.168366 |        -0.252691 |         -3.337385 |             85.202267 |                 100.0 |               112.102145 |               98.814229 |  0.00973 |  0.00973 | -0.009528 | -0.009528 |      0.0 | -0.004653 | -0.000114 |    -8.424076 |     0.755046 |      0.0 |      0.0 | 1.075649 |      0.0 | -0.012411 | -0.052931 |  0.187349 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.069528 | 20.410201 | 20.410201 | -0.359267 | 0.852444 | -1.590581 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 03:00:00 |  0.0 | -6.2 |              0.0 |             91.3 |       3.0 |               101.9 | nan | nan | -2.420833 |                96.325 |                  4.41 |   0.0 | -6.2 | -2.420833 |    2.404774 |         2.602449 |           2.404774 |          3.0 |                 0.383833 |                      0.510972 |                     0.292382 |                       0.02954 |                           0.038053 |             0.350439 |                  0.492194 |     101.549561 |   1.328036 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 |  -0.01405 | -5.301937 |  0.094612 |                   0.0 |                       0.0 |                   0.703735 |                  4.052814 |                 0.243242 |    -0.168867 |        -0.243242 |         -3.349078 |             71.001889 |                 100.0 |               112.102145 |               98.814229 | 0.009528 | 0.009528 | -0.009847 | -0.009847 |      0.0 | -0.008064 | -0.000123 |    -9.682119 |     0.754925 |      0.0 |      0.0 | 1.237144 |      0.0 | -0.021511 | -0.073304 |  0.148428 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.069847 | 20.427793 | 20.427793 | -0.413206 | 0.853761 | -1.654293 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 04:00:00 |  0.0 | -7.3 |              0.0 |             94.2 |       2.5 |               102.0 | nan | nan | -2.558333 |               96.2125 |                  4.41 |   0.0 | -7.3 | -2.558333 |    2.003978 |         2.623448 |           2.003978 |          2.5 |                 0.352531 |                      0.505764 |                     0.268489 |                      0.027395 |                            0.03771 |             0.332084 |                  0.486608 |     101.667916 |   1.334932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 |  -0.01425 |  -5.51431 |  0.113115 |                   0.0 |                       0.0 |                   0.704027 |                  4.064535 |                 0.244652 |    -0.169356 |        -0.244652 |         -3.360508 |             85.202267 |                 100.0 |               112.102145 |               98.814229 | 0.009847 | 0.009847 | -0.013222 | -0.013222 |      0.0 | -0.007865 |  -0.00015 |   -10.751122 |     0.754823 |      0.0 |      0.0 | 1.374942 |      0.0 | -0.020979 | -0.064684 |   0.15899 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.073222 | 20.445505 | 20.445505 | -0.459231 | 0.854871 | -1.736508 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 05:00:00 |  0.0 | -8.3 |              0.0 |             93.2 |       2.3 |               102.0 | nan | nan |    -2.725 |             96.004167 |                  4.41 |   0.0 | -8.3 |    -2.725 |     1.84366 |         2.596101 |            1.84366 |          2.3 |                 0.326061 |                      0.499513 |                      0.24802 |                      0.025563 |                           0.037298 |             0.303889 |                  0.479553 |     101.696111 |   1.340112 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | -0.014518 | -5.788361 |  0.128478 |                   0.0 |                       0.0 |                   0.704206 |                  4.080005 |                 0.249609 |        -0.17 |        -0.249609 |         -3.375799 |             92.611159 |                 100.0 |               112.102145 |               98.814229 | 0.013222 | 0.013222 | -0.012048 | -0.012048 |      0.0 | -0.006569 | -0.000229 |    -11.73641 |     0.754761 |      0.0 |      0.0 | 1.502497 |      0.0 | -0.017522 | -0.061235 |  0.170852 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.072048 | 20.465296 | 20.465296 | -0.501834 | 0.855551 | -1.834086 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 06:00:00 |  0.0 | -8.6 |              0.0 |             96.5 |       2.1 |               102.1 | nan | nan |    -2.925 |             96.004167 |                  4.41 |   0.0 | -8.6 |    -2.925 |    1.683342 |          2.56624 |           1.683342 |          2.1 |                 0.318472 |                      0.492103 |                     0.236581 |                      0.025034 |                           0.036808 |             0.307325 |                  0.472439 |     101.792675 |   1.342932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | -0.014756 |  -6.11362 |  0.134036 |                   0.0 |                       0.0 |                    0.70438 |                  4.093695 |                 0.243083 |    -0.170571 |        -0.243083 |         -3.389315 |             101.43127 |                 100.0 |               112.102145 |               98.814229 | 0.012048 | 0.012048 | -0.014544 | -0.014544 |      0.0 | -0.007887 | -0.000217 |   -12.318991 |       0.7547 |      0.0 |      0.0 | 1.578515 |      0.0 | -0.021037 | -0.062836 |  0.159209 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.074544 | 20.485231 | 20.485231 | -0.527224 | 0.856213 | -1.937222 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 07:00:00 |  0.0 | -8.5 |              0.0 |             95.9 |       1.4 |               102.1 | nan | nan |   -3.1625 |               96.0625 |                  3.76 |   0.0 | -8.5 |   -3.1625 |    1.122228 |         2.521333 |           1.122228 |          1.4 |                 0.320984 |                      0.483429 |                     0.229955 |                       0.02521 |                           0.036234 |             0.307823 |                  0.464394 |     101.792177 |   1.342422 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.641605 |       0.0 |             0.0 |                     0.0 | -0.008525 | -6.457407 |  0.134149 |                   0.0 |                       0.0 |                   0.648077 |                  3.682358 |                 0.225099 |    -0.153432 |        -0.225099 |         -3.034281 |            152.146905 |                 100.0 |               112.102145 |               98.814229 | 0.014544 | 0.014544 | -0.011735 | -0.011735 |      0.0 |  -0.00702 | -0.000258 |   -12.668002 |     0.754666 |      0.0 |      0.0 | 1.625014 |      0.0 | -0.018726 | -0.056952 |  0.149421 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.071735 | 20.506795 | 20.506795 | -0.542755 | 0.856594 | -2.040471 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 08:00:00 |  0.0 | -9.5 |              0.0 |             96.8 |       1.1 |               102.2 | nan | nan | -3.483333 |               96.2625 |                  2.76 |   0.0 | -9.5 | -3.483333 |     0.88175 |         2.408072 |            0.88175 |          1.1 |                 0.296635 |                      0.471927 |                     0.221964 |                      0.023505 |                            0.03547 |             0.287143 |                  0.454289 |     101.912857 |   1.348938 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.355232 |                       7.41713 |             2.287909 |  0.266196 |        0.003492 |                0.003987 |  0.035333 |  -6.80157 |  0.136068 |              0.000978 |                  0.000978 |                   0.561424 |                  3.039527 |                 0.214181 |    -0.125668 |        -0.213202 |         -2.478102 |            193.641515 |                 100.0 |               112.102145 |               99.616909 | 0.011735 | 0.011735 |  -0.00564 |  -0.00564 |      0.0 | -0.005281 | -0.000233 |   -13.101007 |     0.754612 |      0.0 |      0.0 | 1.682096 |      0.0 | -0.014086 | -0.044216 |    0.1549 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  0.06564 | 20.523811 | 20.523811 |  -0.56182 | 0.857176 | -2.145639 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 09:00:00 |  0.0 | -9.2 |              0.0 |             94.3 |       1.0 |               102.2 | nan | nan | -3.816667 |             96.383333 |                  1.79 |   0.0 | -9.2 | -3.816667 |    0.801591 |         2.278972 |           0.801591 |          1.0 |                 0.303761 |                      0.460234 |                     0.225825 |                      0.024006 |                           0.034691 |             0.286447 |                  0.443589 |     101.913553 |   1.347408 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.944823 |  4.919884 |        0.064534 |                0.067188 |  0.075902 | -7.152129 |   0.12394 |              0.016493 |                  0.016493 |                   0.477423 |                  2.412046 |                 0.187789 |    -0.084008 |        -0.171296 |         -1.934623 |            213.005666 |                 100.0 |               112.102145 |              101.107639 |  0.00564 |  0.00564 |  0.001398 |  0.001398 |      0.0 | -0.004727 | -0.000108 |   -12.890094 |     0.754516 |      0.0 |      0.0 | 1.655897 |      0.0 | -0.012608 | -0.043606 |  0.115081 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.058602 | 20.534177 | 20.534177 |  -0.55307 | 0.858233 | -2.238679 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 10:00:00 |  0.0 | -6.9 |              0.0 |             96.7 |       1.4 |               102.2 | nan | nan |   -4.0625 |               96.5375 |                  0.88 |   0.0 | -6.9 |   -4.0625 |    1.122228 |         2.200731 |           1.122228 |          1.4 |                 0.363641 |                      0.451775 |                      0.26039 |                      0.028159 |                           0.034126 |             0.351641 |                  0.436132 |     101.848359 |   1.335446 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.622959 | 14.873457 |        0.195096 |                0.202564 |  0.055588 | -7.462263 |  0.079281 |              0.049755 |                  0.049755 |                   0.398642 |                  1.815487 |                 0.133523 |     -0.02589 |        -0.083768 |         -1.416845 |            152.146905 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.004568 |  0.004568 |      0.0 | -0.008227 |       0.0 |   -11.132699 |     0.754373 |      0.0 |      0.0 | 1.430466 |      0.0 | -0.021945 | -0.057836 |  0.003988 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.054034 | 20.542404 | 20.542404 | -0.477776 | 0.859795 |  -2.28706 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 11:00:00 |  0.0 | -4.0 |              0.1 |             97.4 |       1.3 |               102.2 | nan | nan | -4.195833 |             96.570833 |                  0.39 |   0.0 | -4.0 | -4.195833 |    1.042069 |         2.098318 |           1.042069 |          1.3 |                 0.453913 |                      0.447245 |                     0.320993 |                      0.034269 |                           0.033823 |             0.442111 |                  0.431908 |     101.757889 |   1.320615 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.449648 | 22.189872 |        0.349279 |                0.385806 |  0.044597 | -7.623534 |  0.018924 |              0.094826 |                  0.094826 |                   0.356305 |                  1.493002 |                 0.088443 |     0.032618 |         0.006383 |         -1.136697 |            163.850513 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.009622 |  0.009622 |      0.0 | -0.010551 |       0.0 |    -8.499633 |     0.754213 |      0.0 |      0.0 | 1.092575 |      0.0 | -0.028143 | -0.059406 | -0.093932 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.044412 | 20.552955 | 20.552955 |  -0.36492 | 0.861559 | -2.275084 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 12:00:00 |  0.0 | -2.8 |              0.0 |             93.9 |       1.3 |               102.2 | nan | nan |     -4.35 |               96.3625 |                  0.14 |   0.0 | -2.8 |     -4.35 |    1.042069 |          2.00007 |           1.042069 |          1.3 |                 0.496723 |                      0.442058 |                     0.362782 |                      0.037113 |                           0.033476 |             0.466423 |                  0.425978 |     101.733577 |   1.314635 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.361224 | 22.197307 |        0.291164 |                0.329304 |  0.039084 | -7.583613 | -0.014106 |              0.081002 |                  0.081002 |                   0.334833 |                  1.331032 |                 0.078298 |     0.025542 |         0.002704 |         -0.996199 |            163.850513 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.011078 |  0.011078 |      0.0 | -0.009028 |       0.0 |    -6.930555 |     0.754021 |      0.0 |      0.0 | 0.891337 |      0.0 | -0.024082 | -0.054533 | -0.081319 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.033335 | 20.561983 | 20.561983 | -0.297707 | 0.863678 | -2.230078 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 13:00:00 |  0.0 | -1.4 |              0.0 |             93.4 |       1.4 |               102.1 | nan | nan | -4.370833 |             96.116667 |                   0.1 |   0.0 | -1.4 | -4.370833 |    1.122228 |         1.900997 |           1.122228 |          1.4 |                 0.551149 |                      0.441361 |                     0.400377 |                      0.040686 |                           0.033429 |             0.514774 |                  0.424221 |     101.585226 |   1.306346 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 20.962863 |        0.274971 |                0.289712 |  0.038233 | -7.433593 | -0.038518 |              0.071325 |                  0.071325 |                   0.331643 |                  1.306141 |                  0.06351 |     0.016903 |         0.007815 |         -0.974498 |            152.146905 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.011509 |  0.011509 |      0.0 | -0.010314 |       0.0 |    -5.650355 |     0.753806 |      0.0 |      0.0 |  0.72701 |      0.0 | -0.027511 | -0.058077 | -0.093403 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.021825 | 20.572297 | 20.572297 | -0.242821 | 0.866052 |  -2.16066 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 14:00:00 |  0.0 | -0.4 |              0.0 |             93.3 |       2.3 |               102.0 | nan | nan | -4.354167 |             95.854167 |                   0.1 |   0.0 | -0.4 | -4.354167 |     1.84366 |         1.856982 |            1.84366 |          2.3 |                 0.593188 |                      0.441918 |                     0.431859 |                      0.043416 |                           0.033467 |             0.553444 |                  0.423597 |     101.446556 |   1.300092 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 11.702396 |        0.153501 |                0.155234 |  0.038266 |   -7.2022 | -0.054959 |              0.038255 |                  0.038255 |                   0.331963 |                  1.307301 |                 0.054746 |    -0.016216 |        -0.016491 |         -0.975338 |             92.611159 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.011319 |  0.011319 |      0.0 | -0.014295 |       0.0 |    -4.657812 |     0.753568 |      0.0 |      0.0 | 0.599604 |      0.0 | -0.038132 | -0.075872 | -0.097513 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.010506 | 20.586592 | 20.586592 | -0.200268 | 0.868682 | -2.074801 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 15:00:00 |  0.0 |  0.5 |              0.0 |             94.1 |       2.8 |               102.0 | nan | nan |      -4.3 |             95.608333 |                   0.1 |   0.0 |  0.5 |      -4.3 |    2.244455 |         1.842168 |           2.244455 |          2.8 |                 0.633417 |                      0.443734 |                     0.459246 |                      0.046005 |                           0.033588 |             0.596046 |                  0.424247 |     101.403954 |   1.295611 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 |  2.880181 |         0.03778 |                0.037152 |  0.038265 | -6.916004 |  -0.06633 |              0.009165 |                  0.009165 |                   0.332312 |                  1.307629 |                 0.043931 |    -0.045319 |        -0.034766 |         -0.975317 |             76.073452 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.008136 |  0.008136 |      0.0 | -0.018168 |       0.0 |    -3.845169 |     0.753309 |      0.0 |      0.0 | 0.495336 |      0.0 | -0.048462 |  -0.08746 | -0.101156 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.002371 |  20.60476 |  20.60476 | -0.165442 | 0.871562 | -1.977571 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 16:00:00 |  0.1 |  1.0 |              0.0 |             94.6 |       3.1 |               101.9 | nan | nan |      -4.2 |             95.395833 |                   0.1 | 0.094 |  1.0 |      -4.2 |    2.484933 |          1.86654 |           2.484933 |          3.1 |                 0.656792 |                      0.447104 |                     0.482681 |                        0.0475 |                           0.033814 |             0.621325 |                  0.426519 |     101.278675 |   1.291856 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.061898 |                       7.41713 |             1.347076 |  0.007844 |        0.000103 |                0.000101 |   0.02828 | -6.591904 | -0.073629 |              0.000025 |                  0.000025 |                   0.332597 |                   1.30686 |                 0.041079 |    -0.054428 |        -0.041055 |         -0.974263 |             68.711505 |                 100.0 |               112.102145 |               98.953162 | 0.036371 | 0.009093 |  0.005495 |  0.005495 |      0.0 |  -0.01968 |  0.000133 |    -3.183165 |     0.753097 |      0.0 |      0.0 | 0.411688 | 0.027278 | -0.052495 | -0.089993 | -0.101434 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.054505 |  20.66081 |  20.66081 | -0.128393 |  0.87392 | -1.873042 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 17:00:00 |  0.0 |  1.1 |              0.0 |             94.4 |       3.5 |               101.8 | nan | nan | -4.091667 |             95.170833 |                   0.1 |   0.0 |  1.1 | -4.091667 |    2.805569 |         1.866772 |           2.805569 |          3.5 |                 0.661557 |                      0.450781 |                     0.501947 |                      0.047804 |                            0.03406 |              0.62451 |                  0.429012 |      101.17549 |     1.2901 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.027647 | -6.243475 | -0.077406 |                   0.0 |                       0.0 |                   0.333023 |                  1.305996 |                 0.047279 |    -0.054417 |        -0.047279 |         -0.972973 |             60.858762 |                 100.0 |               112.102145 |               98.814229 |      0.0 |      0.0 |  0.006891 |  0.006891 |      0.0 | -0.018891 |       0.0 |    -2.659875 |     0.752781 |      0.0 |      0.0 | 0.343881 |      0.0 |  -0.05039 | -0.087831 | -0.090942 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.047614 | 20.679701 | 20.679701 | -0.114856 | 0.877451 | -1.764737 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 18:00:00 |  0.3 |  1.0 |              0.0 |             96.9 |       4.4 |               101.6 | nan | nan |     -3.95 |             95.070833 |                   0.1 | 0.282 |  1.0 |     -3.95 |    3.527001 |         1.942897 |           3.527001 |          4.4 |                 0.656792 |                      0.455629 |                     0.517988 |                        0.0475 |                           0.034384 |             0.636431 |                   0.43317 |     100.963569 |   1.287972 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.027682 | -5.882456 | -0.078729 |                   0.0 |                       0.0 |                   0.332764 |                  1.303739 |                 0.052764 |    -0.054322 |        -0.052764 |         -0.970975 |             48.410379 |                 100.0 |               112.102145 |               98.814229 | 0.269614 | 0.067403 |  0.003343 |  0.003343 |      0.0 | -0.021503 |  0.000987 |    -2.237585 |     0.752973 |      0.0 |      0.0 | 0.296166 |  0.20221 | -0.057359 | -0.089084 | -0.093679 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.056657 | 20.970818 | 20.970818 | -0.031381 |   0.8753 | -1.655107 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 19:00:00 |  0.2 |  1.4 |              0.0 |             98.3 |       3.8 |               101.5 | nan | nan | -3.816667 |             95.029167 |                   0.1 | 0.188 |  1.4 | -3.816667 |    3.046047 |         1.973983 |           3.046047 |          3.8 |                 0.676036 |                      0.460234 |                     0.553668 |                      0.048726 |                           0.034691 |             0.664544 |                  0.437356 |     100.835456 |   1.284691 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.027732 | -5.517025 | -0.090262 |                   0.0 |                       0.0 |                   0.333173 |                  1.301296 |                 0.055014 |    -0.054221 |        -0.055014 |         -0.968122 |             56.054123 |                 100.0 |               112.102145 |               98.814229 | 0.184657 | 0.027698 |  -0.00049 |  -0.00049 |      0.0 | -0.018103 |  0.001001 |    -1.338225 |     0.752669 |      0.0 |      0.0 | 0.178473 | 0.156958 | -0.048288 | -0.067758 | -0.061032 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  0.06049 | 21.173577 | 21.173577 | -0.007186 | 0.878699 | -1.533945 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 20:00:00 |  0.4 |  1.6 |              0.0 |             98.7 |       5.7 |               101.4 | nan | nan |    -3.675 |             94.991667 |                   0.1 | 0.376 |  1.6 |    -3.675 |     4.56907 |         2.056027 |            4.56907 |          5.7 |                 0.685844 |                      0.465172 |                     0.582167 |                      0.049349 |                            0.03502 |             0.676928 |                  0.441874 |     100.723072 |   1.282429 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.027786 | -5.113151 | -0.096276 |                   0.0 |                       0.0 |                   0.333621 |                  1.298642 |                 0.060459 |     -0.05411 |        -0.060459 |         -0.965021 |             37.369415 |                 100.0 |               112.102145 |               98.814229 |  0.37649 | 0.037649 |  -0.00016 |  -0.00016 |      0.0 | -0.020957 |  0.002396 |    -0.655946 |     0.752337 |      0.0 |      0.0 | 0.089849 | 0.089849 | -0.055901 | -0.075614 | -0.071056 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  0.06016 |  21.32179 | 21.571024 |       0.0 | 0.882429 |  -1.40677 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 21:00:00 |  0.7 |  2.0 |              0.0 |             99.6 |       5.6 |               101.3 | nan | nan |      -3.5 |             94.991667 |                   0.1 | 0.658 |  2.0 |      -3.5 |    4.488911 |         2.138899 |           4.488911 |          5.6 |                 0.705836 |                      0.471336 |                     0.599971 |                      0.050616 |                           0.035431 |             0.703012 |                   0.44773 |     100.596988 |   1.279174 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.027865 | -4.689232 | -0.096005 |                   0.0 |                       0.0 |                   0.334496 |                  1.295068 |                 0.055477 |    -0.053961 |        -0.055477 |         -0.960572 |             38.036726 |                 100.0 |               112.102145 |               98.814229 |  0.65816 |      0.0 | -0.003032 | -0.003032 |      0.0 | -0.022474 |  0.005511 |     -0.24455 |     0.751687 |      0.0 |      0.0 |  0.03541 |  0.03541 | -0.059948 | -0.074196 | -0.078667 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.063032 | 21.378792 | 22.251658 |       0.0 | 0.889765 | -1.279865 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 22:00:00 |  0.5 |  2.4 |              0.0 |             99.4 |       5.8 |               101.1 | nan | nan | -3.291667 |             95.008333 |                   0.1 |  0.47 |  2.4 | -3.291667 |    4.649229 |         2.249283 |           4.649229 |          5.8 |                  0.72634 |                      0.478769 |                      0.61078 |                       0.05191 |                           0.035924 |             0.721982 |                   0.45487 |     100.378018 |   1.274698 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.027957 | -4.266215 |  -0.09215 |                   0.0 |                       0.0 |                   0.335364 |                  1.290674 |                 0.049365 |    -0.053778 |        -0.049365 |          -0.95531 |             36.725115 |                 100.0 |               112.102145 |               98.814229 | 0.473032 |      0.0 | -0.002331 | -0.002331 |      0.0 | -0.024931 |  0.004753 |          0.0 |     0.751043 |  0.03381 |  0.03381 |      0.0 |      0.0 | -0.066503 | -0.081551 |  -0.09869 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.062331 |   21.3684 | 22.749622 |       0.0 | 0.897096 | -1.156814 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 23:00:00 |  1.0 |  2.5 |              0.0 |             99.2 |       6.2 |               101.0 | nan | nan | -3.070833 |                  95.0 |                   0.1 |  0.94 |  2.5 | -3.070833 |    4.969866 |         2.364694 |           4.969866 |          6.2 |                 0.731547 |                      0.486761 |                      0.61078 |                      0.052238 |                           0.036454 |             0.725695 |                  0.462423 |     100.274305 |   1.272954 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.028052 | -3.856048 | -0.083291 |                   0.0 |                       0.0 |                   0.336227 |                   1.28612 |                 0.047044 |    -0.053588 |        -0.047044 |         -0.949893 |             34.355753 |                 100.0 |               112.102145 |               98.814229 | 0.942331 |      0.0 | -0.001468 | -0.001468 |      0.0 | -0.027164 |  0.009863 |          0.0 |     0.750402 | 0.124411 | 0.124411 |      0.0 |      0.0 | -0.072459 | -0.089566 | -0.114981 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.061468 | 21.268346 | 23.719116 |       0.0 | 0.904448 | -1.042624 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 00:00:00 |  2.0 |  2.7 |              0.0 |             99.2 |       5.6 |               100.8 | nan | nan | -2.804167 |               95.4875 |                   0.1 |  1.88 |  2.7 | -2.804167 |    4.488911 |         2.438173 |           4.488911 |          5.6 |                 0.742061 |                      0.496568 |                      0.61078 |                      0.052899 |                           0.037103 |             0.736124 |                   0.47416 |     100.063876 |   1.269455 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.028146 | -3.475412 | -0.075069 |                   0.0 |                       0.0 |                   0.334774 |                  1.277745 |                 0.041776 |    -0.053239 |        -0.041776 |         -0.942971 |             38.036726 |                 100.0 |               112.102145 |               98.814229 | 1.881468 |      0.0 | -0.001641 | -0.001641 |      0.0 | -0.027338 |  0.021269 |          0.0 |     0.749763 |   0.1994 |   0.1994 |      0.0 |      0.0 | -0.072924 | -0.089252 |   -0.1204 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.061641 | 21.091445 | 25.627923 |       0.0 | 0.911838 | -0.936655 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 01:00:00 |  2.1 |  3.2 |              0.0 |             99.6 |       5.2 |               100.7 | nan | nan | -2.470833 |             95.858333 |                   0.1 | 1.974 |  3.2 | -2.470833 |    4.168274 |         2.514993 |           4.168274 |          5.2 |                  0.76893 |                      0.509073 |                      0.61078 |                      0.054584 |                           0.037928 |             0.765854 |                  0.487989 |      99.934146 |   1.265756 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.028325 | -3.122183 | -0.067439 |                   0.0 |                       0.0 |                   0.335624 |                  1.268373 |                 0.027852 |    -0.052849 |        -0.027852 |          -0.93275 |             40.962628 |                 100.0 |               112.102145 |               98.814229 | 1.975641 |      0.0 | -0.002958 | -0.002958 |      0.0 | -0.031933 |  0.026469 |          0.0 |     0.749128 | 0.347986 | 0.347986 |      0.0 |      0.0 |  -0.08518 |  -0.09987 | -0.157197 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.062958 | 20.767456 | 27.635496 |       0.0 | 0.919237 | -0.838316 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 02:00:00 |  1.8 |  3.9 |              0.0 |             99.8 |       5.7 |               100.6 | nan | nan |    -2.075 |             96.279167 |                   0.1 | 1.692 |  3.9 |    -2.075 |     4.56907 |         2.621871 |            4.56907 |          5.7 |                 0.807986 |                      0.524283 |                      0.61078 |                       0.05702 |                           0.038928 |              0.80637 |                  0.504776 |       99.79363 |   1.261108 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.028536 | -2.794386 | -0.060359 |                   0.0 |                       0.0 |                   0.336466 |                  1.257174 |                 0.008709 |    -0.052382 |        -0.008709 |         -0.920708 |             37.369415 |                 100.0 |               112.102145 |               98.814229 | 1.694958 |      0.0 | -0.003511 | -0.003511 |      0.0 | -0.043255 |  0.027676 |          0.0 |     0.748498 | 0.612901 | 0.612901 |      0.0 |      0.0 | -0.115381 |  -0.13072 | -0.237391 |      1.0 |      0.3 | 0.510192 | 0.174162 |  0.0 |  0.0 | 0.0 |  0.0 | 0.174162 | nan | nan | 0.063511 | 20.184278 | 28.863518 |       0.0 |  0.92663 | -0.747057 | 71.782289 | 0.174162 |      0.0 |   0.0 |   0.0 |  0.0 | 0.002259 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 03:00:00 |  1.3 |  4.4 |              0.0 |             99.8 |       6.6 |               100.5 | nan | nan | -1.633333 |             96.633333 |                   0.1 | 1.222 |  4.4 | -1.633333 |    5.290502 |          2.74211 |           5.290502 |          6.6 |                 0.836943 |                      0.541728 |                      0.61078 |                      0.058816 |                           0.040071 |             0.835269 |                   0.52349 |      99.664731 |   1.257444 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.028761 |  -2.49019 | -0.053788 |                   0.0 |                       0.0 |                   0.337293 |                  1.245134 |                -0.004886 |    -0.051881 |         0.004886 |         -0.907841 |             32.273586 |                 100.0 |               111.620785 |               98.814229 | 1.225511 |      0.0 | -0.003372 | -0.003372 |      0.0 | -0.055801 |       0.0 |          0.0 |      0.74788 | 0.795532 | 0.795532 |      0.0 |      0.0 | -0.148847 | -0.165762 | -0.319496 |      1.0 |      0.3 | 2.365322 | 0.809277 |  0.0 |  0.0 | 0.0 |  0.0 | 0.809277 | nan | nan | 0.063372 |  19.42623 | 27.779508 |       0.0 | 0.933938 | -0.662369 | 73.338335 |      0.5 | 0.309277 |   0.0 |   0.0 |  0.0 | 0.010926 | 0.015637 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 04:00:00 |  0.5 |  4.8 |              0.0 |             99.0 |       7.0 |               100.5 | nan | nan | -1.129167 |             96.833333 |                   0.1 |  0.47 |  4.8 | -1.129167 |    5.611139 |         2.892408 |           5.611139 |          7.0 |                 0.860763 |                      0.562266 |                      0.61078 |                      0.060288 |                            0.04141 |             0.852155 |                  0.544461 |      99.647845 |   1.255555 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |     0.029 | -2.207896 | -0.047691 |                   0.0 |                       0.0 |                   0.338113 |                  1.232336 |                -0.014528 |    -0.051347 |         0.014528 |         -0.894223 |             30.429381 |                 100.0 |                109.40389 |               98.814229 | 0.473372 |      0.0 |  0.000247 |  0.000247 |      0.0 |  -0.06294 |       0.0 |          0.0 |     0.747267 |  0.97133 |  0.97133 |      0.0 |      0.0 |  -0.16789 | -0.189697 | -0.372115 |      1.0 |      0.3 |  1.86552 |  0.63986 |  0.0 |  0.0 | 0.0 |  0.0 |  0.63986 | nan | nan | 0.059753 | 18.496713 |   26.4503 |       0.0 | 0.941246 | -0.583778 | 74.563995 |      0.5 |  0.13986 |   0.0 |   0.0 |  0.0 | 0.023559 | 0.036256 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 05:00:00 |  0.2 |  4.8 |              0.0 |             97.6 |       7.3 |               100.4 | nan | nan | -0.583333 |             97.016667 |                   0.1 | 0.188 |  4.8 | -0.583333 |    5.851616 |         3.059407 |           5.851616 |          7.3 |                 0.860763 |                      0.585275 |                      0.61078 |                      0.060288 |                           0.042904 |             0.840104 |                  0.567814 |      99.559896 |   1.254358 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |   0.02926 | -1.945928 | -0.042032 |                   0.0 |                       0.0 |                   0.338927 |                  1.218349 |                -0.012189 |    -0.050765 |         0.012189 |         -0.879421 |             29.178858 |                 100.0 |               107.671064 |               98.814229 | 0.187753 |      0.0 |  0.006913 |  0.006913 |      0.0 | -0.061894 |       0.0 |          0.0 |     0.746658 | 0.992818 | 0.992818 |      0.0 |      0.0 | -0.165098 | -0.196345 | -0.373633 |      1.0 |      0.3 | 1.610804 | 0.553646 |  0.0 |  0.0 | 0.0 |  0.0 | 0.553646 | nan | nan | 0.053087 | 17.544855 | 25.089142 |       0.0 | 0.948554 | -0.510846 | 75.621154 |      0.5 | 0.053646 |   0.0 |   0.0 |  0.0 | 0.035866 | 0.042194 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 06:00:00 |  0.5 |  5.1 |              0.0 |             97.4 |       7.2 |               100.4 | nan | nan |   -0.0125 |             97.054167 |                   0.1 |  0.47 |  5.1 |   -0.0125 |    5.771457 |         3.229745 |           5.771457 |          7.2 |                 0.879017 |                      0.610223 |                      0.61078 |                      0.061412 |                           0.044515 |             0.856163 |                  0.592247 |      99.543837 |    1.25293 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |   0.02952 | -1.702821 | -0.036781 |                   0.0 |                       0.0 |                   0.339735 |                  1.204312 |                -0.020102 |     -0.05018 |         0.020102 |         -0.864577 |              29.58412 |                 100.0 |               106.185377 |               98.814229 | 0.463087 |      0.0 |  0.007922 |  0.007922 |      0.0 |  -0.06548 |       0.0 |          0.0 |     0.746054 | 1.090564 | 1.090564 |      0.0 |      0.0 | -0.174665 | -0.206262 | -0.401029 |      1.0 |      0.3 | 2.026663 | 0.698226 |  0.0 |  0.0 | 0.0 |  0.0 | 0.698226 | nan | nan | 0.052078 | 16.497235 | 23.591046 |       0.0 | 0.955862 | -0.443165 |  76.94959 |      0.5 | 0.198226 |   0.0 |   0.0 |  0.0 | 0.047855 | 0.050642 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 07:00:00 |  0.2 |  5.1 |              0.0 |             96.3 |       7.6 |               100.4 | nan | nan |  0.554167 |             97.070833 |                   0.1 | 0.188 |  5.1 |  0.554167 |    6.092093 |         3.436822 |           6.092093 |          7.6 |                 0.879017 |                      0.635914 |                      0.61078 |                      0.061412 |                           0.046165 |             0.846494 |                  0.617287 |      99.553506 |   1.252976 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.029782 | -1.477218 | -0.031908 |                   0.0 |                       0.0 |                   0.340537 |                  1.190153 |                -0.018227 |     -0.04959 |         0.018227 |         -0.849615 |             28.027061 |                 100.0 |                104.32952 |               98.814229 | 0.180078 |      0.0 |  0.013937 |  0.013937 |      0.0 | -0.065772 |       0.0 |          0.0 |     0.745455 | 1.130074 | 1.130074 |      0.0 |      0.0 | -0.175444 | -0.215681 | -0.409353 |      1.0 |      0.3 | 1.800589 | 0.621774 |  0.0 |  0.0 | 0.0 |  0.0 | 0.621774 | nan | nan | 0.046063 | 15.410005 | 22.036307 |       0.0 |  0.96317 | -0.380357 | 78.128406 |      0.5 | 0.121774 |   0.0 |   0.0 |  0.0 | 0.059535 | 0.061445 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 08:00:00 |  0.1 |  5.4 |              0.0 |             95.2 |       7.3 |               100.3 | nan | nan |     1.175 |             97.004167 |                   0.1 | 0.094 |  5.4 |     1.175 |    5.851616 |           3.6439 |           5.851616 |          7.3 |                 0.897612 |                      0.665151 |                      0.61078 |                      0.062554 |                           0.048033 |             0.854526 |                  0.645224 |      99.445474 |   1.250338 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.347041 |                      7.400749 |             1.337827 |  0.254871 |         0.00332 |                0.003243 |  0.031005 | -1.267858 | -0.027386 |              0.000827 |                  0.000827 |                   0.341334 |                  1.174779 |                -0.024601 |    -0.048122 |         0.025428 |         -0.833446 |             29.178858 |                 100.0 |               102.692474 |               99.313528 | 0.080063 |      0.0 |  0.019026 |  0.019026 |      0.0 | -0.065786 |       0.0 |          0.0 |      0.74486 | 1.180875 | 1.180875 |      0.0 |      0.0 | -0.175481 | -0.220889 | -0.421798 |      1.0 |      0.3 | 1.773755 |  0.61386 |  0.0 |  0.0 | 0.0 |  0.0 |  0.61386 | nan | nan | 0.040974 | 14.271609 | 20.408401 |       0.0 | 0.970478 | -0.322072 | 79.288301 |      0.5 |  0.11386 |   0.0 |   0.0 |  0.0 | 0.070912 | 0.067041 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 09:00:00 |  0.3 |  5.3 |              0.0 |             96.3 |       7.1 |               100.3 | nan | nan |  1.779167 |               97.0875 |                   0.1 | 0.282 |  5.3 |  1.779167 |    5.691298 |         3.847638 |           5.691298 |          7.1 |                 0.891375 |                      0.694736 |                      0.61078 |                      0.062171 |                           0.049913 |             0.858394 |                  0.674502 |      99.441606 |   1.250768 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  4.882681 |        0.063603 |                0.062163 |  0.032098 | -1.073572 | -0.023189 |              0.015897 |                  0.015897 |                   0.342124 |                   1.15845 |                -0.023742 |    -0.032372 |         0.039639 |         -0.816327 |             30.000798 |                 100.0 |               101.090212 |              100.109558 | 0.262974 |      0.0 |  0.016582 |  0.016582 |      0.0 | -0.065321 |       0.0 |          0.0 |     0.744269 | 1.205376 | 1.205376 |      0.0 |      0.0 | -0.174242 | -0.211904 | -0.425785 |      1.0 |      0.3 | 1.992179 | 0.691104 |  0.0 |  0.0 | 0.0 |  0.0 | 0.691104 | nan | nan | 0.043418 | 13.108054 | 18.744517 |       0.0 | 0.977786 | -0.267983 | 80.589376 |      0.5 | 0.191104 |   0.0 |   0.0 |  0.0 | 0.081996 |   0.0756 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 10:00:00 |  0.1 |  5.5 |              0.0 |             95.5 |       7.1 |               100.3 | nan | nan |  2.295833 |               97.0375 |                   0.1 | 0.094 |  5.5 |  2.295833 |    5.691298 |         4.038016 |           5.691298 |          7.1 |                 0.903886 |                       0.72095 |                      0.61078 |                      0.062939 |                            0.05157 |             0.863211 |                  0.699592 |      99.436789 |   1.249848 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 | 14.862249 |        0.193599 |                0.189423 |  0.031537 | -0.893275 | -0.019295 |              0.048553 |                  0.048553 |                   0.342908 |                  1.144953 |                 -0.02789 |     0.000846 |         0.076442 |         -0.802045 |             30.000798 |                 100.0 |                 99.30245 |               99.928769 | 0.077418 |      0.0 |  0.026484 |  0.026484 |      0.0 | -0.066592 |       0.0 |          0.0 |     0.743683 | 1.361314 | 1.361314 |      0.0 |      0.0 | -0.177631 |   -0.2199 | -0.473974 |      1.0 |      0.3 | 2.031013 | 0.706343 |  0.0 |  0.0 | 0.0 |  0.0 | 0.706343 | nan | nan | 0.033516 | 11.788471 | 16.857514 |       0.0 | 0.985094 | -0.217788 | 81.914045 |      0.5 | 0.206343 |   0.0 |   0.0 |  0.0 | 0.092793 |  0.08785 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 11:00:00 |  0.1 |  6.1 |              0.0 |             91.6 |       8.9 |               100.2 | nan | nan |  2.716667 |             96.795833 |                   0.0 | 0.094 |  6.1 |  2.716667 |    7.134162 |         4.291853 |           7.134162 |          8.9 |                 0.942351 |                      0.742943 |                      0.61078 |                       0.06529 |                           0.052955 |             0.863194 |                  0.719138 |      99.336806 |   1.245915 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 |  22.23196 |        0.289599 |                0.288488 |  0.029185 | -0.725959 | -0.015681 |              0.074112 |                  0.074112 |                   0.334643 |                  1.076886 |                -0.043295 |     0.029242 |         0.117407 |         -0.742243 |             23.933221 |                 100.0 |                97.492069 |               99.739637 | 0.067516 |      0.0 |  0.059612 |  0.059612 |      0.0 | -0.080429 |       0.0 |          0.0 |     0.743101 | 1.828898 | 1.828898 |      0.0 |      0.0 |  -0.21454 | -0.294586 | -0.626533 |      1.0 |      0.3 | 2.695318 | 0.940294 |  0.0 |  0.0 | 0.0 |  0.0 | 0.940294 | nan | nan | 0.000388 | 10.007092 | 14.310141 |       0.0 | 0.992402 | -0.171207 | 83.669069 |      0.5 | 0.440294 |   0.0 |   0.0 |  0.0 | 0.103312 | 0.111456 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 12:00:00 |  0.0 |  6.4 |              0.0 |             90.0 |       9.5 |               100.2 | nan | nan |       3.1 |             96.633333 |                   0.0 |   0.0 |  6.4 |       3.1 |    7.615117 |          4.56573 |           7.615117 |          9.5 |                 0.962118 |                      0.763489 |                      0.61078 |                      0.066493 |                           0.054244 |             0.865906 |                  0.737784 |      99.334094 |   1.244565 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 22.247803 |        0.289805 |                0.288994 |  0.028801 |  -0.57069 | -0.012327 |              0.074409 |                  0.074409 |                   0.335396 |                  1.067855 |                -0.048721 |     0.029915 |          0.12313 |         -0.732459 |             22.421649 |                 100.0 |                95.107864 |               99.480738 |      0.0 |      0.0 |  0.073403 |  0.000388 |      0.0 | -0.085956 |       0.0 |          0.0 |     0.742523 | 1.996682 | 1.996682 |      0.0 |      0.0 | -0.229285 | -0.326804 | -0.679219 |      1.0 |      0.3 | 2.872406 | 1.005591 |  0.0 |  0.0 | 0.0 |  0.0 | 1.005591 | nan | nan |      0.0 |  8.058526 | 11.523692 |       0.0 |  0.99971 |  -0.12798 | 85.535884 |      0.5 | 0.505591 |   0.0 |   0.0 |  0.0 | 0.113559 |  0.14744 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 13:00:00 |  0.1 |  6.4 |              0.0 |             85.7 |       7.9 |               100.2 | nan | nan |     3.425 |               96.3125 |                   0.0 | 0.094 |  6.4 |     3.425 |    6.332571 |         4.782828 |           6.332571 |          7.9 |                 0.962118 |                      0.781298 |                      0.61078 |                      0.066493 |                           0.055357 |             0.824535 |                  0.752488 |      99.375465 |    1.24476 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 20.992726 |        0.273456 |                0.273007 |  0.028513 | -0.426601 | -0.009215 |               0.07045 |                   0.07045 |                   0.336143 |                   1.06128 |                -0.040462 |      0.02623 |         0.110912 |         -0.725137 |             26.962743 |                 100.0 |                92.588538 |               99.194253 |    0.034 |      0.0 |  0.084807 |      0.06 |      0.0 | -0.061599 |       0.0 |          0.0 |     0.741949 | 1.633332 | 1.633332 |      0.0 |      0.0 | -0.164312 | -0.279524 | -0.554747 |      1.0 |      0.3 |  2.38215 | 0.836578 |  0.0 |  0.0 | 0.0 |  0.0 | 0.836578 | nan | nan |      0.0 |  6.459539 |  9.237141 |       0.0 | 1.007018 | -0.087866 | 87.081456 |      0.5 | 0.336578 |   0.0 |   0.0 |  0.0 | 0.123541 |  0.17449 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 14:00:00 |  0.0 |  6.5 |              0.0 |             82.3 |       9.6 |               100.2 | nan | nan |    3.7125 |             95.854167 |                   0.0 |   0.0 |  6.5 |    3.7125 |    7.695276 |         5.026645 |           7.695276 |          9.6 |                 0.968788 |                      0.797357 |                      0.61078 |                      0.066898 |                           0.056358 |             0.797312 |                    0.7643 |      99.402688 |   1.244443 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 11.677032 |        0.152108 |                 0.15202 |  0.028297 | -0.292885 | -0.006326 |              0.039316 |                  0.039316 |                   0.336885 |                  1.056546 |                -0.036448 |    -0.004707 |         0.075764 |         -0.719661 |              22.18809 |                 100.0 |                90.514924 |               98.947814 |      0.0 |      0.0 |  0.115707 |       0.0 |      0.0 | -0.063414 |       0.0 |          0.0 |      0.74138 | 1.717075 | 1.717075 |      0.0 |      0.0 | -0.169154 | -0.334911 | -0.579829 |      1.0 |      0.3 | 2.472274 | 0.870875 |  0.0 |  0.0 | 0.0 |  0.0 | 0.870875 | nan | nan |      0.0 |  4.775022 |  6.828281 |       0.0 | 1.014326 | -0.050639 | 88.682855 |      0.5 | 0.370875 |   0.0 |   0.0 |  0.0 | 0.133265 | 0.192334 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 15:00:00 |  0.1 |  6.3 |              0.1 |             82.5 |       9.1 |               100.3 | nan | nan |  3.954167 |             95.370833 |                   0.1 | 0.094 |  6.3 |  3.954167 |     7.29448 |         5.237063 |            7.29448 |          9.1 |                 0.955489 |                       0.81108 |                      0.61078 |                       0.06609 |                           0.057212 |             0.788278 |                  0.773534 |      99.511722 |   1.246623 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  2.844748 |        0.044468 |                 0.04541 |  0.030016 | -0.168798 | -0.003646 |               0.01177 |                   0.01177 |                   0.346745 |                  1.110113 |                -0.025663 |    -0.034485 |         0.037432 |         -0.763368 |             23.407216 |                 100.0 |                88.377311 |               98.683036 |    0.034 |      0.0 |  0.102764 |      0.06 |      0.0 | -0.057639 |       0.0 |          0.0 |     0.740815 | 1.489816 | 1.489816 |      0.0 |      0.0 |  -0.15375 | -0.310062 | -0.501245 |      1.0 |      0.3 | 2.182421 | 0.770943 |  0.0 |  0.0 | 0.0 |  0.0 | 0.770943 | nan | nan |      0.0 |  3.312937 |    4.7375 |       0.0 | 1.021634 | -0.016093 | 90.094333 |      0.5 | 0.270943 |   0.0 |   0.0 |  0.0 | 0.142738 | 0.205026 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 16:00:00 |  0.0 |  5.9 |              0.0 |             82.2 |       8.7 |               100.3 | nan | nan |  4.158333 |             94.854167 |                   0.1 |   0.0 |  5.9 |  4.158333 |    6.973844 |         5.424101 |           6.973844 |          8.7 |                 0.929372 |                      0.822835 |                      0.61078 |                      0.064498 |                           0.057942 |             0.763944 |                  0.780493 |      99.536056 |   1.248525 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.053708 |                      7.400749 |             1.337827 |   0.00593 |        0.000077 |                0.000079 |  0.031269 | -0.053644 | -0.001159 |               0.00002 |                   0.00002 |                   0.347496 |                  1.107879 |                -0.014096 |    -0.046141 |         0.014116 |         -0.760383 |              24.48341 |                 100.0 |                86.501887 |               98.794123 |      0.0 |      0.0 |  0.096614 |       0.0 |      0.0 | -0.047871 |       0.0 |          0.0 |     0.740253 | 1.257874 | 1.257874 |      0.0 |      0.0 | -0.127693 | -0.279479 | -0.421289 | 0.999402 | 0.299642 | 1.816936 | 0.642772 |  0.0 |  0.0 | 0.0 |  0.0 | 0.642772 | nan | nan |      0.0 |  2.075829 |  2.968435 |       0.0 | 1.028942 |  0.015965 | 91.268496 |      0.5 | 0.142772 |   0.0 |   0.0 |  0.0 | 0.151966 | 0.205097 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 17:00:00 |  0.0 |  5.5 |              0.0 |             76.5 |       8.9 |               100.3 | nan | nan |  4.341667 |             94.108333 |                   0.1 |   0.0 |  5.5 |  4.341667 |    7.134162 |         5.604459 |           7.134162 |          8.9 |                 0.903886 |                      0.833518 |                      0.61078 |                      0.062939 |                           0.058604 |             0.691473 |                   0.78441 |      99.608527 |   1.250659 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.031359 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.348242 |                   1.10789 |                 0.008426 |    -0.046162 |        -0.008426 |         -0.759648 |             23.933221 |                 100.0 |                84.947655 |               98.814229 |      0.0 |      0.0 |  0.129445 |       0.0 |      0.0 | -0.025712 |       0.0 |          0.0 |     0.739696 | 0.975359 | 0.975359 |      0.0 |      0.0 | -0.068585 |  -0.26561 |  -0.32577 | 0.998246 | 0.298949 | 1.406844 | 0.497745 |  0.0 |  0.0 | 0.0 |  0.0 | 0.497745 | nan | nan |      0.0 |  1.110002 |  1.587303 |       0.0 |  1.03625 |  0.046865 | 92.177595 | 0.497745 |      0.0 |   0.0 |   0.0 |  0.0 | 0.160927 | 0.191684 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 18:00:00 |  0.0 |  5.1 |              0.0 |             81.7 |       7.9 |               100.3 | nan | nan |    4.5125 |                93.475 |                   0.1 |   0.0 |  5.1 |    4.5125 |    6.332571 |         5.721357 |           6.332571 |          7.9 |                 0.879017 |                      0.843583 |                      0.61078 |                      0.061412 |                           0.059227 |             0.718157 |                  0.788539 |      99.581843 |   1.252331 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.031379 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.348982 |                  1.107494 |                 0.008574 |    -0.046146 |        -0.008574 |         -0.758512 |             26.962743 |                 100.0 |                83.747795 |               98.814229 |      0.0 |      0.0 |  0.087698 |       0.0 |      0.0 | -0.030943 |       0.0 |          0.0 |     0.739143 | 0.888357 | 0.888357 |      0.0 |      0.0 |  -0.08254 | -0.222746 | -0.296711 |  0.99709 | 0.298256 | 1.295115 | 0.458153 |  0.0 |  0.0 | 0.0 |  0.0 | 0.458153 | nan | nan |      0.0 |  0.225966 |  0.323132 |       0.0 | 1.043558 |  0.077765 | 93.014557 | 0.458153 |      0.0 |   0.0 |   0.0 |  0.0 | 0.169114 | 0.172633 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 19:00:00 |  0.0 |  4.8 |              0.0 |             83.4 |       7.1 |               100.3 | nan | nan |  4.654167 |             92.854167 |                   0.1 |   0.0 |  4.8 |  4.654167 |    5.691298 |         5.831576 |           5.691298 |          7.1 |                 0.860763 |                       0.85201 |                      0.61078 |                      0.060288 |                           0.059747 |             0.717876 |                  0.791127 |      99.582124 |   1.253684 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.031385 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.349716 |                  1.107885 |                  0.01332 |    -0.046162 |         -0.01332 |         -0.758168 |             30.000798 |                 100.0 |                82.645754 |               98.814229 |      0.0 |      0.0 |  0.070509 |       0.0 | 0.025175 | -0.028252 |       0.0 |          0.0 |     0.738594 | 0.760344 | 0.225966 |      0.0 |      0.0 | -0.075361 | -0.191913 | -0.253955 | 0.995933 | 0.297565 | 0.351384 | 0.124178 |  0.0 |  0.0 | 0.0 |  0.0 | 0.124178 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 | 1.050866 |  0.108665 | 93.216589 | 0.124178 |      0.0 |   0.0 |   0.0 |  0.0 | 0.172248 | 0.155475 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 20:00:00 |  0.0 |  4.9 |              0.0 |             82.8 |       6.4 |               100.3 | nan | nan |  4.791667 |             92.191667 |                   0.1 |   0.0 |  4.9 |  4.791667 |    5.130184 |         5.854956 |           5.130184 |          6.4 |                  0.86681 |                       0.86026 |                          0.0 |                       0.06066 |                           0.060257 |             0.717719 |                  0.793088 |      99.582281 |   1.253234 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.046348 |       0.0 |  -0.10584 |                   0.0 |                       0.0 |                   1.204044 |                  1.108737 |                      0.0 |    -0.046197 |              0.0 |          0.095308 |             33.282135 |                 100.0 |                82.380103 |               98.814229 |      0.0 |      0.0 |  0.069028 |       0.0 | 0.026279 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.990816 | 0.294515 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  0.245405 |  93.19031 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.169395 | 0.140023 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 21:00:00 |  0.0 |  4.9 |              0.0 |             85.8 |       7.2 |               100.3 | nan | nan |    4.9125 |             91.616667 |                   0.1 |   0.0 |  4.9 |    4.9125 |    5.771457 |         5.908395 |           5.771457 |          7.2 |                  0.86681 |                      0.867569 |                          0.0 |                       0.06066 |                           0.060707 |             0.743723 |                  0.794837 |      99.556277 |   1.253111 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.046335 |       0.0 |  -0.10584 |                   0.0 |                       0.0 |                   1.204044 |                  1.109465 |                      0.0 |    -0.046228 |              0.0 |          0.094579 |              29.58412 |                 100.0 |                82.414649 |               98.814229 |      0.0 |      0.0 |  0.063746 |       0.0 | 0.022605 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.985698 |  0.29148 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  0.382145 | 93.167705 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.16502 | 0.126106 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 22:00:00 |  0.1 |  4.7 |              0.0 |             87.6 |       7.0 |               100.3 | nan | nan |  5.008333 |                91.125 |                   0.1 | 0.094 |  4.7 |  5.008333 |    5.611139 |         5.948475 |           5.611139 |          7.0 |                 0.854752 |                      0.873404 |                          0.0 |                      0.059917 |                           0.061066 |             0.748763 |                  0.795889 |      99.551237 |   1.253989 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |   0.04632 |       0.0 |  -0.10152 |                   0.0 |                       0.0 |                   1.204044 |                   1.11033 |                      0.0 |    -0.046264 |              0.0 |          0.093714 |             30.429381 |                 100.0 |                82.444367 |               98.814229 |    0.034 |      0.0 |  0.053758 |  0.053758 |      0.0 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.980742 | 0.288557 |    0.034 | 0.011734 |  0.0 |  0.0 | 0.0 |  0.0 | 0.011734 | nan | nan | 0.006242 |       0.0 |       0.0 |       0.0 |      nan |  0.514565 | 93.189971 | 0.011734 |      0.0 |   0.0 |   0.0 |  0.0 | 0.160909 | 0.113573 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 23:00:00 |  0.1 |  3.8 |              0.0 |             85.6 |       6.8 |               100.2 | nan | nan |    5.0625 |             90.558333 |                   0.1 | 0.094 |  3.8 |    5.0625 |     5.45082 |         5.968515 |            5.45082 |          6.8 |                 0.802302 |                      0.876717 |                          0.0 |                      0.056666 |                            0.06127 |              0.68677 |                   0.79394 |       99.51323 |   1.257101 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.046273 |       0.0 |  -0.08208 |                   0.0 |                       0.0 |                   1.204044 |                  1.112997 |                      0.0 |    -0.046375 |              0.0 |          0.091047 |             31.324363 |                 100.0 |                82.415094 |               98.814229 | 0.040242 |      0.0 |  0.058512 |  0.058512 |      0.0 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.976514 | 0.286074 | 0.040242 | 0.013797 |  0.0 |  0.0 | 0.0 |  0.0 | 0.013797 | nan | nan | 0.001488 |       0.0 |       0.0 |       0.0 |      nan |  0.627545 | 93.216416 | 0.013797 |      0.0 |   0.0 |   0.0 |  0.0 | 0.157083 | 0.102285 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 00:00:00 |  0.0 |  3.9 |              0.0 |             85.3 |       6.4 |               100.2 | nan | nan |    5.1125 |             89.979167 |                   0.1 |   0.0 |  3.9 |    5.1125 |    5.130184 |         5.995234 |           5.130184 |          6.4 |                 0.807986 |                      0.879785 |                          0.0 |                       0.05702 |                           0.061459 |             0.689212 |                  0.791623 |      99.510788 |   1.256635 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |   0.04605 |       0.0 |  -0.08424 |                   0.0 |                       0.0 |                   1.196665 |                  1.116023 |                      0.0 |    -0.046501 |              0.0 |          0.080642 |             33.282135 |                 100.0 |                82.380329 |               98.814229 |      0.0 |      0.0 |  0.056539 |  0.001488 |  0.02061 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.972205 | 0.283555 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  0.742685 | 93.195806 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.153203 | 0.092119 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 01:00:00 |  0.0 |  3.7 |              0.0 |             86.2 |       6.3 |               100.2 | nan | nan |  5.133333 |             89.420833 |                   0.1 |   0.0 |  3.7 |  5.133333 |    5.050025 |         6.031974 |           5.050025 |          6.3 |                 0.796653 |                      0.881066 |                          0.0 |                      0.056315 |                           0.061538 |             0.686715 |                  0.787857 |      99.513285 |   1.257555 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.045986 |       0.0 |  -0.07992 |                   0.0 |                       0.0 |                   1.196665 |                  1.119656 |                      0.0 |    -0.046652 |              0.0 |          0.077009 |             33.810423 |                 100.0 |                82.407423 |               98.814229 |      0.0 |      0.0 |  0.051821 |       0.0 | 0.019519 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.968057 | 0.281141 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  0.853505 | 93.176287 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.149245 | 0.082963 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 02:00:00 |  0.1 |  3.5 |              0.0 |             89.3 |       5.7 |               100.2 | nan | nan |  5.116667 |             88.983333 |                   0.1 | 0.094 |  3.5 |  5.116667 |     4.56907 |         6.031974 |            4.56907 |          5.7 |                  0.78546 |                      0.880041 |                          0.0 |                      0.055617 |                           0.061475 |             0.701415 |                   0.78309 |      99.498585 |   1.258394 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.045919 |       0.0 |   -0.0756 |                   0.0 |                       0.0 |                   1.196665 |                  1.123475 |                      0.0 |    -0.046811 |              0.0 |           0.07319 |             37.369415 |                 100.0 |                82.433085 |               98.814229 |    0.034 |      0.0 |  0.036159 |  0.036159 |      0.0 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.964072 |  0.27883 |    0.034 | 0.011429 |  0.0 |  0.0 | 0.0 |  0.0 | 0.011429 | nan | nan | 0.023841 |       0.0 |       0.0 |       0.0 |      nan |  0.960005 | 93.198857 | 0.011429 |      0.0 |   0.0 |   0.0 |  0.0 | 0.145538 | 0.074718 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 03:00:00 |  0.0 |  3.3 |              0.0 |             91.4 |       5.2 |               100.3 | nan | nan |  5.070833 |             88.633333 |                   0.1 |   0.0 |  3.3 |  5.070833 |    4.168274 |         5.985215 |           4.168274 |          5.2 |                 0.774405 |                      0.877228 |                          0.0 |                      0.054927 |                           0.061302 |             0.707807 |                  0.777516 |      99.592193 |   1.260534 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.045849 |       0.0 |  -0.07128 |                   0.0 |                       0.0 |                   1.196665 |                  1.127474 |                      0.0 |    -0.046978 |              0.0 |           0.06919 |             40.962628 |                 100.0 |                82.403412 |               98.814229 |      0.0 |      0.0 |  0.026365 |  0.023841 | 0.001055 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.960248 | 0.276623 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.062185 | 93.197803 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.141926 | 0.067292 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 04:00:00 |  0.0 |  3.3 |              0.0 |             90.9 |       4.6 |               100.4 | nan | nan |  5.008333 |             88.295833 |                   0.1 |   0.0 |  3.3 |  5.008333 |     3.68732 |         5.905055 |            3.68732 |          4.6 |                 0.774405 |                      0.873404 |                          0.0 |                      0.054927 |                           0.061066 |             0.703935 |                  0.771179 |      99.696065 |   1.261813 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.045773 |       0.0 |  -0.07128 |                   0.0 |                       0.0 |                   1.196665 |                  1.131856 |                      0.0 |    -0.047161 |              0.0 |          0.064808 |              46.30558 |                 100.0 |                82.404798 |               98.814229 |      0.0 |      0.0 |  0.024822 |       0.0 | 0.011075 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.956423 | 0.274424 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.164365 | 93.186728 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.13826 | 0.060604 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 05:00:00 |  0.0 |  3.0 |              0.0 |             89.2 |       4.9 |               100.5 | nan | nan |  4.933333 |             87.945833 |                   0.1 |   0.0 |  3.0 |  4.933333 |    3.927797 |         5.824896 |           3.927797 |          4.9 |                 0.758081 |                      0.868834 |                          0.0 |                      0.053905 |                           0.060785 |             0.676208 |                  0.764103 |      99.823792 |   1.264577 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.045688 |       0.0 |   -0.0648 |                   0.0 |                       0.0 |                   1.196665 |                  1.136687 |                      0.0 |    -0.047362 |              0.0 |          0.059978 |             43.470544 |                 100.0 |                82.419358 |               98.814229 |      0.0 |      0.0 |  0.030917 |       0.0 | 0.013284 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.952842 | 0.272372 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.260065 | 93.173444 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.134688 |  0.05458 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 06:00:00 |  0.0 |  1.9 |              0.0 |             85.6 |       6.3 |               100.6 | nan | nan |       4.8 |             87.454167 |                   0.1 |   0.0 |  1.9 |       4.8 |    5.050025 |         5.794837 |           5.050025 |          6.3 |                  0.70079 |                      0.860763 |                          0.0 |                      0.050296 |                           0.060288 |             0.599876 |                  0.752773 |     100.000124 |   1.271267 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.045555 |       0.0 |  -0.04104 |                   0.0 |                       0.0 |                   1.196665 |                  1.144252 |                      0.0 |    -0.047677 |              0.0 |          0.052412 |             33.810423 |                 100.0 |                82.436821 |               98.814229 |      0.0 |      0.0 |  0.050119 |       0.0 | 0.018339 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.950149 | 0.270835 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.332005 | 93.155106 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.131209 | 0.049156 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 07:00:00 |  0.0 |  1.2 |              0.0 |             86.3 |       5.8 |               100.8 | nan | nan |    4.6375 |               87.0375 |                   0.1 |   0.0 |  1.2 |    4.6375 |    4.649229 |         5.734717 |           4.649229 |          5.8 |                 0.666353 |                      0.851015 |                          0.0 |                       0.04811 |                           0.059686 |             0.575062 |                  0.740702 |     100.224938 |   1.277169 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.045419 |       0.0 |  -0.02592 |                   0.0 |                       0.0 |                   1.196665 |                  1.152031 |                      0.0 |    -0.048001 |              0.0 |          0.044634 |             36.725115 |                 100.0 |                82.460932 |               98.814229 |      0.0 |      0.0 |  0.042707 |       0.0 | 0.016244 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.948023 | 0.269624 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.388825 | 93.138862 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.12782 |  0.04427 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 08:00:00 |  0.0 |  0.8 |              0.0 |             91.2 |       6.1 |               100.9 | nan | nan |  4.445833 |             86.870833 |                   0.1 |   0.0 |  0.8 |  4.445833 |    4.889707 |         5.694638 |           4.889707 |          6.1 |                 0.647352 |                      0.839643 |                          0.0 |                      0.046897 |                           0.058983 |             0.590385 |                  0.729404 |     100.309615 |   1.280232 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.339738 |                      7.386143 |             1.329627 |  0.244956 |        0.003171 |                0.003219 |  0.029403 |       0.0 |  -0.01728 |              0.002897 |                  0.002897 |                   1.196665 |                  1.158654 |                      0.0 |     -0.04538 |              0.0 |          0.038011 |             34.918962 |                 100.0 |                82.482289 |               98.498383 |      0.0 |      0.0 |  0.026053 |       0.0 | 0.009575 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |  0.94622 |   0.2686 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.437005 | 93.129287 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.124518 |  0.03987 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 09:00:00 |  0.0 |  0.8 |              0.0 |             87.5 |       5.9 |               101.0 | nan | nan |  4.258333 |             86.504167 |                   0.1 |   0.0 |  0.8 |  4.258333 |    4.729388 |         5.654558 |           4.729388 |          5.9 |                 0.647352 |                      0.828647 |                          0.0 |                      0.046897 |                           0.058302 |             0.566433 |                  0.716814 |     100.433567 |   1.281619 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 |  4.849391 |        0.062779 |                0.063765 | -0.001183 |       0.0 |  -0.01728 |              0.057389 |                  0.057389 |                   1.196665 |                  1.166643 |                      0.0 |     0.008779 |              0.0 |          0.030022 |             36.102655 |                 100.0 |                82.494879 |               97.892093 |      0.0 |      0.0 |  0.040566 |       0.0 |  0.01527 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.944417 | 0.267577 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.485185 | 93.114017 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.121302 | 0.035908 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 10:00:00 |  0.1 |  0.5 |              0.0 |             84.5 |       6.4 |               101.1 | nan | nan |      4.05 |             86.045833 |                   0.1 | 0.094 |  0.5 |      4.05 |    5.130184 |         5.631178 |           5.130184 |          6.4 |                 0.633417 |                      0.816579 |                     0.555459 |                      0.046005 |                           0.057554 |             0.535238 |                  0.702632 |     100.564762 |   1.284447 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 14.852029 |         0.19227 |                0.195489 |  0.040966 |       0.0 |   0.02796 |              0.020003 |                  0.020003 |                   0.136052 |                  1.175846 |                 0.100918 |     -0.02899 |        -0.080915 |         -1.039794 |             33.282135 |                 100.0 |                82.514957 |               97.894961 |    0.034 |  0.01275 |  0.054006 |  0.054006 |      0.0 |  0.004903 |  0.000025 |    -1.294436 |     0.897677 |      0.0 |      0.0 | 0.000192 | 0.000192 |  0.013079 | -0.065945 |  0.028049 | 0.944307 | 0.267515 | 0.015768 | 0.005135 |  0.0 |  0.0 | 0.0 |  0.0 | 0.005135 | nan | nan | 0.005994 |  0.009462 |  0.013329 |       0.0 |  0.00743 |  1.488126 | 93.124649 | 0.005135 |      0.0 |   0.0 |   0.0 |  0.0 | 0.118235 | 0.032339 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 11:00:00 |  0.0 |  0.3 |              0.0 |             76.7 |       6.8 |               101.2 | nan | nan |  3.808333 |                85.425 |                   0.1 |   0.0 |  0.3 |  3.808333 |     5.45082 |         5.561039 |            5.45082 |          6.8 |                 0.624275 |                      0.802774 |                     0.535467 |                      0.045418 |                           0.056696 |             0.478819 |                   0.68577 |     100.721181 |   1.286932 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.269583 |        0.288296 |                0.293415 |  0.041292 |       0.0 |  0.038672 |              0.030684 |                  0.030684 |                   0.139046 |                   1.18712 |                 0.111622 |    -0.018779 |        -0.080938 |         -1.048074 |             31.324363 |                 100.0 |                82.500977 |               97.892964 |      0.0 |      0.0 |  0.085936 |  0.005994 | 0.027365 |  0.013329 |       0.0 |    -1.790392 |     0.895425 |      0.0 |      0.0 | 0.000193 | 0.000193 |  0.038481 | -0.080682 |  0.038737 | 0.944598 | 0.267679 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 | 0.014738 |  1.480353 | 93.097283 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.115247 | 0.029125 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 12:00:00 |  0.0 | -0.1 |              0.0 |             74.8 |       7.0 |               101.3 | nan | nan |    3.5375 |             84.791667 |                   0.1 |   0.0 | -0.1 |    3.5375 |    5.611139 |          5.47754 |           5.611139 |          7.0 |                 0.606339 |                      0.787548 |                          0.0 |                      0.044264 |                           0.055747 |             0.453542 |                  0.667775 |     100.846458 |   1.290215 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.293018 |        0.288599 |                0.293995 |  0.000099 |       0.0 |   0.00216 |              0.264596 |                  0.264596 |                   1.196665 |                  1.199167 |                      0.0 |      0.21463 |              0.0 |         -0.002502 |             30.429381 |                 100.0 |                82.536961 |               97.898103 |      0.0 |      0.0 |  0.123245 |       0.0 | 0.041152 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.943522 |  0.26707 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.509093 | 93.056132 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.11227 |  0.02623 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 13:00:00 |  0.0 | -0.4 |              0.0 |             69.7 |       6.3 |               101.4 | nan | nan |  3.254167 |                84.125 |                   0.1 |   0.0 | -0.4 |  3.254167 |    5.050025 |         5.424101 |           5.050025 |          6.3 |                 0.593188 |                      0.771892 |                          0.0 |                      0.043416 |                           0.054769 |             0.413452 |                  0.649354 |     100.986548 |   1.293105 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 21.019348 |        0.272111 |                0.277484 |  0.000592 |       0.0 |   0.00864 |              0.249736 |                  0.249736 |                   1.196665 |                   1.21168 |                      0.0 |     0.199249 |              0.0 |         -0.015015 |             33.810423 |                 100.0 |                82.591077 |               97.905825 |      0.0 |      0.0 |  0.126948 |       0.0 | 0.045092 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.942689 | 0.266599 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.531353 |  93.01104 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.10937 | 0.023623 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 14:00:00 |  0.0 | -0.6 |              0.2 |             66.9 |       7.3 |               101.5 | nan | nan |  2.958333 |             83.483333 |                   0.3 |   0.0 | -0.6 |  2.958333 |    5.851616 |         5.347282 |           5.851616 |          7.3 |                  0.58456 |                      0.755838 |                          0.0 |                      0.042857 |                           0.053764 |             0.391071 |                  0.630999 |     101.108929 |    1.29544 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.399736 | 11.654171 |         0.21122 |                0.216915 |  0.003554 |       0.0 |   0.01296 |              0.195223 |                  0.195223 |                   1.259762 |                  1.349977 |                      0.0 |     0.138974 |              0.0 |         -0.090215 |             29.178858 |                 100.0 |                 82.65038 |               97.914276 |      0.0 |      0.0 |   0.14235 |       0.0 | 0.045883 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.942017 | 0.266219 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.549293 | 92.965157 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.106545 | 0.021275 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 15:00:00 |  0.0 | -0.8 |              0.2 |             64.4 |       7.5 |               101.5 | nan | nan |    2.6625 |             82.729167 |                   0.4 |   0.0 | -0.8 |    2.6625 |    6.011934 |         5.293842 |           6.011934 |          7.5 |                 0.576044 |                       0.74008 |                          0.0 |                      0.042305 |                           0.052775 |             0.370972 |                  0.612262 |     101.129028 |   1.296488 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |              1.43479 |  2.813062 |        0.050984 |                0.053414 |  0.005391 |       0.0 |   0.01728 |              0.048073 |                  0.048073 |                   1.291311 |                  1.428152 |                      0.0 |    -0.011433 |              0.0 |         -0.136842 |             28.400756 |                 100.0 |                82.710732 |               97.922866 |      0.0 |      0.0 |   0.13117 |       0.0 | 0.041397 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.941508 | 0.265931 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.562913 | 92.923761 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.103793 | 0.019161 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 16:00:00 |  0.0 | -1.3 |              0.2 |             65.4 |       7.0 |               101.7 | nan | nan |    2.3625 |             82.029167 |                   0.6 |   0.0 | -1.3 |    2.3625 |    5.611139 |         5.237063 |           5.611139 |          7.0 |                 0.555231 |                      0.724396 |                          0.0 |                      0.040952 |                           0.051788 |             0.363121 |                  0.594216 |     101.336879 |   1.301474 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.046405 |                      7.386143 |             1.504898 |  0.004444 |        0.000553 |                0.000608 |  0.039339 |       0.0 |   0.02808 |              0.000547 |                  0.000547 |                   1.354408 |                  1.571379 |                      0.0 |    -0.064927 |              0.0 |         -0.216971 |             30.429381 |                 100.0 |                82.765189 |               98.772872 |      0.0 |      0.0 |  0.113782 |       0.0 | 0.037061 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.941402 | 0.265871 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.565733 | 92.886699 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.101112 | 0.017256 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 17:00:00 |  0.0 | -1.6 |              0.0 |             67.6 |       6.0 |               101.7 | nan | nan |  2.066667 |             81.658333 |                   0.6 |   0.0 | -1.6 |  2.066667 |    4.809547 |         5.140204 |           4.809547 |          6.0 |                 0.543065 |                      0.709217 |                          0.0 |                      0.040158 |                           0.050829 |             0.367112 |                  0.579135 |     101.332888 |   1.302892 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.040616 |       0.0 |   0.03456 |                   0.0 |                       0.0 |                   1.354408 |                  1.583999 |                      0.0 |       -0.066 |              0.0 |          -0.22959 |             35.500944 |                 100.0 |                82.813948 |               98.814229 |      0.0 |      0.0 |  0.089868 |       0.0 |  0.03213 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.941539 | 0.265949 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.562073 | 92.854569 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0985 | 0.015541 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 18:00:00 |  0.0 | -2.0 |              0.0 |             70.2 |       4.8 |               101.8 | nan | nan |  1.770833 |             81.179167 |                   0.6 |   0.0 | -2.0 |  1.770833 |    3.847638 |         5.036665 |           3.847638 |          4.8 |                  0.52721 |                       0.69432 |                          0.0 |                       0.03912 |                           0.049887 |             0.370102 |                  0.563643 |     101.429898 |   1.306085 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |   0.04038 |       0.0 |    0.0432 |                   0.0 |                       0.0 |                   1.354408 |                    1.5975 |                      0.0 |    -0.066562 |              0.0 |         -0.243092 |             44.376181 |                 100.0 |                82.856223 |               98.814229 |      0.0 |      0.0 |  0.064528 |       0.0 | 0.026194 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |    0.942 | 0.266209 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.549773 | 92.828376 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.095956 | 0.013997 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 19:00:00 |  0.0 | -2.1 |              0.0 |             70.4 |       4.2 |               101.9 | nan | nan |  1.483333 |               80.6375 |                   0.6 |   0.0 | -2.1 |  1.483333 |    3.366683 |         4.939806 |           3.366683 |          4.2 |                 0.523311 |                      0.680107 |                          0.0 |                      0.038864 |                           0.048985 |             0.368411 |                  0.548422 |     101.531589 |    1.30786 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |   0.04014 |       0.0 |   0.04536 |                   0.0 |                       0.0 |                   1.354408 |                  1.611195 |                      0.0 |    -0.067133 |              0.0 |         -0.256786 |             50.715635 |                 100.0 |                 82.89069 |               98.814229 |      0.0 |      0.0 |   0.05568 |       0.0 | 0.024303 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.942541 | 0.266515 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.535313 | 92.804073 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.093477 | 0.012606 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 20:00:00 |  0.0 | -2.2 |              0.0 |             69.5 |       4.9 |               101.9 | nan | nan |    1.1875 |             80.083333 |                   0.6 |   0.0 | -2.2 |    1.1875 |    3.927797 |         4.889707 |           3.927797 |          4.9 |                 0.519437 |                      0.665752 |                          0.0 |                       0.03861 |                           0.048071 |             0.361009 |                  0.533156 |     101.538991 |   1.308379 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.039896 |       0.0 |   0.04752 |                   0.0 |                       0.0 |                   1.354408 |                  1.625096 |                      0.0 |    -0.067712 |              0.0 |         -0.270688 |             43.470544 |                 100.0 |                82.922671 |               98.814229 |      0.0 |      0.0 |  0.066696 |       0.0 | 0.026688 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.943163 | 0.266867 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.518693 | 92.777385 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.091063 | 0.011353 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 21:00:00 |  0.0 | -2.6 |              0.0 |             71.6 |       4.0 |               102.0 | nan | nan |     0.875 |             79.491667 |                   0.6 |   0.0 | -2.6 |     0.875 |    3.206365 |         4.782828 |           3.206365 |          4.0 |                 0.504195 |                      0.650878 |                          0.0 |                      0.037607 |                           0.047123 |             0.361003 |                  0.517394 |     101.638997 |   1.311601 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.039641 |       0.0 |   0.05616 |                   0.0 |                       0.0 |                   1.354408 |                  1.639665 |                      0.0 |    -0.068319 |              0.0 |         -0.285257 |             53.251417 |                 100.0 |                82.957793 |               98.814229 |      0.0 |      0.0 |  0.049278 |       0.0 | 0.021944 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.944108 | 0.267402 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.493433 | 92.755441 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.08871 | 0.010224 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 22:00:00 |  0.0 | -3.1 |              0.0 |             74.5 |       3.4 |               102.0 | nan | nan |      0.55 |             78.945833 |                   0.6 |   0.0 | -3.1 |      0.55 |     2.72541 |         4.662589 |            2.72541 |          3.4 |                 0.485698 |                      0.635721 |                          0.0 |                      0.036384 |                           0.046153 |             0.361845 |                  0.501875 |     101.638155 |   1.314025 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.039391 |       0.0 |   0.06696 |                   0.0 |                       0.0 |                   1.354408 |                  1.653965 |                      0.0 |    -0.068915 |              0.0 |         -0.299557 |             62.648725 |                 100.0 |                82.986674 |               98.814229 |      0.0 |      0.0 |  0.036056 |       0.0 | 0.017347 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.945458 | 0.268167 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.457373 | 92.738094 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.086419 | 0.009208 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 23:00:00 |  0.0 | -3.9 |              0.0 |             80.2 |       3.4 |               102.0 | nan | nan |  0.229167 |             78.720833 |                   0.6 |   0.0 | -3.9 |  0.229167 |     2.72541 |          4.54903 |            2.72541 |          3.4 |                 0.457351 |                      0.621065 |                          0.0 |                      0.034499 |                           0.045212 |             0.366795 |                  0.488908 |     101.633205 |   1.317905 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.039199 |       0.0 |   0.08424 |                   0.0 |                       0.0 |                   1.354408 |                  1.664915 |                      0.0 |    -0.069371 |              0.0 |         -0.310507 |             62.648725 |                 100.0 |                83.009506 |               98.814229 |      0.0 |      0.0 |  0.025949 |       0.0 | 0.012381 |       0.0 |       0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.947454 | 0.269301 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.404033 | 92.725713 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.084187 | 0.008293 |   0.0 |   0.0 |  0.0 |   nan |    nan |

.. _lland_v3_nadelw_winter_hourly:

conifers (winter)
-----------------

Through comparing the hourly simulation results for |ACKER|
(:ref:`lland_v3_acker_winter_hourly`) and |NADELW| (this example), you can
see the shading effect of the tree-canopies.  Initially, during the
nighttime, the decreased longwave radiation loss (|NetLongwaveRadiationSnow|)
under the canopy results in an earlier start of snow-melt.  But then, during
the daytime, the decreased shortwave radiation gain (|NetShortwaveRadiationSnow|)
results in a lower melting rate, so that the complete melting process finishes
even later for |NADELW| than for |ACKER|:

.. integration-test::

    >>> lnk(NADELW)
    >>> test('lland_v3_nadelw_winter_hourly',
    ...      axis1=(inputs.nied, fluxes.wada),
    ...      axis2=(states.waes, states.wats),
    ...      use_conditions=conditions_nadelw_winter)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration |  nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |         g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes |     sbes |      evpo |       evi |      evb |       evs |     wnied | tempssurface | actualalbedo |  schmpot |     schm |  gefrpot |     gefr |  wlatsnow | wsenssnow |     wsurf |      sff |      fvg |     wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz | qah |  qa |     inzp |      wats |      waes |     esnow |     taus |      ebdn |      bowa |    qdgz1 |    qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet | outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-12-10 00:00:00 |  0.0 | -3.7 |              0.0 |             87.5 |       3.4 |               101.7 | nan | nan |    -2.025 |             96.870833 |                  4.41 |   0.0 | -3.7 |    -2.025 |     2.72541 |         2.684392 |           1.206967 |          3.4 |                 0.464297 |                      0.526233 |                     0.529351 |                      0.034962 |                           0.039056 |              0.40626 |                  0.509766 |      101.29374 |   1.312855 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 |  0.050674 |       0.0 |   0.04202 |                   0.0 |                       0.0 |                   0.723841 |                  4.014396 |                 0.048267 |    -0.167266 |        -0.048267 |         -3.290555 |             27.647059 |                 100.0 |              3113.156712 |               69.444444 | 0.145708 | 0.145708 |  0.025898 |  0.025898 |      0.0 |  0.011494 | -0.001127 |    -1.945364 |     0.747923 |      0.0 |      0.0 |  0.18822 |  0.18822 |  0.030659 |  0.024832 |  0.103759 | 0.990633 | 0.294406 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.174102 | 11.726023 | 13.490212 |       0.0 | 0.933427 |  0.250282 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 01:00:00 |  0.0 | -4.8 |              0.0 |             90.7 |       2.9 |               101.8 | nan | nan | -2.158333 |                96.825 |                  4.41 |   0.0 | -4.8 | -2.158333 |    2.324615 |         2.647918 |           1.029472 |          2.9 |                 0.427219 |                      0.521048 |                     0.521839 |                      0.032479 |                           0.038716 |             0.387487 |                  0.504505 |     101.412513 |   1.319627 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 |  0.050689 |       0.0 |  0.046179 |                   0.0 |                       0.0 |                   0.725602 |                  4.024777 |                 0.048365 |    -0.167699 |        -0.048365 |         -3.299174 |             32.413793 |                 100.0 |             20046.551039 |               69.444444 |      0.0 |      0.0 |  0.010274 |  0.010274 |      0.0 |  0.011639 |       0.0 |     -2.13791 |      0.74731 |      0.0 |      0.0 | 0.204149 | 0.204149 |  0.031047 |  0.034953 |  0.114365 | 0.991205 | 0.294746 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.163828 | 11.919879 | 13.478573 |       0.0 | 0.940735 |  0.235003 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 02:00:00 |  0.0 | -5.6 |              0.0 |             89.7 |       2.5 |               101.8 | nan | nan |    -2.325 |             96.604167 |                  4.41 |   0.0 | -5.6 |    -2.325 |    2.003978 |         2.606417 |           0.887476 |          2.5 |                 0.401924 |                      0.514631 |                     0.514853 |                      0.030771 |                           0.038294 |             0.360526 |                  0.497155 |     101.439474 |   1.323706 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 |  0.050715 |       0.0 |  0.050095 |                   0.0 |                       0.0 |                   0.727351 |                  4.040774 |                 0.049173 |    -0.168366 |        -0.049173 |         -3.313423 |                  37.6 |                 100.0 |             28069.018521 |               69.444444 |      0.0 |      0.0 |  0.008863 |  0.008863 |      0.0 |  0.012537 |       0.0 |    -2.319202 |     0.746701 |      0.0 |      0.0 |   0.2183 |   0.2183 |  0.033441 |  0.040393 |  0.123007 | 0.991923 | 0.295174 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.154965 | 12.126889 | 13.466036 |       0.0 | 0.948043 |  0.215808 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 03:00:00 |  0.0 | -6.2 |              0.0 |             91.3 |       3.0 |               101.9 | nan | nan | -2.420833 |                96.325 |                  4.41 |   0.0 | -6.2 | -2.420833 |    2.404774 |         2.602449 |           1.064971 |          3.0 |                 0.383833 |                      0.510972 |                     0.505858 |                       0.02954 |                           0.038053 |             0.350439 |                  0.492194 |     101.549561 |   1.328036 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 |  0.050734 |       0.0 |  0.055206 |                   0.0 |                       0.0 |                   0.729086 |                  4.052814 |                 0.049248 |    -0.168867 |        -0.049248 |         -3.323727 |             31.333333 |                 100.0 |             28069.018521 |               69.444444 |      0.0 |      0.0 |   0.00851 |   0.00851 |      0.0 |  0.013674 |       0.0 |    -2.555835 |     0.746096 |      0.0 |      0.0 |  0.23685 |  0.23685 |  0.036474 |  0.048592 |  0.134314 | 0.992833 | 0.295715 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.146455 | 12.351185 | 13.452362 |       0.0 | 0.955351 |  0.191502 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 04:00:00 |  0.0 | -7.3 |              0.0 |             94.2 |       2.5 |               102.0 | nan | nan | -2.558333 |               96.2125 |                  4.41 |   0.0 | -7.3 | -2.558333 |    2.003978 |         2.623448 |           0.887476 |          2.5 |                 0.352531 |                      0.505764 |                     0.499391 |                      0.027395 |                            0.03771 |             0.332084 |                  0.486608 |     101.667916 |   1.334932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 |  0.050752 |       0.0 |  0.058931 |                   0.0 |                       0.0 |                   0.730809 |                  4.064535 |                 0.049397 |    -0.169356 |        -0.049397 |         -3.333726 |                  37.6 |                 100.0 |             28069.018521 |               69.444444 |      0.0 |      0.0 | -0.001315 | -0.001315 |      0.0 |  0.013591 |       0.0 |    -2.728282 |     0.745497 |      0.0 |      0.0 | 0.248522 | 0.248522 |  0.036253 |  0.056286 |  0.141937 | 0.993882 |  0.29634 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.147769 | 12.586977 | 13.438771 |       0.0 | 0.962659 |  0.163472 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 05:00:00 |  0.0 | -8.3 |              0.0 |             93.2 |       2.3 |               102.0 | nan | nan |    -2.725 |             96.004167 |                  4.41 |   0.0 | -8.3 |    -2.725 |     1.84366 |         2.596101 |           0.816478 |          2.3 |                 0.326061 |                      0.499513 |                     0.490498 |                      0.025563 |                           0.037298 |             0.303889 |                  0.479553 |     101.696111 |   1.340112 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 |  0.050777 |       0.0 |  0.064123 |                   0.0 |                       0.0 |                   0.732518 |                  4.080005 |                 0.050254 |        -0.17 |        -0.050254 |         -3.347487 |             40.869565 |                 100.0 |             28069.018521 |               69.444444 |      0.0 |      0.0 | -0.000401 | -0.000401 |      0.0 |  0.014656 |       0.0 |     -2.96865 |     0.744901 |      0.0 |      0.0 | 0.265517 | 0.265517 |  0.039093 |  0.063459 |  0.152806 | 0.995125 | 0.297082 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.148171 | 12.838478 | 13.424116 |       0.0 | 0.969967 |  0.130249 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 06:00:00 |  0.0 | -8.6 |              0.0 |             96.5 |       2.1 |               102.1 | nan | nan |    -2.925 |             96.004167 |                  4.41 |   0.0 | -8.6 |    -2.925 |    1.683342 |          2.56624 |            0.74548 |          2.1 |                 0.318472 |                      0.492103 |                     0.490225 |                      0.025034 |                           0.036808 |             0.307325 |                  0.472439 |     101.792675 |   1.342932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 |  0.050798 |       0.0 |  0.064284 |                   0.0 |                       0.0 |                   0.734215 |                  4.093695 |                 0.049796 |    -0.170571 |        -0.049796 |          -3.35948 |             44.761905 |                 100.0 |             28069.018521 |               69.444444 |      0.0 |      0.0 | -0.006667 | -0.006667 |      0.0 |  0.013871 |       0.0 |    -2.976098 |      0.74431 |      0.0 |      0.0 | 0.260938 | 0.260938 |     0.037 |  0.064641 |  0.151437 | 0.996375 | 0.297829 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.154837 | 13.085881 | 13.410245 |       0.0 | 0.977275 |  0.096865 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 07:00:00 |  0.0 | -8.5 |              0.0 |             95.9 |       1.4 |               102.1 | nan | nan |   -3.1625 |               96.0625 |                  3.76 |   0.0 | -8.5 |   -3.1625 |    1.122228 |         2.521333 |           0.496987 |          1.4 |                 0.320984 |                      0.483429 |                     0.498045 |                       0.02521 |                           0.036234 |             0.307823 |                  0.464394 |     101.792177 |   1.342422 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.641605 |       0.0 |             0.0 |                     0.0 |  0.050158 |       0.0 |  0.059711 |                   0.0 |                       0.0 |                   0.676981 |                  3.682358 |                 0.047666 |    -0.153432 |        -0.047666 |         -3.005377 |             67.142857 |                 100.0 |             28069.018521 |               69.444444 |      0.0 |      0.0 | -0.005725 | -0.005725 |      0.0 |   0.01263 |       0.0 |    -2.764404 |     0.743724 |      0.0 |      0.0 | 0.237602 | 0.237602 |  0.033689 |  0.057715 |   0.13907 | 0.997453 | 0.298474 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.160562 | 13.310935 | 13.397615 |       0.0 | 0.984583 |  0.068054 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 08:00:00 |  0.0 | -9.5 |              0.0 |             96.8 |       1.1 |               102.2 | nan | nan | -3.483333 |               96.2625 |                  2.76 |   0.0 | -9.5 | -3.483333 |     0.88175 |         2.408072 |           0.390489 |          1.1 |                 0.296635 |                      0.471927 |                     0.495211 |                      0.023505 |                            0.03547 |             0.287143 |                  0.454289 |     101.912857 |   1.348938 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.355232 |                       7.41713 |             2.287909 |  0.266196 |        0.003492 |                0.003987 |  0.028172 |       0.0 |  0.061361 |              0.001024 |                   0.00019 |                    0.58767 |                  3.039527 |                 0.044085 |    -0.125623 |        -0.043895 |         -2.451857 |             85.454545 |                 100.0 |             28069.018521 |              106.754223 |      0.0 |      0.0 | -0.006222 | -0.006222 |      0.0 |  0.012972 |       0.0 |     -2.84079 |     0.743141 |      0.0 |      0.0 | 0.239705 |  0.08668 |  0.034603 |  0.062924 |  0.141422 | 0.998593 | 0.299156 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.166785 | 13.384643 | 13.384643 |  -0.05111 | 0.991891 |  0.037593 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 09:00:00 |  0.0 | -9.2 |              0.0 |             94.3 |       1.0 |               102.2 | nan | nan | -3.816667 |             96.383333 |                  1.79 |   0.0 | -9.2 | -3.816667 |    0.801591 |         2.278972 |            0.35499 |          1.0 |                 0.303761 |                      0.460234 |                     0.473204 |                      0.024006 |                           0.034691 |             0.286447 |                  0.443589 |     101.913553 |   1.347408 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.944823 |  4.919884 |        0.064534 |                0.067188 | -0.007733 |       0.0 |  0.074463 |              0.017281 |                  0.003209 |                   0.500214 |                  2.412046 |                 0.039431 |    -0.083221 |        -0.036222 |         -1.911833 |                  94.0 |                 100.0 |             28069.018521 |             4298.215505 |      0.0 |      0.0 | -0.003922 | -0.003922 |      0.0 |  0.011392 |       0.0 |    -3.447378 |     0.742797 |      0.0 |      0.0 | 0.288732 |      0.0 |  0.030387 |  0.053182 |   0.11979 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.170707 | 13.373251 | 13.373251 | -0.096437 | 0.996231 | -0.005971 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 10:00:00 |  0.0 | -6.9 |              0.0 |             96.7 |       1.4 |               102.2 | nan | nan |   -4.0625 |               96.5375 |                  0.88 |   0.0 | -6.9 |   -4.0625 |    1.122228 |         2.200731 |           0.496987 |          1.4 |                 0.363641 |                      0.451775 |                     0.471295 |                      0.028159 |                           0.034126 |             0.351641 |                  0.436132 |     101.848359 |   1.335446 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.622959 | 14.873457 |        0.195096 |                0.202564 | -0.005653 | -0.019902 |  0.075195 |              0.052153 |                  0.009686 |                   0.417852 |                  1.815487 |                  0.03091 |    -0.023492 |        -0.021225 |         -1.397635 |             67.142857 |                 100.0 |             28069.018521 |             4298.215505 |      0.0 |      0.0 |  0.000829 |  0.000829 |      0.0 |  0.007944 |       0.0 |    -3.501169 |     0.742537 |      0.0 |      0.0 | 0.292988 |      0.0 |  0.021191 |  0.034201 |  0.076617 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.169878 | 13.365307 | 13.365307 | -0.097858 | 0.999534 | -0.050266 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 11:00:00 |  0.0 | -4.0 |              0.1 |             97.4 |       1.3 |               102.2 | nan | nan | -4.195833 |             96.570833 |                  0.39 |   0.0 | -4.0 | -4.195833 |    1.042069 |         2.098318 |           0.461488 |          1.3 |                 0.453913 |                      0.447245 |                     0.502925 |                      0.034269 |                           0.033823 |             0.442111 |                  0.431908 |     101.757889 |   1.320615 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.449648 | 22.189872 |        0.349279 |                0.385806 | -0.004528 | -0.167553 |  0.053271 |               0.09943 |                  0.018466 |                   0.373605 |                  1.493002 |                  0.02343 |     0.037222 |        -0.004965 |         -1.119397 |             72.307692 |                 100.0 |              3952.988639 |             2243.642922 |      0.0 |      0.0 |  0.007356 |  0.007356 |      0.0 |  0.003956 |       0.0 |    -2.633802 |     0.742279 |      0.0 |      0.0 | 0.220273 |      0.0 |  0.010551 |  0.013468 |  0.028984 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.162521 | 13.361352 | 13.361352 | -0.073571 | 1.002813 | -0.072637 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 12:00:00 |  0.0 | -2.8 |              0.0 |             93.9 |       1.3 |               102.2 | nan | nan |     -4.35 |               96.3625 |                  0.14 |   0.0 | -2.8 |     -4.35 |    1.042069 |          2.00007 |           0.461488 |          1.3 |                 0.496723 |                      0.442058 |                     0.525459 |                      0.037113 |                           0.033476 |             0.466423 |                  0.425978 |     101.733577 |   1.314635 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.361224 | 22.197307 |        0.291164 |                0.329304 | -0.003963 | -0.242123 |  0.038938 |              0.084965 |                  0.015779 |                   0.351214 |                  1.331032 |                 0.020561 |     0.029505 |        -0.004782 |         -0.979819 |             72.307692 |                 100.0 |              1813.687715 |             1354.658798 |      0.0 |      0.0 |  0.011733 |  0.011733 |      0.0 |   0.00384 |       0.0 |    -2.044829 |     0.741987 |      0.0 |      0.0 | 0.170965 |      0.0 |  0.010243 |  0.007445 |  0.022469 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.150789 | 13.357512 | 13.357512 | -0.057102 | 1.006539 | -0.080675 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 13:00:00 |  0.0 | -1.4 |              0.0 |             93.4 |       1.4 |               102.1 | nan | nan | -4.370833 |             96.116667 |                   0.1 |   0.0 | -1.4 | -4.370833 |    1.122228 |         1.900997 |           0.496987 |          1.4 |                 0.551149 |                      0.441361 |                     0.548501 |                      0.040686 |                           0.033429 |             0.514774 |                  0.424221 |     101.585226 |   1.306346 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 20.962863 |        0.274971 |                0.289712 | -0.003875 | -0.268918 |   0.02584 |              0.074843 |                  0.013899 |                      0.348 |                  1.306141 |                 0.018226 |     0.020421 |        -0.004327 |         -0.958141 |             67.142857 |                 100.0 |              1111.794179 |              926.445145 |      0.0 |      0.0 |  0.012789 |  0.012789 |      0.0 |  0.002239 |       0.0 |    -1.465237 |     0.741663 |      0.0 |      0.0 | 0.122471 |      0.0 |  0.005973 | -0.000656 |  0.009643 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |    2.138 | 13.355272 | 13.355272 | -0.040905 | 1.010687 | -0.075616 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 14:00:00 |  0.0 | -0.4 |              0.0 |             93.3 |       2.3 |               102.0 | nan | nan | -4.354167 |             95.854167 |                   0.1 |   0.0 | -0.4 | -4.354167 |     1.84366 |         1.856982 |           0.816478 |          2.3 |                 0.593188 |                      0.441918 |                     0.566519 |                      0.043416 |                           0.033467 |             0.553444 |                  0.423597 |     101.446556 |   1.300092 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 11.702396 |        0.153501 |                0.155234 | -0.003878 | -0.252053 |  0.016735 |               0.04016 |                  0.007458 |                   0.348493 |                  1.307301 |                  0.01667 |    -0.014311 |        -0.009212 |         -0.958808 |             40.869565 |                 100.0 |               871.594617 |              756.218729 |      0.0 |      0.0 |  0.015194 |  0.015194 |      0.0 |  0.001027 |       0.0 |    -1.026803 |     0.741297 |      0.0 |      0.0 |  0.08581 |      0.0 |  0.002739 | -0.007461 |   0.00449 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.122806 | 13.354245 | 13.354245 | -0.028661 | 1.015398 |  -0.06145 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 15:00:00 |  0.0 |  0.5 |              0.0 |             94.1 |       2.8 |               102.0 | nan | nan |      -4.3 |             95.608333 |                   0.1 |   0.0 |  0.5 |      -4.3 |    2.244455 |         1.842168 |           0.993973 |          2.8 |                 0.633417 |                      0.443734 |                     0.584724 |                      0.046005 |                           0.033588 |             0.596046 |                  0.424247 |     101.403954 |   1.295611 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 |  2.880181 |         0.03778 |                0.037152 | -0.003877 | -0.204835 |  0.008453 |              0.009627 |                  0.001788 |                   0.349042 |                  1.307629 |                 0.014965 |    -0.044858 |        -0.013177 |         -0.958586 |             33.571429 |                 100.0 |                728.08904 |              647.654425 |      0.0 |      0.0 |  0.012187 |  0.012187 |      0.0 | -0.000966 |       0.0 |    -0.596195 |     0.740889 |      0.0 |      0.0 |  0.04982 |      0.0 | -0.002576 | -0.014169 | -0.003567 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.110619 | 13.355211 | 13.355211 |  -0.01664 |  1.02067 | -0.039004 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 16:00:00 |  0.1 |  1.0 |              0.0 |             94.6 |       3.1 |               101.9 | nan | nan |      -4.2 |             95.395833 |                   0.1 | 0.094 |  1.0 |      -4.2 |    2.484933 |          1.86654 |            1.10047 |          3.1 |                 0.656792 |                      0.447104 |                     0.600356 |                        0.0475 |                           0.033814 |             0.621325 |                  0.426519 |     101.278675 |   1.291856 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.061898 |                       7.41713 |             1.347076 |  0.007844 |        0.000103 |                0.000101 |  0.043338 | -0.130013 |  0.002284 |              0.000026 |                  0.000005 |                    0.34965 |                   1.30686 |                 0.013978 |    -0.054426 |        -0.013974 |         -0.957211 |             30.322581 |                 100.0 |               666.767813 |               73.465121 | 0.004619 | 0.001155 |  0.018669 |  0.018669 |      0.0 | -0.001873 |  0.000017 |    -0.235772 |     0.740438 |      0.0 |      0.0 | 0.019715 | 0.003464 | -0.004997 | -0.016731 | -0.007754 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.181331 | 13.361703 | 13.361703 | -0.005428 | 1.026533 | -0.010388 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 17:00:00 |  0.0 |  1.1 |              0.0 |             94.4 |       3.5 |               101.8 | nan | nan | -4.091667 |             95.170833 |                   0.1 |   0.0 |  1.1 | -4.091667 |    2.805569 |         1.866772 |           1.242466 |          3.5 |                 0.661557 |                      0.450781 |                     0.610317 |                      0.047804 |                            0.03406 |              0.62451 |                  0.429012 |      101.17549 |     1.2901 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |   0.04645 | -0.034627 | -0.000523 |                   0.0 |                       0.0 |                   0.350352 |                  1.305996 |                 0.013867 |    -0.054417 |        -0.013867 |         -0.955644 |             26.857143 |                 100.0 |               656.365228 |               69.444444 |      0.0 |      0.0 |  0.022632 |  0.022632 |      0.0 | -0.001344 |       0.0 |    -0.010403 |     0.739916 |      0.0 |      0.0 |  0.00087 |      0.0 | -0.003586 | -0.015942 | -0.005661 | 0.999213 | 0.299528 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.158698 | 13.363048 | 13.363048 | -0.000291 | 1.033354 |  0.021035 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 18:00:00 |  0.3 |  1.0 |              0.0 |             96.9 |       4.4 |               101.6 | nan | nan |     -3.95 |             95.070833 |                   0.1 | 0.282 |  1.0 |     -3.95 |    3.527001 |         1.942897 |           1.561958 |          4.4 |                 0.656792 |                      0.455629 |                      0.61078 |                        0.0475 |                           0.034384 |             0.636431 |                   0.43317 |     100.963569 |   1.287972 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.046445 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.350465 |                  1.303739 |                 0.013313 |    -0.054322 |        -0.013313 |         -0.953274 |             21.363636 |                 100.0 |               661.679848 |               69.444444 | 0.240698 | 0.060175 |  0.015142 |  0.015142 |      0.0 | -0.002741 |  0.000882 |          0.0 |     0.739833 | 0.032299 | 0.032299 |      0.0 |      0.0 | -0.007312 | -0.016197 | -0.010197 | 0.998056 | 0.298835 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.184858 | 13.393622 | 13.606487 |       0.0 | 1.034449 |  0.051935 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 19:00:00 |  0.2 |  1.4 |              0.0 |             98.3 |       3.8 |               101.5 | nan | nan | -3.816667 |             95.029167 |                   0.1 | 0.188 |  1.4 | -3.816667 |    3.046047 |         1.973983 |           1.348964 |          3.8 |                 0.676036 |                      0.460234 |                      0.61078 |                      0.048726 |                           0.034691 |             0.664544 |                  0.437356 |     100.835456 |   1.284691 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |   0.04644 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.350939 |                  1.301296 |                  0.01222 |    -0.054221 |         -0.01222 |         -0.950357 |             24.736842 |                 100.0 |               617.558461 |               69.444444 | 0.172858 | 0.025929 |  0.006659 |  0.006659 |      0.0 |  -0.00531 |  0.000937 |          0.0 |     0.739481 | 0.071378 | 0.071378 |      0.0 |      0.0 | -0.014165 | -0.020958 | -0.022903 |   0.9969 | 0.298143 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.193341 | 13.353317 | 13.784655 |       0.0 | 1.039083 |  0.082835 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 20:00:00 |  0.4 |  1.6 |              0.0 |             98.7 |       5.7 |               101.4 | nan | nan |    -3.675 |             94.991667 |                   0.1 | 0.376 |  1.6 |    -3.675 |     4.56907 |         2.056027 |           2.023445 |          5.7 |                 0.685844 |                      0.465172 |                      0.61078 |                      0.049349 |                            0.03502 |             0.676928 |                  0.441874 |     100.723072 |   1.282429 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.046435 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.351294 |                  1.298642 |                 0.011755 |     -0.05411 |        -0.011755 |         -0.947348 |             16.491228 |                 100.0 |               598.069608 |               69.444444 | 0.369341 | 0.036934 |  0.007784 |  0.007784 |      0.0 | -0.008229 |   0.00235 |          0.0 |     0.739217 | 0.127886 | 0.127886 |      0.0 |      0.0 | -0.021951 | -0.030168 | -0.040364 | 0.995743 | 0.297452 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.192216 | 13.270076 | 14.162225 |       0.0 | 1.042571 |  0.113735 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 21:00:00 |  0.7 |  2.0 |              0.0 |             99.6 |       5.6 |               101.3 | nan | nan |      -3.5 |             94.991667 |                   0.1 | 0.658 |  2.0 |      -3.5 |    4.488911 |         2.138899 |           1.987946 |          5.6 |                 0.705836 |                      0.471336 |                      0.61078 |                      0.050616 |                           0.035431 |             0.703012 |                   0.44773 |     100.596988 |   1.279174 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.046427 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.352039 |                  1.295068 |                 0.010788 |    -0.053961 |        -0.010788 |         -0.943029 |             16.785714 |                 100.0 |               562.173975 |               69.444444 | 0.650216 |      0.0 |  0.001499 |  0.001499 |      0.0 |  -0.01135 |  0.005445 |          0.0 |     0.738665 | 0.186326 | 0.186326 |      0.0 |      0.0 | -0.030275 | -0.037301 | -0.056788 | 0.994587 | 0.296761 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.198501 | 13.093775 | 14.823791 |       0.0 | 1.049918 |  0.144635 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 22:00:00 |  0.5 |  2.4 |              0.0 |             99.4 |       5.8 |               101.1 | nan | nan | -3.291667 |             95.008333 |                   0.1 |  0.47 |  2.4 | -3.291667 |    4.649229 |         2.249283 |           2.058944 |          5.8 |                  0.72634 |                      0.478769 |                      0.61078 |                       0.05191 |                           0.035924 |             0.721982 |                   0.45487 |     100.378018 |   1.274698 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.046418 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.352777 |                  1.290674 |                 0.010126 |    -0.053778 |        -0.010126 |         -0.937897 |             16.206897 |                 100.0 |               532.195149 |               69.444444 | 0.468501 |      0.0 |  0.003113 |  0.003113 |      0.0 | -0.013984 |  0.004708 |          0.0 |     0.738117 | 0.232415 | 0.232415 |      0.0 |      0.0 | -0.037302 | -0.045743 | -0.072919 | 0.993431 | 0.296071 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.196887 | 12.873122 | 15.306277 |       0.0 | 1.057257 |  0.175535 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 23:00:00 |  1.0 |  2.5 |              0.0 |             99.2 |       6.2 |               101.0 | nan | nan | -3.070833 |                  95.0 |                   0.1 |  0.94 |  2.5 | -3.070833 |    4.969866 |         2.364694 |           2.200941 |          6.2 |                 0.731547 |                      0.486761 |                      0.61078 |                      0.052238 |                           0.036454 |             0.725695 |                  0.462423 |     100.274305 |   1.272954 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.046408 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.353513 |                   1.28612 |                 0.010003 |    -0.053588 |        -0.010003 |         -0.932608 |              15.16129 |                 100.0 |               525.492703 |               69.444444 | 0.936887 |      0.0 |  0.005017 |  0.005017 |      0.0 | -0.015071 |  0.009806 |          0.0 |      0.73757 | 0.268559 | 0.268559 |      0.0 |      0.0 | -0.040202 | -0.049694 | -0.079892 | 0.992274 | 0.295382 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 2.194983 | 12.616259 | 16.258235 |       0.0 | 1.064623 |  0.206435 | 71.325008 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 00:00:00 |  2.0 |  2.7 |              0.0 |             99.2 |       5.6 |               100.8 | nan | nan | -2.804167 |               95.4875 |                   0.1 |  1.88 |  2.7 | -2.804167 |    4.488911 |         2.438173 |           1.987946 |          5.6 |                 0.742061 |                      0.496568 |                      0.61078 |                      0.052899 |                           0.037103 |             0.736124 |                   0.47416 |     100.063876 |   1.269455 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |   0.04635 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.351815 |                  1.277745 |                 0.009642 |    -0.053239 |        -0.009642 |         -0.925929 |             16.785714 |                 100.0 |               511.865132 |               69.444444 | 1.874983 |      0.0 |  0.004505 |  0.004505 |      0.0 | -0.015425 |  0.020156 |          0.0 |     0.737025 | 0.305434 | 0.305434 |      0.0 |      0.0 | -0.041144 | -0.050357 | -0.081859 | 0.991118 | 0.294694 | 0.529112 | 0.177941 |  0.0 |  0.0 | 0.0 |  0.0 | 0.177941 | nan | nan | 2.195495 |  12.32135 |  17.61953 |       0.0 | 1.072031 |  0.237335 | 71.676178 | 0.177941 |      0.0 |   0.0 |   0.0 |  0.0 | 0.002308 |      0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 01:00:00 |  2.1 |  3.2 |              0.0 |             99.6 |       5.2 |               100.7 | nan | nan | -2.470833 |             95.858333 |                   0.1 | 1.974 |  3.2 | -2.470833 |    4.168274 |         2.514993 |            1.84595 |          5.2 |                  0.76893 |                      0.509073 |                      0.61078 |                      0.054584 |                           0.037928 |             0.765854 |                  0.487989 |      99.934146 |   1.265756 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.046332 |       0.0 |       0.0 |                   0.0 |                       0.0 |                    0.35253 |                  1.268373 |                 0.008604 |    -0.052849 |        -0.008604 |         -0.915843 |             18.076923 |                 100.0 |               477.813345 |               69.444444 | 1.969495 |      0.0 |  0.001548 |  0.001548 |      0.0 | -0.018246 |       0.0 |          0.0 |      0.73649 | 0.290813 | 0.290813 |      0.0 |      0.0 | -0.048671 | -0.057065 | -0.097132 | 0.989961 | 0.294007 | 2.385789 | 0.802693 |  0.0 |  0.0 | 0.0 |  0.0 | 0.802693 | nan | nan | 2.198452 | 12.042995 | 17.221483 |       0.0 | 1.079339 |  0.268235 | 73.259274 |      0.5 | 0.302693 |   0.0 |   0.0 |  0.0 | 0.011023 | 0.015304 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 02:00:00 |  1.8 |  3.9 |              0.0 |             99.8 |       5.7 |               100.6 | nan | nan |    -2.075 |             96.279167 |                   0.1 | 1.692 |  3.9 |    -2.075 |     4.56907 |         2.621871 |           2.023445 |          5.7 |                 0.807986 |                      0.524283 |                      0.61078 |                       0.05702 |                           0.038928 |              0.80637 |                  0.504776 |       99.79363 |   1.261108 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |   0.04631 |       0.0 |       0.0 |                   0.0 |                       0.0 |                    0.35324 |                  1.257174 |                  0.00723 |    -0.052382 |         -0.00723 |         -0.903934 |             16.491228 |                 100.0 |               431.029398 |               69.444444 | 1.690452 |      0.0 |   0.00041 |   0.00041 |      0.0 | -0.024333 |       0.0 |          0.0 |      0.73596 | 0.392848 | 0.392848 |      0.0 |      0.0 | -0.064906 | -0.073535 | -0.131211 | 0.988805 |  0.29332 | 2.253018 | 0.758795 |  0.0 |  0.0 | 0.0 |  0.0 | 0.758795 | nan | nan |  2.19959 | 11.666608 |  16.68325 |       0.0 | 1.086647 |  0.299135 | 74.753497 |      0.5 | 0.258795 |   0.0 |   0.0 |  0.0 | 0.023653 | 0.041648 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 03:00:00 |  1.3 |  4.4 |              0.0 |             99.8 |       6.6 |               100.5 | nan | nan | -1.633333 |             96.633333 |                   0.1 | 1.222 |  4.4 | -1.633333 |    5.290502 |          2.74211 |           2.342937 |          6.6 |                 0.836943 |                      0.541728 |                      0.61078 |                      0.058816 |                           0.040071 |             0.835269 |                   0.52349 |      99.664731 |   1.257444 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.046287 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.353945 |                  1.245134 |                 0.006273 |    -0.051881 |        -0.006273 |         -0.891189 |             14.242424 |                 100.0 |               400.244058 |               69.444444 |  1.22159 |      0.0 |  0.000742 |  0.000742 |      0.0 | -0.030654 |       0.0 |          0.0 |     0.735433 | 0.498663 | 0.498663 |      0.0 |      0.0 | -0.081767 | -0.091059 | -0.166554 | 0.987648 | 0.292635 | 1.935989 | 0.652479 |  0.0 |  0.0 | 0.0 |  0.0 | 0.652479 | nan | nan | 2.199258 | 11.188465 | 15.999505 |       0.0 | 1.093955 |  0.330035 | 76.037007 |      0.5 | 0.152479 |   0.0 |   0.0 |  0.0 | 0.035958 | 0.057855 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 04:00:00 |  0.5 |  4.8 |              0.0 |             99.0 |       7.0 |               100.5 | nan | nan | -1.129167 |             96.833333 |                   0.1 |  0.47 |  4.8 | -1.129167 |    5.611139 |         2.892408 |           2.484933 |          7.0 |                 0.860763 |                      0.562266 |                      0.61078 |                      0.060288 |                            0.04141 |             0.852155 |                  0.544461 |      99.647845 |   1.255555 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.046263 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.354645 |                  1.232336 |                 0.005743 |    -0.051347 |        -0.005743 |         -0.877692 |             13.428571 |                 100.0 |                378.78423 |               69.444444 | 0.469258 |      0.0 |  0.008652 |  0.008652 |      0.0 | -0.034262 |       0.0 |          0.0 |      0.73491 | 0.565607 | 0.565607 |      0.0 |      0.0 | -0.091392 | -0.103263 | -0.188913 | 0.986492 |  0.29195 | 1.279808 | 0.431328 |  0.0 |  0.0 | 0.0 |  0.0 | 0.431328 | nan | nan | 2.191348 | 10.645606 | 15.223216 |       0.0 | 1.101263 |  0.360935 | 76.885487 | 0.431328 |      0.0 |   0.0 |   0.0 |  0.0 | 0.047054 |  0.05955 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 05:00:00 |  0.2 |  4.8 |              0.0 |             97.6 |       7.3 |               100.4 | nan | nan | -0.583333 |             97.016667 |                   0.1 | 0.188 |  4.8 | -0.583333 |    5.851616 |         3.059407 |            2.59143 |          7.3 |                 0.860763 |                      0.585275 |                      0.61078 |                      0.060288 |                           0.042904 |             0.840104 |                  0.567814 |      99.559896 |   1.254358 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.046236 |       0.0 |       0.0 |                   0.0 |                       0.0 |                    0.35534 |                  1.218349 |                 0.006177 |    -0.050765 |        -0.006177 |         -0.863009 |             12.876712 |                 100.0 |                376.81648 |               69.444444 | 0.179348 |      0.0 |  0.023166 |  0.023166 |      0.0 |  -0.03348 |       0.0 |          0.0 |      0.73439 | 0.566874 | 0.566874 |      0.0 |      0.0 | -0.089306 | -0.106208 | -0.189336 | 0.985336 | 0.291266 | 0.991761 |  0.33408 |  0.0 |  0.0 | 0.0 |  0.0 |  0.33408 | nan | nan | 2.176834 | 10.100897 | 14.444283 |       0.0 | 1.108571 |  0.391835 | 77.543168 |  0.33408 |      0.0 |   0.0 |   0.0 |  0.0 | 0.055718 | 0.053631 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 06:00:00 |  0.5 |  5.1 |              0.0 |             97.4 |       7.2 |               100.4 | nan | nan |   -0.0125 |             97.054167 |                   0.1 |  0.47 |  5.1 |   -0.0125 |    5.771457 |         3.229745 |           2.555931 |          7.2 |                 0.879017 |                      0.610223 |                      0.61078 |                      0.061412 |                           0.044515 |             0.856163 |                  0.592247 |      99.543837 |    1.25293 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |   0.04621 |       0.0 |       0.0 |                   0.0 |                       0.0 |                    0.35603 |                  1.204312 |                 0.005657 |     -0.05018 |        -0.005657 |         -0.848282 |             13.055556 |                 100.0 |               362.823601 |               69.444444 | 0.446834 |      0.0 |  0.025209 |  0.025209 |      0.0 | -0.035493 |       0.0 |          0.0 |     0.733874 | 0.601263 | 0.601263 |      0.0 |      0.0 | -0.094676 | -0.111803 | -0.200822 | 0.984179 | 0.290583 | 1.308754 | 0.440722 |  0.0 |  0.0 | 0.0 |  0.0 | 0.440722 | nan | nan | 2.174791 |  9.522977 | 13.617857 |       0.0 | 1.115879 |  0.422735 |   78.4112 | 0.440722 |      0.0 |   0.0 |   0.0 |  0.0 | 0.064292 | 0.048301 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 07:00:00 |  0.2 |  5.1 |              0.0 |             96.3 |       7.6 |               100.4 | nan | nan |  0.554167 |             97.070833 |                   0.1 | 0.188 |  5.1 |  0.554167 |    6.092093 |         3.436822 |           2.697927 |          7.6 |                 0.879017 |                      0.635914 |                      0.61078 |                      0.061412 |                           0.046165 |             0.846494 |                  0.617287 |      99.553506 |   1.252976 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.046183 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.356716 |                  1.190153 |                 0.006005 |     -0.04959 |        -0.006005 |         -0.833437 |             12.368421 |                 100.0 |               360.368897 |               69.444444 | 0.162791 |      0.0 |  0.038313 |  0.038313 |      0.0 | -0.035367 |       0.0 |          0.0 |     0.733362 | 0.611699 | 0.611699 |      0.0 |      0.0 | -0.094338 | -0.115974 | -0.204308 | 0.983023 |   0.2899 | 1.039793 | 0.349999 |  0.0 |  0.0 | 0.0 |  0.0 | 0.349999 | nan | nan | 2.161687 |  8.934421 | 12.776222 |       0.0 | 1.123187 |  0.453635 | 79.100994 | 0.349999 |      0.0 |   0.0 |   0.0 |  0.0 | 0.072839 |   0.0435 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 08:00:00 |  0.1 |  5.4 |              0.0 |             95.2 |       7.3 |               100.3 | nan | nan |     1.175 |             97.004167 |                   0.1 | 0.094 |  5.4 |     1.175 |    5.851616 |           3.6439 |            2.59143 |          7.3 |                 0.897612 |                      0.665151 |                      0.61078 |                      0.062554 |                           0.048033 |             0.854526 |                  0.645224 |      99.445474 |   1.250338 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.347041 |                      7.400749 |             1.337827 |  0.254871 |         0.00332 |                0.003243 |  0.028987 |       0.0 |       0.0 |              0.000866 |                  0.000161 |                   0.357397 |                  1.174779 |                 0.005774 |    -0.048083 |        -0.005613 |         -0.817383 |             12.876712 |                 100.0 |               348.585823 |               95.720243 | 0.055687 |      0.0 |  0.045397 |  0.045397 |      0.0 | -0.035585 |       0.0 |          0.0 |     0.732853 | 0.625128 | 0.625128 |      0.0 |      0.0 | -0.094922 | -0.119484 | -0.208793 | 0.981866 | 0.289218 | 0.952111 |  0.32029 |  0.0 |  0.0 | 0.0 |  0.0 |  0.32029 | nan | nan | 2.154603 |  8.332436 | 11.915384 |       0.0 | 1.130495 |  0.484535 | 79.732815 |  0.32029 |      0.0 |   0.0 |   0.0 |  0.0 | 0.079613 | 0.039177 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 09:00:00 |  0.3 |  5.3 |              0.0 |             96.3 |       7.1 |               100.3 | nan | nan |  1.779167 |               97.0875 |                   0.1 | 0.282 |  5.3 |  1.779167 |    5.691298 |         3.847638 |           2.520432 |          7.1 |                 0.891375 |                      0.694736 |                      0.61078 |                      0.062171 |                           0.049913 |             0.858394 |                  0.674502 |      99.441606 |   1.250768 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  4.882681 |        0.063603 |                0.062163 | -0.003244 |       0.0 |       0.0 |              0.016638 |                   0.00309 |                   0.358073 |                   1.15845 |                 0.005615 |    -0.031631 |        -0.002525 |         -0.800378 |             13.239437 |                 100.0 |               347.155921 |              330.980632 | 0.236603 |      0.0 |  0.029965 |  0.029965 |      0.0 | -0.035482 |       0.0 |          0.0 |     0.732348 | 0.620432 | 0.620432 |      0.0 |      0.0 | -0.094646 | -0.115104 | -0.207224 |  0.98071 | 0.288537 | 1.126463 | 0.378763 |  0.0 |  0.0 | 0.0 |  0.0 | 0.378763 | nan | nan | 2.170035 |  7.734969 | 11.061005 |       0.0 | 1.137803 |  0.515435 | 80.480514 | 0.378763 |      0.0 |   0.0 |   0.0 |  0.0 | 0.086588 | 0.035283 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 10:00:00 |  0.1 |  5.5 |              0.0 |             95.5 |       7.1 |               100.3 | nan | nan |  2.295833 |               97.0375 |                   0.1 | 0.094 |  5.5 |  2.295833 |    5.691298 |         4.038016 |           2.520432 |          7.1 |                 0.903886 |                       0.72095 |                      0.61078 |                      0.062939 |                            0.05157 |             0.863211 |                  0.699592 |      99.436789 |   1.249848 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 | 14.862249 |        0.193599 |                0.189423 | -0.003187 |       0.0 |       0.0 |              0.050795 |                  0.009433 |                   0.358744 |                  1.144953 |                 0.005481 |     0.003088 |         0.003952 |         -0.786209 |             13.239437 |                 100.0 |               338.387467 |              323.154916 | 0.064035 |      0.0 |  0.044939 |  0.044939 |      0.0 | -0.036172 |       0.0 |          0.0 |     0.731846 | 0.658344 | 0.658344 |      0.0 |      0.0 | -0.096487 | -0.119447 | -0.219887 | 0.979553 | 0.287857 | 1.008546 | 0.338943 |  0.0 |  0.0 | 0.0 |  0.0 | 0.338943 | nan | nan | 2.155061 |  7.099767 | 10.152667 |       0.0 | 1.145111 |  0.546335 | 81.150117 | 0.338943 |      0.0 |   0.0 |   0.0 |  0.0 | 0.093619 | 0.031776 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 11:00:00 |  0.1 |  6.1 |              0.0 |             91.6 |       8.9 |               100.2 | nan | nan |  2.716667 |             96.795833 |                   0.0 | 0.094 |  6.1 |  2.716667 |    7.134162 |         4.291853 |           3.159415 |          8.9 |                 0.942351 |                      0.742943 |                      0.61078 |                       0.06529 |                           0.052955 |             0.863194 |                  0.719138 |      99.336806 |   1.245915 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 |  22.23196 |        0.289599 |                0.288488 | -0.002947 |       0.0 |       0.0 |              0.077503 |                  0.014393 |                   0.349954 |                  1.076886 |                 0.004536 |     0.032633 |         0.009858 |         -0.726932 |             10.561798 |                 100.0 |               323.499422 |              309.808033 | 0.049061 |      0.0 |  0.112834 |  0.112834 |      0.0 | -0.042299 |       0.0 |          0.0 |     0.731347 | 0.831192 | 0.831192 |      0.0 |      0.0 | -0.112831 | -0.154929 | -0.277618 | 0.978397 | 0.287178 | 1.242618 | 0.417472 |  0.0 |  0.0 | 0.0 |  0.0 | 0.417472 | nan | nan | 2.087166 |  6.294692 |  9.001409 |       0.0 | 1.152419 |  0.577235 | 81.975263 | 0.417472 |      0.0 |   0.0 |   0.0 |  0.0 | 0.100974 | 0.028618 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 12:00:00 |  0.0 |  6.4 |              0.0 |             90.0 |       9.5 |               100.2 | nan | nan |       3.1 |             96.633333 |                   0.0 |   0.0 |  6.4 |       3.1 |    7.615117 |          4.56573 |           3.372409 |          9.5 |                 0.962118 |                      0.763489 |                      0.61078 |                      0.066493 |                           0.054244 |             0.865906 |                  0.737784 |      99.334094 |   1.244565 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 22.247803 |        0.289805 |                0.288994 | -0.002908 |       0.0 |       0.0 |              0.077782 |                  0.014445 |                   0.350599 |                  1.067855 |                  0.00449 |     0.033288 |         0.009955 |         -0.717256 |              9.894737 |                 100.0 |               314.175289 |               301.41071 |      0.0 |      0.0 |  0.143294 |  0.143294 |      0.0 | -0.044819 |       0.0 |          0.0 |     0.730852 | 0.897928 | 0.897928 |      0.0 |      0.0 | -0.119553 |   -0.1704 | -0.299908 |  0.97724 |   0.2865 | 1.290431 | 0.433459 |  0.0 |  0.0 | 0.0 |  0.0 | 0.433459 | nan | nan | 1.943872 |  5.423634 |  7.755797 |       0.0 | 1.159727 |  0.608135 | 82.832235 | 0.433459 |      0.0 |   0.0 |   0.0 |  0.0 | 0.109357 | 0.025774 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 13:00:00 |  0.1 |  6.4 |              0.0 |             85.7 |       7.9 |               100.2 | nan | nan |     3.425 |               96.3125 |                   0.0 | 0.094 |  6.4 |     3.425 |    6.332571 |         4.782828 |           2.804424 |          7.9 |                 0.962118 |                      0.781298 |                      0.61078 |                      0.066493 |                           0.055357 |             0.824535 |                  0.752488 |      99.375465 |    1.24476 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 20.992726 |        0.273456 |                0.273007 | -0.002878 |       0.0 |       0.0 |              0.073614 |                  0.013671 |                    0.35124 |                   1.06128 |                 0.006024 |     0.029394 |         0.007647 |          -0.71004 |             11.898734 |                 100.0 |               317.310308 |              304.237424 |      0.0 |      0.0 |  0.169058 |  0.169058 |      0.0 | -0.032937 |       0.0 |          0.0 |      0.73036 | 0.733435 | 0.733435 |      0.0 |      0.0 | -0.087858 | -0.149462 | -0.244967 | 0.976084 | 0.285822 | 1.053267 | 0.353663 |  0.0 |  0.0 | 0.0 |  0.0 | 0.353663 | nan | nan | 1.868814 |  4.710117 |  6.735468 |       0.0 | 1.167035 |  0.639035 | 83.531838 | 0.353663 |      0.0 |   0.0 |   0.0 |  0.0 | 0.116694 | 0.023212 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 14:00:00 |  0.0 |  6.5 |              0.0 |             82.3 |       9.6 |               100.2 | nan | nan |    3.7125 |             95.854167 |                   0.0 |   0.0 |  6.5 |    3.7125 |    7.695276 |         5.026645 |           3.407908 |          9.6 |                 0.968788 |                      0.797357 |                      0.61078 |                      0.066898 |                           0.056358 |             0.797312 |                    0.7643 |      99.402688 |   1.244443 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 11.677032 |        0.152108 |                 0.15202 | -0.002856 |       0.0 |       0.0 |              0.041065 |                  0.007626 |                   0.351876 |                  1.056546 |                 0.007091 |    -0.002958 |         0.000536 |          -0.70467 |              9.791667 |                 100.0 |               317.222305 |              304.158121 |      0.0 |      0.0 |  0.245106 |  0.245106 |      0.0 | -0.033021 |       0.0 |          0.0 |     0.729871 | 0.787449 | 0.787449 |      0.0 |      0.0 | -0.088081 | -0.174392 | -0.263008 | 0.974928 | 0.285145 | 1.131573 | 0.379811 |  0.0 |  0.0 | 0.0 |  0.0 | 0.379811 | nan | nan | 1.623707 |  3.941899 |  5.636915 |       0.0 | 1.174343 |  0.669935 |   84.2836 | 0.379811 |      0.0 |   0.0 |   0.0 |  0.0 | 0.123154 | 0.020905 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 15:00:00 |  0.1 |  6.3 |              0.1 |             82.5 |       9.1 |               100.3 | nan | nan |  3.954167 |             95.370833 |                   0.1 | 0.094 |  6.3 |  3.954167 |     7.29448 |         5.237063 |           3.230413 |          9.1 |                 0.955489 |                       0.81108 |                      0.61078 |                       0.06609 |                           0.057212 |             0.788278 |                  0.773534 |      99.511722 |   1.246623 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  2.844748 |        0.044468 |                 0.04541 | -0.003032 |       0.0 |       0.0 |              0.012289 |                  0.002282 |                   0.362035 |                  1.110113 |                 0.008451 |    -0.033966 |        -0.006169 |         -0.748079 |              10.32967 |                 100.0 |               318.648113 |              305.442644 |      0.0 |      0.0 |   0.22222 |   0.22222 |      0.0 | -0.030224 |       0.0 |          0.0 |     0.729386 | 0.709692 | 0.709692 |      0.0 |      0.0 | -0.080621 | -0.162585 | -0.237037 | 0.973771 | 0.284469 |   1.0203 | 0.342309 |  0.0 |  0.0 | 0.0 |  0.0 | 0.342309 | nan | nan | 1.495487 |  3.249537 |  4.646839 |       0.0 | 1.181651 |  0.700835 | 84.961591 | 0.342309 |      0.0 |   0.0 |   0.0 |  0.0 | 0.129297 | 0.018827 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 16:00:00 |  0.0 |  5.9 |              0.0 |             82.2 |       8.7 |               100.3 | nan | nan |  4.158333 |             94.854167 |                   0.1 |   0.0 |  5.9 |  4.158333 |    6.973844 |         5.424101 |           3.088417 |          8.7 |                 0.929372 |                      0.822835 |                      0.61078 |                      0.064498 |                           0.057942 |             0.763944 |                  0.780493 |      99.536056 |   1.248525 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.053708 |                      7.400749 |             1.337827 |   0.00593 |        0.000077 |                0.000079 |  0.043389 |       0.0 |       0.0 |              0.000021 |                  0.000004 |                    0.36268 |                  1.107879 |                 0.009318 |     -0.04614 |        -0.009314 |           -0.7452 |             10.804598 |                 100.0 |               326.818078 |               72.472631 |      0.0 |      0.0 |  0.219751 |  0.219751 |      0.0 | -0.025254 |       0.0 |          0.0 |     0.728904 | 0.615227 | 0.615227 |      0.0 |      0.0 | -0.067363 | -0.147437 | -0.205486 | 0.972615 | 0.283794 | 0.884556 | 0.296585 |  0.0 |  0.0 | 0.0 |  0.0 | 0.296585 | nan | nan | 1.275735 |  2.648627 |  3.787537 |       0.0 | 1.188959 |  0.731735 | 85.549562 | 0.296585 |      0.0 |   0.0 |   0.0 |  0.0 | 0.134206 | 0.016956 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 17:00:00 |  0.0 |  5.5 |              0.0 |             76.5 |       8.9 |               100.3 | nan | nan |  4.341667 |             94.108333 |                   0.1 |   0.0 |  5.5 |  4.341667 |    7.134162 |         5.604459 |           3.159415 |          8.9 |                 0.903886 |                      0.833518 |                      0.61078 |                      0.062939 |                           0.058604 |             0.691473 |                   0.78441 |      99.608527 |   1.250659 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.046022 |       0.0 |       0.0 |                   0.0 |                       0.0 |                    0.36332 |                   1.10789 |                 0.012225 |    -0.046162 |        -0.012225 |          -0.74457 |             10.561798 |                 100.0 |               345.233185 |               69.444444 |      0.0 |      0.0 |  0.293126 |  0.293126 |      0.0 | -0.013522 |       0.0 |          0.0 |     0.728425 | 0.489627 | 0.489627 |      0.0 |      0.0 |  -0.03607 |  -0.13969 | -0.163536 | 0.971458 | 0.283119 | 0.702667 | 0.235402 |  0.0 |  0.0 | 0.0 |  0.0 | 0.235402 | nan | nan |  0.98261 |  2.166708 |  3.098392 |       0.0 | 1.196267 |  0.762635 | 86.016826 | 0.235402 |      0.0 |   0.0 |   0.0 |  0.0 | 0.137607 | 0.015271 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 18:00:00 |  0.0 |  5.1 |              0.0 |             81.7 |       7.9 |               100.3 | nan | nan |    4.5125 |                93.475 |                   0.1 |   0.0 |  5.1 |    4.5125 |    6.332571 |         5.721357 |           2.804424 |          7.9 |                 0.879017 |                      0.843583 |                      0.61078 |                      0.061412 |                           0.059227 |             0.718157 |                  0.788539 |      99.581843 |   1.252331 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.046021 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.363956 |                  1.107494 |                 0.010983 |    -0.046146 |        -0.010983 |         -0.743538 |             11.898734 |                 100.0 |               346.478663 |               69.444444 |      0.0 |      0.0 |  0.200054 |  0.200054 |      0.0 | -0.016545 |       0.0 |          0.0 |      0.72795 | 0.455852 | 0.455852 |      0.0 |      0.0 | -0.044134 | -0.119103 | -0.152254 | 0.970302 | 0.282446 | 0.655349 | 0.219342 |  0.0 |  0.0 | 0.0 |  0.0 | 0.219342 | nan | nan | 0.782556 |  1.719992 |  2.459589 |       0.0 | 1.203575 |  0.793535 | 86.452834 | 0.219342 |      0.0 |   0.0 |   0.0 |  0.0 | 0.139925 | 0.013753 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 19:00:00 |  0.0 |  4.8 |              0.0 |             83.4 |       7.1 |               100.3 | nan | nan |  4.654167 |             92.854167 |                   0.1 |   0.0 |  4.8 |  4.654167 |    5.691298 |         5.831576 |           2.520432 |          7.1 |                 0.860763 |                       0.85201 |                      0.61078 |                      0.060288 |                           0.059747 |             0.717876 |                  0.791127 |      99.582124 |   1.253684 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |   0.04602 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.364589 |                  1.107885 |                 0.010915 |    -0.046162 |        -0.010915 |         -0.743296 |             13.239437 |                 100.0 |               351.378767 |               69.444444 |      0.0 |      0.0 |   0.16161 |   0.16161 |      0.0 | -0.015346 |       0.0 |          0.0 |     0.727477 | 0.401993 | 0.401993 |      0.0 |      0.0 | -0.040935 | -0.104245 | -0.134266 | 0.969145 | 0.281773 | 0.578436 | 0.193398 |  0.0 |  0.0 | 0.0 |  0.0 | 0.193398 | nan | nan | 0.620945 |  1.326223 |  1.896499 |       0.0 | 1.210883 |  0.824435 | 86.837872 | 0.193398 |      0.0 |   0.0 |   0.0 |  0.0 |  0.14164 | 0.012386 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 20:00:00 |  0.0 |  4.9 |              0.0 |             82.8 |       6.4 |               100.3 | nan | nan |  4.791667 |             92.191667 |                   0.1 |   0.0 |  4.9 |  4.791667 |    5.130184 |         5.854956 |           2.271939 |          6.4 |                  0.86681 |                       0.86026 |                      0.61078 |                       0.06066 |                           0.060257 |             0.717719 |                  0.793088 |      99.582281 |   1.253234 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |   0.04602 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.365217 |                  1.108737 |                 0.010948 |    -0.046197 |        -0.010948 |          -0.74352 |               14.6875 |                 100.0 |               347.024298 |               69.444444 |      0.0 |      0.0 |  0.151867 |  0.151867 |      0.0 | -0.014314 |       0.0 |          0.0 |     0.727008 | 0.379152 | 0.379152 |      0.0 |      0.0 | -0.038181 | -0.099403 | -0.126637 | 0.967989 | 0.281101 |  0.54628 | 0.182447 |  0.0 |  0.0 | 0.0 |  0.0 | 0.182447 | nan | nan | 0.469079 |  0.954219 |  1.364533 |       0.0 | 1.218191 |  0.855335 | 87.201705 | 0.182447 |      0.0 |   0.0 |   0.0 |  0.0 | 0.142835 | 0.011155 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 21:00:00 |  0.0 |  4.9 |              0.0 |             85.8 |       7.2 |               100.3 | nan | nan |    4.9125 |             91.616667 |                   0.1 |   0.0 |  4.9 |    4.9125 |    5.771457 |         5.908395 |           2.555931 |          7.2 |                  0.86681 |                      0.867569 |                      0.61078 |                       0.06066 |                           0.060707 |             0.743723 |                  0.794837 |      99.556277 |   1.253111 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.046021 |       0.0 |       0.0 |                   0.0 |                       0.0 |                    0.36584 |                  1.109465 |                 0.009883 |    -0.046228 |        -0.009883 |         -0.743624 |             13.055556 |                 100.0 |               340.381741 |               69.444444 |      0.0 |      0.0 |  0.140652 |  0.140652 |      0.0 | -0.019229 |       0.0 |          0.0 |     0.726541 | 0.445595 | 0.445595 |      0.0 |      0.0 | -0.051293 | -0.107419 | -0.148829 | 0.966833 |  0.28043 |  0.64618 | 0.215593 |  0.0 |  0.0 | 0.0 |  0.0 | 0.215593 | nan | nan | 0.328427 |  0.515791 |  0.737582 |       0.0 | 1.225499 |  0.886235 | 87.632292 | 0.215593 |      0.0 |   0.0 |   0.0 |  0.0 | 0.144288 | 0.010047 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 22:00:00 |  0.1 |  4.7 |              0.0 |             87.6 |       7.0 |               100.3 | nan | nan |  5.008333 |                91.125 |                   0.1 | 0.094 |  4.7 |  5.008333 |    5.611139 |         5.948475 |           2.484933 |          7.0 |                 0.854752 |                      0.873404 |                      0.61078 |                      0.059917 |                           0.061066 |             0.748763 |                  0.795889 |      99.551237 |   1.253989 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.046021 |       0.0 |       0.0 |                   0.0 |                       0.0 |                    0.36646 |                   1.11033 |                 0.009631 |    -0.046264 |        -0.009631 |          -0.74387 |             13.428571 |                 100.0 |               342.070698 |               69.444444 |      0.0 |      0.0 |  0.118566 |  0.118566 |      0.0 | -0.019586 |       0.0 |          0.0 |     0.726078 | 0.430317 | 0.430317 |      0.0 |      0.0 | -0.052245 | -0.101112 | -0.143726 | 0.965676 | 0.279759 | 0.631693 | 0.210554 |  0.0 |  0.0 | 0.0 |  0.0 | 0.210554 | nan | nan | 0.303861 |  0.087745 |  0.125475 |       0.0 | 1.232807 |  0.917135 | 88.053431 | 0.210554 |      0.0 |   0.0 |   0.0 |  0.0 | 0.146064 | 0.009048 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 23:00:00 |  0.1 |  3.8 |              0.0 |             85.6 |       6.8 |               100.2 | nan | nan |    5.0625 |             90.558333 |                   0.1 | 0.094 |  3.8 |    5.0625 |     5.45082 |         5.968515 |           2.413935 |          6.8 |                 0.802302 |                      0.876717 |                      0.61078 |                      0.056666 |                            0.06127 |              0.68677 |                   0.79394 |       99.51323 |   1.257101 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.046025 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.367076 |                  1.112997 |                  0.01195 |    -0.046375 |         -0.01195 |         -0.745921 |             13.823529 |                 100.0 |                376.59127 |               69.444444 |      0.0 |      0.0 |  0.129154 |  0.129154 |      0.0 | -0.010581 |       0.0 |          0.0 |     0.725618 | 0.288837 | 0.087745 |      0.0 |      0.0 | -0.028225 | -0.080196 | -0.096472 |  0.96452 | 0.279089 | 0.136056 | 0.045284 |  0.0 |  0.0 | 0.0 |  0.0 | 0.045284 | nan | nan | 0.268706 |       0.0 |       0.0 |       0.0 | 1.240115 |  0.948035 | 88.144204 | 0.045284 |      0.0 |   0.0 |   0.0 |  0.0 | 0.145586 | 0.008149 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 00:00:00 |  0.0 |  3.9 |              0.0 |             85.3 |       6.4 |               100.2 | nan | nan |    5.1125 |             89.979167 |                   0.1 |   0.0 |  3.9 |    5.1125 |    5.130184 |         5.995234 |           2.271939 |          6.4 |                 0.807986 |                      0.879785 |                          0.0 |                       0.05702 |                           0.061459 |             0.689212 |                  0.791623 |      99.510788 |   1.256635 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |   0.04454 |       0.0 |  -0.08424 |                   0.0 |                       0.0 |                   1.170072 |                  1.116023 |                      0.0 |    -0.046501 |              0.0 |          0.054049 |               14.6875 |                 100.0 |               372.495476 |               69.444444 |      0.0 |      0.0 |  0.124445 |  0.124445 |      0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.960211 | 0.276601 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.144261 |       0.0 |       0.0 |       0.0 |      nan |  1.063175 | 88.144204 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.142408 | 0.007339 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 01:00:00 |  0.0 |  3.7 |              0.0 |             86.2 |       6.3 |               100.2 | nan | nan |  5.133333 |             89.420833 |                   0.1 |   0.0 |  3.7 |  5.133333 |    5.050025 |         6.031974 |            2.23644 |          6.3 |                 0.796653 |                      0.881066 |                          0.0 |                      0.056315 |                           0.061538 |             0.686715 |                  0.787857 |      99.513285 |   1.257555 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.044546 |       0.0 |  -0.07992 |                   0.0 |                       0.0 |                   1.170072 |                  1.119656 |                      0.0 |    -0.046652 |              0.0 |          0.050416 |             14.920635 |                 100.0 |               379.277175 |               69.444444 |      0.0 |      0.0 |  0.114089 |  0.114089 |      0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.956063 | 0.274217 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.030172 |       0.0 |       0.0 |       0.0 |      nan |  1.173995 | 88.144204 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.138729 | 0.006609 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 02:00:00 |  0.1 |  3.5 |              0.0 |             89.3 |       5.7 |               100.2 | nan | nan |  5.116667 |             88.983333 |                   0.1 | 0.094 |  3.5 |  5.116667 |     4.56907 |         6.031974 |           2.023445 |          5.7 |                  0.78546 |                      0.880041 |                          0.0 |                      0.055617 |                           0.061475 |             0.701415 |                   0.78309 |      99.498585 |   1.258394 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.044553 |       0.0 |   -0.0756 |                   0.0 |                       0.0 |                   1.170072 |                  1.123475 |                      0.0 |    -0.046811 |              0.0 |          0.046597 |             16.491228 |                 100.0 |               382.954855 |               69.444444 |      0.0 |      0.0 |  0.079451 |  0.079451 |      0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.952077 | 0.271935 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.044722 |       0.0 |       0.0 |       0.0 |      nan |  1.280495 | 88.144204 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.135146 | 0.005953 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 03:00:00 |  0.0 |  3.3 |              0.0 |             91.4 |       5.2 |               100.3 | nan | nan |  5.070833 |             88.633333 |                   0.1 |   0.0 |  3.3 |  5.070833 |    4.168274 |         5.985215 |            1.84595 |          5.2 |                 0.774405 |                      0.877228 |                          0.0 |                      0.054927 |                           0.061302 |             0.707807 |                  0.777516 |      99.592193 |   1.260534 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.044561 |       0.0 |  -0.07128 |                   0.0 |                       0.0 |                   1.170072 |                  1.127474 |                      0.0 |    -0.046978 |              0.0 |          0.042598 |             18.076923 |                 100.0 |                388.64397 |               69.444444 |      0.0 |      0.0 |   0.05785 |  0.044722 | 0.004169 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.948253 | 0.269755 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.382675 | 88.140035 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.131655 | 0.005361 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 04:00:00 |  0.0 |  3.3 |              0.0 |             90.9 |       4.6 |               100.4 | nan | nan |  5.008333 |             88.295833 |                   0.1 |   0.0 |  3.3 |  5.008333 |     3.68732 |         5.905055 |           1.632956 |          4.6 |                 0.774405 |                      0.873404 |                          0.0 |                      0.054927 |                           0.061066 |             0.703935 |                  0.771179 |      99.696065 |   1.261813 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.044568 |       0.0 |  -0.07128 |                   0.0 |                       0.0 |                   1.170072 |                  1.131856 |                      0.0 |    -0.047161 |              0.0 |          0.038216 |             20.434783 |                 100.0 |               389.448172 |               69.444444 |      0.0 |      0.0 |  0.054338 |       0.0 | 0.018687 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.944429 | 0.267584 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.484855 | 88.121348 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.128254 | 0.004828 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 05:00:00 |  0.0 |  3.0 |              0.0 |             89.2 |       4.9 |               100.5 | nan | nan |  4.933333 |             87.945833 |                   0.1 |   0.0 |  3.0 |  4.933333 |    3.927797 |         5.824896 |           1.739453 |          4.9 |                 0.758081 |                      0.868834 |                          0.0 |                      0.053905 |                           0.060785 |             0.676208 |                  0.764103 |      99.823792 |   1.264577 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.044577 |       0.0 |   -0.0648 |                   0.0 |                       0.0 |                   1.170072 |                  1.136687 |                      0.0 |    -0.047362 |              0.0 |          0.033385 |             19.183673 |                 100.0 |               406.567973 |               69.444444 |      0.0 |      0.0 |  0.067898 |       0.0 | 0.022299 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.940847 | 0.265558 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.580555 |  88.09905 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.124941 | 0.004348 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 06:00:00 |  0.0 |  1.9 |              0.0 |             85.6 |       6.3 |               100.6 | nan | nan |       4.8 |             87.454167 |                   0.1 |   0.0 |  1.9 |       4.8 |    5.050025 |         5.794837 |            2.23644 |          6.3 |                  0.70079 |                      0.860763 |                          0.0 |                      0.050296 |                           0.060288 |             0.599876 |                  0.752773 |     100.000124 |   1.271267 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.044591 |       0.0 |  -0.04104 |                   0.0 |                       0.0 |                   1.170072 |                  1.144252 |                      0.0 |    -0.047677 |              0.0 |           0.02582 |             14.920635 |                 100.0 |               476.268326 |               69.444444 |      0.0 |      0.0 |  0.110978 |       0.0 | 0.030069 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.938155 | 0.264041 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.652495 | 88.068981 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.121714 | 0.003916 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 07:00:00 |  0.0 |  1.2 |              0.0 |             86.3 |       5.8 |               100.8 | nan | nan |    4.6375 |               87.0375 |                   0.1 |   0.0 |  1.2 |    4.6375 |    4.649229 |         5.734717 |           2.058944 |          5.8 |                 0.666353 |                      0.851015 |                          0.0 |                       0.04811 |                           0.059686 |             0.575062 |                  0.740702 |     100.224938 |   1.277169 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.044605 |       0.0 |  -0.02592 |                   0.0 |                       0.0 |                   1.170072 |                  1.152031 |                      0.0 |    -0.048001 |              0.0 |          0.018042 |             16.206897 |                 100.0 |               527.605431 |               69.444444 |      0.0 |      0.0 |  0.094618 |       0.0 | 0.026824 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.936029 | 0.262845 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.709315 | 88.042156 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.11857 | 0.003527 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 08:00:00 |  0.0 |  0.8 |              0.0 |             91.2 |       6.1 |               100.9 | nan | nan |  4.445833 |             86.870833 |                   0.1 |   0.0 |  0.8 |  4.445833 |    4.889707 |         5.694638 |           2.165441 |          6.1 |                 0.647352 |                      0.839643 |                          0.0 |                      0.046897 |                           0.058983 |             0.590385 |                  0.729404 |     100.309615 |   1.280232 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.339738 |                      7.386143 |             1.329627 |  0.244956 |        0.003171 |                0.003219 |  0.029475 |       0.0 |  -0.01728 |              0.002833 |                  0.000526 |                   1.170072 |                  1.158654 |                      0.0 |    -0.045445 |              0.0 |          0.011418 |             15.409836 |                 100.0 |               554.251478 |               98.272434 |      0.0 |      0.0 |  0.060572 |       0.0 | 0.012629 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.934225 | 0.261833 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.757495 | 88.029527 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.115507 | 0.003176 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 09:00:00 |  0.0 |  0.8 |              0.0 |             87.5 |       5.9 |               101.0 | nan | nan |  4.258333 |             86.504167 |                   0.1 |   0.0 |  0.8 |  4.258333 |    4.729388 |         5.654558 |           2.094443 |          5.9 |                 0.647352 |                      0.828647 |                          0.0 |                      0.046897 |                           0.058302 |             0.566433 |                  0.716814 |     100.433567 |   1.281619 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 |  4.849391 |        0.062779 |                0.063765 |  0.000014 |       0.0 |  -0.01728 |              0.056113 |                  0.010421 |                   1.170072 |                  1.166643 |                      0.0 |     0.007503 |              0.0 |          0.003429 |             15.932203 |                 100.0 |               561.275052 |              514.360654 |      0.0 |      0.0 |  0.088211 |       0.0 |  0.00436 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.932422 | 0.260823 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.805675 | 88.025167 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.112524 | 0.002861 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 10:00:00 |  0.1 |  0.5 |              0.0 |             84.5 |       6.4 |               101.1 | nan | nan |      4.05 |             86.045833 |                   0.1 | 0.094 |  0.5 |      4.05 |    5.130184 |         5.631178 |           2.271939 |          6.4 |                 0.633417 |                      0.816579 |                          0.0 |                      0.046005 |                           0.057554 |             0.535238 |                  0.702632 |     100.564762 |   1.284447 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 14.852029 |         0.19227 |                0.195489 | -0.000023 |       0.0 |   -0.0108 |              0.172031 |                  0.031949 |                   1.170072 |                  1.175846 |                      0.0 |     0.123037 |              0.0 |         -0.005774 |               14.6875 |                 100.0 |               597.300895 |              543.825136 |      0.0 |      0.0 |  0.135435 |     0.094 | 0.001787 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.930862 | 0.259951 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.847375 |  88.02338 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.109617 | 0.002576 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 11:00:00 |  0.0 |  0.3 |              0.0 |             76.7 |       6.8 |               101.2 | nan | nan |  3.808333 |                85.425 |                   0.1 |   0.0 |  0.3 |  3.808333 |     5.45082 |         5.561039 |           2.413935 |          6.8 |                 0.624275 |                      0.802774 |                          0.0 |                      0.045418 |                           0.056696 |             0.478819 |                   0.68577 |     100.721181 |   1.286932 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.269583 |        0.288296 |                0.293415 | -0.000069 |       0.0 |  -0.00648 |              0.258205 |                  0.047952 |                   1.170072 |                   1.18712 |                      0.0 |     0.208742 |              0.0 |         -0.017048 |             13.823529 |                 100.0 |               635.658629 |              574.779578 |      0.0 |      0.0 |  0.216507 |       0.0 | 0.008323 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.929463 |  0.25917 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.884755 | 88.015057 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.106786 |  0.00232 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 12:00:00 |  0.0 | -0.1 |              0.0 |             74.8 |       7.0 |               101.3 | nan | nan |    3.5375 |             84.791667 |                   0.1 |   0.0 | -0.1 |    3.5375 |    5.611139 |          5.47754 |           2.484933 |          7.0 |                 0.606339 |                      0.787548 |                          0.0 |                      0.044264 |                           0.055747 |             0.453542 |                  0.667775 |     100.846458 |   1.290215 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.293018 |        0.288599 |                0.293995 | -0.000118 |       0.0 |   0.00216 |              0.258716 |                  0.048047 |                   1.170072 |                  1.199167 |                      0.0 |      0.20875 |              0.0 |         -0.029095 |             13.428571 |                 100.0 |               690.367564 |              618.201803 |      0.0 |      0.0 |  0.233286 |       0.0 | 0.008055 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.928387 | 0.258571 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.913495 | 88.007002 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.104027 |  0.00209 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 13:00:00 |  0.0 | -0.4 |              0.0 |             69.7 |       6.3 |               101.4 | nan | nan |  3.254167 |                84.125 |                   0.1 |   0.0 | -0.4 |  3.254167 |    5.050025 |         5.424101 |            2.23644 |          6.3 |                 0.593188 |                      0.771892 |                          0.0 |                      0.043416 |                           0.054769 |             0.413452 |                  0.649354 |     100.986548 |   1.293105 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 21.019348 |        0.272111 |                0.277484 | -0.000169 |       0.0 |   0.00864 |              0.244186 |                  0.045349 |                   1.170072 |                   1.21168 |                      0.0 |     0.193699 |              0.0 |         -0.041608 |             14.920635 |                 100.0 |               746.364552 |              661.783703 |      0.0 |      0.0 |  0.244629 |       0.0 | 0.008677 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.927554 | 0.258107 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.935755 | 87.998325 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.10134 | 0.001882 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 14:00:00 |  0.0 | -0.6 |              0.2 |             66.9 |       7.3 |               101.5 | nan | nan |  2.958333 |             83.483333 |                   0.3 |   0.0 | -0.6 |  2.958333 |    5.851616 |         5.347282 |            2.59143 |          7.3 |                  0.58456 |                      0.755838 |                          0.0 |                      0.042857 |                           0.053764 |             0.391071 |                  0.630999 |     101.108929 |    1.29544 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.399736 | 11.654171 |         0.21122 |                0.216915 |  -0.00048 |       0.0 |   0.01296 |              0.190885 |                   0.03545 |                   1.231767 |                  1.349977 |                      0.0 |     0.134636 |              0.0 |          -0.11821 |             12.876712 |                 100.0 |               786.332824 |              692.370878 |      0.0 |      0.0 |   0.28934 |       0.0 | 0.008499 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.926883 | 0.257733 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.953695 | 87.989826 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.098722 | 0.001695 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 15:00:00 |  0.0 | -0.8 |              0.2 |             64.4 |       7.5 |               101.5 | nan | nan |    2.6625 |             82.729167 |                   0.4 |   0.0 | -0.8 |    2.6625 |    6.011934 |         5.293842 |           2.662428 |          7.5 |                 0.576044 |                       0.74008 |                          0.0 |                      0.042305 |                           0.052775 |             0.370972 |                  0.612262 |     101.129028 |   1.296488 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |              1.43479 |  2.813062 |        0.050984 |                0.053414 | -0.000672 |       0.0 |   0.01728 |              0.047005 |                  0.008729 |                   1.262615 |                  1.428152 |                      0.0 |    -0.012502 |              0.0 |         -0.165538 |             12.533333 |                 100.0 |               829.198189 |              724.705569 |      0.0 |      0.0 |  0.291704 |       0.0 |  0.00795 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.926373 |  0.25745 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.967315 | 87.981876 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.096172 | 0.001526 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 16:00:00 |  0.0 | -1.3 |              0.2 |             65.4 |       7.0 |               101.7 | nan | nan |    2.3625 |             82.029167 |                   0.6 |   0.0 | -1.3 |    2.3625 |    5.611139 |         5.237063 |           2.484933 |          7.0 |                 0.555231 |                      0.724396 |                          0.0 |                      0.040952 |                           0.051788 |             0.363121 |                  0.594216 |     101.336879 |   1.301474 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.046405 |                      7.386143 |             1.504898 |  0.004444 |        0.000553 |                0.000608 |  0.042945 |       0.0 |   0.02808 |              0.000535 |                  0.000099 |                    1.32431 |                  1.571379 |                      0.0 |    -0.064939 |              0.0 |         -0.247069 |             13.428571 |                 100.0 |               934.614685 |               72.518345 |      0.0 |      0.0 |  0.257919 |       0.0 | 0.058911 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.926267 | 0.257391 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.970135 | 87.922965 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.093688 | 0.001375 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 17:00:00 |  0.0 | -1.6 |              0.0 |             67.6 |       6.0 |               101.7 | nan | nan |  2.066667 |             81.658333 |                   0.6 |   0.0 | -1.6 |  2.066667 |    4.809547 |         5.140204 |           2.129942 |          6.0 |                 0.543065 |                      0.709217 |                          0.0 |                      0.040158 |                           0.050829 |             0.367112 |                  0.579135 |     101.332888 |   1.302892 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.045106 |       0.0 |   0.03456 |                   0.0 |                       0.0 |                    1.32431 |                  1.583999 |                      0.0 |       -0.066 |              0.0 |         -0.259688 |             15.666667 |                 100.0 |              1008.959016 |               69.444444 |      0.0 |      0.0 |   0.20423 |       0.0 | 0.053743 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.926404 | 0.257468 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.966475 | 87.869222 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.091268 | 0.001238 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 18:00:00 |  0.0 | -2.0 |              0.0 |             70.2 |       4.8 |               101.8 | nan | nan |  1.770833 |             81.179167 |                   0.6 |   0.0 | -2.0 |  1.770833 |    3.847638 |         5.036665 |           1.703954 |          4.8 |                  0.52721 |                       0.69432 |                          0.0 |                       0.03912 |                           0.049887 |             0.370102 |                  0.563643 |     101.429898 |   1.306085 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.045131 |       0.0 |    0.0432 |                   0.0 |                       0.0 |                    1.32431 |                    1.5975 |                      0.0 |    -0.066562 |              0.0 |          -0.27319 |             19.583333 |                 100.0 |              1132.702415 |               69.444444 |      0.0 |      0.0 |  0.147417 |       0.0 | 0.045064 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.926865 | 0.257723 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.954175 | 87.824158 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.08891 | 0.001115 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 19:00:00 |  0.0 | -2.1 |              0.0 |             70.4 |       4.2 |               101.9 | nan | nan |  1.483333 |               80.6375 |                   0.6 |   0.0 | -2.1 |  1.483333 |    3.366683 |         4.939806 |            1.49096 |          4.2 |                 0.523311 |                      0.680107 |                          0.0 |                      0.038864 |                           0.048985 |             0.368411 |                  0.548422 |     101.531589 |    1.30786 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.045155 |       0.0 |   0.04536 |                   0.0 |                       0.0 |                    1.32431 |                  1.611195 |                      0.0 |    -0.067133 |              0.0 |         -0.286884 |             22.380952 |                 100.0 |              1171.145536 |               69.444444 |      0.0 |      0.0 |  0.127573 |       0.0 | 0.042552 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.927406 | 0.258024 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.939715 | 87.781605 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.086614 | 0.001004 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 20:00:00 |  0.0 | -2.2 |              0.0 |             69.5 |       4.9 |               101.9 | nan | nan |    1.1875 |             80.083333 |                   0.6 |   0.0 | -2.2 |    1.1875 |    3.927797 |         4.889707 |           1.739453 |          4.9 |                 0.519437 |                      0.665752 |                          0.0 |                       0.03861 |                           0.048071 |             0.361009 |                  0.533156 |     101.538991 |   1.308379 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.045181 |       0.0 |   0.04752 |                   0.0 |                       0.0 |                    1.32431 |                  1.625096 |                      0.0 |    -0.067712 |              0.0 |         -0.300786 |             19.183673 |                 100.0 |              1216.067398 |               69.444444 |      0.0 |      0.0 |  0.152631 |       0.0 | 0.045859 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.928028 | 0.258371 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.923095 | 87.735747 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.084377 | 0.000904 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 21:00:00 |  0.0 | -2.6 |              0.0 |             71.6 |       4.0 |               102.0 | nan | nan |     0.875 |             79.491667 |                   0.6 |   0.0 | -2.6 |     0.875 |    3.206365 |         4.782828 |           1.419962 |          4.0 |                 0.504195 |                      0.650878 |                          0.0 |                      0.037607 |                           0.047123 |             0.361003 |                  0.517394 |     101.638997 |   1.311601 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.045207 |       0.0 |   0.05616 |                   0.0 |                       0.0 |                    1.32431 |                  1.639665 |                      0.0 |    -0.068319 |              0.0 |         -0.315355 |                  23.5 |                 100.0 |              1408.064904 |               69.444444 |      0.0 |      0.0 |  0.113635 |       0.0 | 0.038824 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.928973 | 0.258897 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.897835 | 87.696923 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.082197 | 0.000815 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 22:00:00 |  0.0 | -3.1 |              0.0 |             74.5 |       3.4 |               102.0 | nan | nan |      0.55 |             78.945833 |                   0.6 |   0.0 | -3.1 |      0.55 |     2.72541 |         4.662589 |           1.206967 |          3.4 |                 0.485698 |                      0.635721 |                          0.0 |                      0.036384 |                           0.046153 |             0.361845 |                  0.501875 |     101.638155 |   1.314025 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.045233 |       0.0 |   0.06696 |                   0.0 |                       0.0 |                    1.32431 |                  1.653965 |                      0.0 |    -0.068915 |              0.0 |         -0.329655 |             27.647059 |                 100.0 |              1761.293678 |               69.444444 |      0.0 |      0.0 |   0.08419 |       0.0 | 0.031612 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.930323 |  0.25965 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.861775 | 87.665311 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.080074 | 0.000734 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 23:00:00 |  0.0 | -3.9 |              0.0 |             80.2 |       3.4 |               102.0 | nan | nan |  0.229167 |             78.720833 |                   0.6 |   0.0 | -3.9 |  0.229167 |     2.72541 |          4.54903 |           1.206967 |          3.4 |                 0.457351 |                      0.621065 |                          0.0 |                      0.034499 |                           0.045212 |             0.366795 |                  0.488908 |     101.633205 |   1.317905 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.045252 |       0.0 |   0.08424 |                   0.0 |                       0.0 |                    1.32431 |                  1.664915 |                      0.0 |    -0.069371 |              0.0 |         -0.340605 |             27.647059 |                 100.0 |              2990.590085 |               69.444444 |      0.0 |      0.0 |  0.062047 |       0.0 | 0.023045 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 | 0.932319 | 0.260766 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 |       0.0 |       0.0 |       0.0 |      nan |  1.808435 | 87.642266 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.078005 | 0.000661 |   0.0 |   0.0 |  0.0 |   nan |    nan |

.. _lland_v3_acker_winter_ktschnee_hourly:

acre (snow surface temperature)
-------------------------------

Due to the reasons discussed in the daily integration test
:ref:`lland_v3_acker_winter_ktschnee_daily`, enabling the iterative
adjustment of the snow surface temperature results both in a more delayed
and smoothed snow-melt, which becomes even more apparent in the following
hourly simulation results:

.. integration-test::

    >>> lnk(ACKER)
    >>> ktschnee(0.018)
    >>> test('lland_v3_acker_winter_ktschnee_hourly',
    ...      axis1=(inputs.nied, fluxes.wada),
    ...      axis2=(states.waes, states.wats),
    ...      use_conditions=conditions_acker_winter_ktschnee)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration |  nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |         g |        tz |        wg | netshortwaveradiation | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationsnow | netradiation | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes |     sbes |      evpo |       evi | evb |       evs |     wnied | tempssurface | actualalbedo |  schmpot |     schm |  gefrpot |     gefr |  wlatsnow | wsenssnow |     wsurf |      sff |      fvg |     wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz | qah |  qa |     inzp |      wats |      waes |     esnow |     taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet | outlet |
    ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-12-10 00:00:00 |  0.0 | -3.7 |              0.0 |             87.5 |       3.4 |               101.7 | nan | nan |    -2.025 |             96.870833 |                  4.41 |   0.0 | -3.7 |    -2.025 |     2.72541 |         2.684392 |            2.72541 |          3.4 |                 0.464297 |                      0.526233 |                     0.335784 |                      0.034962 |                           0.039056 |              0.40626 |                  0.509766 |      101.29374 |   1.312855 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 |  -0.01337 | -2.473858 | -0.003625 |                   0.0 |                       0.0 |                   0.704039 |                  4.014396 |                 0.226273 |    -0.167266 |        -0.226273 |         -3.310357 |             62.648725 |                 100.0 |               112.102145 |               98.814229 | 0.234493 | 0.234493 | -0.003699 | -0.003699 | 0.0 | -0.010648 | -0.001813 |    -7.924586 |     0.754819 |      0.0 |      0.0 | 0.319079 | 0.319079 | -0.028403 | -0.096736 |  0.101134 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.063699 | 19.132901 | 20.466014 |       0.0 | 0.854914 | -0.707632 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 01:00:00 |  0.0 | -4.8 |              0.0 |             90.7 |       2.9 |               101.8 | nan | nan | -2.158333 |                96.825 |                  4.41 |   0.0 | -4.8 | -2.158333 |    2.324615 |         2.647918 |           2.324615 |          2.9 |                 0.427219 |                      0.521048 |                      0.31051 |                      0.032479 |                           0.038716 |             0.387487 |                  0.504505 |     101.412513 |   1.319627 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 |  -0.01352 | -2.358774 |  0.002704 |                   0.0 |                       0.0 |                   0.705869 |                  4.024777 |                 0.228637 |    -0.167699 |        -0.228637 |         -3.318908 |              73.45023 |                 100.0 |               112.102145 |               98.814229 | 0.003699 | 0.003699 | -0.009724 | -0.009724 | 0.0 | -0.010458 | -0.000037 |    -8.921628 |     0.754182 |      0.0 |      0.0 | 0.338956 | 0.338956 | -0.027895 | -0.084863 |  0.115878 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.069724 | 19.485506 | 20.480171 |       0.0 | 0.861903 | -0.679436 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 02:00:00 |  0.0 | -5.6 |              0.0 |             89.7 |       2.5 |               101.8 | nan | nan |    -2.325 |             96.604167 |                  4.41 |   0.0 | -5.6 |    -2.325 |    2.003978 |         2.606417 |           2.003978 |          2.5 |                 0.401924 |                      0.514631 |                     0.289023 |                      0.030771 |                           0.038294 |             0.360526 |                  0.497155 |     101.439474 |   1.323706 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | -0.013771 | -2.264788 |  0.008827 |                   0.0 |                       0.0 |                    0.70755 |                  4.040774 |                   0.2316 |    -0.168366 |          -0.2316 |         -3.333224 |             85.202267 |                 100.0 |               112.102145 |               98.814229 | 0.009724 | 0.009724 |  -0.00952 |  -0.00952 | 0.0 | -0.008843 | -0.000114 |    -9.827606 |     0.753596 |      0.0 |      0.0 | 0.359468 | 0.359468 | -0.023587 | -0.079238 |  0.128775 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  0.06952 | 19.863267 | 20.498738 |       0.0 | 0.868367 | -0.657363 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 03:00:00 |  0.0 | -6.2 |              0.0 |             91.3 |       3.0 |               101.9 | nan | nan | -2.420833 |                96.325 |                  4.41 |   0.0 | -6.2 | -2.420833 |    2.404774 |         2.602449 |           2.404774 |          3.0 |                 0.383833 |                      0.510972 |                     0.283865 |                       0.02954 |                           0.038053 |             0.350439 |                  0.492194 |     101.549561 |   1.328036 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | -0.013953 |  -2.19121 |  0.011552 |                   0.0 |                       0.0 |                   0.709221 |                  4.052814 |                 0.237702 |    -0.168867 |        -0.237702 |         -3.343592 |             71.001889 |                 100.0 |               112.102145 |               98.814229 |  0.00952 |  0.00952 | -0.009837 | -0.009837 | 0.0 | -0.009247 | -0.000123 |   -10.053971 |     0.753014 |      0.0 |      0.0 | 0.360704 | 0.360704 | -0.024666 | -0.081132 |  0.131903 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.069837 | 20.242613 | 20.517505 |       0.0 |  0.87484 | -0.638015 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 04:00:00 |  0.0 | -7.3 |              0.0 |             94.2 |       2.5 |               102.0 | nan | nan | -2.558333 |               96.2125 |                  4.41 |   0.0 | -7.3 | -2.558333 |    2.003978 |         2.623448 |           2.003978 |          2.5 |                 0.352531 |                      0.505764 |                     0.261715 |                      0.027395 |                            0.03771 |             0.332084 |                  0.486608 |     101.667916 |   1.334932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 |  -0.01413 | -2.126715 |   0.01785 |                   0.0 |                       0.0 |                   0.710873 |                  4.064535 |                 0.239963 |    -0.169356 |        -0.239963 |         -3.353663 |             85.202267 |                 100.0 |               112.102145 |               98.814229 | 0.009837 | 0.009837 |  -0.01321 |  -0.01321 | 0.0 | -0.008702 |  -0.00015 |   -11.069587 |     0.752439 |      0.0 |      0.0 | 0.384421 | 0.274891 | -0.023213 | -0.070653 |  0.146097 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  0.07321 | 20.536044 | 20.536044 | -0.036583 | 0.881279 | -0.624965 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 05:00:00 |  0.0 | -8.3 |              0.0 |             93.2 |       2.3 |               102.0 | nan | nan |    -2.725 |             96.004167 |                  4.41 |   0.0 | -8.3 |    -2.725 |     1.84366 |         2.596101 |            1.84366 |          2.3 |                 0.326061 |                      0.499513 |                      0.23921 |                      0.025563 |                           0.037298 |             0.303889 |                  0.479553 |     101.696111 |   1.340112 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | -0.014382 | -2.083215 |  0.033791 |                   0.0 |                       0.0 |                   0.711986 |                  4.080005 |                 0.243113 |        -0.17 |        -0.243113 |          -3.36802 |             92.611159 |                 100.0 |               112.102145 |               98.814229 |  0.01321 |  0.01321 | -0.012035 | -0.012035 | 0.0 | -0.007605 | -0.000229 |   -12.182931 |     0.752052 |      0.0 |      0.0 | 0.469033 |      0.0 | -0.020285 | -0.069192 |  0.153636 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.072035 | 20.556859 | 20.556859 | -0.156657 | 0.885644 | -0.627855 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 06:00:00 |  0.0 | -8.6 |              0.0 |             96.5 |       2.1 |               102.1 | nan | nan |    -2.925 |             96.004167 |                  4.41 |   0.0 | -8.6 |    -2.925 |    1.683342 |          2.56624 |           1.683342 |          2.1 |                 0.318472 |                      0.492103 |                      0.22557 |                      0.025034 |                           0.036808 |             0.307325 |                  0.472439 |     101.792675 |   1.342932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | -0.014613 | -2.092851 |  0.068393 |                   0.0 |                       0.0 |                   0.712533 |                  4.093695 |                 0.234636 |    -0.170571 |        -0.234636 |         -3.381162 |             101.43127 |                 100.0 |               112.102145 |               98.814229 | 0.012035 | 0.012035 | -0.014531 | -0.014531 | 0.0 | -0.009114 | -0.000216 |   -12.903922 |     0.751861 |      0.0 |      0.0 | 0.676906 |      0.0 | -0.024312 | -0.072719 |  0.137606 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.074531 | 20.578009 | 20.578009 | -0.226086 | 0.887798 | -0.665348 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 07:00:00 |  0.0 | -8.5 |              0.0 |             95.9 |       1.4 |               102.1 | nan | nan |   -3.1625 |               96.0625 |                  3.76 |   0.0 | -8.5 |   -3.1625 |    1.122228 |         2.521333 |           1.122228 |          1.4 |                 0.320984 |                      0.483429 |                     0.217837 |                       0.02521 |                           0.036234 |             0.307823 |                  0.464394 |     101.792177 |   1.342422 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.641605 |       0.0 |             0.0 |                     0.0 |  -0.00839 | -2.217826 |  0.086494 |                   0.0 |                       0.0 |                   0.655819 |                  3.682358 |                 0.215581 |    -0.153432 |        -0.215581 |         -3.026539 |            152.146905 |                 100.0 |               112.102145 |               98.814229 | 0.014531 | 0.014531 | -0.011723 | -0.011723 | 0.0 | -0.008113 | -0.000258 |   -13.330091 |     0.751735 |      0.0 |      0.0 | 0.801775 |      0.0 | -0.021641 | -0.065999 |  0.127942 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.071723 | 20.600652 | 20.600652 | -0.267793 | 0.889228 | -0.720942 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 08:00:00 |  0.0 | -9.5 |              0.0 |             96.8 |       1.1 |               102.2 | nan | nan | -3.483333 |               96.2625 |                  2.76 |   0.0 | -9.5 | -3.483333 |     0.88175 |         2.408072 |            0.88175 |          1.1 |                 0.296635 |                      0.471927 |                     0.206116 |                      0.023505 |                            0.03547 |             0.287143 |                  0.454289 |     101.912857 |   1.348938 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.355232 |                       7.41713 |             2.287909 |  0.266196 |        0.003492 |                0.003987 |  0.035315 |  -2.40314 |   0.09752 |               0.00099 |                   0.00099 |                   0.568304 |                  3.039527 |                 0.201305 |    -0.125657 |        -0.200315 |         -2.471223 |            193.641515 |                 100.0 |               112.102145 |               99.616909 | 0.011723 | 0.011723 |  -0.00564 |  -0.00564 | 0.0 | -0.006565 | -0.000233 |   -14.002402 |     0.751605 |      0.0 |      0.0 | 0.892291 |      0.0 | -0.017511 | -0.055284 |   0.12752 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |  0.06564 |  20.61894 |  20.61894 | -0.298025 | 0.890694 | -0.787562 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 09:00:00 |  0.0 | -9.2 |              0.0 |             94.3 |       1.0 |               102.2 | nan | nan | -3.816667 |             96.383333 |                  1.79 |   0.0 | -9.2 | -3.816667 |    0.801591 |         2.278972 |           0.801591 |          1.0 |                 0.303761 |                      0.460234 |                     0.221436 |                      0.024006 |                           0.034691 |             0.286447 |                  0.443589 |     101.913553 |   1.347408 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.944823 |  4.919884 |        0.064534 |                0.067188 |  0.075667 | -2.625208 |  0.097693 |                0.0167 |                    0.0167 |                   0.483404 |                  2.412046 |                  0.18434 |    -0.083802 |         -0.16764 |         -1.928642 |            213.005666 |                 100.0 |               112.102145 |              101.107639 |  0.00564 |  0.00564 |  0.001396 |  0.001396 | 0.0 | -0.005069 | -0.000108 |   -13.130061 |     0.751441 |      0.0 |      0.0 | 0.922508 |      0.0 | -0.013521 | -0.046442 |  0.107677 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.058604 | 20.629649 | 20.629649 | -0.308118 | 0.892568 | -0.854355 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 10:00:00 |  0.0 | -6.9 |              0.0 |             96.7 |       1.4 |               102.2 | nan | nan |   -4.0625 |               96.5375 |                  0.88 |   0.0 | -6.9 |   -4.0625 |    1.122228 |         2.200731 |           1.122228 |          1.4 |                 0.363641 |                      0.451775 |                     0.294101 |                      0.028159 |                           0.034126 |             0.351641 |                  0.436132 |     101.848359 |   1.335446 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.622959 | 14.873457 |        0.195096 |                0.202564 |  0.055388 |  -2.84785 |  0.080173 |              0.050391 |                  0.050391 |                   0.403735 |                  1.815487 |                  0.15611 |    -0.025254 |        -0.105719 |         -1.411752 |            152.146905 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.004608 |  0.004608 | 0.0 | -0.005188 |       0.0 |    -9.608213 |     0.751235 |      0.0 |      0.0 |  0.84677 |      0.0 | -0.013838 | -0.037005 |  0.054876 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.053996 | 20.634837 | 20.634837 | -0.282821 | 0.894904 | -0.903628 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 11:00:00 |  0.0 | -4.0 |              0.1 |             97.4 |       1.3 |               102.2 | nan | nan | -4.195833 |             96.570833 |                  0.39 |   0.0 | -4.0 | -4.195833 |    1.042069 |         2.098318 |           1.042069 |          1.3 |                 0.453913 |                      0.447245 |                     0.399863 |                      0.034269 |                           0.033823 |             0.442111 |                  0.431908 |     101.757889 |   1.320615 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.449648 | 22.189872 |        0.349279 |                0.385806 |  0.044415 | -3.012092 |  0.052405 |              0.096057 |                  0.096057 |                   0.360931 |                  1.493002 |                 0.132004 |     0.033849 |        -0.035946 |          -1.13207 |            163.850513 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.009731 |  0.009731 | 0.0 |  -0.00368 |       0.0 |    -5.667135 |     0.751021 |      0.0 |      0.0 | 0.702201 |      0.0 | -0.009817 |  -0.02201 |   0.00412 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.044265 | 20.638517 | 20.638517 | -0.234535 | 0.897347 | -0.925133 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 12:00:00 |  0.0 | -2.8 |              0.0 |             93.9 |       1.3 |               102.2 | nan | nan |     -4.35 |               96.3625 |                  0.14 |   0.0 | -2.8 |     -4.35 |    1.042069 |          2.00007 |           1.042069 |          1.3 |                 0.496723 |                      0.442058 |                     0.432472 |                      0.037113 |                           0.033476 |             0.466423 |                  0.425978 |     101.733577 |   1.314635 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.361224 | 22.197307 |        0.291164 |                0.329304 |  0.038912 | -3.083777 |  0.033813 |              0.082067 |                  0.082067 |                   0.339234 |                  1.331032 |                 0.114055 |     0.026607 |        -0.031989 |         -0.991798 |            163.850513 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.011176 |  0.011176 | 0.0 | -0.002957 |       0.0 |     -4.63914 |     0.750787 |      0.0 |      0.0 | 0.600422 |      0.0 | -0.007889 | -0.024281 | -0.000181 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.033089 | 20.641474 | 20.641474 | -0.200541 | 0.900024 | -0.928046 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 13:00:00 |  0.0 | -1.4 |              0.0 |             93.4 |       1.4 |               102.1 | nan | nan | -4.370833 |             96.116667 |                   0.1 |   0.0 | -1.4 | -4.370833 |    1.122228 |         1.900997 |           1.122228 |          1.4 |                 0.551149 |                      0.441361 |                     0.474747 |                      0.040686 |                           0.033429 |             0.514774 |                  0.424221 |     101.585226 |   1.306346 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 20.962863 |        0.274971 |                0.289712 |   0.03806 | -3.093487 |  0.018972 |              0.072273 |                  0.072273 |                   0.336047 |                  1.306141 |                 0.099061 |      0.01785 |        -0.026789 |         -0.970094 |            152.146905 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.011601 |  0.011601 | 0.0 | -0.003609 |       0.0 |    -3.404039 |     0.750536 |      0.0 |      0.0 | 0.513018 |      0.0 | -0.009626 | -0.027383 |  -0.01022 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.021488 | 20.645083 | 20.645083 | -0.171348 | 0.902906 | -0.916119 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 14:00:00 |  0.0 | -0.4 |              0.0 |             93.3 |       2.3 |               102.0 | nan | nan | -4.354167 |             95.854167 |                   0.1 |   0.0 | -0.4 | -4.354167 |     1.84366 |         1.856982 |            1.84366 |          2.3 |                 0.593188 |                      0.441918 |                     0.498421 |                      0.043416 |                           0.033467 |             0.553444 |                  0.423597 |     101.446556 |   1.300092 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 11.702396 |        0.153501 |                0.155234 |  0.038091 | -3.053729 |  0.008926 |              0.038767 |                  0.038767 |                   0.336409 |                  1.307301 |                 0.085148 |    -0.015704 |        -0.046381 |         -0.970892 |             92.611159 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.011365 |  0.011365 | 0.0 | -0.006469 |       0.0 |    -2.754317 |     0.750267 |      0.0 |      0.0 | 0.447886 |      0.0 | -0.017257 | -0.041953 | -0.012828 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.010123 | 20.651552 | 20.651552 | -0.149594 | 0.906004 | -0.894145 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 15:00:00 |  0.0 |  0.5 |              0.0 |             94.1 |       2.8 |               102.0 | nan | nan |      -4.3 |             95.608333 |                   0.1 |   0.0 |  0.5 |      -4.3 |    2.244455 |         1.842168 |           2.244455 |          2.8 |                 0.633417 |                      0.443734 |                     0.522342 |                      0.046005 |                           0.033588 |             0.596046 |                  0.424247 |     101.403954 |   1.295611 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 |  2.880181 |         0.03778 |                0.037152 |  0.038089 | -2.980483 |  0.001449 |              0.009289 |                  0.009289 |                   0.336793 |                  1.307629 |                 0.071627 |    -0.045196 |        -0.062339 |         -0.970835 |             76.073452 |                 100.0 |               112.102145 |              101.107639 |      0.0 |      0.0 |  0.008128 |  0.008128 | 0.0 | -0.009788 |       0.0 |    -2.124962 |     0.749982 |      0.0 |      0.0 | 0.393827 |      0.0 |  -0.02611 | -0.052835 | -0.016607 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.001995 | 20.661341 | 20.661341 | -0.131538 |   0.9093 | -0.864694 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 16:00:00 |  0.1 |  1.0 |              0.0 |             94.6 |       3.1 |               101.9 | nan | nan |      -4.2 |             95.395833 |                   0.1 | 0.094 |  1.0 |      -4.2 |    2.484933 |          1.86654 |           2.484933 |          3.1 |                 0.656792 |                      0.447104 |                     0.542682 |                        0.0475 |                           0.033814 |             0.621325 |                  0.426519 |     101.278675 |   1.291856 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.061898 |                       7.41713 |             1.347076 |  0.007844 |        0.000103 |                0.000101 |  0.028344 | -2.882313 | -0.004259 |              0.000025 |                  0.000025 |                   0.337105 |                   1.30686 |                 0.066581 |    -0.054427 |        -0.066556 |         -0.969755 |             68.711505 |                 100.0 |               112.102145 |               98.953162 | 0.035995 | 0.008999 |  0.005505 |  0.005505 | 0.0 | -0.011163 |  0.000132 |    -1.609548 |     0.749751 |      0.0 |      0.0 | 0.348217 | 0.026996 | -0.029777 |  -0.05614 | -0.019361 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.054495 | 20.708499 | 20.708499 | -0.107288 | 0.911981 | -0.829535 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 17:00:00 |  0.0 |  1.1 |              0.0 |             94.4 |       3.5 |               101.8 | nan | nan | -4.091667 |             95.170833 |                   0.1 |   0.0 |  1.1 | -4.091667 |    2.805569 |         1.866772 |           2.805569 |          3.5 |                 0.661557 |                      0.450781 |                     0.553596 |                      0.047804 |                            0.03406 |              0.62451 |                  0.429012 |      101.17549 |     1.2901 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.027726 | -2.765117 | -0.009914 |                   0.0 |                       0.0 |                   0.337548 |                  1.305996 |                 0.068763 |    -0.054417 |        -0.068763 |         -0.968448 |             60.858762 |                 100.0 |               112.102145 |               98.814229 |      0.0 |      0.0 |  0.006903 |  0.006903 | 0.0 |  -0.01093 |       0.0 |    -1.339983 |     0.749422 |      0.0 |      0.0 | 0.298836 |      0.0 | -0.029155 | -0.056998 | -0.017391 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.047592 | 20.719429 | 20.719429 | -0.099811 | 0.915807 | -0.788721 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 18:00:00 |  0.3 |  1.0 |              0.0 |             96.9 |       4.4 |               101.6 | nan | nan |     -3.95 |             95.070833 |                   0.1 | 0.282 |  1.0 |     -3.95 |    3.527001 |         1.942897 |           3.527001 |          4.4 |                 0.656792 |                      0.455629 |                     0.565562 |                        0.0475 |                           0.034384 |             0.636431 |                   0.43317 |     100.963569 |   1.287972 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.027762 |  -2.62907 | -0.011856 |                   0.0 |                       0.0 |                   0.337297 |                  1.303739 |                 0.072174 |    -0.054322 |        -0.072174 |         -0.966442 |             48.410379 |                 100.0 |               112.102145 |               98.814229 | 0.269592 | 0.067398 |  0.003356 |  0.003356 | 0.0 | -0.012866 |  0.000987 |    -1.049789 |     0.749608 |      0.0 |      0.0 | 0.275847 | 0.202194 |  -0.03432 | -0.056401 | -0.018547 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.056644 | 21.001887 | 21.001887 |   -0.0246 | 0.913636 | -0.745965 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 19:00:00 |  0.2 |  1.4 |              0.0 |             98.3 |       3.8 |               101.5 | nan | nan | -3.816667 |             95.029167 |                   0.1 | 0.188 |  1.4 | -3.816667 |    3.046047 |         1.973983 |           3.046047 |          3.8 |                 0.676036 |                      0.460234 |                     0.589852 |                      0.048726 |                           0.034691 |             0.664544 |                  0.437356 |     100.835456 |   1.284691 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.027812 | -2.486549 | -0.031677 |                   0.0 |                       0.0 |                    0.33771 |                  1.301296 |                 0.069202 |    -0.054221 |        -0.069202 |         -0.963585 |             56.054123 |                 100.0 |               112.102145 |               98.814229 | 0.184644 | 0.027697 | -0.000478 | -0.000478 | 0.0 | -0.012195 |  0.001001 |    -0.477013 |     0.749301 |      0.0 |      0.0 | 0.136234 | 0.136234 | -0.032529 | -0.046447 | -0.009774 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.060478 |    21.178 | 21.198726 |       0.0 | 0.917211 | -0.683388 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 20:00:00 |  0.4 |  1.6 |              0.0 |             98.7 |       5.7 |               101.4 | nan | nan |    -3.675 |             94.991667 |                   0.1 | 0.376 |  1.6 |    -3.675 |     4.56907 |         2.056027 |            4.56907 |          5.7 |                 0.685844 |                      0.465172 |                     0.616425 |                      0.049349 |                            0.03502 |             0.676928 |                  0.441874 |     100.723072 |   1.282429 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.027866 | -2.277959 | -0.037945 |                   0.0 |                       0.0 |                   0.338156 |                  1.298642 |                 0.073436 |     -0.05411 |        -0.073436 |         -0.960486 |             37.369415 |                 100.0 |               112.102145 |               98.814229 | 0.376478 | 0.037648 | -0.000147 | -0.000147 | 0.0 |  -0.01338 |  0.002396 |     0.126205 |      0.74897 |      0.0 |      0.0 | 0.071545 | 0.071545 | -0.035691 | -0.049398 | -0.011654 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.060147 | 21.300395 | 21.588584 |       0.0 | 0.921079 | -0.614542 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 21:00:00 |  0.7 |  2.0 |              0.0 |             99.6 |       5.6 |               101.3 | nan | nan |      -3.5 |             94.991667 |                   0.1 | 0.658 |  2.0 |      -3.5 |    4.488911 |         2.138899 |           4.488911 |          5.6 |                 0.705836 |                      0.471336 |                     0.637024 |                      0.050616 |                           0.035431 |             0.703012 |                   0.44773 |     100.596988 |   1.279174 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.027943 | -2.048474 |   -0.0373 |                   0.0 |                       0.0 |                   0.338996 |                  1.295068 |                 0.069193 |    -0.053961 |        -0.069193 |         -0.956071 |             38.036726 |                 100.0 |               112.102145 |               98.814229 | 0.658147 |      0.0 | -0.003019 | -0.003019 | 0.0 | -0.014392 |  0.005511 |     0.578207 |     0.748347 |      0.0 |      0.0 | 0.046683 | 0.046683 | -0.038391 | -0.046999 | -0.016197 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.063019 | 21.360888 | 22.261124 |       0.0 | 0.928414 | -0.546342 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 22:00:00 |  0.5 |  2.4 |              0.0 |             99.4 |       5.8 |               101.1 | nan | nan | -3.291667 |             95.008333 |                   0.1 |  0.47 |  2.4 | -3.291667 |    4.649229 |         2.249283 |           4.649229 |          5.8 |                  0.72634 |                      0.478769 |                     0.656358 |                       0.05191 |                           0.035924 |             0.721982 |                   0.45487 |     100.378018 |   1.274698 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |  0.028035 | -1.821141 | -0.035317 |                   0.0 |                       0.0 |                    0.33983 |                  1.290674 |                 0.065942 |    -0.053778 |        -0.065942 |         -0.950844 |             36.725115 |                 100.0 |               112.102145 |               98.814229 | 0.473019 |      0.0 | -0.002318 | -0.002318 | 0.0 | -0.014713 |  0.004753 |     0.990862 |     0.747728 |      0.0 |      0.0 | 0.028079 | 0.028079 | -0.039246 | -0.047882 | -0.021185 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.062318 |  21.40281 | 22.748855 |       0.0 | 0.935745 | -0.480125 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-10 23:00:00 |  1.0 |  2.5 |              0.0 |             99.2 |       6.2 |               101.0 | nan | nan | -3.070833 |                  95.0 |                   0.1 |  0.94 |  2.5 | -3.070833 |    4.969866 |         2.364694 |           4.969866 |          6.2 |                 0.731547 |                      0.486761 |                     0.663719 |                      0.052238 |                           0.036454 |             0.725695 |                  0.462423 |     100.274305 |   1.272954 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 |   0.02813 | -1.600418 | -0.033628 |                   0.0 |                       0.0 |                   0.340659 |                   1.28612 |                  0.06622 |    -0.053588 |         -0.06622 |         -0.945461 |             34.355753 |                 100.0 |               112.102145 |               98.814229 | 0.942318 |      0.0 | -0.001455 | -0.001455 | 0.0 |  -0.01465 |  0.009863 |     1.145165 |     0.747112 |      0.0 |      0.0 | 0.007087 | 0.007087 | -0.039078 | -0.048539 | -0.021397 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.061455 | 21.423136 | 23.705823 |       0.0 | 0.943097 | -0.415598 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 00:00:00 |  2.0 |  2.7 |              0.0 |             99.2 |       5.6 |               100.8 | nan | nan | -2.804167 |               95.4875 |                   0.1 |  1.88 |  2.7 | -2.804167 |    4.488911 |         2.438173 |           4.488911 |          5.6 |                 0.742061 |                      0.496568 |                     0.669719 |                      0.052899 |                           0.037103 |             0.736124 |                   0.47416 |     100.063876 |   1.269455 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.028222 | -1.385326 | -0.029923 |                   0.0 |                       0.0 |                   0.339142 |                  1.277745 |                 0.063053 |    -0.053239 |        -0.063053 |         -0.938603 |             38.036726 |                 100.0 |               112.102145 |               98.814229 | 1.881455 |      0.0 | -0.001628 | -0.001628 | 0.0 | -0.014483 |  0.021269 |     1.269821 |     0.746498 | 0.042522 | 0.042522 |      0.0 |      0.0 | -0.038634 | -0.047276 | -0.022857 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.061628 | 21.392716 | 25.601761 |       0.0 | 0.950489 | -0.354775 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 01:00:00 |  2.1 |  3.2 |              0.0 |             99.6 |       5.2 |               100.7 | nan | nan | -2.470833 |             95.858333 |                   0.1 | 1.974 |  3.2 | -2.470833 |    4.168274 |         2.514993 |           4.168274 |          5.2 |                  0.76893 |                      0.509073 |                     0.691346 |                      0.054584 |                           0.037928 |             0.765854 |                  0.487989 |      99.934146 |   1.265756 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.028401 | -1.182582 | -0.025544 |                   0.0 |                       0.0 |                   0.339958 |                  1.268373 |                 0.056593 |    -0.052849 |        -0.056593 |         -0.928415 |             40.962628 |                 100.0 |               112.102145 |               98.814229 | 1.975628 |      0.0 | -0.002945 | -0.002945 | 0.0 | -0.015343 |  0.026469 |      1.71111 |     0.745888 | 0.094986 | 0.094986 |      0.0 |      0.0 | -0.040926 | -0.046467 |   -0.0308 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.062945 | 21.309579 | 27.592732 |       0.0 | 0.957888 | -0.298331 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 02:00:00 |  1.8 |  3.9 |              0.0 |             99.8 |       5.7 |               100.6 | nan | nan |    -2.075 |             96.279167 |                   0.1 | 1.692 |  3.9 |    -2.075 |     4.56907 |         2.621871 |            4.56907 |          5.7 |                 0.807986 |                      0.524283 |                     0.729018 |                       0.05702 |                           0.038928 |              0.80637 |                  0.504776 |       99.79363 |   1.261108 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.028611 | -0.994436 |  -0.02148 |                   0.0 |                       0.0 |                   0.340768 |                  1.257174 |                 0.050055 |    -0.052382 |        -0.050055 |         -0.916406 |             37.369415 |                 100.0 |               112.102145 |               98.814229 | 1.694945 |      0.0 | -0.003498 | -0.003498 | 0.0 | -0.017107 |  0.027676 |     2.451504 |     0.745282 | 0.150668 | 0.150668 |      0.0 |      0.0 | -0.045631 |  -0.04855 | -0.044127 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.063498 |  21.17127 | 29.304783 |       0.0 | 0.965283 | -0.245951 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 03:00:00 |  1.3 |  4.4 |              0.0 |             99.8 |       6.6 |               100.5 | nan | nan | -1.633333 |             96.633333 |                   0.1 | 1.222 |  4.4 | -1.633333 |    5.290502 |          2.74211 |           5.290502 |          6.6 |                 0.836943 |                      0.541728 |                     0.760098 |                      0.058816 |                           0.040071 |             0.835269 |                   0.52349 |      99.664731 |   1.257444 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.028836 | -0.819837 | -0.017708 |                   0.0 |                       0.0 |                   0.341568 |                  1.245134 |                 0.046505 |    -0.051881 |        -0.046505 |         -0.903566 |             32.273586 |                 100.0 |               112.102145 |               98.814229 | 1.225498 |      0.0 | -0.003358 | -0.003358 | 0.0 | -0.018685 |  0.017872 |     3.037367 |     0.744684 | 0.164179 | 0.164179 |      0.0 |      0.0 | -0.049842 | -0.051335 | -0.054673 |      1.0 |      0.3 | 0.490286 | 0.167364 |  0.0 |  0.0 | 0.0 |  0.0 | 0.167364 | nan | nan | 0.063358 | 21.020056 |  30.05868 |       0.0 | 0.972643 | -0.197343 | 71.769182 | 0.167364 |   0.0 |   0.0 |   0.0 |  0.0 | 0.002171 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 04:00:00 |  0.5 |  4.8 |              0.0 |             99.0 |       7.0 |               100.5 | nan | nan | -1.129167 |             96.833333 |                   0.1 |  0.47 |  4.8 | -1.129167 |    5.611139 |         2.892408 |           5.611139 |          7.0 |                 0.860763 |                      0.562266 |                     0.780532 |                      0.060288 |                            0.04141 |             0.852155 |                  0.544461 |      99.647845 |   1.255555 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.029075 | -0.657809 | -0.014209 |                   0.0 |                       0.0 |                   0.342357 |                  1.232336 |                 0.043305 |    -0.051347 |        -0.043305 |         -0.889979 |             30.429381 |                 100.0 |               111.639544 |               98.814229 | 0.473358 |      0.0 |  0.000261 |  0.000261 | 0.0 | -0.018676 |       0.0 |     3.411151 |     0.744095 | 0.141294 | 0.141294 |      0.0 |      0.0 | -0.049818 | -0.054888 | -0.061401 |      1.0 |      0.3 | 0.675533 | 0.230778 |  0.0 |  0.0 | 0.0 |  0.0 | 0.230778 | nan | nan | 0.059739 | 20.891735 | 29.875181 |       0.0 | 0.979951 | -0.152234 | 72.213937 | 0.230778 |   0.0 |   0.0 |   0.0 |  0.0 | 0.007261 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 05:00:00 |  0.2 |  4.8 |              0.0 |             97.6 |       7.3 |               100.4 | nan | nan | -0.583333 |             97.016667 |                   0.1 | 0.188 |  4.8 | -0.583333 |    5.851616 |         3.059407 |           5.851616 |          7.3 |                 0.860763 |                      0.585275 |                     0.777034 |                      0.060288 |                           0.042904 |             0.840104 |                  0.567814 |      99.559896 |   1.254358 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.029334 | -0.507446 | -0.010961 |                   0.0 |                       0.0 |                    0.34314 |                  1.218349 |                  0.04455 |    -0.050765 |         -0.04455 |         -0.875209 |             29.178858 |                 100.0 |               111.003842 |               98.814229 | 0.187739 |      0.0 |  0.006927 |  0.006927 | 0.0 | -0.017022 |       0.0 |     3.347783 |      0.74351 | 0.147603 | 0.147603 |      0.0 |      0.0 | -0.045407 | -0.059403 |  -0.06026 |      1.0 |      0.3 | 0.398931 | 0.136365 |  0.0 |  0.0 | 0.0 |  0.0 | 0.136365 | nan | nan | 0.053073 | 20.755952 | 29.681012 |       0.0 | 0.987259 | -0.110373 | 72.476503 | 0.136365 |   0.0 |   0.0 |   0.0 |  0.0 |  0.01181 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 06:00:00 |  0.5 |  5.1 |              0.0 |             97.4 |       7.2 |               100.4 | nan | nan |   -0.0125 |             97.054167 |                   0.1 |  0.47 |  5.1 |   -0.0125 |    5.771457 |         3.229745 |           5.771457 |          7.2 |                 0.879017 |                      0.610223 |                     0.791632 |                      0.061412 |                           0.044515 |             0.856163 |                  0.592247 |      99.543837 |    1.25293 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.029593 |  -0.36791 | -0.007947 |                   0.0 |                       0.0 |                   0.343917 |                  1.204312 |                  0.04118 |     -0.05018 |         -0.04118 |         -0.860396 |              29.58412 |                 100.0 |               110.629315 |               98.814229 | 0.463073 |      0.0 |  0.007936 |  0.007936 | 0.0 |  -0.01722 |       0.0 |     3.610594 |     0.742929 |  0.17079 |  0.17079 |      0.0 |      0.0 | -0.045934 | -0.060237 | -0.064991 |      1.0 |      0.3 | 0.707445 | 0.242008 |  0.0 |  0.0 | 0.0 |  0.0 | 0.242008 | nan | nan | 0.052064 | 20.597105 |  29.45386 |       0.0 | 0.994567 | -0.071526 |  72.94194 | 0.242008 |   0.0 |   0.0 |   0.0 |  0.0 | 0.016397 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 07:00:00 |  0.2 |  5.1 |              0.0 |             96.3 |       7.6 |               100.4 | nan | nan |  0.554167 |             97.070833 |                   0.1 | 0.188 |  5.1 |  0.554167 |    6.092093 |         3.436822 |           6.092093 |          7.6 |                 0.879017 |                      0.635914 |                      0.78965 |                      0.061412 |                           0.046165 |             0.846494 |                  0.617287 |      99.553506 |   1.252976 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.029855 | -0.238421 |  -0.00515 |                   0.0 |                       0.0 |                   0.344688 |                  1.190153 |                 0.042442 |     -0.04959 |        -0.042442 |         -0.845465 |             28.027061 |                 100.0 |               109.966783 |               98.814229 | 0.180064 |      0.0 |   0.01395 |   0.01395 | 0.0 | -0.015861 |       0.0 |     3.575173 |     0.742352 | 0.177255 | 0.177255 |      0.0 |      0.0 | -0.042309 | -0.064486 | -0.064353 |      1.0 |      0.3 | 0.433676 | 0.148449 |  0.0 |  0.0 | 0.0 |  0.0 | 0.148449 | nan | nan |  0.04605 | 20.430846 |  29.21611 |       0.0 | 1.001875 | -0.035476 | 73.227167 | 0.148449 |   0.0 |   0.0 |   0.0 |  0.0 | 0.021011 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 08:00:00 |  0.1 |  5.4 |              0.0 |             95.2 |       7.3 |               100.3 | nan | nan |     1.175 |             97.004167 |                   0.1 | 0.094 |  5.4 |     1.175 |    5.851616 |           3.6439 |           5.851616 |          7.3 |                 0.897612 |                      0.665151 |                     0.800135 |                      0.062554 |                           0.048033 |             0.854526 |                  0.645224 |      99.445474 |   1.250338 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.347041 |                      7.400749 |             1.337827 |  0.254871 |         0.00332 |                0.003243 |  0.030996 | -0.118254 | -0.002554 |              0.000837 |                  0.000837 |                   0.345454 |                  1.174779 |                 0.039299 |    -0.048112 |        -0.038461 |         -0.829326 |             29.178858 |                 100.0 |               109.561627 |               99.526047 |  0.08005 |      0.0 |  0.019027 |  0.019027 | 0.0 |  -0.01468 |       0.0 |     3.761721 |      0.74178 |  0.19508 |  0.19508 |      0.0 |      0.0 | -0.039158 | -0.067014 | -0.067711 |      1.0 |      0.3 | 0.359154 | 0.122998 |  0.0 |  0.0 | 0.0 |  0.0 | 0.122998 | nan | nan | 0.040973 | 20.245934 | 28.951686 |       0.0 | 1.009183 | -0.002022 | 73.463323 | 0.122998 |   0.0 |   0.0 |   0.0 |  0.0 | 0.023973 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 09:00:00 |  0.3 |  5.3 |              0.0 |             96.3 |       7.1 |               100.3 | nan | nan |  1.779167 |               97.0875 |                   0.1 | 0.282 |  5.3 |  1.779167 |    5.691298 |         3.847638 |           5.691298 |          7.1 |                 0.891375 |                      0.694736 |                     0.805409 |                      0.062171 |                           0.049913 |             0.858394 |                  0.674502 |      99.441606 |   1.250768 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  4.882681 |        0.063603 |                0.062163 |  0.031937 |  -0.00674 | -0.000146 |              0.016087 |                  0.016087 |                   0.346214 |                   1.15845 |                 0.041771 |    -0.032182 |        -0.025684 |         -0.812237 |             30.000798 |                 100.0 |               109.226657 |               100.86461 | 0.262973 |      0.0 |  0.016588 |  0.016588 | 0.0 | -0.013978 |       0.0 |     3.854743 |     0.741212 | 0.207305 | 0.207305 |      0.0 |      0.0 | -0.037285 | -0.057784 | -0.069385 | 0.998914 | 0.299349 | 0.559562 | 0.191407 |  0.0 |  0.0 | 0.0 |  0.0 | 0.191407 | nan | nan | 0.043412 | 20.048304 | 28.669075 |       0.0 | 1.016491 |  0.029024 | 73.831478 | 0.191407 |   0.0 |   0.0 |   0.0 |  0.0 | 0.027418 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 10:00:00 |  0.1 |  5.5 |              0.0 |             95.5 |       7.1 |               100.3 | nan | nan |  2.295833 |               97.0375 |                   0.1 | 0.094 |  5.5 |  2.295833 |    5.691298 |         4.038016 |           5.691298 |          7.1 |                 0.903886 |                       0.72095 |                     0.829549 |                      0.062939 |                            0.05157 |             0.863211 |                  0.699592 |      99.436789 |   1.249848 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 | 14.862249 |        0.193599 |                0.189423 |  0.031377 |       0.0 |       0.0 |              0.049127 |                  0.049127 |                   0.346968 |                  1.144953 |                 0.044912 |     0.001421 |         0.004215 |         -0.797985 |             30.000798 |                 100.0 |               108.705317 |                100.8193 | 0.077412 |      0.0 |  0.026562 |  0.026562 | 0.0 |  -0.00888 |       0.0 |     4.273799 |     0.740648 | 0.230325 | 0.230325 |      0.0 |      0.0 | -0.023687 | -0.049026 | -0.076928 | 0.997757 | 0.298656 | 0.406878 | 0.138992 |  0.0 |  0.0 | 0.0 |  0.0 | 0.138992 | nan | nan | 0.033438 | 19.824118 | 28.348489 |       0.0 | 1.023799 |  0.059924 | 74.099364 | 0.138992 |   0.0 |   0.0 |   0.0 |  0.0 | 0.030974 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 11:00:00 |  0.1 |  6.1 |              0.0 |             91.6 |       8.9 |               100.2 | nan | nan |  2.716667 |             96.795833 |                   0.0 | 0.094 |  6.1 |  2.716667 |    7.134162 |         4.291853 |           7.134162 |          8.9 |                 0.942351 |                      0.742943 |                     0.867347 |                       0.06529 |                           0.052955 |             0.863194 |                  0.719138 |      99.336806 |   1.245915 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 |  22.23196 |        0.289599 |                0.288488 |  0.029031 |       0.0 |       0.0 |              0.074982 |                  0.074982 |                   0.338568 |                  1.076886 |                 0.040616 |     0.030111 |         0.034365 |         -0.738318 |             23.933221 |                 100.0 |               108.326618 |               100.78614 | 0.067438 |      0.0 |  0.059751 |  0.059751 | 0.0 |  0.001323 |       0.0 |     4.908848 |     0.740088 | 0.264549 | 0.264549 |      0.0 |      0.0 |   0.00353 | -0.057524 | -0.088359 | 0.996601 | 0.297964 | 0.445725 | 0.152056 |  0.0 |  0.0 | 0.0 |  0.0 | 0.152056 | nan | nan | 0.000249 | 19.558656 | 27.968878 |       0.0 | 1.031107 |  0.090824 | 74.393033 | 0.152056 |   0.0 |   0.0 |   0.0 |  0.0 | 0.033934 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 12:00:00 |  0.0 |  6.4 |              0.0 |             90.0 |       9.5 |               100.2 | nan | nan |       3.1 |             96.633333 |                   0.0 |   0.0 |  6.4 |       3.1 |    7.615117 |          4.56573 |           7.615117 |          9.5 |                 0.962118 |                      0.763489 |                     0.878811 |                      0.066493 |                           0.054244 |             0.865906 |                  0.737784 |      99.334094 |   1.244565 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 22.247803 |        0.289805 |                0.288994 |  0.028647 |       0.0 |       0.0 |              0.075274 |                  0.075274 |                   0.339292 |                  1.067855 |                  0.03849 |      0.03078 |         0.036784 |         -0.728563 |             22.421649 |                 100.0 |                107.91209 |              100.749601 |      0.0 |      0.0 |  0.073542 |  0.000249 | 0.0 |  0.004348 |       0.0 |     5.096646 |     0.739532 |  0.27467 |  0.27467 |      0.0 |      0.0 |  0.011598 | -0.066553 |  -0.09174 | 0.995444 | 0.297273 | 0.392716 | 0.133786 |  0.0 |  0.0 | 0.0 |  0.0 | 0.133786 | nan | nan |      0.0 | 19.280989 | 27.571814 |       0.0 | 1.038415 |  0.121724 | 74.651964 | 0.133786 |   0.0 |   0.0 |   0.0 |  0.0 | 0.036748 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 13:00:00 |  0.1 |  6.4 |              0.0 |             85.7 |       7.9 |               100.2 | nan | nan |     3.425 |               96.3125 |                   0.0 | 0.094 |  6.4 |     3.425 |    6.332571 |         4.782828 |           6.332571 |          7.9 |                 0.962118 |                      0.781298 |                     0.853745 |                      0.066493 |                           0.055357 |             0.824535 |                  0.752488 |      99.375465 |    1.24476 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 20.992726 |        0.273456 |                0.273007 |  0.028361 |       0.0 |       0.0 |               0.07126 |                   0.07126 |                   0.340011 |                   1.06128 |                 0.039492 |      0.02704 |         0.031768 |         -0.721269 |             26.962743 |                 100.0 |               107.547128 |               100.71722 |    0.034 |      0.0 |  0.084935 |      0.06 | 0.0 |  0.008418 |       0.0 |     4.683185 |      0.73898 | 0.252387 | 0.252387 |      0.0 |      0.0 |  0.022454 | -0.074983 | -0.084297 | 0.994288 | 0.296583 | 0.394804 | 0.134308 |  0.0 |  0.0 | 0.0 |  0.0 | 0.134308 | nan | nan |      0.0 | 19.022792 | 27.202593 |       0.0 | 1.045723 |  0.152624 | 74.912459 | 0.134308 |   0.0 |   0.0 |   0.0 |  0.0 | 0.039262 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 14:00:00 |  0.0 |  6.5 |              0.0 |             82.3 |       9.6 |               100.2 | nan | nan |    3.7125 |             95.854167 |                   0.0 |   0.0 |  6.5 |    3.7125 |    7.695276 |         5.026645 |           7.695276 |          9.6 |                 0.968788 |                      0.797357 |                     0.834544 |                      0.066898 |                           0.056358 |             0.797312 |                    0.7643 |      99.402688 |   1.244443 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 11.677032 |        0.152108 |                 0.15202 |  0.028146 |       0.0 |       0.0 |              0.039764 |                  0.039764 |                   0.340725 |                  1.056546 |                 0.037842 |    -0.004259 |         0.001922 |         -0.715821 |              22.18809 |                 100.0 |               107.180457 |              100.684486 |      0.0 |      0.0 |  0.115766 |       0.0 | 0.0 |  0.012657 |       0.0 |     4.359159 |     0.738432 | 0.234925 | 0.234925 |      0.0 |      0.0 |  0.033763 | -0.110306 | -0.078465 | 0.993132 | 0.295893 | 0.335786 | 0.114065 |  0.0 |  0.0 | 0.0 |  0.0 | 0.114065 | nan | nan |      0.0 | 18.779125 | 26.854149 |       0.0 | 1.053031 |  0.183524 | 75.134181 | 0.114065 |   0.0 |   0.0 |   0.0 |  0.0 | 0.041454 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 15:00:00 |  0.1 |  6.3 |              0.1 |             82.5 |       9.1 |               100.3 | nan | nan |  3.954167 |             95.370833 |                   0.1 | 0.094 |  6.3 |  3.954167 |     7.29448 |         5.237063 |            7.29448 |          9.1 |                 0.955489 |                       0.81108 |                     0.808886 |                       0.06609 |                           0.057212 |             0.788278 |                  0.773534 |      99.511722 |   1.246623 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  2.844748 |        0.044468 |                 0.04541 |  0.029862 |       0.0 |       0.0 |              0.011903 |                  0.011903 |                   0.350661 |                  1.110113 |                  0.04091 |    -0.034352 |        -0.029007 |         -0.759452 |             23.407216 |                 100.0 |               106.868752 |                100.6565 |    0.034 |      0.0 |   0.10276 |      0.06 | 0.0 |  0.006692 |       0.0 |     3.915783 |     0.737887 |  0.21103 |  0.21103 |      0.0 |      0.0 |  0.017851 | -0.117342 | -0.070484 | 0.991975 | 0.295204 | 0.335698 | 0.113866 |  0.0 |  0.0 | 0.0 |  0.0 | 0.113866 | nan | nan |      0.0 | 18.563468 | 26.545759 |       0.0 | 1.060339 |  0.214424 | 75.356012 | 0.113866 |   0.0 |   0.0 |   0.0 |  0.0 | 0.043327 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 16:00:00 |  0.0 |  5.9 |              0.0 |             82.2 |       8.7 |               100.3 | nan | nan |  4.158333 |             94.854167 |                   0.1 |   0.0 |  5.9 |  4.158333 |    6.973844 |         5.424101 |           6.973844 |          8.7 |                 0.929372 |                      0.822835 |                      0.78008 |                      0.064498 |                           0.057942 |             0.763944 |                  0.780493 |      99.536056 |   1.248525 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.053708 |                      7.400749 |             1.337827 |   0.00593 |        0.000077 |                0.000079 |  0.031325 |       0.0 |       0.0 |              0.000021 |                  0.000021 |                   0.351384 |                  1.107879 |                 0.043596 |    -0.046141 |        -0.043575 |         -0.756495 |              24.48341 |                 100.0 |               106.557245 |                   98.91 |      0.0 |      0.0 |  0.096625 |       0.0 | 0.0 |  0.005043 |       0.0 |     3.402986 |     0.737347 | 0.183394 | 0.183394 |      0.0 |      0.0 |  0.013453 | -0.118282 | -0.061254 | 0.990819 | 0.294517 | 0.262204 | 0.088801 |  0.0 |  0.0 | 0.0 |  0.0 | 0.088801 | nan | nan |      0.0 | 18.376582 | 26.278512 |       0.0 | 1.067647 |  0.245324 | 75.529415 | 0.088801 |   0.0 |   0.0 |   0.0 |  0.0 | 0.044824 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 17:00:00 |  0.0 |  5.5 |              0.0 |             76.5 |       8.9 |               100.3 | nan | nan |  4.341667 |             94.108333 |                   0.1 |   0.0 |  5.5 |  4.341667 |    7.134162 |         5.604459 |           7.134162 |          8.9 |                 0.903886 |                      0.833518 |                     0.737782 |                      0.062939 |                           0.058604 |             0.691473 |                   0.78441 |      99.608527 |   1.250659 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.031427 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.352102 |                   1.10789 |                 0.052635 |    -0.046162 |        -0.052635 |         -0.755788 |             23.933221 |                 100.0 |               106.313986 |               98.814229 |      0.0 |      0.0 |  0.129458 |       0.0 | 0.0 |  0.014756 |       0.0 |     2.618911 |      0.73681 | 0.141139 | 0.141139 |      0.0 |      0.0 |  0.039361 | -0.139136 |  -0.04714 | 0.989662 | 0.293829 | 0.201715 | 0.068204 |  0.0 |  0.0 | 0.0 |  0.0 | 0.068204 | nan | nan |      0.0 | 18.225203 |  26.06204 |       0.0 | 1.074955 |  0.276224 | 75.662926 | 0.068204 |   0.0 |   0.0 |   0.0 |  0.0 | 0.045693 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 18:00:00 |  0.0 |  5.1 |              0.0 |             81.7 |       7.9 |               100.3 | nan | nan |    4.5125 |                93.475 |                   0.1 |   0.0 |  5.1 |    4.5125 |    6.332571 |         5.721357 |           6.332571 |          7.9 |                 0.879017 |                      0.843583 |                     0.734852 |                      0.061412 |                           0.059227 |             0.718157 |                  0.788539 |      99.581843 |   1.252331 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.031446 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.352815 |                  1.107494 |                 0.051829 |    -0.046146 |        -0.051829 |         -0.754679 |             26.962743 |                 100.0 |               106.126835 |               98.814229 |      0.0 |      0.0 |   0.08771 |       0.0 | 0.0 |  0.004811 |       0.0 |     2.563144 |     0.736278 | 0.138133 | 0.138133 |      0.0 |      0.0 |  0.012834 | -0.110799 | -0.046137 | 0.988506 | 0.293143 | 0.197494 | 0.066667 |  0.0 |  0.0 | 0.0 |  0.0 | 0.066667 | nan | nan |      0.0 | 18.083731 | 25.859735 |       0.0 | 1.082263 |  0.307124 | 75.793754 | 0.066667 |   0.0 |   0.0 |   0.0 |  0.0 | 0.046254 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 19:00:00 |  0.0 |  4.8 |              0.0 |             83.4 |       7.1 |               100.3 | nan | nan |  4.654167 |             92.854167 |                   0.1 |   0.0 |  4.8 |  4.654167 |    5.691298 |         5.831576 |           5.691298 |          7.1 |                 0.860763 |                       0.85201 |                     0.723095 |                      0.060288 |                           0.059747 |             0.717876 |                  0.791127 |      99.582124 |   1.253684 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.031452 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.353523 |                  1.107885 |                 0.052715 |    -0.046162 |        -0.052715 |         -0.754362 |             30.000798 |                 100.0 |               105.943567 |               98.814229 |      0.0 |      0.0 |  0.070521 |       0.0 | 0.0 |  0.001377 |       0.0 |     2.337372 |     0.735749 | 0.125966 | 0.125966 |      0.0 |      0.0 |  0.003673 |  -0.09846 | -0.042073 | 0.987349 | 0.292458 | 0.180122 | 0.060702 |  0.0 |  0.0 | 0.0 |  0.0 | 0.060702 | nan | nan |      0.0 | 17.956808 | 25.678236 |       0.0 | 1.089571 |  0.338024 | 75.913174 | 0.060702 |   0.0 |   0.0 |   0.0 |  0.0 | 0.046704 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 20:00:00 |  0.0 |  4.9 |              0.0 |             82.8 |       6.4 |               100.3 | nan | nan |  4.791667 |             92.191667 |                   0.1 |   0.0 |  4.9 |  4.791667 |    5.130184 |         5.854956 |           5.130184 |          6.4 |                  0.86681 |                       0.86026 |                     0.721836 |                       0.06066 |                           0.060257 |             0.717719 |                  0.793088 |      99.582281 |   1.253234 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.031449 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.354225 |                  1.108737 |                 0.050775 |    -0.046197 |        -0.050775 |         -0.754511 |             33.282135 |                 100.0 |               105.776382 |               98.814229 |      0.0 |      0.0 |  0.066292 |       0.0 | 0.0 |  0.000998 |       0.0 |     2.312996 |     0.735223 | 0.124653 | 0.124653 |      0.0 |      0.0 |  0.002663 | -0.095072 | -0.041634 | 0.986193 | 0.291773 | 0.178246 | 0.059969 |  0.0 |  0.0 | 0.0 |  0.0 | 0.059969 | nan | nan |      0.0 | 17.831463 | 25.498991 |       0.0 | 1.096879 |  0.368924 | 76.031452 | 0.059969 |   0.0 |   0.0 |   0.0 |  0.0 | 0.047056 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 21:00:00 |  0.0 |  4.9 |              0.0 |             85.8 |       7.2 |               100.3 | nan | nan |    4.9125 |             91.616667 |                   0.1 |   0.0 |  4.9 |    4.9125 |    5.771457 |         5.908395 |           5.771457 |          7.2 |                  0.86681 |                      0.867569 |                      0.73735 |                       0.06066 |                           0.060707 |             0.743723 |                  0.794837 |      99.556277 |   1.253111 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.031449 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.354923 |                  1.109465 |                 0.050132 |    -0.046228 |        -0.050132 |         -0.754541 |              29.58412 |                 100.0 |               105.610894 |               98.814229 |      0.0 |      0.0 |  0.060995 |       0.0 | 0.0 | -0.001701 |       0.0 |     2.610692 |     0.734701 | 0.140696 | 0.140696 |      0.0 |      0.0 | -0.004537 | -0.092588 | -0.046992 | 0.985037 | 0.291089 | 0.201209 | 0.067583 |  0.0 |  0.0 | 0.0 |  0.0 | 0.067583 | nan | nan |      0.0 | 17.691947 | 25.299483 |       0.0 | 1.104187 |  0.399824 | 76.165078 | 0.067583 |   0.0 |   0.0 |   0.0 |  0.0 | 0.047489 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 22:00:00 |  0.1 |  4.7 |              0.0 |             87.6 |       7.0 |               100.3 | nan | nan |  5.008333 |                91.125 |                   0.1 | 0.094 |  4.7 |  5.008333 |    5.611139 |         5.948475 |           5.611139 |          7.0 |                 0.854752 |                      0.873404 |                     0.733261 |                      0.059917 |                           0.061066 |             0.748763 |                  0.795889 |      99.551237 |   1.253989 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.031446 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.355617 |                   1.11033 |                 0.050843 |    -0.046264 |        -0.050843 |         -0.754713 |             30.429381 |                 100.0 |               105.424048 |               98.814229 |    0.034 |      0.0 |  0.051031 |  0.051031 | 0.0 | -0.004042 |       0.0 |     2.532763 |     0.734183 | 0.136496 | 0.136496 |      0.0 |      0.0 | -0.010783 |  -0.08565 |  -0.04559 |  0.98388 | 0.290406 | 0.229221 | 0.076867 |  0.0 |  0.0 | 0.0 |  0.0 | 0.076867 | nan | nan | 0.008969 | 17.558255 | 25.108305 |       0.0 | 1.111495 |  0.430724 | 76.317432 | 0.076867 |   0.0 |   0.0 |   0.0 |  0.0 | 0.048128 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-11 23:00:00 |  0.1 |  3.8 |              0.0 |             85.6 |       6.8 |               100.2 | nan | nan |    5.0625 |             90.558333 |                   0.1 | 0.094 |  3.8 |    5.0625 |     5.45082 |         5.968515 |            5.45082 |          6.8 |                 0.802302 |                      0.876717 |                     0.683115 |                      0.056666 |                            0.06127 |              0.68677 |                   0.79394 |       99.51323 |   1.257101 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 |  0.031411 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.356305 |                  1.112997 |                  0.06173 |    -0.046375 |         -0.06173 |         -0.756692 |             31.324363 |                 100.0 |               105.211163 |               98.814229 | 0.042969 |      0.0 |  0.055873 |  0.055873 | 0.0 | -0.000931 |       0.0 |     1.544614 |     0.733669 | 0.083243 | 0.083243 |      0.0 |      0.0 | -0.002483 |  -0.08705 | -0.027803 | 0.982724 | 0.289724 |  0.16201 | 0.054238 |  0.0 |  0.0 | 0.0 |  0.0 | 0.054238 | nan | nan | 0.004127 |  17.47566 | 24.990194 |       0.0 | 1.118803 |  0.461624 | 76.425204 | 0.054238 |   0.0 |   0.0 |   0.0 |  0.0 | 0.048577 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 00:00:00 |  0.0 |  3.9 |              0.0 |             85.3 |       6.4 |               100.2 | nan | nan |    5.1125 |             89.979167 |                   0.1 |   0.0 |  3.9 |    5.1125 |    5.130184 |         5.995234 |           5.130184 |          6.4 |                 0.807986 |                      0.879785 |                     0.684566 |                       0.05702 |                           0.061459 |             0.689212 |                  0.791623 |      99.510788 |   1.256635 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.031304 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.354801 |                  1.116023 |                 0.060148 |    -0.046501 |        -0.060148 |         -0.761222 |             33.282135 |                 100.0 |               105.060668 |               98.814229 |      0.0 |      0.0 |  0.053921 |  0.004127 | 0.0 | -0.001126 |       0.0 |     1.574098 |     0.733158 | 0.084832 | 0.084832 |      0.0 |      0.0 | -0.003005 | -0.085477 | -0.028334 | 0.981567 | 0.289042 | 0.121315 | 0.040543 |  0.0 |  0.0 | 0.0 |  0.0 | 0.040543 | nan | nan |      0.0 | 17.391613 | 24.870006 |       0.0 | 1.126111 |  0.492524 | 76.505975 | 0.040543 |   0.0 |   0.0 |   0.0 |  0.0 | 0.048546 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 01:00:00 |  0.0 |  3.7 |              0.0 |             86.2 |       6.3 |               100.2 | nan | nan |  5.133333 |             89.420833 |                   0.1 |   0.0 |  3.7 |  5.133333 |    5.050025 |         6.031974 |           5.050025 |          6.3 |                 0.796653 |                      0.881066 |                     0.678059 |                      0.056315 |                           0.061538 |             0.686715 |                  0.787857 |      99.513285 |   1.257555 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.031252 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.355475 |                  1.119656 |                 0.061541 |    -0.046652 |        -0.061541 |          -0.76418 |             33.810423 |                 100.0 |               104.947929 |               98.814229 |      0.0 |      0.0 |  0.049225 |       0.0 | 0.0 | -0.002072 |       0.0 |     1.441453 |      0.73265 | 0.077683 | 0.077683 |      0.0 |      0.0 | -0.005528 | -0.081958 | -0.025946 | 0.980411 | 0.288362 | 0.111096 | 0.037063 |  0.0 |  0.0 | 0.0 |  0.0 | 0.037063 | nan | nan |      0.0 | 17.315372 | 24.760983 |       0.0 | 1.133419 |  0.523424 | 76.580008 | 0.037063 |   0.0 |   0.0 |   0.0 |  0.0 | 0.048294 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 02:00:00 |  0.1 |  3.5 |              0.0 |             89.3 |       5.7 |               100.2 | nan | nan |  5.116667 |             88.983333 |                   0.1 | 0.094 |  3.5 |  5.116667 |     4.56907 |         6.031974 |            4.56907 |          5.7 |                  0.78546 |                      0.880041 |                     0.675826 |                      0.055617 |                           0.061475 |             0.701415 |                   0.78309 |      99.498585 |   1.258394 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.031197 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.356146 |                  1.123475 |                 0.060505 |    -0.046811 |        -0.060505 |         -0.767329 |             37.369415 |                 100.0 |               104.844634 |               98.814229 |    0.034 |      0.0 |    0.0336 |    0.0336 | 0.0 | -0.005659 |       0.0 |     1.395689 |     0.732146 | 0.075217 | 0.075217 |      0.0 |      0.0 | -0.015095 | -0.070532 | -0.025122 | 0.979254 | 0.287682 | 0.141585 | 0.047153 |  0.0 |  0.0 | 0.0 |  0.0 | 0.047153 | nan | nan |   0.0264 | 17.244096 | 24.659057 |       0.0 | 1.140727 |  0.554324 |  76.67444 | 0.047153 |   0.0 |   0.0 |   0.0 |  0.0 | 0.048135 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 03:00:00 |  0.0 |  3.3 |              0.0 |             91.4 |       5.2 |               100.3 | nan | nan |  5.070833 |             88.633333 |                   0.1 |   0.0 |  3.3 |  5.070833 |    4.168274 |         5.985215 |           4.168274 |          5.2 |                 0.774405 |                      0.877228 |                     0.670452 |                      0.054927 |                           0.061302 |             0.707807 |                  0.777516 |      99.592193 |   1.260534 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.031139 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.356811 |                  1.127474 |                 0.060276 |    -0.046978 |        -0.060276 |         -0.770663 |             40.962628 |                 100.0 |               104.712931 |               98.814229 |      0.0 |      0.0 |  0.023843 |  0.023843 | 0.0 | -0.007692 |       0.0 |     1.284975 |     0.731646 |  0.06925 |  0.06925 |      0.0 |      0.0 | -0.020518 | -0.062887 |  -0.02313 | 0.978098 | 0.287003 | 0.099059 | 0.032932 |  0.0 |  0.0 | 0.0 |  0.0 | 0.032932 | nan | nan | 0.002557 | 17.180203 | 24.567691 |       0.0 | 1.148035 |  0.585224 | 76.740567 | 0.032932 |   0.0 |   0.0 |   0.0 |  0.0 | 0.047925 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 04:00:00 |  0.0 |  3.3 |              0.0 |             90.9 |       4.6 |               100.4 | nan | nan |  5.008333 |             88.295833 |                   0.1 |   0.0 |  3.3 |  5.008333 |     3.68732 |         5.905055 |            3.68732 |          4.6 |                 0.774405 |                      0.873404 |                     0.664829 |                      0.054927 |                           0.061066 |             0.703935 |                  0.771179 |      99.696065 |   1.261813 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.031074 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.357472 |                  1.131856 |                 0.059168 |    -0.047161 |        -0.059168 |         -0.774384 |              46.30558 |                 100.0 |                104.62074 |               98.814229 |      0.0 |      0.0 |  0.022326 |  0.002557 | 0.0 | -0.007338 |       0.0 |     1.168289 |     0.731148 | 0.062962 | 0.062962 |      0.0 |      0.0 | -0.019574 | -0.060624 | -0.021029 | 0.976941 | 0.286324 | 0.090062 | 0.029888 |  0.0 |  0.0 | 0.0 |  0.0 | 0.029888 | nan | nan |      0.0 | 17.122354 | 24.484966 |       0.0 | 1.155343 |  0.616124 | 76.800741 | 0.029888 |   0.0 |   0.0 |   0.0 |  0.0 | 0.047498 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 05:00:00 |  0.0 |  3.0 |              0.0 |             89.2 |       4.9 |               100.5 | nan | nan |  4.933333 |             87.945833 |                   0.1 |   0.0 |  3.0 |  4.933333 |    3.927797 |         5.824896 |           3.927797 |          4.9 |                 0.758081 |                      0.868834 |                     0.648694 |                      0.053905 |                           0.060785 |             0.676208 |                  0.764103 |      99.823792 |   1.264577 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.031001 |       0.0 |       0.0 |                   0.0 |                       0.0 |                   0.358129 |                  1.136687 |                 0.064288 |    -0.047362 |        -0.064288 |         -0.778558 |             43.470544 |                 100.0 |               104.536873 |               98.814229 |      0.0 |      0.0 |  0.028435 |       0.0 | 0.0 | -0.005414 |       0.0 |     0.828592 |     0.730655 | 0.044655 | 0.044655 |      0.0 |      0.0 | -0.014442 |  -0.06476 | -0.014915 | 0.975785 | 0.285647 |  0.06387 | 0.021157 |  0.0 |  0.0 | 0.0 |  0.0 | 0.021157 | nan | nan |      0.0 | 17.081476 |  24.42651 |       0.0 | 1.162651 |  0.647024 | 76.843454 | 0.021157 |   0.0 |   0.0 |   0.0 |  0.0 |  0.04693 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 06:00:00 |  0.0 |  1.9 |              0.0 |             85.6 |       6.3 |               100.6 | nan | nan |       4.8 |             87.454167 |                   0.1 |   0.0 |  1.9 |       4.8 |    5.050025 |         5.794837 |           5.050025 |          6.3 |                  0.70079 |                      0.860763 |                     0.600114 |                      0.050296 |                           0.060288 |             0.599876 |                  0.752773 |     100.000124 |   1.271267 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |   0.03088 |       0.0 |  0.000885 |                   0.0 |                       0.0 |                   0.358781 |                  1.144252 |                 0.081157 |    -0.047677 |        -0.081157 |         -0.785471 |             33.810423 |                 100.0 |               104.477357 |               98.814229 |      0.0 |      0.0 |  0.047693 |       0.0 | 0.0 |  0.000057 |       0.0 |    -0.241282 |     0.730164 |      0.0 |      0.0 | 0.008148 | 0.008148 |  0.000152 | -0.077703 |  0.003606 | 0.974662 |  0.28499 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 17.089584 | 24.426454 |       0.0 | 1.169959 |  0.677039 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.04599 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 07:00:00 |  0.0 |  1.2 |              0.0 |             86.3 |       5.8 |               100.8 | nan | nan |    4.6375 |               87.0375 |                   0.1 |   0.0 |  1.2 |    4.6375 |    4.649229 |         5.734717 |           4.649229 |          5.8 |                 0.666353 |                      0.851015 |                     0.571796 |                       0.04811 |                           0.059686 |             0.575062 |                  0.740702 |     100.224938 |   1.277169 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 |  0.030755 |       0.0 |  0.003303 |                   0.0 |                       0.0 |                   0.359429 |                  1.152031 |                 0.086796 |    -0.048001 |        -0.086796 |         -0.792601 |             36.725115 |                 100.0 |               104.477357 |               98.814229 |      0.0 |      0.0 |  0.040358 |       0.0 | 0.0 | -0.000732 |       0.0 |    -0.900735 |     0.729677 |      0.0 |      0.0 | 0.030414 | 0.030414 | -0.001953 | -0.071382 |  0.013461 | 0.973629 | 0.284386 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 17.120511 | 24.427186 |       0.0 | 1.177267 |  0.704636 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.044802 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 08:00:00 |  0.0 |  0.8 |              0.0 |             91.2 |       6.1 |               100.9 | nan | nan |  4.445833 |             86.870833 |                   0.1 |   0.0 |  0.8 |  4.445833 |    4.889707 |         5.694638 |           4.889707 |          6.1 |                 0.647352 |                      0.839643 |                     0.570584 |                      0.046897 |                           0.058983 |             0.590385 |                  0.729404 |     100.309615 |   1.280232 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.339738 |                      7.386143 |             1.329627 |  0.244956 |        0.003171 |                0.003219 |  0.030926 |       0.0 |  0.003411 |              0.000872 |                  0.000872 |                   0.360073 |                  1.158654 |                 0.088254 |    -0.047406 |        -0.087383 |         -0.798581 |             34.918962 |                 100.0 |               104.477357 |               99.359558 |      0.0 |      0.0 |  0.025973 |       0.0 | 0.0 |  -0.00462 |       0.0 |    -0.929612 |     0.729193 |      0.0 |      0.0 | 0.031378 | 0.031378 | -0.012325 | -0.061167 |  0.013891 |   0.9726 | 0.283785 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 17.155133 | 24.431806 |       0.0 | 1.184575 |  0.732126 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.043644 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 09:00:00 |  0.0 |  0.8 |              0.0 |             87.5 |       5.9 |               101.0 | nan | nan |  4.258333 |             86.504167 |                   0.1 |   0.0 |  0.8 |  4.258333 |    4.729388 |         5.654558 |           4.729388 |          5.9 |                 0.647352 |                      0.828647 |                     0.567783 |                      0.046897 |                           0.058302 |             0.566433 |                  0.716814 |     100.433567 |   1.281619 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 |  4.849391 |        0.062779 |                0.063765 |  0.031752 |       0.0 |  0.003658 |              0.017299 |                  0.017299 |                   0.360712 |                  1.166643 |                 0.093244 |    -0.031311 |        -0.075946 |         -0.805931 |             36.102655 |                 100.0 |               104.477357 |              100.436775 |      0.0 |      0.0 |  0.039435 |       0.0 | 0.0 |  0.000307 |       0.0 |    -0.996527 |     0.728712 |      0.0 |      0.0 | 0.033627 | 0.033627 |  0.000818 | -0.061875 |  0.014889 | 0.971581 | 0.283191 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 17.188544 | 24.431499 |       0.0 | 1.191883 |  0.759368 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.042517 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 10:00:00 |  0.1 |  0.5 |              0.0 |             84.5 |       6.4 |               101.1 | nan | nan |      4.05 |             86.045833 |                   0.1 | 0.094 |  0.5 |      4.05 |    5.130184 |         5.631178 |           5.130184 |          6.4 |                 0.633417 |                      0.816579 |                     0.566091 |                      0.046005 |                           0.057554 |             0.535238 |                  0.702632 |     100.564762 |   1.284447 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 14.852029 |         0.19227 |                0.195489 |  0.032095 |       0.0 |  0.003793 |              0.053108 |                  0.053108 |                   0.361215 |                  1.175846 |                 0.105146 |     0.004114 |        -0.052038 |         -0.814631 |             33.282135 |                 100.0 |               104.477357 |              100.436775 |    0.034 |  0.01275 |  0.057812 |  0.057812 | 0.0 |  0.007481 |  0.000058 |    -1.037081 |     0.728333 |      0.0 |      0.0 | 0.034895 | 0.034895 |  0.019956 | -0.056487 |  0.015506 | 0.970566 |   0.2826 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan | 0.002188 | 17.230919 | 24.458018 |       0.0 | 1.197672 |  0.786474 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.041419 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 11:00:00 |  0.0 |  0.3 |              0.0 |             76.7 |       6.8 |               101.2 | nan | nan |  3.808333 |                85.425 |                   0.1 |   0.0 |  0.3 |  3.808333 |     5.45082 |         5.561039 |            5.45082 |          6.8 |                 0.624275 |                      0.802774 |                     0.550833 |                      0.045418 |                           0.056696 |             0.478819 |                   0.68577 |     100.721181 |   1.286932 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.269583 |        0.288296 |                0.293415 |  0.032514 |       0.0 |   0.00517 |               0.07985 |                   0.07985 |                   0.361847 |                   1.18712 |                 0.117878 |     0.030387 |        -0.038027 |         -0.825273 |             31.324363 |                 100.0 |               104.477357 |              100.436775 |      0.0 |      0.0 |  0.092255 |  0.002188 | 0.0 |  0.018339 |       0.0 |     -1.40778 |     0.727858 |      0.0 |      0.0 | 0.047491 | 0.047491 |  0.048919 | -0.065915 |  0.021032 | 0.969603 | 0.282039 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 17.265453 | 24.439679 |       0.0 |  1.20498 |  0.812204 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.040349 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 12:00:00 |  0.0 | -0.1 |              0.0 |             74.8 |       7.0 |               101.3 | nan | nan |    3.5375 |             84.791667 |                   0.1 |   0.0 | -0.1 |    3.5375 |    5.611139 |          5.47754 |           5.611139 |          7.0 |                 0.606339 |                      0.787548 |                     0.532185 |                      0.044264 |                           0.055747 |             0.453542 |                  0.667775 |     100.846458 |   1.290215 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.293018 |        0.288599 |                0.293995 |  0.032964 |       0.0 |  0.006889 |              0.080147 |                  0.080147 |                   0.362474 |                  1.199167 |                  0.12351 |     0.030182 |        -0.043363 |         -0.836693 |             30.429381 |                 100.0 |               104.477357 |              100.436775 |      0.0 |      0.0 |  0.099857 |       0.0 | 0.0 |  0.020507 |       0.0 |    -1.873367 |     0.727387 |      0.0 |      0.0 | 0.063144 | 0.063144 |    0.0547 | -0.070084 |  0.027979 | 0.968705 | 0.281517 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 17.314057 | 24.419172 |       0.0 | 1.212288 |  0.836215 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.039307 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 13:00:00 |  0.0 | -0.4 |              0.0 |             69.7 |       6.3 |               101.4 | nan | nan |  3.254167 |                84.125 |                   0.1 |   0.0 | -0.4 |  3.254167 |    5.050025 |         5.424101 |           5.050025 |          6.3 |                 0.593188 |                      0.771892 |                     0.507395 |                      0.043416 |                           0.054769 |             0.413452 |                  0.649354 |     100.986548 |   1.293105 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 21.019348 |        0.272111 |                0.277484 |  0.033433 |       0.0 |  0.009266 |              0.075776 |                  0.075776 |                   0.363098 |                   1.21168 |                 0.130083 |     0.025289 |        -0.054307 |         -0.848582 |             33.810423 |                 100.0 |               104.477357 |              100.436775 |      0.0 |      0.0 |  0.105566 |       0.0 | 0.0 |  0.022493 |       0.0 |    -2.515142 |     0.726918 |      0.0 |      0.0 | 0.084685 | 0.084685 |  0.059999 | -0.076755 |  0.037551 | 0.967895 | 0.281046 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 17.382715 | 24.396679 |       0.0 | 1.219596 |  0.857849 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.038291 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 14:00:00 |  0.0 | -0.6 |              0.2 |             66.9 |       7.3 |               101.5 | nan | nan |  2.958333 |             83.483333 |                   0.3 |   0.0 | -0.6 |  2.958333 |    5.851616 |         5.347282 |           5.851616 |          7.3 |                  0.58456 |                      0.755838 |                      0.48556 |                      0.042857 |                           0.053764 |             0.391071 |                  0.630999 |     101.108929 |    1.29544 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.399736 | 11.654171 |         0.21122 |                0.216915 |  0.038101 |       0.0 |  0.011461 |              0.059337 |                  0.059337 |                   0.382895 |                  1.349977 |                 0.140047 |     0.003088 |         -0.08071 |         -0.967082 |             29.178858 |                 100.0 |               104.477357 |              100.436775 |      0.0 |      0.0 |  0.126981 |       0.0 | 0.0 |  0.025502 |       0.0 |    -3.103793 |     0.726452 |      0.0 |      0.0 | 0.104363 | 0.104363 |  0.068026 | -0.102418 |  0.046318 | 0.967167 | 0.280624 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 17.468799 | 24.371177 |       0.0 | 1.226904 |  0.877288 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.037302 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 15:00:00 |  0.0 | -0.8 |              0.2 |             64.4 |       7.5 |               101.5 | nan | nan |    2.6625 |             82.729167 |                   0.4 |   0.0 | -0.8 |    2.6625 |    6.011934 |         5.293842 |           6.011934 |          7.5 |                 0.576044 |                       0.74008 |                     0.456441 |                      0.042305 |                           0.052775 |             0.370972 |                  0.612262 |     101.129028 |   1.296488 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |              1.43479 |  2.813062 |        0.050984 |                0.053414 |  0.040777 |       0.0 |  0.014538 |              0.014636 |                  0.014636 |                   0.393148 |                  1.428152 |                 0.141049 |     -0.04487 |        -0.126413 |         -1.035005 |             28.400756 |                 100.0 |               104.477357 |              100.436775 |      0.0 |      0.0 |  0.131464 |       0.0 | 0.0 |  0.023588 |       0.0 |    -3.926396 |     0.725989 |      0.0 |      0.0 | 0.131805 | 0.131805 |  0.062921 | -0.130773 |   0.05856 | 0.966555 | 0.280269 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 17.583568 | 24.347589 |       0.0 | 1.234212 |  0.893651 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.036339 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 16:00:00 |  0.0 | -1.3 |              0.2 |             65.4 |       7.0 |               101.7 | nan | nan |    2.3625 |             82.029167 |                   0.6 |   0.0 | -1.3 |    2.3625 |    5.611139 |         5.237063 |           5.611139 |          7.0 |                 0.555231 |                      0.724396 |                      0.43436 |                      0.040952 |                           0.051788 |             0.363121 |                  0.594216 |     101.336879 |   1.301474 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.046405 |                      7.386143 |             1.504898 |  0.004444 |        0.000553 |                0.000608 |  0.025337 |       0.0 |  0.017019 |              0.000167 |                  0.000167 |                    0.41305 |                  1.571379 |                 0.148601 |    -0.065307 |        -0.148434 |         -1.158329 |             30.429381 |                 100.0 |               104.477357 |               98.888363 |      0.0 |      0.0 |  0.111666 |       0.0 | 0.0 |  0.018576 |       0.0 |    -4.581748 |      0.72553 |      0.0 |      0.0 | 0.153502 | 0.153502 |   0.04955 | -0.129695 |  0.068289 | 0.966036 | 0.279967 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 17.723538 | 24.329013 |       0.0 |  1.24152 |  0.907531 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0354 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 17:00:00 |  0.0 | -1.6 |              0.0 |             67.6 |       6.0 |               101.7 | nan | nan |  2.066667 |             81.658333 |                   0.6 |   0.0 | -1.6 |  2.066667 |    4.809547 |         5.140204 |           4.809547 |          6.0 |                 0.543065 |                      0.709217 |                     0.426703 |                      0.040158 |                           0.050829 |             0.367112 |                  0.579135 |     101.332888 |   1.302892 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |   0.02414 |       0.0 |  0.017953 |                   0.0 |                       0.0 |                   0.413737 |                  1.583999 |                 0.147363 |       -0.066 |        -0.147363 |         -1.170261 |             35.500944 |                 100.0 |               104.477357 |               98.814229 |      0.0 |      0.0 |  0.087497 |       0.0 | 0.0 |  0.013723 |       0.0 |     -4.81588 |     0.725073 |      0.0 |      0.0 | 0.160997 | 0.160997 |  0.036606 | -0.112244 |  0.071725 | 0.965551 | 0.279687 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 17.874446 | 24.315289 |       0.0 | 1.248828 |  0.920479 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.034486 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 18:00:00 |  0.0 | -2.0 |              0.0 |             70.2 |       4.8 |               101.8 | nan | nan |  1.770833 |             81.179167 |                   0.6 |   0.0 | -2.0 |  1.770833 |    3.847638 |         5.036665 |           3.847638 |          4.8 |                  0.52721 |                       0.69432 |                     0.415655 |                       0.03912 |                           0.049887 |             0.370102 |                  0.563643 |     101.429898 |   1.306085 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.023915 |       0.0 |  0.019306 |                   0.0 |                       0.0 |                    0.41442 |                    1.5975 |                 0.146053 |    -0.066562 |        -0.146053 |          -1.18308 |             44.376181 |                 100.0 |               104.477357 |               98.814229 |      0.0 |      0.0 |  0.062236 |       0.0 | 0.0 |  0.008825 |       0.0 |     -5.16037 |     0.724619 |      0.0 |      0.0 | 0.172133 | 0.172133 |  0.023541 | -0.092796 |  0.076798 | 0.965117 | 0.279435 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 18.040029 | 24.306464 |       0.0 | 1.256136 |  0.932073 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.033595 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 19:00:00 |  0.0 | -2.1 |              0.0 |             70.4 |       4.2 |               101.9 | nan | nan |  1.483333 |               80.6375 |                   0.6 |   0.0 | -2.1 |  1.483333 |    3.366683 |         4.939806 |           3.366683 |          4.2 |                 0.523311 |                      0.680107 |                     0.410428 |                      0.038864 |                           0.048985 |             0.368411 |                  0.548422 |     101.531589 |    1.30786 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.023687 |       0.0 |      0.02 |                   0.0 |                       0.0 |                   0.415098 |                  1.611195 |                 0.145331 |    -0.067133 |        -0.145331 |         -1.196096 |             50.715635 |                 100.0 |               104.477357 |               98.814229 |      0.0 |      0.0 |  0.053427 |       0.0 | 0.0 |  0.007372 |       0.0 |    -5.326196 |     0.724168 |      0.0 |      0.0 | 0.177259 | 0.177259 |  0.019665 | -0.085791 |  0.079205 | 0.964709 | 0.279199 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 18.211762 | 24.299092 |       0.0 | 1.263444 |  0.942973 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.032727 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 20:00:00 |  0.0 | -2.2 |              0.0 |             69.5 |       4.9 |               101.9 | nan | nan |    1.1875 |             80.083333 |                   0.6 |   0.0 | -2.2 |    1.1875 |    3.927797 |         4.889707 |           3.927797 |          4.9 |                 0.519437 |                      0.665752 |                     0.409287 |                       0.03861 |                           0.048071 |             0.361009 |                  0.533156 |     101.538991 |   1.308379 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.023456 |       0.0 |  0.020213 |                   0.0 |                       0.0 |                   0.415772 |                  1.625096 |                 0.148663 |    -0.067712 |        -0.148663 |         -1.209323 |             43.470544 |                 100.0 |               104.477357 |               98.814229 |      0.0 |      0.0 |  0.064423 |       0.0 | 0.0 |    0.0095 |       0.0 |    -5.362617 |     0.723721 |      0.0 |      0.0 | 0.178052 | 0.178052 |  0.025342 | -0.094322 |  0.079683 | 0.964309 | 0.278968 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 18.382624 | 24.289592 |       0.0 | 1.270752 |  0.953659 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.031882 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 21:00:00 |  0.0 | -2.6 |              0.0 |             71.6 |       4.0 |               102.0 | nan | nan |     0.875 |             79.491667 |                   0.6 |   0.0 | -2.6 |     0.875 |    3.206365 |         4.782828 |           3.206365 |          4.0 |                 0.504195 |                      0.650878 |                     0.398243 |                      0.037607 |                           0.047123 |             0.361003 |                  0.517394 |     101.638997 |   1.311601 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.023212 |       0.0 |  0.021644 |                   0.0 |                       0.0 |                   0.416442 |                  1.639665 |                 0.148191 |    -0.068319 |        -0.148191 |         -1.223223 |             53.251417 |                 100.0 |               104.477357 |               98.814229 |      0.0 |      0.0 |  0.047085 |       0.0 | 0.0 |  0.006307 |       0.0 |     -5.72012 |     0.723276 |      0.0 |      0.0 | 0.189463 | 0.189463 |  0.016824 | -0.080089 |  0.084925 | 0.963963 | 0.278767 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 18.567265 | 24.283285 |       0.0 |  1.27806 |  0.962915 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.031058 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 22:00:00 |  0.0 | -3.1 |              0.0 |             74.5 |       3.4 |               102.0 | nan | nan |      0.55 |             78.945833 |                   0.6 |   0.0 | -3.1 |      0.55 |     2.72541 |         4.662589 |            2.72541 |          3.4 |                 0.485698 |                      0.635721 |                     0.387154 |                      0.036384 |                           0.046153 |             0.361845 |                  0.501875 |     101.638155 |   1.314025 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.022973 |       0.0 |   0.02313 |                   0.0 |                       0.0 |                   0.417107 |                  1.653965 |                 0.148531 |    -0.068915 |        -0.148531 |         -1.236857 |             62.648725 |                 100.0 |               104.477357 |               98.814229 |      0.0 |      0.0 |  0.033948 |       0.0 | 0.0 |  0.003824 |       0.0 |    -6.088014 |     0.722833 |      0.0 |      0.0 | 0.201138 | 0.201138 |    0.0102 | -0.068421 |   0.09031 | 0.963672 | 0.278599 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 18.765448 | 24.279461 |       0.0 | 1.285368 |  0.970685 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.030256 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
    | 2010-12-12 23:00:00 |  0.0 | -3.9 |              0.0 |             80.2 |       3.4 |               102.0 | nan | nan |  0.229167 |             78.720833 |                   0.6 |   0.0 | -3.9 |  0.229167 |     2.72541 |          4.54903 |            2.72541 |          3.4 |                 0.457351 |                      0.621065 |                     0.375947 |                      0.034499 |                           0.045212 |             0.366795 |                  0.488908 |     101.633205 |   1.317905 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 |  0.022793 |       0.0 |  0.024683 |                   0.0 |                       0.0 |                   0.417769 |                  1.664915 |                 0.151027 |    -0.069371 |        -0.151027 |         -1.247147 |             62.648725 |                 100.0 |               104.477357 |               98.814229 |      0.0 |      0.0 |  0.023916 |       0.0 | 0.0 |  0.001383 |       0.0 |    -6.469402 |     0.722394 |      0.0 |      0.0 | 0.213167 | 0.213167 |  0.003688 | -0.058835 |   0.09588 | 0.963439 | 0.278465 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | nan | nan |      0.0 | 18.977534 | 24.278078 |       0.0 | 1.292676 |  0.976903 | 76.843454 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.029474 |   0.0 |   0.0 |   0.0 |  0.0 |   nan |    nan |
"""
# import...
# ...from HydPy
from hydpy.exe.modelimports import *
from hydpy.core import masktools
from hydpy.core import modeltools
# ...from lland
from hydpy.models.lland import lland_model
from hydpy.models.lland import lland_masks
from hydpy.models.lland.lland_constants import *


class Model(modeltools.AdHocModel):
    """Penman-Monteith and Knauf version of HydPy-L-Land (|lland_v3|)."""
    INLET_METHODS = (
        lland_model.Pick_QZ_V1,
    )
    RECEIVER_METHODS = ()
    ADD_METHODS = (
        lland_model.Return_AdjustedWindSpeed_V1,
        lland_model.Return_ActualVapourPressure_V1,
        lland_model.Return_NetLongwaveRadiationSnow_V1,
        lland_model.Return_DailyGlobalRadiation_V1,
        lland_model.Return_Penman_V1,
        lland_model.Return_PenmanMonteith_V1,
        lland_model.Return_EnergyGainSnowSurface_V1,
        lland_model.Return_SaturationVapourPressure_V1,
        lland_model.Return_SaturationVapourPressureSlope_V1,
        lland_model.Return_NetRadiation_V1,
        lland_model.Return_WSensSnow_V1,
        lland_model.Return_WLatSnow_V1,
        lland_model.Return_WSurf_V1,
        lland_model.Return_TempS_V1,
        lland_model.Return_WG_V1,
        lland_model.Return_BackwardEulerError_V1,
        lland_model.Return_ESnow_V1,
        lland_model.Return_TempSSurface_V1,
    )
    RUN_METHODS = (
        lland_model.Calc_QZH_V1,
        lland_model.Update_LoggedTemL_V1,
        lland_model.Calc_TemLTag_V1,
        lland_model.Update_LoggedRelativeHumidity_V1,
        lland_model.Calc_DailyRelativeHumidity_V1,
        lland_model.Update_LoggedSunshineDuration_V1,
        lland_model.Calc_NKor_V1,
        lland_model.Calc_TKor_V1,
        lland_model.Calc_TKorTag_V1,
        lland_model.Calc_WindSpeed2m_V1,
        lland_model.Calc_ReducedWindSpeed2m_V1,
        lland_model.Update_LoggedWindSpeed2m_V1,
        lland_model.Calc_DailyWindSpeed2m_V1,
        lland_model.Calc_WindSpeed10m_V1,
        lland_model.Calc_SolarDeclination_V1,
        lland_model.Calc_TSA_TSU_V1,
        lland_model.Calc_DailySunshineDuration_V1,
        lland_model.Calc_EarthSunDistance_V1,
        lland_model.Calc_ExtraterrestrialRadiation_V1,
        lland_model.Calc_PossibleSunshineDuration_V1,
        lland_model.Calc_DailyPossibleSunshineDuration_V1,
        lland_model.Calc_SP_V1,
        lland_model.Calc_GlobalRadiation_V1,
        lland_model.Update_LoggedGlobalRadiation_V1,
        lland_model.Calc_DailyGlobalRadiation_V1,
        lland_model.Calc_AdjustedGlobalRadiation_V1,
        lland_model.Calc_SaturationVapourPressure_V1,
        lland_model.Calc_DailySaturationVapourPressure_V1,
        lland_model.Calc_SaturationVapourPressureSlope_V1,
        lland_model.Calc_DailySaturationVapourPressureSlope_V1,
        lland_model.Calc_ActualVapourPressure_V1,
        lland_model.Calc_DailyActualVapourPressure_V1,
        lland_model.Calc_NBes_Inzp_V1,
        lland_model.Calc_SNRatio_V1,
        lland_model.Calc_SBes_V1,
        lland_model.Calc_WATS_V1,
        lland_model.Calc_WaDa_WAeS_V1,
        lland_model.Calc_WNied_ESnow_V1,
        lland_model.Calc_TempS_V1,
        lland_model.Update_TauS_V1,
        lland_model.Calc_ActualAlbedo_V1,
        lland_model.Calc_NetShortwaveRadiationSnow_V1,
        lland_model.Calc_TZ_V1,
        lland_model.Calc_WG_V1,
        lland_model.Update_ESnow_V1,
        lland_model.Calc_SchmPot_V2,
        lland_model.Calc_Schm_WATS_V1,
        lland_model.Calc_GefrPot_V1,
        lland_model.Calc_Gefr_WATS_V1,
        lland_model.Calc_EvS_WAeS_WATS_V1,
        lland_model.Update_WaDa_WAeS_V1,
        lland_model.Update_ESnow_V2,
        lland_model.Calc_NetShortwaveRadiation_V1,
        lland_model.Calc_DailyNetShortwaveRadiation_V1,
        lland_model.Calc_DailyNetLongwaveRadiation_V1,
        lland_model.Calc_NetRadiation_V1,
        lland_model.Calc_DailyNetRadiation_V1,
        lland_model.Calc_DryAirPressure_V1,
        lland_model.Calc_DensityAir_V1,
        lland_model.Calc_AerodynamicResistance_V1,
        lland_model.Calc_SoilSurfaceResistance_V1,
        lland_model.Calc_LanduseSurfaceResistance_V1,
        lland_model.Calc_ActualSurfaceResistance_V1,
        lland_model.Calc_G_V1,
        lland_model.Calc_EvPo_V2,
        lland_model.Calc_EvI_Inzp_V1,
        lland_model.Calc_EvB_V2,
        lland_model.Update_EBdn_V1,
        lland_model.Calc_SFF_V1,
        lland_model.Calc_FVG_V1,
        lland_model.Calc_QKap_V1,
        lland_model.Calc_QBB_V1,
        lland_model.Calc_QIB1_V1,
        lland_model.Calc_QIB2_V1,
        lland_model.Calc_QDB_V1,
        lland_model.Update_QDB_V1,
        lland_model.Calc_BoWa_V1,
        lland_model.Calc_QBGZ_V1,
        lland_model.Calc_QIGZ1_V1,
        lland_model.Calc_QIGZ2_V1,
        lland_model.Calc_QDGZ_V1,
        lland_model.Calc_QDGZ1_QDGZ2_V1,
        lland_model.Calc_QBGA_V1,
        lland_model.Calc_QIGA1_V1,
        lland_model.Calc_QIGA2_V1,
        lland_model.Calc_QDGA1_V1,
        lland_model.Calc_QDGA2_V1,
        lland_model.Calc_QAH_V1,
        lland_model.Calc_QA_V1,
    )
    OUTLET_METHODS = (
        lland_model.Pass_QA_V1,
    )
    SENDER_METHODS = ()
    SUBMODELS = (
        lland_model.PegasusESnow,
        lland_model.PegasusTempSSurface,
    )
    idx_hru = modeltools.Idx_HRU()


class Masks(masktools.Masks):
    """Masks applicable to |lland_v3|."""
    # pylint: disable=no-member
    # bug of pylint 2.4?
    CLASSES = lland_masks.Masks.CLASSES


tester = Tester()
cythonizer = Cythonizer()
cythonizer.finalise()
